(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005, 2006
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

(* a basic implementation of a discrimination tree term indexing tree.
   Reference: Peter Graf, Term Indexing, Springer-Verlag, New York, NY, 1995

   Copyright (C) 2004 John Wheeler 
   
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*)

(* This implementation is based on code found in the file
   dtIndex.ml contained in the source code for KRHyper, copyright
   2003, by Universitaet Koblenz-Landau.  The file dtIndex.ml
   included the following copyright notice:

   * Copyright (C) 2003 Universitaet Koblenz-Landau
   * 
   * This file is part of KRHyper.
   * 
   * KRHyper is free software; you can redistribute it and/or modify it under the
   * terms of the GNU General Public License as published by the Free Software
   * Foundation; either version 2 of the License, or (at your option) any later
   * version.
   * 
   * KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
   * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
   * details.
   * 
   * You should have received a copy of the GNU General Public License along with
   * KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
   * Place, Suite 330, Boston, MA 02111-1307 USA
   * 
   * Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(* The current author (John Wheer) has adapted the basic implementation of discrimination tree
   term indexing found in the dtIndex.ml file for use in the darwin program and 
   extended the implementation to handle both universal and paremetric variables.  
   This adaptation and extension retained the basic outlines of implementing functions, 
   but rewrote both the call interfaces to match the darwin term indexing 
   interface (see term_index.ml* for details) and the code that handles the term and
   data items that are stored in the darwin term index.  Approximately 70% 
   of the code in the file dtIndex.ml has been modified or rewritten and approximately 
   300 lines of code of additional code were created to implement extensions to 
   address universal and parametric variables.
*)

(* History:
   the code was originally taken from KRHyper by Christoph Wernhard,
   then integrated into Darwin by John Wheer,
   and then modifyied by Alexander Fuchs:
   transformation into a class, coding style harmonization,
   some performance enhancements, mostly compression of tails,
   implementation of the productivity check
*)

open Bp

let psubst s =
	print_string (Subst.subst_to_string s)	

(*** types ***)

type symbol = Symbol.symbol
type term = Term.term
type var = Var.var
type literal = Term.literal
type clause = Term.clause
type 'a data = 'a Term_indexing.data
type 'a iterator = 'a Term_indexing.iterator
type entry_data = Indexing_misc.entry_data

exception ITERATOR_EMPTY = Term_indexing.ITERATOR_EMPTY
exception SET_FAIL = Subst.SET_FAIL
exception UNIFICATION_FAIL = Unification.UNIFICATION_FAIL
exception FOUND

(* abstraction of term position within the discrimination tree. *)
type hkey =
  | Variable 
  | Parameter
  | Symbol of symbol
  | Var of var		(* for perfect indexing of variables *) 

(* a flattened term, i.a. the term representation within the discrimination tree. *)
type fterm =
    hkey list

(* Debug: used for the substitution computations in perfect discrimination tree indexing;
I'm not sure the recompute-argument of the Subst.set' functions should be true or false
in that type of indexing, so this can be switched here. *)
let global_recompute = false




(*** hkey functions ***)


let hkey_equal ?(distinguish_perfect_vars:bool = true) (k1: hkey) (k2: hkey) : bool =
  match k1, k2 with
    | Variable, Variable ->
	true
    | Parameter, Parameter ->
	true
    | Symbol s1, Symbol s2 ->
	Symbol.equal s1 s2
	| Var v1, Var v2 ->
		if distinguish_perfect_vars then
			(v1 = v2)
		else true
    | _ ->
	false
	
let hkey_equal' (k1: hkey) (k2: hkey) : bool =
	hkey_equal k1 k2 	
	  
let hkey_hash (k1: hkey) : int =
  match k1 with
    | Variable ->
	0
    | Parameter ->
	1
    | Symbol s ->
	2 + (2 * Symbol.id s)
	| Var v ->
	3 + (2 * (Var.hash_of_var v))
(* While OCaml generally allows hash tables with multiple
values assigned to a single key, the functions/methods below
in the original Darwin assume one value per key. The odd hash 
calculation above ensures a unique hash key for each
unique hkey. *)

(* mapping from hkeys *)
module Table =
  Hashtbl.Make (
    struct
      type t = hkey
      let equal = hkey_equal'
      let hash = hkey_hash
    end
  )

  

(* stored entry, term and data *)
type 'data entry = {
  term: term;
  data: 'data;
  entry_data: entry_data;
}
	
	
(* leaf node - containing data *)      
type 'data leaf = {
  (* all entries abstracting to the same fterm *)
  mutable entries: 'data entry list;

  (* abstraction of the term entries,
     used as a prefilter for productivity related tests. *)
  fterm: fterm;
}

(* splitting towards subnodes via their hkey. *)
type 'data inner_node = {
  mutable next: 'data node Table.t;
}

(* encapsulates the prefix towards a subnode *)
and 'data node = {
  (* the common prefix of all subnodes of variant *)
  mutable prefix: fterm;

  (* the subnode *)
  mutable variant: 'data variant;
}

(* a tree node or a leaf *)
and 'data variant =
  | DtInnerNode of 'data inner_node
  | DtLeafNode of 'data leaf

      
(* a discrimination tree *)
type 'data tree = {
  (* the discrimination tree *)
  mutable root: 'data node;
  
  (* its size *)
  mutable size: int;

  (* ugly: this simulates raise (FOUND 'data), see exception FOUND *)
  mutable found: 'data option;
}


(* a check function done on a leaf node, see function create_leaf_check *)
type ('data, 'a) leaf_check = 'data leaf -> 'a list -> 'a list

(* analogous, but for perfect discrimination trees; hence takes two
additional substitution arguments *)
type ('data, 'a) perfect_leaf_check =
	'data leaf -> Subst.subst -> 'a list -> 'a list


(*** functions ***)


(*** representation ***)
let hkey_to_string hkey : string =
  match hkey with
      Symbol sym ->
	(Symbol.name sym) ^ "/" ^ (string_of_int (Symbol.arity sym))
    | Variable ->
	"*U/0"
    | Parameter ->
	"*P/0"
	| Var var -> "*X_" ^ (string_of_int (Var.hash_of_var var))

let rec fterm_to_string x : string =
    match x with
      | [] ->
	  ""

      | h :: t ->
	  hkey_to_string h ^ " " ^  fterm_to_string t


let leaf_to_string (leaf: 'data leaf) (depth: int) : string =
  let depth_prefix =
    String.make (2 * depth) ' ' ^ string_of_int depth ^ ": "
  in
    String.concat ""
      (List.map
	 (fun entry ->
	    depth_prefix ^ Term.term_to_string entry.term ^ "\n"
	 )
	 leaf.entries
      )

let rec node_to_string (node: 'data node) (depth: int) : string = 
  let depth_prefix =
    String.make (2 * depth) ' ' ^ string_of_int depth ^ ": "
  in
    depth_prefix ^ 
    "prefix : " ^ fterm_to_string node.prefix ^ "\n" ^
  match node.variant with
    | DtInnerNode table ->
	Table.fold
	  (fun hkey node acc ->
	     depth_prefix ^ 
	       "hkey : " ^ (hkey_to_string hkey) ^ "\n" ^
	       (node_to_string node (depth + 1)) ^ 
	       acc
	  )
	  table.next
	  ""

    | DtLeafNode leaf ->
	leaf_to_string leaf depth


let tree_to_string (tree: 'data tree) : string =
  node_to_string tree.root 0


	


(*** creation ***)

let create_inner_node () = {
  next = Table.create 8;
}

let create_node () = {
  prefix = [];
  variant = DtInnerNode (create_inner_node ());
}

let create_entry 
	?(entry_data: Indexing_misc.entry_data = Indexing_misc.empty_entry_data)
	(term: term)
	(data: 'data) 
	: 'data entry =
  {
    term = term;
    data = data;
    entry_data = entry_data;
  }

let create_tree () = {
  root = create_node ();
  size = 0;
  found = None;
}


let clear (tree: 'data tree) : unit =
  tree.root <- create_node ();
  tree.size <- 0


(*** data access ***)
let get_size (tree: 'data tree) : int =
  tree.size

let inc_size (tree: 'data tree) : unit =
  tree.size <- tree.size + 1

let dec_size (tree: 'data tree) : unit =
  tree.size <- tree.size - 1



(*** hkey ***)

let hkey_arity (hkey: hkey) : int =
  match hkey with
    | Symbol s ->
	Symbol.arity s
    | _ ->
	0

let term_hkey ?(perfect: bool = false) (term: Term.term ): hkey =
  match term with
    | Term.Var var ->
    	if perfect then
    		(Var var)
    	else
			(if Var.is_universal var then
			  Variable
			else
			  Parameter)
    | Term.Const const ->
	Symbol const
    | Term.Func func ->
	Symbol func.Term.symbol
	| Term.Eq _ ->
	Symbol Symbol.equality






(*** flat terms *)

let flatten_term ?(perfect: bool = false) (_term: term) : fterm =
  let rec flatten_term' (term: term) (acc: fterm) : fterm =
    match term with
      | Term.Var var ->
      	if perfect then
      		(Var var :: acc)
      	else
		  (if Var.is_universal var then
		    Variable :: acc
		  else
		    Parameter :: acc)
	      
	| Term.Const symbol ->
	    Symbol symbol :: acc

	| Term.Func func(*(symbol, terms, _)*) ->
	    (* flatten subterms from right to left *)
	    let new_acc =
	      Array.fold_right
		(fun term acc ->
		   flatten_term' term acc
		)
		func.Term.subterms
		acc
	    in
	      Symbol func.Term.symbol :: new_acc
	| Term.Eq eq ->
		flatten_term' eq.Term.term acc
  in
  let flattened =
    flatten_term' _term []
  in
    match flattened with
      | _propositional_atom :: [] ->
	  (* a propositional predicate, keep it -
	     though this term index is not the adequate data structure... *)
	  flattened

      | _ :: tail ->
	  (* as each index is for one predicate symbol only,
	     we can just drop this (first order) symbol from all terms *)
	  tail

      | [] ->
	  failwith "Discrimination_tree.flatten_term"

(* UNDER CONSTRUCTION - this might be unnecessary
(* works like flatten_term with ~perfect set to true, but in addition normalizes
the flat term (renumbers the variables) and returns both the normalized flat term
and a substitution for undoing the renumbering *)	  
let perfect_flatten_term (_term: term) : (fterm * Subst.subst) =
	let var_cnt = ref 0
	in
  let rec perfect_flatten_term' 
  	(term: term) (acc: fterm) (subst: Subst.subst)
  	: (fterm * Subst.subst) =
    match term with
      | Term.Var var ->
		(let normalized_var, subst', var_cnt_inc =
			(match (Subst.get' subst var 0) with
				| None ->
					(let normalized_var =
						Var.clone_renumbered var !var_cnt
					in
					let subst' =
						Subst.set' ~recompute:false subst var 0 normalized_var 1
					in
						(var_cnt := !var_cnt + 1;
						(normalized_var, subst')))
				| Some s_term ->
					(match sterm.Subst.st_term with
						| Term.Var var' -> 
							(var', subst)
						| _ -> 
							(failwith 
								"Discrimination_tree.perfect_flatten_term: bad normalization.")))
		in
		((Var normalized_var :: acc), subst'))
	      
	| Term.Const symbol ->
	    ((Symbol symbol :: acc), subst)

	| Term.Func func ->
	    (* flatten subterms from right to left *)
	    let new_acc =
	      Array.fold_right
			(fun term acc ->
			   flatten_term' term acc
			)
			func.Term.subterms
			acc
	    in
	      Symbol func.Term.symbol :: new_acc
	| Term.Eq eq ->
		(perfect_flatten_term' eq.Term.term acc subst)
  in
  let flattened, subst =
    (perfect_flatten_term' _term [] 0 Subst.empty)
  in
  	match flattened with
      | _ :: tail ->
	  (* as each index is for one predicate symbol only,
	     we can just drop this (first order) symbol from all terms *)
	  (* DANGER, DANGER! This might not work well with perfect indexing.
	  We return the full fterm for now, the original optimization can
	  be tested later on. Better not introduce too many uncertainties
	  at this stage... *)
	  (flattened, subst)

      | [] ->
	  failwith "Discrimination_tree.perfect_flatten_term"
*)	  
	  
(* prints a flattended term *)					 
let print_fterm (fterm: fterm) : unit =
	print_string (fterm_to_string fterm)	

(* analogous to skip_terms_fterm, but returns the skipped part separately *)	  
let partition_fterm (fterm: fterm) (skip: int) : (fterm * fterm) =
	let rec partition_fterm' (skipped: fterm) (fterm': fterm) (skip': int) : (fterm * fterm) =
		if skip' = 0 then
			(skipped, fterm')
		else
			match fterm' with
				| hkey :: tail -> 
					partition_fterm' (hkey :: skipped) tail (skip' - 1 + (hkey_arity hkey))
				| _ ->
					(failwith "Discrimination_tree.partition_fterm: Bad flattened term structure.")
	in
	let skipped, remainder =
		partition_fterm' [] fterm skip
	in
		((List.rev skipped), remainder)
	
(* analogous to skip_terms_fterm_partial, but returns the skipped part separately *)			
let partition_fterm_partial (fterm: fterm) (skip: int) : (fterm * fterm * int) =		
	let rec partition_fterm_partial' (skipped: fterm) (fterm: fterm) (skip: int) : (fterm * fterm * int) =
		if skip = 0 then
			((List.rev skipped), fterm, 0)
		else
			(match fterm with
				| hkey :: tail ->
					partition_fterm_partial'
						(hkey :: skipped)
						tail
						(skip - 1 + (hkey_arity hkey))
				| [] ->
					((List.rev skipped), [], skip))
	in
		partition_fterm_partial' [] fterm skip
			
			
		
(* the inverse of flatten_term *)	  
(* undefined for non-perfect fterms *)
let unflatten_fterm (fterm: fterm) : term = 
	let make_term (hkey: hkey) (subterms: term list) : term =
		if (List.length subterms) = (hkey_arity hkey) then
			(match hkey with
				| Symbol symbol ->
					(if (List.length subterms) = 0 then
						(Term.request_const symbol)
					else 
						(if symbol == Symbol.equality then
							(let left, right =
								match subterms with
									| left :: right :: [] ->
										(left, right)
									| _ -> (raise SET_FAIL)
							in
								Term.request_eq (left, right))
						else
							(Term.request_func (symbol, (Array.of_list subterms)))))
				| Var var ->
					(Term.request_var var)
				| _ -> 
					(raise SET_FAIL))
		else
			(raise SET_FAIL)
	in
(* assumes fterm to be a list/concatenation of flattened terms; a list of the unflattened
terms is returned (i.e. [f; x; g; a; b] -> [f(x); g(a, b)]*)	
	let rec make_term_list (fterm: fterm) : (term list) =
		match fterm with
			| hkey :: tail ->
				let subterms_fterm, rest_fterm = 
					partition_fterm tail (hkey_arity hkey)
				in 
					let term = make_term hkey (make_term_list subterms_fterm)
					in
						(term :: (make_term_list rest_fterm))
			| _ -> []
	in		
(* the initial call; similar to make_term_list, but here fterm represents
a single term, not a concatenation, so only one unflattened term is returned *)		
	match fterm with	
		| hkey :: tail -> 
			let subterms_fterm, rest_fterm = 
				partition_fterm tail (hkey_arity hkey)
			in 
				if (rest_fterm = []) then
					(make_term hkey (make_term_list subterms_fterm))
				else
					(failwith "Discrimination_tree.unflatten_fterm: bad fterm")
		| _ -> failwith "Discrimination_tree.unflatten_fterm: empy fterm"
				
    
(* skip some terms in fterm. *)
let rec skip_terms_fterm (fterm: fterm) (skip: int) : fterm =
  if skip = 0 then
    fterm

  else
    match fterm with 
      | hkey :: tail ->
	  (* one symbol skipped, but also skip the subterm corresponding to it,
	     e.g. if f is skipped in f(a) then a must be skipped as well. *)
	  skip_terms_fterm tail (skip - 1 +  (hkey_arity hkey))

      | _ ->
	  failwith "Discrimination_tree.skip_terms_fterm: Bad flattened term structure."

let rec skip_terms_fterm_partial (fterm: fterm) (skip: int) : fterm * int =
  if skip = 0 then
    fterm, 0

  else
    match fterm with 
      | hkey :: tail ->
	  (* one symbol skipped, but also skip the subterm corresponding to it,
	     e.g. if f is skipped in f(a) then a must be skipped as well. *)
	  skip_terms_fterm_partial tail (skip - 1 +  (hkey_arity hkey))

      | [] ->
	  [], skip

  
	  
(* flip the sides of an equational fterm *)
let flip_fterm (fterm: fterm) : fterm = 
	match fterm with
		| [] -> fterm
		| hkey :: tail ->
			let rec split_fterm (skipped: fterm) (fterm': fterm) (skip: int) : fterm * fterm =
				if skip = 0 then
					(skipped, fterm')
				else
					(match fterm' with
						| hkey' :: tail' ->
							split_fterm (hkey' :: skipped) tail' (skip - 1 + (hkey_arity hkey'))
						| [] -> (skipped, []))
			in
				let rev_left, right = split_fterm [] fterm 1
				in		
					List.append right (List.rev rev_left)



(*** fterm comparisons based on (abstracted) unification ***)

(* removes prefix from fterm, raises Exit if not a proper prefix *)
let rec shorten_fterm ?(distinguish_perfect_vars: bool = true) (prefix: fterm) (fterm: fterm) : fterm =
  match prefix, fterm with
    | [], _ ->
	fterm

    | xh :: xt, yh :: yt when
	hkey_equal ~distinguish_perfect_vars xh yh ->
	shorten_fterm ~distinguish_perfect_vars xt yt

    | _, _ ->
	raise Exit

(* like above, but also returns a substitution that maps distinguished variables
to each other (if the distinguish_perfect_vars-parameter is set to true; otherwise
the returned substitutions are empty *)
let rec shorten_fterm_subst 
	?(subst:Subst.subst = Subst.empty) 
	?(distinguish_perfect_vars: bool = true)
	(prefix: fterm) 
	(fterm: fterm) 
		: (fterm * Subst.subst) = 
	  match prefix, fterm with
	    | [], _ ->
		fterm, subst
	
		| Variable :: xt, Variable :: yt ->
			shorten_fterm_subst 
				~subst:subst
				~distinguish_perfect_vars:distinguish_perfect_vars
				xt 
				yt
		| Parameter :: xt, Parameter :: yt ->
			shorten_fterm_subst 
				~subst:subst
				~distinguish_perfect_vars:distinguish_perfect_vars
				xt 
				yt
		| Symbol s1 :: xt, Symbol s2 :: yt when
			Symbol.equal s1 s2 ->
			shorten_fterm_subst 
				~subst:subst
				~distinguish_perfect_vars:distinguish_perfect_vars 
				xt 
				yt									
	    | Var v1 :: xt, Var v2 :: yt ->
	    	if distinguish_perfect_vars then
				(try
					(let v1t = Term.request_var v1
					and v2t = Term.request_var v2
					in
					let subst_xy = Subst.set' ~recompute:global_recompute ~force_check:true Subst.empty v1 0 v2t 1
					and subst_yx = Subst.set' ~recompute:global_recompute ~force_check:true Subst.empty v2 1 v1t 0
					in
					let subst' = Unification.merge_substs subst subst_xy
					in
					let subst'' = Unification.merge_substs subst' subst_yx
					in
						shorten_fterm_subst
							~subst:subst''
							~distinguish_perfect_vars:distinguish_perfect_vars
							xt 
							yt)
				with
					| SET_FAIL 
					| Unification.UNIFICATION_FAIL -> raise Exit)
			else
				(shorten_fterm_subst
					~subst:subst
					~distinguish_perfect_vars:distinguish_perfect_vars  
					xt 
					yt)
	
	    | _ ->
		raise Exit
		
		
(* find common prefix *)
let find_prefix (_x: fterm) (_y: fterm) : fterm * fterm * fterm =

  let rec find_prefix' (prefix: fterm) (x: fterm) (y: fterm) : fterm * fterm * fterm =
    match x, y with
      | xh :: xt, yh :: yt when
	  hkey_equal xh yh ->
	  find_prefix' (xh :: prefix) xt yt
	    
      | _ ->
	  List.rev prefix, x, y
  in
    find_prefix' [] _x _y

(* strictly identical - will fail for perfect flat terms which differ only in variable names;
i.e. f(x, y) and f(y, x) are not fterm_equal *)
let rec fterm_equal (x: fterm) (y: fterm) : bool =
  x == y
  ||
  match x, y with
    | [], [] ->
	true

    | xh :: xt, yh :: yt when
	hkey_equal xh yh ->
	fterm_equal xt yt

    | _ ->
	false

(* variant check - intended as fterm-equality for perfect fterms;
will work like fterm_equal for non-perfect flat terms, just slower *)
let fterm_variant (x: fterm) (y: fterm) : bool =
	let rec fterm_variant' (bindings: (int * int) list) (x: fterm) (y: fterm) : bool =
		x == y
		||
		(match x, y with
		    | [], [] ->
				true
		
		    | xh :: xt, yh :: yt ->
		    	(match xh, yh with
		    		| Var v1, Var v2 ->
		    			(let v1 = Var.hash_of_var v1
		    			and v2 = Var.hash_of_var v2
		    			in
		    			try
		    				(let no_binding = ((-1), (-1))
		    				in
		    				let b1 = 
		    					(try
		    						(List.find (fun (v1', _) -> v1 = v1') bindings)
		    					with
		    						| Not_found -> no_binding)
		    				and b2 = 
		    					(try
		    						(List.find (fun (_, v2') -> v2 = v2') bindings)
		    					with
		    						| Not_found -> no_binding)
		    				in
		    					(* v1 and v2 are bound to each other or to nothing *)
		    					if (b1 == b2) then
		    						(* v1 and v2 are bound to nothing; raise exception to
		    						add binding *)
		    						(if (b1 == no_binding) then
		    							(raise Not_found)
		    						(* v1 and v2 are bound to each other; all is well, continue *)
		    						else
		    							(fterm_variant' bindings xt yt))
		    					(* v1 is bound to something other than v2 (or vice versa) *)		
		    					else
		    						(false))
		    			with
		    				| Not_found ->
		    					(fterm_variant' ((v1, v2) :: bindings) xt yt))
		    		| _, _ ->
		    			(if (hkey_equal xh yh) then
		    				(fterm_variant' bindings xt yt)
		    			else
		    				(false)))
		    | _ ->
				false)
	in
		fterm_variant' [] x y

(* x is more general than y,
   i.e. the term abstracted to x can be instantiated to the term abstracted to y *)
   (* only intended for non-perfect flat terms, undefined for perfect ones;
   use perfect_fterm_equal_generalization there instead *)
let rec fterm_equal_generalization ~(p_preserving:bool) (x: fterm) (y: fterm) : bool =
  match x, y with
    | [], [] ->
	true

    | xh :: xt, yh :: yt ->
	(* equal, so compare tails *)
	if hkey_equal xh yh then begin
	  fterm_equal_generalization ~p_preserving:p_preserving xt yt
	end
	  
	(* xh more general, so skip instance in y and compare rest *)
	else if
	  xh = Variable
	  ||
	  (
	    not p_preserving
	    &&
	    xh = Parameter
	  )
	then begin
	  fterm_equal_generalization ~p_preserving:p_preserving xt (skip_terms_fterm y 1)
	end

	else begin
	  false
	end

    | _ ->
	print_endline (fterm_to_string x);
	print_endline (fterm_to_string y);
	failwith "Discrimination_tree.fterm_equal_generalization: Bad flattened term structure."


(* a generalized version of a function that is found in slightly different forms
as subfunctions throughout the module...
Used to assign Subst.vars to integers. A list is used instead of a table,
as tables would be overkill for the small numbers of bindings that usually need to
be kept track of.
However, this generalized function is a bit unwieldy, and so far task-specific, optimized
subfunctions are preferred. *)
(* let get_subst_var (var_subst_var_list: ((int * Subst.var) list)) (v: int) (offset: int) 
	: (Subst.var * ((int * Subst.var) list)) =
	try
		(let _, _, subst_var =
			(List.find
				(fun v' offset' _ ->
					(v' = v) && (offset' = offset))
				var_subst_var_list,
			var_subst_var_list)	
		in
			subst_var)
	with Not_found ->
		let subst_var =
			Subst.make_var offset v
		in
			(subst_var,
			(v, offset, subst_var) :: var_subst_var_list)
*)	
	
			
			
			
(* as above, but intended for perfect flat terms *)
(* The main difference is that this function does not return a boolean. Rather, it will either
succeed and return a substitution that turns x into y, or it will fail with a SET_FAIL exception. *)		
let rec perfect_fterm_equal_generalization_subst (subst:Subst.subst) ~(p_preserving:bool) (x: fterm) (y: fterm) : Subst.subst =
	match x, y with
	| [], [] ->
		subst
	
	| xh :: xt, yh :: yt ->
		(match xh, yh with
			| Var v1, Var v2 -> 
				(let t1 = Subst.get' subst v1 0
				in
					match t1 with
						| None ->
							(Subst.set' ~recompute:global_recompute ~force_check:true subst v1 0 (Term.request_var v2) 1)
						| Some st ->
							(match st.Subst.st_term with
								| Term.Var stv ->
									(if (stv = v2) then
(* Only st_terms are compared, not the offsets, since we only ever bind subterms from y
to variables from x. In a more general substitution the offsets would have to be compared. *)										
										(perfect_fterm_equal_generalization_subst
											subst 
											~p_preserving:p_preserving 
											xt 
											yt)
									else
										(raise SET_FAIL))
								| _ -> raise SET_FAIL))
			| Var v1, _ -> 
				(let y_fterm, y_rest_fterm = partition_fterm yt (hkey_arity yh)
				in 
				let y_term = unflatten_fterm y_fterm
				and t1 = Subst.get' subst v1 0
				in 
					match t1 with
						| None ->
							(Subst.set' ~recompute:global_recompute ~force_check:true subst v1 0 y_term 1)
						| Some st ->
							(if (Term.term_equal st.Subst.st_term y_term) then
								(perfect_fterm_equal_generalization_subst
									subst
									~p_preserving:p_preserving
									xt
									y_rest_fterm)
							else
								(raise SET_FAIL)))
			| _, Var v2 -> 
				(raise SET_FAIL)	(* non-variable cannot be substituted to become variable *)
			| _, _ ->
				(if (hkey_equal xh yh) then
					(perfect_fterm_equal_generalization_subst
						subst 
						~p_preserving:p_preserving 
						xt 
						yt)
				else
					(raise SET_FAIL)))
	| _, _ -> raise SET_FAIL


(*** returning fast when finding _one_ entry ***)

(* an element matching the request has been found in the index.
   unfortunately, just returning it with (FOUND 'data)
   is not possible due to the polymorphic exception.

   therefore, stick it into a tree variable,
   raise the exception, and retrieve it afterwards.
*)
exception FOUND

let set_found (tree: 'data tree) (data: 'data) : unit =
  tree.found <- Some data

let get_found (tree: 'data tree) : 'data =
  match tree.found with
    | None ->
	failwith "Discrimination_tree.get_found"
	  
    | Some data ->
	tree.found <- None;
	data

	  




(*** add an entry to a tree ***)


(* a new entry shares as much of its prefix with existing entries
   as possible.

   e.g. if previously the tree contained
   f(a, b) and f(b, a)
   it contained two leafs with the prefix [f] and the tails [a, b] resp. [b, a].
   now if f(a, a) is added,
   a new inner node with the prefix [f, a] is created,
   f(a, b) is stored below it in a leaf with tail [b],
   and f(a, a) is stored below it in a leaf with tail [b].
*)

(* no_duplicates: see term_indexing
   don't add the same term twice (even if it carries different data). *)
let rec add_term ~(no_duplicates: bool) (_data: 'data data)
    (entry: 'data entry) (new_term: fterm) (fterm: fterm) (node: 'data inner_node) : bool =

  match fterm with
    | [] ->
	failwith "discrimination_tree:add_term: Bad index structure 1."
	  
    | hkey :: tail ->
	begin
	  try 
	    let next =
	      Table.find node.next hkey
	    in
	      begin
		(* check if node compression matches remaining fterm *)
		try
		  let tail =
		    shorten_fterm next.prefix tail
		  in
		    (* ok, compatible prefix, continue *)
		    match next.variant with
		      | DtLeafNode entries ->
			  (* already at the leaf node *)
			  if Const.debug && tail != [] then
			    failwith "add_term";

			  if no_duplicates then begin
			    (* entry does already exist *)
			    if
			      List.exists
				(fun entry' ->
				   Term.term_equal entry.term entry'.term
				)
				entries.entries
			    then
			      false

			    (* first entry under this term *)
			    else begin
			      entries.entries <- entry :: entries.entries;
			      true
			    end
			  end
			    
			  else begin
			    entries.entries <- entry :: entries.entries;
			    true
			  end
			    
		      | DtInnerNode table1 ->
			  (* continue traversal *)
			  add_term ~no_duplicates:no_duplicates _data entry new_term tail table1

		with
		  | Exit ->
		      (* incompatability in the remainder of the path to this node.
			 so, split into a common node with two subnodes *)
		      let common_prefix, prefix', fterm' =
			find_prefix next.prefix tail
		      in

		      (* create a new node with the content of the current node,
			 but a shortened prefix *)
		      let new_next = {
			prefix = List.tl prefix';
			variant = next.variant;
		      }
		      in

		      (* create a new leaf node for the new entry *)
		      let new_leaf_node = {
			prefix = List.tl fterm';
			variant = DtLeafNode {
			  entries = [ entry ];
			  fterm = new_term;
			}
		      }
		      in

		      (* transform the previous node to a common prefix node *)
		      let common_inner_node =
			create_inner_node ()
		      in
			Table.add common_inner_node.next (List.hd prefix') new_next;
			Table.add common_inner_node.next (List.hd fterm') new_leaf_node;

			next.prefix <- common_prefix;
			next.variant <- DtInnerNode common_inner_node;
			true
	      end
	  with
	    | Not_found ->
		(* hkey, does not exist, so insert a new leaf node *)
		let new_leaf = {
		  prefix = tail;
		  variant = DtLeafNode {
		    entries = [ entry ];
		    fterm = new_term;
		  }
		}
		in
		  Table.add node.next hkey new_leaf;
		  true
	end




(*** remove ***)


(* remove entry from a leaf.
   raise Not_found if the leaf does not contain a corresponding entry. *)
let remove_leaf ~(equal: bool)
    (* the entry to remove *)
    (term: term) (data: 'data option) (_data: 'data data)
    (leaf: 'data leaf) : unit =

  (* returns the list of remaining entries, without the removed one *)
  let rec list_remove_variant' (entries: 'data entry list) : 'data entry list =
    match entries with
      | [] ->
	  (* no suitable entry found *)
	  raise Not_found
	    
      | entry :: tail ->
	  if
	    (equal && Term.term_equal entry.term term)
	    ||
	    (not equal && Term.are_terms_variants entry.term term)
	  then begin
	    match data with
	      | Some termdata when 
		  not (_data#is_equal entry.data termdata) ->
		  (* continue search ... *)
		  entry :: list_remove_variant' tail
		    
	      | _ ->
		    (* success, remove entry *)
		  tail
	  end
	    
	  else
	    (* continue search ... *)
	    entry :: list_remove_variant' tail
  in
    (* remove a matching entry *)
    leaf.entries <- list_remove_variant' leaf.entries



(* raises Not_found if no matching entry is found.
   returns true if all entries of the subnode have been removed. *)
let rec remove_node ~(equal: bool) (term: term) (fterm: fterm) (data: 'data option) (_data: 'data data) (node: 'data node) : bool =
  try
    let fterm =
      shorten_fterm node.prefix fterm
    in
      match node.variant with
	| DtInnerNode table ->
	    (* node, so continue traversal *)
	    begin
	      match fterm with
		| [] ->
		    raise (Failure "Discrimination_tree.remove_node: Bad index structure.")
		      
		| hkey :: fterm1 ->
		    begin
		      try
			let empty =
			  remove_node ~equal:equal term fterm1 data _data (Table.find table.next hkey)
			in
			  (* the hkey node is empty now, so remove it *) 
			  if empty then begin
			    Table.remove table.next hkey;
			    (* is this node now empty as well? *)
			    Table.length table.next == 0;
			  end
			    
			  else begin
			    false
			  end
		      with
			| Not_found ->
			    raise Not_found
		    end
	    end
	      
	| DtLeafNode leaf ->
	    if fterm != [] then
	      failwith "remove_node";

	    (* leaf, so try to remove *)
	    remove_leaf ~equal:equal term data _data leaf;
	    (* no entries left? *)
	    (leaf.entries = [])

 with
    | Exit ->
	(* prefix does not match *)
	raise Not_found









let remove_variant ?(perfect:bool = false) ~(equal:bool) (tree: 'data tree) (term: term) (data: 'data option) (_data: 'data data) : bool =
  let fterm =
    flatten_term ~perfect term
  in 
    try
      ignore (remove_node ~equal:equal term fterm data _data tree.root : bool);
      dec_size tree;
      (match term with
      	| Term.Eq eq ->
      		(let flipped_term = Term.request_eq ~not_orient:true (eq.Term.right, eq.Term.left)
	      	in
	      	let flipped_fterm = flatten_term flipped_term
      		in 
      			ignore (remove_node ~equal:equal flipped_term flipped_fterm data _data tree.root : bool);
      			dec_size tree)
      	| _ -> ());
      true
    with
      | Not_found ->
	  false

	  
	  
	  
	  
	  
(* analogous to remove_leaf, but removes clauses. *)
let remove_clause_leaf ~(equal: bool)
    (clause: clause) (leaf: 'data leaf) : unit =

  (* returns the list of remaining entries, without the removed one *)
  let rec list_remove_variant' (entries: 'data entry list) : 'data entry list =
    match entries with
      | [] ->
	  (* no suitable entry found *)
	  	(raise Not_found)
	    
      | entry :: tail ->
	  if
	    (equal && (clause == (Indexing_misc.get_clause entry.entry_data)))
	    ||
	    (not equal && (false))	(* currently only tests object identity; replace false with a clause variant check *)
	  then 
	    (tail)
	  else
	    (* continue search ... *)
	    (entry :: list_remove_variant' tail)
  in
    (* remove a matching entry *)
    leaf.entries <- list_remove_variant' leaf.entries
	  
    
(* analogous to remove_clause_leaf, but returns the data of the removed leaf. *)
let extract_clause_leaf ~(equal: bool) 
    (clause: clause) (leaf: 'data leaf) : 'data option =
  let extracted_data = ref None
  in
  (* returns the list of remaining entries, without the removed one *)
  let rec list_remove_variant' (entries: 'data entry list) : 'data entry list =
    match entries with
      | [] ->
	  (* no suitable entry found *)
	  	(raise Not_found)
	    
      | entry :: tail ->
	  if
	    (equal && (clause == (Indexing_misc.get_clause entry.entry_data)))
	    ||
	    (not equal && (false))	(* currently only tests object identity; replace false with a clause variant check *)
	  then 
	    (extracted_data := Some entry.data;
	    tail)
	  else
	    (* continue search ... *)
	    (entry :: list_remove_variant' tail)
  in
    (* remove a matching entry *)
    leaf.entries <- list_remove_variant' leaf.entries;
    !extracted_data	  
    

let rec remove_clause_node ~(equal:bool) (fterm: fterm) (clause: clause) (node: 'data node) : bool =	  
	try
	    (let fterm =
	     	shorten_fterm node.prefix fterm
	    in
	      	match node.variant with
				| DtInnerNode table ->
			    (* node, so continue traversal *)
			    	(match fterm with
						| [] ->
							raise (Failure "Discrimination_tree.remove_clause_node: Bad index structure.")
						| hkey :: fterm1 ->
							(try
								(let empty =
									remove_clause_node ~equal:equal fterm1 clause (Table.find table.next hkey)
								in
								(* the hkey node is empty now, so remove it *) 
								if empty then
									(Table.remove table.next hkey;
								(* is this node now empty as well? *)
									Table.length table.next == 0)
								else
									(false))
							with
							| Not_found ->
							raise Not_found))
				| DtLeafNode leaf ->
	    			if fterm != [] then
	      				(failwith "Discrimination_tree.remove_clause_node: fterm too long.");
				(* leaf, so try to remove *)
	    			remove_clause_leaf ~equal:equal clause leaf;
	    		(* no entries left? *)
	    			(leaf.entries = []))
	with
		| Exit ->
			(raise Not_found)

			
let rec extract_clause_node ~(equal:bool) (fterm: fterm) (clause: clause) (node: 'data node) : bool * ('data option) =	  
	try
	    (let fterm =
	     	shorten_fterm node.prefix fterm
	    in
	      	match node.variant with
				| DtInnerNode table ->
			    (* node, so continue traversal *)
			    	(match fterm with
						| [] ->
							raise (Failure "Discrimination_tree.remove_clause_node: Bad index structure.")
						| hkey :: fterm1 ->
							(try
								(let empty, data =
									extract_clause_node ~equal:equal fterm1 clause (Table.find table.next hkey)
								in
								(* the hkey node is empty now, so remove it *) 
								if empty then
									(Table.remove table.next hkey;
								(* is this node now empty as well? *)
									Table.length table.next == 0, data)
								else
									(false, data))
							with
							| Not_found ->
							raise Not_found))
				| DtLeafNode leaf ->
	    			if fterm != [] then
	      				(failwith "Discrimination_tree.remove_clause_node: fterm too long.");
				(* leaf, so try to remove *)
	    			let data = extract_clause_leaf ~equal:equal clause leaf
	    			in
	    		(* no entries left? *)
	    			(leaf.entries = []), data)
	with
		| Exit ->
			(raise Not_found)
			
			
				  
let remove_clause ?(perfect:bool = false) ~(equal:bool) (tree: 'data tree) (term: term) (clause: clause) : bool =
	let fterm =
		flatten_term ~perfect term
	in
		try
			(ignore
				(remove_clause_node ~equal:equal fterm clause tree.root);
			dec_size tree;
			true)
		with
			| Not_found ->
				(false)

let extract_clause  ?(perfect:bool = false) ~(equal:bool) (tree: 'data tree) (term: term) (clause: clause) : bool * ('data option) =
	let fterm =
		flatten_term ~perfect term
	in
		try
			(let _, data =
				(extract_clause_node ~equal:equal fterm clause tree.root)
			in
			dec_size tree;
			true, data)
		with
			| Not_found ->
				(false, None)



(*** tree traversal ***)


let rec iter (func: term -> 'data -> unit) (node: 'data node) : unit =
  match node.variant with
    | DtInnerNode table ->
	Table.iter 
	  (fun _ node' ->
	     iter func node')
	  table.next

    | DtLeafNode entries ->
	List.iter
	  (fun entry -> func entry.term entry.data )
	  entries.entries
	

let rec fold (func: 'a -> term -> 'data -> 'a) (acc: 'a) (node: 'data node) : 'a =
  match node.variant with
    | DtInnerNode table ->
	Table.fold
	  (fun _ node acc ->
	     fold func acc node
	  )
	  table.next
	  acc

    | DtLeafNode entries ->
	List.fold_left
	  (fun acc entry ->
	     func acc entry.term entry.data
	  )
	  acc
	  entries.entries
	




(*** leaf check predicates for retrieval functions ***)

(* check if a leaf node contains searched for entries *)
let create_leaf_check
    (* find all or only the first suitable entry? *)
    ?(find_all: bool = true)
    (* search happens within this index *)
    (tree: 'data tree)
    (* check on the leaf fterm *)
    (check_fterm: fterm -> bool)
    (* check on each entry's term *)
    (check_term: term -> bool)

    (* the leaf to check *)
    (leaf: 'data leaf)
    (* the currently found results *)
    (acc: 'data list)
    : 'data list =

  if not (check_fterm leaf.fterm) then begin
    acc
  end
    
  (* add all suitable entries to the results *)
  else begin	
    List.fold_left
      (fun acc entry ->	
	 if check_term entry.term then begin	
	   if find_all then
	     entry.data :: acc
	   else begin
	     set_found tree entry.data;
	     raise FOUND;
	   end
	 end
	 else
	   acc
      )
      acc
      leaf.entries
  end

  
(*** leaf check predicates for retrieval functions ***)

(* as above, but returns full entries, not just 'data *)
let create_leaf_check_entry_data
    (* find all or only the first suitable entry? *)
    ?(find_all: bool = true)
    (* search happens within this index *)
    (tree: 'data tree)
    (* check on the leaf fterm *)
    (check_fterm: fterm -> bool)
    (* check on each entry's term *)
    (check_term: term -> bool)

    (* the leaf to check *)
    (leaf: 'data leaf)
    (* the currently found results *)
    (acc: ('data * entry_data) list)
    : ('data * entry_data) list =

  if not (check_fterm leaf.fterm) then begin
    acc
  end
    
  (* add all suitable entries to the results *)
  else begin	
    List.fold_left
      (fun acc entry ->
	 if check_term entry.term then begin
	   if find_all then
	     (entry.data, entry.entry_data) :: acc
	   else begin
	     set_found tree entry.data;
	     raise FOUND;
	   end
	 end
	 else
	   acc
      )
      acc
      leaf.entries
  end
  
  
  
  
  
let create_leaf_check_subst
    (* find all or only the first suitable entry? *)
(*    ?(find_all: bool = true)*)
    (* search happens within this index *)
    (_tree: 'data tree)
    (* check on the leaf fterm *)
    (check_fterm: fterm -> bool)
    (* check on each entry's term *)
    (check_term: term -> Subst.subst)

    (* the leaf to check *)
    (leaf: 'data leaf)
    (* the currently found results *)
    (acc: ('data * Subst.subst) list)
    : ('data * Subst.subst) list =

  if not (check_fterm leaf.fterm) then begin
    acc
  end
    
  (* add all suitable entries to the results *)
  else begin	
    List.fold_left
      (fun acc entry ->
	 try
	   let subst =
	     check_term entry.term
	   in
(*	     if find_all then*)
	       (entry.data, subst) :: acc
(*	     else begin
	       failwith "Discrimation_tree.create_leaf_check_subst";
	       (*
	       set_found tree entry.data;
	       raise FOUND;
	       *)
	     end*)
	 with
	   | Unification.UNIFICATION_FAIL  ->
	       acc
      )
      acc
      leaf.entries
  end

  

  
  
  
let create_leaf_check_entry_data_subst
    (* find all or only the first suitable entry? *)
    ?(find_all: bool = true)
    (* search happens within this index *)
    (_tree: 'data tree)
    (* check on the leaf fterm *)
    (check_fterm: fterm -> bool)
    (* check on each entry's term *)
    (check_term: term -> Subst.subst)

    (* the leaf to check *)
    (leaf: 'data leaf)
    (* the currently found results *)
    (acc: ('data * entry_data * Subst.subst) list)
    : ('data * entry_data * Subst.subst) list =

  if not (check_fterm leaf.fterm) then begin
    acc
  end
    																	
  (* add all suitable entries to the results *)
  else begin	
    List.fold_left
      (fun acc entry ->
	 try
	   let subst =
	     check_term entry.term
	   in
	     if find_all then
	       (entry.data, entry.entry_data, subst) :: acc
	     else begin
	       set_found _tree entry.data;
	       raise FOUND;
	     end
	 with
	   | Unification.UNIFICATION_FAIL  ->
	       acc
      )
      acc
      leaf.entries
  end

  

  
  
  
  
  
  
  
  
  
let create_perfect_leaf_check
    (* search happens within this index *)
    (_tree: 'data tree)
    (* check on the leaf fterm *)	(* likely some trivial (fun _ -> true) *)
    (check_fterm: fterm -> bool)
    (* check on each entry's term *)
(*    (check_term: term -> Subst.subst)	*)
	(* whether redundant entries should be considered at all *)
	(dismiss_redundant_entries: bool)
    (* the leaf to check *)
    (leaf: 'data leaf)
	(* the substitution/unifier *)
	(subst: Subst.subst)
    (* the currently found results *)
    (acc: ('data * entry_data * Subst.subst) list)
    : ('data * entry_data * Subst.subst) list =

  if not (check_fterm leaf.fterm) then
    (acc)
    																	
  (* add all suitable entries to the results *)
  else 	
		(let acc =
			if dismiss_redundant_entries then
			  	(List.fold_left
					(fun acc entry ->
						if (Indexing_misc.is_redundant entry.entry_data) then
							(acc)
						else 
							(entry.data, entry.entry_data, subst) :: acc)
					acc
					leaf.entries)			
			else
			  	(List.fold_left
					(fun acc entry -> 
						(entry.data, entry.entry_data, subst) :: acc)
					acc
					leaf.entries)
		in	
			acc)
 
  
  
  
(*** tree traversal based on fterms ***)

(* descend into the variant of the given hkey.
   tail is the rest of the fterm towards the potential leaf of the query tree. *)
let descend_variant ~(p_preserving: bool)
    (func: p_preserving:bool -> ('data, 'a) leaf_check -> 'a list -> 'fterm -> 'data node -> fterm -> 'a list)
    (leaf_check: ('data, 'a) leaf_check)
    (acc: 'a list) 
    (hkey: hkey) (tail: fterm) (table: 'data inner_node)
    : 'a list
    = 		
  let node =
    try
      Some (Table.find table.next hkey)
    with
      | Not_found ->
	  (None)
  in	
    match node with
      | None ->
	  (acc)
	    
      | Some node' ->
	  (func ~p_preserving:p_preserving leaf_check acc tail node' node'.prefix)

(* descend into the generalization of the given hkey.
   tail is the rest of the fterm towards the potential leaf of the query tree,
   where fterm = hkey :: tail. *)
let descend_generalized ~(p_preserving: bool)
    (func: p_preserving:bool -> ('data, 'a) leaf_check -> 'a list -> 'fterm -> 'data node -> fterm -> 'a list)
    (leaf_check: ('data, 'a) leaf_check)
    (acc: 'a list) 
    (hkey: hkey) (fterm: fterm) (table: 'data inner_node)
    : 'a list
    =		
  let node =
    try
      Some (Table.find table.next hkey)
    with
      | Not_found ->
	  None
  in
    match node with
      | None ->
	  acc
	    
      | Some node' ->
	  func ~p_preserving:p_preserving leaf_check acc (skip_terms_fterm fterm 1) node' node'.prefix
	  

(* descend into an instance of the given hkey.
   tail is the rest of the fterm towards the potential leaf of the query tree. *)
let rec descend_instance ?(perfect:bool = false) ~(p_preserving:bool)
    (func: p_preserving:bool -> ('data, 'a) leaf_check -> 'a list -> 'fterm -> 'data node -> fterm -> 'a list)
    (leaf_check: ('data, 'a) leaf_check)
    (acc: 'a list) 
    (query_tail: fterm) (table: 'data inner_node) (skip: int)
    : 'a list
    =
  Table.fold
    (fun hkey node acc ->
       let skip' =
	 skip - 1 + (hkey_arity hkey)
       in
	 
       (* now skip also the prefix of this node *)
       let prefix, skip' =
	 skip_terms_fterm_partial node.prefix skip'
       in
		 (if skip' = 0 then
		   func ~p_preserving:p_preserving leaf_check acc query_tail node prefix
	
		 else
		   match node.variant with
		     | DtLeafNode _ ->
			 failwith "Discrimination_tree.descend_instance"
			 
		     | DtInnerNode table' ->
			 descend_instance ~p_preserving:p_preserving func leaf_check acc query_tail table' skip')

    )
    table.next
    acc

(*** perfect tree traversal based on perfect fterms ***)

(* descend into the variant of the given hkey.
   tail is the rest of the fterm towards the potential leaf of the query tree. *)
let perfect_descend_variant (subst:Subst.subst) (unify:bool) ~(p_preserving: bool)
    (func: Subst.subst -> p_preserving:bool -> ('data, 'a) perfect_leaf_check -> 'a list -> 'fterm -> 'data node -> fterm -> 'a list)
    (leaf_check: ('data, 'a) perfect_leaf_check)
    (acc: 'a list) 
    (hkey: hkey) (tail: fterm) (table: 'data inner_node)
    : 'a list
    =
    match hkey with
    	| Var v1 ->	(* this never happens, right? wrong, perfect_find_generalized *)
    		(Table.fold
    			(fun hkey' node' acc' ->
    				match hkey' with
    					| Var v2 -> 
    						(try
    							(let subst_yx = 
    								Subst.set' 
    									~recompute:global_recompute 
    									Subst.empty v2 1 (Term.request_var v1) 0
    							in	
    							let subst' =
    								if unify then
    									(Unification.unify_substs
    										~recompute:global_recompute
    										~p_preserving:p_preserving
    										subst subst_yx)
    								else
	    								(Unification.merge_substs
	    									subst subst_yx)
    							in	
    								List.rev_append
	    								(func 
	    									subst'
	    									~p_preserving:p_preserving 
	    									leaf_check 
	    									acc tail 
	    									node' node'.prefix)
	    								acc')
    						with
    							| SET_FAIL | UNIFICATION_FAIL -> acc')
    					| _ ->
    						acc')
    			table.next
    			acc)
    	| _ ->
		  (let node =
		    try
		      Some (Table.find table.next hkey)
		    with
		      | Not_found ->
			  (None)
		  in	
		    match node with
		      | None ->
			  (acc)
			    
		      | Some node' ->
			  (func subst ~p_preserving:p_preserving leaf_check acc tail node' node'.prefix))

(* descend into the generalization of the given hkey.
   tail is the rest of the fterm towards the potential leaf of the query tree,
   where fterm = hkey :: tail. *)
let perfect_descend_generalized (subst:Subst.subst) (unify:bool) ~(p_preserving: bool)
    (func: Subst.subst -> p_preserving:bool -> ('data, 'a) perfect_leaf_check -> 'a list -> 'fterm -> 'data node -> fterm -> 'a list)
    (leaf_check: ('data, 'a) perfect_leaf_check)
    (acc: 'a list) 
    (hkey: hkey) (fterm: fterm) (table: 'data inner_node)
    : 'a list
    =		
  let node =	
    try
      Some (Table.find table.next hkey)
    with
      | Not_found ->
	  None
  in
    match node with
      | None ->
	  acc
	    
      | Some node' ->
		  (match hkey with
		  	| Var var ->
			  (let skipped, rest_fterm =
			  	partition_fterm fterm 1
			  in	
			  let bound_term = 
			  	unflatten_fterm skipped
			  in	
			  try	
			  	(let subst_xy =
			  		Subst.set'
			  			~recompute:global_recompute
			  			Subst.empty
			  			var 1 bound_term 0
			  	in		
			  	let subst' =
			  		if unify then
				  		(Unification.unify_substs
				  			~recompute:global_recompute
				  			~p_preserving:p_preserving
				  			subst subst_xy)
				  	else
				  		(Unification.merge_substs
				  			subst subst_xy)
			  	in	
					func 
						subst'
						~p_preserving:p_preserving 
						leaf_check 
						acc 
						rest_fterm 
						node' 
						node'.prefix)
			  with
			  	| SET_FAIL | UNIFICATION_FAIL -> acc)
		  	| _ ->
		  		failwith "Discrimination_tree.perfect_descend_generalized: non-variable node")
	  

(* descend into an instance of the given hkey.
   tail is the rest of the fterm towards the potential leaf of the query tree. *)
let rec perfect_descend_instance (subst:Subst.subst) (unify:bool)
	?(perfect:bool = false) ?(var: Var.var=(Var.create_universal 0)) ?(skipped: fterm=[]) ~(p_preserving:bool)
    (func: Subst.subst -> p_preserving:bool -> ('data, 'a) perfect_leaf_check -> 'a list -> 'fterm -> 'data node -> fterm -> 'a list)
    (leaf_check: ('data, 'a) perfect_leaf_check)
    (acc: 'a list) 
    (query_tail: fterm) (table: 'data inner_node) (skip: int)
    : 'a list
    =	
  Table.fold
    (fun hkey node acc ->	
       let skip' =
	 		skip - 1 + (hkey_arity hkey)
       in

       (* now skip also the prefix of this node *)
	 	let skipped', prefix', skip' =
	 		partition_fterm_partial node.prefix skip'
       in	(* print_fterm skipped;pe " <SKIPPED";print_fterm [hkey];pe " <HKEY";print_fterm skipped';pe " <SKIPPED'";*)
		 (if skip' = 0 then
		 	(try
			   (let bound_term = 
			   		unflatten_fterm
			   			(List.rev_append skipped (hkey :: skipped')) 
			   	in	
			   	let subst_xy = 
			   		Subst.set' 
			   			~recompute:global_recompute 
			   			Subst.empty var 0 bound_term 1
			   	in	
			   	let subst' =
			   		if unify then
			   			(Unification.unify_substs
			   				~recompute:global_recompute
			   				~p_preserving:p_preserving
			   				subst subst_xy)
			   		else
				   		(Unification.merge_substs
				   			subst subst_xy)
			   	in 
			   		func subst' ~p_preserving:p_preserving leaf_check acc query_tail node prefix')
			with
				| SET_FAIL | UNIFICATION_FAIL -> acc)
		 else
		   (match node.variant with
		     | DtLeafNode _ ->
			 failwith "Discrimination_tree.perfect_descend_instance"
			 
		     | DtInnerNode table' ->
				 (perfect_descend_instance 
				 	subst
				 	unify
				 	~var:var
				 	~skipped: (List.rev_append skipped' (hkey :: skipped)) 
				 	~p_preserving:p_preserving 
				 	func 
				 	leaf_check 
				 	acc 
				 	query_tail 
				 	table' 
				 	skip')))
    )
    table.next
    acc


(*** variant retrieval ***)


let rec find_variants (leaf_check: ('data, 'data) leaf_check)
    (fterm: fterm) (node: 'data node) : 'data list =

  try
    let fterm =
      shorten_fterm node.prefix fterm
    in
      begin
	match fterm, node.variant with
	  | hkey :: fterm', DtInnerNode table ->
	      begin
		try
		  let node' =
		    Table.find table.next hkey
		  in
		    find_variants leaf_check fterm' node'
		with
		  | Not_found ->
		      []
	      end
		
	  | [], DtLeafNode leaf ->
	      leaf_check leaf []
		
	  | _ ->
	      failwith "Discrimination_tree.find_variants"
      end
  with
    | Exit ->
	[]



(*** generalization retrieval ***)

let rec find_generalizations ~(p_preserving: bool)
    (leaf_check: ('data, 'data) leaf_check) (acc: 'data list)
    (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : 'data list =

  (* performance: order so that elements with most variation come first *)
  match query_fterm, tree_fterm, node.variant with
      (* equal keys *)
    | query_key :: query_tail, tree_key :: tree_tail, _ when
	hkey_equal query_key tree_key ->
	(* variant *)
	find_generalizations ~p_preserving:p_preserving
	  leaf_check acc
	  query_tail node tree_tail

    (* query_fterm empty, so check the node itself *)
    | [], [], DtLeafNode leaf ->
	leaf_check leaf acc

    | Symbol _ as hkey :: query_tail, [], DtInnerNode table ->
	let acc =
	  descend_variant ~p_preserving:p_preserving
	    find_generalizations leaf_check acc
	    hkey query_tail table		    
	in
	let acc =
	  descend_generalized ~p_preserving:p_preserving
	    find_generalizations leaf_check acc
	    Variable query_fterm table
	in
	  if p_preserving then
	    acc
	  else
	    descend_generalized ~p_preserving:p_preserving 
	      find_generalizations leaf_check acc
	      Parameter query_fterm table
	      
    | Variable :: query_tail, [], DtInnerNode table ->
	let acc =
	  descend_variant ~p_preserving:p_preserving 
	    find_generalizations leaf_check acc
	    Variable query_tail table
	in
	  if p_preserving then
	    acc
	  else
	    descend_variant ~p_preserving:p_preserving
	      find_generalizations leaf_check acc
	      Parameter query_tail table
	      
    | Parameter :: query_tail, [], DtInnerNode table ->
	let acc =
	  descend_variant ~p_preserving:p_preserving
	    find_generalizations leaf_check acc
	    Variable query_tail table
	in
	  descend_variant ~p_preserving:p_preserving
	    find_generalizations leaf_check acc
	    Parameter query_tail table
	    
    (* symbol or parameter <-> variable *)
    | _ :: _, Variable :: tree_tail, _ ->
 	(* generalization *)
	find_generalizations ~p_preserving:p_preserving
	  leaf_check acc
	  (skip_terms_fterm query_fterm 1) node tree_tail
    
    (* symbol or variable <-> parameter *)
    | _ :: _, Parameter :: tree_tail, _ when not p_preserving ->
	find_generalizations ~p_preserving:p_preserving
	  leaf_check acc
	  (skip_terms_fterm query_fterm 1) node tree_tail

    (* mismatch *)
    | _ :: _, _ :: _, _ ->
	acc

    | [], _ :: _, _ ->
	failwith "Discrimination_tree.find_generalizations 1"

    | _ :: _, [], DtLeafNode _ ->
	failwith "Discrimination_tree.find_generalizations 2"

    | [], [], DtInnerNode _ ->
	failwith "Discrimination_tree.find_generalizations 3"
	
    | _::_, [], DtInnerNode _ ->
	failwith "Discrimination_tree.find_generalizations 4"	

	




(*** unifiable retrieval ***)

let rec find_unifiables ~(p_preserving: bool)
    (leaf_check: ('data, 'data) leaf_check) (acc: 'data list)
    (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : 'data list =	
  (* first handle the tree_term *)	
  									
  match query_fterm, tree_fterm, node.variant  with
      (* equal keys *)
    | query_key :: query_tail, tree_key :: tree_tail, _ when
	hkey_equal query_key tree_key ->							
	(* variant *)
	find_unifiables ~p_preserving:p_preserving
	  leaf_check acc
	  query_tail node tree_tail


    (* tree_fterm empty, so check the node itself *)
    | [], [] , DtLeafNode leaf ->									
	leaf_check leaf acc
		  
    | Symbol _ as hkey :: query_tail, [], DtInnerNode table ->							

    let acc =
	  descend_variant ~p_preserving:p_preserving
	    find_unifiables leaf_check acc
	    hkey query_tail table
	in				
	let acc =
	  descend_generalized ~p_preserving:p_preserving
	    find_unifiables leaf_check acc
	    Variable query_fterm table
	in			
	  if p_preserving then
	    (acc)
	  else
	    (descend_generalized ~p_preserving:p_preserving
	      find_unifiables leaf_check acc
	      Parameter query_fterm table)
		
    | Variable :: query_tail, [], DtInnerNode table ->							
	descend_instance ~p_preserving:p_preserving
	  find_unifiables leaf_check acc
	  query_tail table 1
	  
    | Parameter :: query_tail, [], DtInnerNode table ->							
	if p_preserving then begin
	  let acc =
	    descend_variant ~p_preserving:p_preserving
	      find_unifiables leaf_check acc
	      Variable query_tail table
	  in
	    descend_variant ~p_preserving:p_preserving
	      find_unifiables leaf_check acc
	      Parameter query_tail table
	end
	  
	else begin
	  descend_instance ~p_preserving:p_preserving
	    find_unifiables leaf_check acc
	    query_tail table 1
	end
	  
    (* variable <-> symbol or parameter *)
    | Variable :: query_tail, _ :: _, _ ->							
	let tree_fterm', skip =
	  skip_terms_fterm_partial tree_fterm 1
	in
	  if skip = 0 then
	    find_unifiables ~p_preserving:p_preserving
	      leaf_check acc
	      query_tail node tree_fterm'
	  else
	    begin
	      match node.variant with
		| DtInnerNode inner_node ->
		    descend_instance ~p_preserving:p_preserving
		      find_unifiables leaf_check acc
		      query_tail inner_node skip

		| DtLeafNode _ ->
		    failwith "find_unifiables 3"
	    end

    (* symbol or parameter <-> variable *)
    | _ :: _, Variable :: tree_tail, _ ->							
	(* generalization *)
	find_unifiables ~p_preserving:p_preserving
	  leaf_check acc
	  (skip_terms_fterm query_fterm 1) node tree_tail

    (* parameter <-> symbol *)
    | Parameter :: query_tail, _ :: _, _ ->							
	(* generalization *)
	if p_preserving then
	  acc
	else
	  let tree_fterm', skip =
	    skip_terms_fterm_partial tree_fterm 1
	  in
	    if skip = 0 then
	      find_unifiables ~p_preserving:p_preserving
		leaf_check acc
		query_tail node tree_fterm'
	    else
	    begin
	      match node.variant with
		| DtInnerNode inner_node ->
		    descend_instance ~p_preserving:p_preserving
		      find_unifiables leaf_check acc
		      query_tail inner_node skip

		| DtLeafNode _ ->
		    failwith "find_unifiables 5"
	    end

    (* symbol <-> parameter *)
    | _ :: _, Parameter :: tree_tail, _ ->							
	(* generalization *)
	if p_preserving then
	  acc
	else
	  find_unifiables ~p_preserving:p_preserving
	    leaf_check acc
	    (skip_terms_fterm query_fterm 1) node tree_tail
    
    (* mismatch *)
    | _ :: _, _ :: _, _ ->							
	acc

    | [], _ :: _, _ ->							
	failwith "Discrimination_tree.find_unifiables 1"

    | [], [], DtInnerNode _ ->					
	failwith "Discrimination_tree.find_unifiables"
	  
    | _ :: _, [], DtLeafNode _ ->							
	failwith "Discrimination_tree.find_unifiables 2"

    | _ :: _, [], DtInnerNode _ ->							
	failwith "Discrimination_tree.find_unifiables 3"	
	
	
	
	
let rec find_unifiables_entry_data ~(p_preserving: bool)
    (leaf_check: ('data, ('data * entry_data)) leaf_check) (acc: ('data * entry_data) list)
    (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : ('data * entry_data) list =	
  (* first handle the tree_term *)
  match query_fterm, tree_fterm, node.variant  with
      (* equal keys *)
    | query_key :: query_tail, tree_key :: tree_tail, _ when
	hkey_equal query_key tree_key ->							
	(* variant *)
	find_unifiables_entry_data ~p_preserving:p_preserving
	  leaf_check acc
	  query_tail node tree_tail


    (* tree_fterm empty, so check the node itself *)
    | [], [] , DtLeafNode leaf ->									
	leaf_check leaf acc
		  
    | Symbol _ as hkey :: query_tail, [], DtInnerNode table ->							
	let acc =
	  descend_variant ~p_preserving:p_preserving
	    find_unifiables_entry_data leaf_check acc
	    hkey query_tail table
	in
	let acc =
	  descend_generalized ~p_preserving:p_preserving
	    find_unifiables_entry_data leaf_check acc
	    Variable query_fterm table
	in
	  if p_preserving then
	    acc
	  else
	    descend_generalized ~p_preserving:p_preserving
	      find_unifiables_entry_data leaf_check acc
	      Parameter query_fterm table
		
    | Variable :: query_tail, [], DtInnerNode table ->							
	descend_instance ~p_preserving:p_preserving
	  find_unifiables_entry_data leaf_check acc
	  query_tail table 1
	  
    | Parameter :: query_tail, [], DtInnerNode table ->							
	if p_preserving then begin
	  let acc =
	    descend_variant ~p_preserving:p_preserving
	      find_unifiables_entry_data leaf_check acc
	      Variable query_tail table
	  in
	    descend_variant ~p_preserving:p_preserving
	      find_unifiables_entry_data leaf_check acc
	      Parameter query_tail table
	end
	  
	else begin
	  descend_instance ~p_preserving:p_preserving
	    find_unifiables_entry_data leaf_check acc
	    query_tail table 1
	end
	  
    (* variable <-> symbol or parameter *)
    | Variable :: query_tail, _ :: _, _ ->							
	let tree_fterm', skip =
	  skip_terms_fterm_partial tree_fterm 1
	in
	  if skip = 0 then
	    find_unifiables_entry_data ~p_preserving:p_preserving
	      leaf_check acc
	      query_tail node tree_fterm'
	  else
	    begin
	      match node.variant with
		| DtInnerNode inner_node ->
		    descend_instance ~p_preserving:p_preserving
		      find_unifiables_entry_data leaf_check acc
		      query_tail inner_node skip

		| DtLeafNode _ ->
		    failwith "find_unifiables_entries 3"
	    end

    (* symbol or parameter <-> variable *)
    | _ :: _, Variable :: tree_tail, _ ->							
	(* generalization *)
	find_unifiables_entry_data ~p_preserving:p_preserving
	  leaf_check acc
	  (skip_terms_fterm query_fterm 1) node tree_tail

    (* parameter <-> symbol *)
    | Parameter :: query_tail, _ :: _, _ ->							
	(* generalization *)
	if p_preserving then
	  acc
	else
	  let tree_fterm', skip =
	    skip_terms_fterm_partial tree_fterm 1
	  in
	    if skip = 0 then
	      find_unifiables_entry_data ~p_preserving:p_preserving
		leaf_check acc
		query_tail node tree_fterm'
	    else
	    begin
	      match node.variant with
		| DtInnerNode inner_node ->
		    descend_instance ~p_preserving:p_preserving
		      find_unifiables_entry_data leaf_check acc
		      query_tail inner_node skip

		| DtLeafNode _ ->
		    failwith "find_unifiables_entries 5"
	    end

    (* symbol <-> parameter *)
    | _ :: _, Parameter :: tree_tail, _ ->							
	(* generalization *)
	if p_preserving then
	  acc
	else
	  find_unifiables_entry_data ~p_preserving:p_preserving
	    leaf_check acc
	    (skip_terms_fterm query_fterm 1) node tree_tail
    
    (* mismatch *)
    | _ :: _, _ :: _, _ ->							
	acc

    | [], _ :: _, _ ->							
	failwith "Discrimination_tree.find_unifiables_entries 1"

    | [], [], DtInnerNode _ ->							
	failwith "Discrimination_tree.find_unifiables_entries"
	  
    | _ :: _, [], DtLeafNode _ ->							
	failwith "Discrimination_tree.find_unifiables_entries 2"
	
    | _ :: _, [], DtInnerNode _ ->							
	failwith "Discrimination_tree.find_unifiables_entries 3"	
	
	
	
let rec find_unifiables_entry_data_subst ~(p_preserving: bool)
    leaf_check (acc: ('data * entry_data * Subst.subst) list)
    (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : ('data * entry_data * Subst.subst) list = 
  (* first handle the tree_term *)
  match query_fterm, tree_fterm, node.variant  with
      (* equal keys *)
    | query_key :: query_tail, tree_key :: tree_tail, _ when
	hkey_equal query_key tree_key ->								
	(* variant *)
	find_unifiables_entry_data_subst ~p_preserving:p_preserving
	  leaf_check acc
	  query_tail node tree_tail


    (* tree_fterm empty, so check the node itself *)
    | [], [] , DtLeafNode leaf ->								
	leaf_check leaf acc
		  
    | Symbol _ as hkey :: query_tail, [], DtInnerNode table ->								
	let acc =
	  descend_variant ~p_preserving:p_preserving
	    find_unifiables_entry_data_subst leaf_check acc
	    hkey query_tail table
	in									
	let acc =
	  descend_generalized ~p_preserving:p_preserving
	    find_unifiables_entry_data_subst leaf_check acc
	    Variable query_fterm table
	in									
	  if p_preserving then
	    (acc)
	  else
	    descend_generalized ~p_preserving:p_preserving
	      find_unifiables_entry_data_subst leaf_check acc
	      Parameter query_fterm table
		
    | Variable :: query_tail, [], DtInnerNode table ->								
	descend_instance ~p_preserving:p_preserving
	  find_unifiables_entry_data_subst leaf_check acc
	  query_tail table 1
	  
    | Parameter :: query_tail, [], DtInnerNode table ->								
	if p_preserving then begin
	  let acc =
	    descend_variant ~p_preserving:p_preserving
	      find_unifiables_entry_data_subst leaf_check acc
	      Variable query_tail table
	  in
	    descend_variant ~p_preserving:p_preserving
	      find_unifiables_entry_data_subst leaf_check acc
	      Parameter query_tail table
	end
	  
	else begin
	  descend_instance ~p_preserving:p_preserving
	    find_unifiables_entry_data_subst leaf_check acc
	    query_tail table 1
	end
	  
    (* variable <-> symbol or parameter *)
    | Variable :: query_tail, _ :: _, _ ->								
	let tree_fterm', skip =
	  skip_terms_fterm_partial tree_fterm 1
	in
	  if skip = 0 then
	    find_unifiables_entry_data_subst ~p_preserving:p_preserving
	      leaf_check acc
	      query_tail node tree_fterm'
	  else
	    begin
	      match node.variant with
		| DtInnerNode inner_node ->
		    descend_instance ~p_preserving:p_preserving
		      find_unifiables_entry_data_subst leaf_check acc
		      query_tail inner_node skip

		| DtLeafNode _ ->
		    failwith "find_unifiables_entry_data_subst 3"
	    end

    (* symbol or parameter <-> variable *)
    | _ :: _, Variable :: tree_tail, _ ->								
	(* generalization *)
	find_unifiables_entry_data_subst ~p_preserving:p_preserving
	  leaf_check acc
	  (skip_terms_fterm query_fterm 1) node tree_tail

    (* parameter <-> symbol *)
    | Parameter :: query_tail, _ :: _, _ ->								
	(* generalization *)
	if p_preserving then
	  acc
	else
	  let tree_fterm', skip =
	    skip_terms_fterm_partial tree_fterm 1
	  in
	    if skip = 0 then
	      find_unifiables_entry_data_subst ~p_preserving:p_preserving
		leaf_check acc
		query_tail node tree_fterm'
	    else
	    begin
	      match node.variant with
		| DtInnerNode inner_node ->
		    descend_instance ~p_preserving:p_preserving
		      find_unifiables_entry_data_subst leaf_check acc
		      query_tail inner_node skip

		| DtLeafNode _ ->
		    failwith "find_unifiables_entry_data_subst 5"
	    end

    (* symbol <-> parameter *)
    | _ :: _, Parameter :: tree_tail, _ ->								
	(* generalization *)
	if p_preserving then
	  acc
	else
	  find_unifiables_entry_data_subst ~p_preserving:p_preserving
	    leaf_check acc
	    (skip_terms_fterm query_fterm 1) node tree_tail
    
    (* mismatch *)
    | _ :: _, _ :: _, _ ->								
	acc

    | [], _ :: _, _ ->								
	failwith "Discrimination_tree.find_unifiables_entry_data_subst 1"

    | [], [], DtInnerNode _ ->								
	failwith "Discrimination_tree.find_unifiables_entry_data_subst"
	  
    | _ :: _, [], DtLeafNode _ ->								
	failwith "Discrimination_tree.find_unifiables_entry_data_subst 2"
	
    | _ :: _, [], DtInnerNode _ ->								
	failwith "Discrimination_tree.find_unifiables_entry_data_subst 3"	
	
	
	
	

let rec find_unifiables_subst ~(p_preserving: bool)
    leaf_check (acc: ('data * Subst.subst) list)
    (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : ('data * Subst.subst) list = 
  (* first handle the tree_term *)
  match query_fterm, tree_fterm, node.variant  with
      (* equal keys *)
    | query_key :: query_tail, tree_key :: tree_tail, _ when
	hkey_equal query_key tree_key ->								
	(* variant *)
	find_unifiables_subst ~p_preserving:p_preserving
	  leaf_check acc
	  query_tail node tree_tail


    (* tree_fterm empty, so check the node itself *)
    | [], [] , DtLeafNode leaf ->								
	leaf_check leaf acc
		  
    | Symbol _ as hkey :: query_tail, [], DtInnerNode table ->								
	let acc =
	  descend_variant ~p_preserving:p_preserving
	    find_unifiables_subst leaf_check acc
	    hkey query_tail table
	in									
	let acc =
	  descend_generalized ~p_preserving:p_preserving
	    find_unifiables_subst leaf_check acc
	    Variable query_fterm table
	in									
	  if p_preserving then
	    (acc)
	  else
	    descend_generalized ~p_preserving:p_preserving
	      find_unifiables_subst leaf_check acc
	      Parameter query_fterm table
		
    | Variable :: query_tail, [], DtInnerNode table ->								
	descend_instance ~p_preserving:p_preserving
	  find_unifiables_subst leaf_check acc
	  query_tail table 1
	  
    | Parameter :: query_tail, [], DtInnerNode table ->								
	if p_preserving then begin
	  let acc =
	    descend_variant ~p_preserving:p_preserving
	      find_unifiables_subst leaf_check acc
	      Variable query_tail table
	  in
	    descend_variant ~p_preserving:p_preserving
	      find_unifiables_subst leaf_check acc
	      Parameter query_tail table
	end
	  
	else begin
	  descend_instance ~p_preserving:p_preserving
	    find_unifiables_subst leaf_check acc
	    query_tail table 1
	end
	  
    (* variable <-> symbol or parameter *)
    | Variable :: query_tail, _ :: _, _ ->								
	let tree_fterm', skip =
	  skip_terms_fterm_partial tree_fterm 1
	in
	  if skip = 0 then
	    find_unifiables_subst ~p_preserving:p_preserving
	      leaf_check acc
	      query_tail node tree_fterm'
	  else
	    begin
	      match node.variant with
		| DtInnerNode inner_node ->
		    descend_instance ~p_preserving:p_preserving
		      find_unifiables_subst leaf_check acc
		      query_tail inner_node skip

		| DtLeafNode _ ->
		    failwith "find_unifiables 3"
	    end

    (* symbol or parameter <-> variable *)
    | _ :: _, Variable :: tree_tail, _ ->								
	(* generalization *)
	find_unifiables_subst ~p_preserving:p_preserving
	  leaf_check acc
	  (skip_terms_fterm query_fterm 1) node tree_tail

    (* parameter <-> symbol *)
    | Parameter :: query_tail, _ :: _, _ ->								
	(* generalization *)
	if p_preserving then
	  acc
	else
	  let tree_fterm', skip =
	    skip_terms_fterm_partial tree_fterm 1
	  in
	    if skip = 0 then
	      find_unifiables_subst ~p_preserving:p_preserving
		leaf_check acc
		query_tail node tree_fterm'
	    else
	    begin
	      match node.variant with
		| DtInnerNode inner_node ->
		    descend_instance ~p_preserving:p_preserving
		      find_unifiables_subst leaf_check acc
		      query_tail inner_node skip

		| DtLeafNode _ ->
		    failwith "find_unifiables 5"
	    end

    (* symbol <-> parameter *)
    | _ :: _, Parameter :: tree_tail, _ ->								
	(* generalization *)
	if p_preserving then
	  acc
	else
	  find_unifiables_subst ~p_preserving:p_preserving
	    leaf_check acc
	    (skip_terms_fterm query_fterm 1) node tree_tail
    
    (* mismatch *)
    | _ :: _, _ :: _, _ ->								
	acc

    | [], _ :: _, _ ->								
	failwith "Discrimination_tree.find_unifiables 1"

    | [], [], DtInnerNode _ ->								
	failwith "Discrimination_tree.find_unifiables"
	  
    | _ :: _, [], DtLeafNode _ ->								
	failwith "Discrimination_tree.find_unifiables 2"
	
    | _ :: _, [], DtInnerNode _ ->								
	failwith "Discrimination_tree.find_unifiables 3"	


(*** instance retrieval ***)

let rec find_instances ~(p_preserving: bool)
    (leaf_check: ('data, 'data) leaf_check) (acc: 'data list)
    (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : 'data list =

  (* first handle the tree_term *)
  match query_fterm, tree_fterm, node.variant with
      (* equal keys *)
    | query_key :: query_tail, tree_key :: tree_tail, _ when
	hkey_equal query_key tree_key ->							
	(* variant *)
	find_instances ~p_preserving:p_preserving
	  leaf_check acc
	  query_tail node tree_tail

    (* query_fterm empty, so check the node itself *)
    | [], [], DtLeafNode leaf ->			
	leaf_check leaf acc
		  
    | Symbol _ as hkey :: query_tail, [], DtInnerNode table ->				
	descend_variant ~p_preserving:p_preserving
	  find_instances leaf_check acc
	  hkey query_tail table
	  
    | Variable :: query_tail, [], DtInnerNode table ->				
	descend_instance ~p_preserving:p_preserving
	  find_instances leaf_check acc
	  query_tail table 1
	  
    | Parameter :: query_tail, [], DtInnerNode table ->			
	if p_preserving then begin
	  descend_variant ~p_preserving:p_preserving
	    find_instances leaf_check acc
	    Parameter query_tail table
	end

	else begin
	  descend_instance ~p_preserving:p_preserving
	    find_instances leaf_check acc
	    query_tail table 1
	end
	  
    (* variable <-> symbol or parameter *)
    | Variable :: query_tail, _ :: _, _ ->							
	let tree_fterm', skip =
	  skip_terms_fterm_partial tree_fterm 1
	in
	  if skip = 0 then
	    find_instances ~p_preserving:p_preserving
	      leaf_check acc
	      query_tail node tree_fterm'
	  else
	    begin
	      match node.variant with
		| DtInnerNode inner_node ->
		    descend_instance ~p_preserving:p_preserving
		      find_instances leaf_check acc
		      query_tail inner_node skip

		| DtLeafNode _ ->
		    failwith "find_instances 3"
	    end

    (* parameter <-> symbol or parameter *)
    | Parameter :: query_tail, _ :: _, _ ->									
	if p_preserving then
	  acc
	else
	  let tree_fterm', skip =
	    skip_terms_fterm_partial tree_fterm 1
	  in
	    if skip = 0 then
	      find_instances ~p_preserving:p_preserving
		leaf_check acc
		query_tail node tree_fterm'
	    else
	    begin
	      match node.variant with
		| DtInnerNode inner_node ->
		    descend_instance ~p_preserving:p_preserving
		      find_instances leaf_check acc
		      query_tail inner_node skip

		| DtLeafNode _ ->
		    failwith "find_instances 4"
	    end

    (* failure *)
    | _ :: _, _ :: _, _ ->															
	acc

    | [], _ :: _, _ ->																
	failwith "Discrimination_tree.find_instances 1"

    | [], [], DtInnerNode _ ->														
	failwith "Discrimination_tree.find_instance 1"

    | _ :: _, [], DtLeafNode _ ->													
	failwith "Discrimination_tree.find_instance 2"
	
    | _ :: _, [], DtInnerNode _ ->													
	failwith "Discrimination_tree.find_instance 3"	





(*** shielding ***)


(* find an entry potentially p-instantiating below the producer term *)
let rec find_shielding ~(p_preserving: bool)
    (leaf_check: ('data, 'data) leaf_check) (acc: 'data list)
    (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : 'data list =

  (* first handle the tree_term *)
  match query_fterm, tree_fterm, node.variant with
      (* equal keys *)
    | query_key :: query_tail, tree_key :: tree_tail, _ when
	hkey_equal query_key tree_key ->
	(* variant *)
	find_shielding ~p_preserving:p_preserving
	  leaf_check acc
	  query_tail node tree_tail

    (* query_fterm empty, so check the node itself *)
    | [], [], DtLeafNode leaf ->
	leaf_check leaf acc

    | Symbol _ as hkey :: query_tail, [], DtInnerNode table ->
	let acc =
	  descend_variant ~p_preserving:p_preserving
	    find_shielding leaf_check acc
	    hkey query_tail table
	in
	  descend_generalized ~p_preserving:p_preserving
	    find_shielding leaf_check acc
	    Variable query_fterm table
	    
    | Variable :: query_tail, [], DtInnerNode table
    | Parameter :: query_tail, [], DtInnerNode table ->
	descend_instance ~p_preserving:p_preserving
	  find_shielding leaf_check acc
	  query_tail table 1
	  
    | Variable :: query_tail, Parameter :: tree_tail, _
    | Parameter :: query_tail, Variable :: tree_tail, _ ->
	find_shielding ~p_preserving:p_preserving
	  leaf_check acc
	  query_tail node tree_tail

    | Variable :: query_tail, Symbol _ :: _, _
    | Parameter :: query_tail, Symbol _ :: _, _ ->
	let tree_fterm', skip =
	  skip_terms_fterm_partial tree_fterm 1
	in
	  if skip = 0 then
	    find_shielding ~p_preserving:p_preserving
	      leaf_check acc
	      query_tail node tree_fterm'
	  else
	    begin
	      match node.variant with
		| DtInnerNode inner_node ->
		    descend_instance ~p_preserving:p_preserving
		      find_shielding leaf_check acc
		      query_tail inner_node skip

		| DtLeafNode _ ->
		    failwith "find_shielding 3"
	    end

    (* symbol <-> variable *)
    | _ :: _, Variable :: tree_tail, _ ->
 	(* generalization *)
	find_shielding ~p_preserving:p_preserving
	  leaf_check acc
	  (skip_terms_fterm query_fterm 1) node tree_tail


    (* mismatch *)
    | _ :: _, _ :: _, _ ->
	acc

    | [], _ :: _, _ ->
	failwith "Discrimination_tree.find_instances 1"

    | [], [], DtInnerNode _ ->
	failwith "Discrimination_tree.find_shielding"

    | _, [], DtLeafNode _ ->
	failwith "Discrimination_tree.find_shielding 2"
	
    | _, [], DtInnerNode _ ->
	failwith "Discrimination_tree.find_shielding 3"	

(* The various 'find'-functions above do not work well for perfect discrimination trees,
if we want to get a substitution with the result. Technically they would work, to a degree 
(at least the '_subst' variants of these functions), and completely with some modifications. 
However, they get their substitutions by first finding candidates, and then computing a 
substitution from key-term and candidate (if possible). While this yields the desired results, 
it does not really make use of the perfect discrimination tree's advantage, where we can 
compute the substitution on the fly as we move down the tree. 
Integrating this behaviour into the existing functions while not messing up their original 
operation would make them unwieldy, requiring lots of boolean flag arguments and if-then-else
decisions. It seems preferrable to use the following specialized 'perfect_find'-functions 
for perfect discrimination trees. *)

(*** variant retrieval ***)



let rec perfect_find_variants' (subst: Subst.subst) 
	(leaf_check: ('data, ('data * entry_data * Subst.subst)) perfect_leaf_check) (fterm: fterm) 
	(node: 'data node) : ('data * entry_data * Subst.subst) list = 
	try
		(let fterm, subst =
			shorten_fterm_subst
				~subst:subst 
				~distinguish_perfect_vars:true 
				node.prefix 
				fterm
		in

		(match fterm, node.variant with
			| Var var :: fterm', DtInnerNode table ->
				(Table.fold
					(fun hkey node' acc ->
						match hkey with
							| Var var' ->
								(try
									(let subst_xy = 
										Subst.set' 
											~recompute:global_recompute
											Subst.empty var 0 (Term.request_var var') 1
									and subst_yx =
										Subst.set' 
											~recompute:global_recompute
											Subst.empty var' 1 (Term.request_var var) 0
									in	
									let subst' =
										Unification.merge_substs
											subst subst_xy
									in
									let subst'' =
										Unification.merge_substs
											subst' subst_yx
									in	
									let acc' =
										perfect_find_variants'
											subst''
											leaf_check
											fterm'
											node'
									in
										List.rev_append acc acc')
								with
									| SET_FAIL | UNIFICATION_FAIL ->
										acc)
							| _ ->
								acc)
					table.next
					[])
			| hkey :: fterm', DtInnerNode table ->
				(try
					(let node' =
						Table.find table.next hkey
					in	
						perfect_find_variants' 
							subst
							leaf_check 
							fterm' 
							node')
					with
						| Not_found ->
						  [])
			| [], DtLeafNode leaf ->
				(leaf_check leaf subst [])
			
			| _ ->
				failwith "Discrimination_tree.perfect_find_variants'")
		)
	with
		| Exit ->
			[]
		
let perfect_find_variants (leaf_check: ('data, ('data * entry_data * Subst.subst)) perfect_leaf_check)
    (fterm: fterm) (node: 'data node) : ('data * entry_data * Subst.subst) list =
	perfect_find_variants' Subst.empty leaf_check fterm node	
		
		
		
			
(*** generalization retrieval ***)

let rec perfect_find_generalizations' (subst: Subst.subst) 
	~(p_preserving: bool) (leaf_check: ('data, ('data * entry_data * Subst.subst)) perfect_leaf_check) 
	(acc: ('data * entry_data * Subst.subst) list)
    (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : 
    ('data * entry_data * Subst.subst) list =

(* performance: order so that elements with most variation come first *)
	  (match query_fterm, tree_fterm, node.variant with
	      (* equal keys *)
	    | query_key :: query_tail, tree_key :: tree_tail, _ when
		hkey_equal ~distinguish_perfect_vars:false query_key tree_key ->
		(* variant *)	
		(* This is the rapid variant check using the tree prefixes, so we are not travelling
		down the tree one node per fterm-hkey, rather we stay at one node and cut off as much 
		from the fterms as is shared, and only then do we take a step down to the next node. 
		For perfect fterms this is a bit tricky, as we distinguish variables. If tree prefix
		and fterm have a variable at the same position, but with different variable numbers,
		then these may still match, as long as this	does not contradict a substitution. Even if
		the variable numbers are identical, we must still check the substitution for a 
		contradiction, and the variables must be bound to each other if they aren't yet. *)
			(match query_key, tree_key with
				| Var v1, Var v2 ->
					(try
						(let subst_yx = 
							Subst.set' 
								~recompute:global_recompute 
								Subst.empty v2 1 (Term.request_var v1) 0
						in	
						let subst' =
							Unification.merge_substs
								subst subst_yx
						in					
							(perfect_find_generalizations' 
								subst' 
								~p_preserving:p_preserving
							  	leaf_check acc
							  	query_tail node tree_tail))
					with
						| SET_FAIL | UNIFICATION_FAIL -> acc)
				| _, _ ->
					(perfect_find_generalizations' 
						subst 
						~p_preserving:p_preserving
					  	leaf_check acc
					  	query_tail node tree_tail))
	
	    (* query_fterm empty, so check the node itself *)
	    | [], [], DtLeafNode leaf -> 
		leaf_check leaf subst acc
	
	    | Symbol _ as hkey :: query_tail, [], DtInnerNode table -> 
		let acc =
		  perfect_descend_variant subst false ~p_preserving:p_preserving
		    perfect_find_generalizations' leaf_check acc
		    hkey query_tail table		    
		in
			Table.fold
				(fun hkey node acc ->
					match hkey with
						| Var var ->
							(perfect_descend_generalized subst false ~p_preserving:p_preserving
		    					perfect_find_generalizations' leaf_check acc
		    					hkey query_fterm table)
						| _ -> acc)
				table.next
				acc

	    | Var _ as hkey :: query_tail, [], DtInnerNode table -> 
		  perfect_descend_variant subst false ~p_preserving:p_preserving
		    perfect_find_generalizations' leaf_check acc
		    hkey query_tail table
		    
	    | _ :: _, Var var :: tree_tail, _  ->
		(let skipped, rest_query_fterm = partition_fterm query_fterm 1
		in	
		let bound_term = unflatten_fterm skipped
		in	
		try
			(let subst_yx = 
				Subst.set' 
					~recompute:global_recompute 
					Subst.empty var 1 bound_term 0
			in 
			let subst' =
				Unification.merge_substs
					subst subst_yx
			in	
				perfect_find_generalizations' subst' ~p_preserving:p_preserving
				  leaf_check acc
				  rest_query_fterm node tree_tail)	  
		with
			| SET_FAIL | UNIFICATION_FAIL -> acc)
	    (* mismatch *)
	    | _ :: _, _ :: _, _ ->
		acc
	
	    | [], _ :: _, _ ->
		failwith "Discrimination_tree.perfect_find_generalizations 1"
	
	    | _ :: _, [], DtLeafNode _ ->
		failwith "Discrimination_tree.perfect_find_generalizations 2"
	
	    | [], [], DtInnerNode _ ->
		failwith "Discrimination_tree.perfect_find_generalizations 3"
		
	    | _::_, [], DtInnerNode _ ->
		failwith "Discrimination_tree.perfect_find_generalizations 4")		

let perfect_find_generalizations ~(p_preserving: bool)
    (leaf_check: ('data, ('data * entry_data * Subst.subst)) perfect_leaf_check) 
    (acc: ('data * entry_data * Subst.subst) list)
    (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : 
    ('data * entry_data * Subst.subst) list =
	perfect_find_generalizations' 
		Subst.empty 
		~p_preserving:p_preserving
		leaf_check 
		acc 
		query_fterm 
		node 
		tree_fterm		
			
			
(*** unifiable retrieval ***)
    
let rec perfect_find_unifiables' (subst:Subst.subst) 
	~(p_preserving: bool) (leaf_check: ('data, ('data * entry_data * Subst.subst)) perfect_leaf_check) 
	(acc: ('data * entry_data * Subst.subst) list)
	(query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : 
	('data * entry_data * Subst.subst) list =

(* first handle the tree_term *)	
	  (match query_fterm, tree_fterm, node.variant  with
	      (* equal keys *)
	    | query_key :: query_tail, tree_key :: tree_tail, _ when
		hkey_equal ~distinguish_perfect_vars:false query_key tree_key ->		
		(* variant *)
			(match query_key, tree_key with
				| Var v1, Var v2 ->
					(try
						(let subst_xy = 
							Subst.set' 
								~recompute:global_recompute 
								Subst.empty v1 0 (Term.request_var v2) 1
						and subst_yx = 
							Subst.set' 
								~recompute:global_recompute 
								Subst.empty v2 1 (Term.request_var v1) 0
						in	
						let subst' =
							Unification.unify_substs
								~recompute:global_recompute
								~p_preserving:p_preserving
								subst subst_xy
						in	
						let subst'' =
							Unification.unify_substs
								~recompute:global_recompute
								~p_preserving:p_preserving
								subst' subst_yx						
						in	
							perfect_find_unifiables' subst'' ~p_preserving:p_preserving
								leaf_check acc
								query_tail node tree_tail)								
								
					with 
						| SET_FAIL | UNIFICATION_FAIL -> acc)
				| _, _ ->
					(perfect_find_unifiables' subst ~p_preserving:p_preserving
						leaf_check acc
						query_tail node tree_tail))
	
	
	    (* tree_fterm empty, so check the node itself *)
	    | [], [] , DtLeafNode leaf ->		
		(leaf_check leaf subst acc)
			  
	    | Symbol _ as hkey :: query_tail, [], DtInnerNode table ->
	
			(let acc =
			  perfect_descend_variant subst true ~p_preserving:p_preserving
					perfect_find_unifiables' leaf_check acc
					hkey query_tail table
			in
				Table.fold
					(fun hkey node acc ->
						match hkey with
							| Var var ->
								(perfect_descend_generalized subst true ~p_preserving:p_preserving
			    					perfect_find_unifiables' leaf_check acc
			    					hkey query_fterm table)
							| _ -> acc)
					table.next
					acc)

					
									
	    | Var var :: query_tail, [], DtInnerNode table ->
		perfect_descend_instance subst true ~var:var ~skipped:[] ~p_preserving:p_preserving
		  perfect_find_unifiables' leaf_check acc
		  query_tail table 1
		  
	    | Var var :: query_tail, _ :: _, _ ->
		let skipped, tree_fterm', skip =
			partition_fterm_partial tree_fterm 1
		in
		let bound_term =
			unflatten_fterm skipped
		in
		(try
			(let subst_xy = 
				Subst.set' 
					~recompute:global_recompute 
					Subst.empty var 0 bound_term 1
			in	
			let subst' =
				Unification.unify_substs
					~recompute:global_recompute
					~p_preserving:p_preserving
					subst subst_xy
			in	
			  if skip = 0 then
			    (perfect_find_unifiables' subst' ~p_preserving:p_preserving
					leaf_check acc
					query_tail node tree_fterm')
			  else
				(match node.variant with
					| DtInnerNode inner_node ->
						(perfect_descend_instance subst' true ~var:var ~skipped:(List.rev skipped) ~p_preserving:p_preserving
							perfect_find_unifiables' leaf_check acc
							query_tail inner_node skip)
							
					| DtLeafNode _ ->
						failwith "perfect_find_unifiables 3"))
		with 
			| SET_FAIL | UNIFICATION_FAIL -> acc)
	
    (* symbol or parameter <-> variable *)
		| _ :: _, Var var :: tree_tail, _ ->		
		(* generalization *)
			(let skipped, rest_fterm =
				partition_fterm query_fterm 1
			in
			let bound_term = 
				unflatten_fterm skipped
			in
				try
					(let subst_yx =
							Subst.set' 
								~recompute:global_recompute 
								Subst.empty var 1 bound_term 0
					in	
					let subst' =
						Unification.unify_substs
							~recompute:global_recompute
							~p_preserving:p_preserving
							subst subst_yx
					in	
						perfect_find_unifiables' subst' ~p_preserving:p_preserving
							leaf_check acc
							rest_fterm node tree_tail)
				with
					| SET_FAIL | UNIFICATION_FAIL ->
						(acc))

	    (* mismatch *)
	    | _ :: _, _ :: _, _ ->					
		acc
	
	    | [], _ :: _, _ ->						
		failwith "Discrimination_tree.perfect_find_unifiables 1"
	
	    | [], [], DtInnerNode _ ->				
		failwith "Discrimination_tree.perfect_find_unifiables"
		  
	    | _ :: _, [], DtLeafNode _ ->			
		failwith "Discrimination_tree.perfect_find_unifiables 2"
		
	    | _ :: _, [], DtInnerNode _ ->			
		failwith "Discrimination_tree.perfect_find_unifiables 4")		

let perfect_find_unifiables 
	~(p_preserving: bool) (leaf_check: ('data, ('data * entry_data * Subst.subst)) perfect_leaf_check) 
	(acc: ('data * entry_data * Subst.subst) list)
    (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : 
    ('data * entry_data * Subst.subst) list =	
	perfect_find_unifiables'
		Subst.empty
		~p_preserving:p_preserving 
		leaf_check
		acc
		query_fterm
		node
		tree_fterm
	
	
(*** instance retrieval ***)

let rec perfect_find_instances'
	(subst:Subst.subst) ~(p_preserving: bool)
    (leaf_check: ('data, ('data * entry_data * Subst.subst)) perfect_leaf_check) 
    (acc: ('data * entry_data * Subst.subst) list)
    (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : 
    ('data * entry_data * Subst.subst) list =
   (* print_fterm query_fterm;pe " <FINDINSTANCES";psubst subst;pe "";*)
	(* first handle the tree_term *)
	  (match query_fterm, tree_fterm, node.variant with
	      (* equal keys *)
	    | query_key :: query_tail, tree_key :: tree_tail, _ when
		hkey_equal ~distinguish_perfect_vars:false query_key tree_key ->							
		(* variant *)
		(match query_key, tree_key with
			| Var v1, Var v2 ->
				(try
					(let subst_xy = 
						Subst.set' 
							~recompute:global_recompute
							Subst.empty v1 0 (Term.request_var v2) 1
					in	
					let subst' =
						Unification.merge_substs
							subst subst_xy
					in
						(perfect_find_instances' subst' 
							~p_preserving:p_preserving
							leaf_check acc
							query_tail node tree_tail))
				with
					| SET_FAIL | UNIFICATION_FAIL -> acc)
			| _ ->
				(perfect_find_instances' subst ~p_preserving:p_preserving
					leaf_check acc
					query_tail node tree_tail))
	
	    (* query_fterm empty, so check the node itself *)
	    | [], [], DtLeafNode leaf ->			
		leaf_check leaf subst acc
			  
	    | Symbol _ as hkey :: query_tail, [], DtInnerNode table ->			
		perfect_descend_variant subst false ~p_preserving:p_preserving
		  perfect_find_instances' leaf_check acc
		  hkey query_tail table

	    | Var var :: query_tail, [], DtInnerNode table ->				
			(perfect_descend_instance subst false
				~var:var ~skipped:[] ~p_preserving:p_preserving
			  	perfect_find_instances' leaf_check acc
			  	query_tail table 1)			  
		  			  
	    (* variable <-> symbol or parameter *)
	    | Var var :: query_tail, _ :: _, _ ->					
		let skipped, tree_fterm', skip =
		  partition_fterm_partial tree_fterm 1
		in	
		  if skip = 0 then
		    (let bound_term = unflatten_fterm skipped
		    in
		    try
			    (let subst_xy = 
			    	Subst.set' 
			    		~recompute:global_recompute
			    		Subst.empty var 0 bound_term 1	
			    in	
			    let subst' =
					Unification.merge_substs
						subst subst_xy
			    in	
				    perfect_find_instances' subst' ~p_preserving:p_preserving
				      leaf_check acc
				      query_tail node tree_fterm')
			with
				| SET_FAIL | UNIFICATION_FAIL -> acc)
		  else
		      (match node.variant with
			| DtInnerNode inner_node ->
			    (perfect_descend_instance subst false
			      ~var:var ~skipped:(List.rev skipped) ~p_preserving:p_preserving
			      perfect_find_instances' leaf_check acc
			      query_tail inner_node skip)
	
			| DtLeafNode _ ->
			    failwith "perfect_find_instances 3")

	
	    (* failure *)
	    | _ :: _, _ :: _, _ ->															
		acc
	
	    | [], _ :: _, _ ->																
		failwith "Discrimination_tree.find_instances 1"
	
	    | [], [], DtInnerNode _ ->														
		failwith "Discrimination_tree.find_instance 1"
	
	    | _ :: _, [], DtLeafNode _ ->													
		failwith "Discrimination_tree.find_instance 2"
		
	    | _ :: _, [], DtInnerNode _ ->													
		failwith "Discrimination_tree.find_instance 3")		

let perfect_find_instances
	~(p_preserving: bool)
    (leaf_check: ('data, ('data * entry_data * Subst.subst)) perfect_leaf_check) 
    (acc: ('data * entry_data * Subst.subst) list)
    (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : 
    ('data * entry_data * Subst.subst) list = 
	(perfect_find_instances'
		Subst.empty
		~p_preserving:p_preserving
		leaf_check
		acc
		query_fterm
		node 
		tree_fterm)			
	
	
	
	
			

(* class index (database: Term.database)

   essentially this class is just a wrapper around the functions defined above.
*)
class ['data] predicate_index (__productivity: bool) (__data: 'data data) =

object (self)

   
  (**** values ***)

  (* object providing functions on the data type *)
  val _data = __data
  val _productivity = __productivity

  val _tree: 'data tree = create_tree ()

	
  (*** creation ***)
      
  method clear =
    clear _tree
      
  method size =
    get_size _tree
 
      


  (*** add ***)

  
  

method add ?(no_duplicates: bool = false) 
	?(redundancy_explanation: State.explanation ref = ref State.invalid_explanation) 
	(term: term) (data: 'data) : unit = 
	match _tree.root.variant with
		| DtInnerNode table ->
			let fterm = flatten_term term
			and entry_data = 
				Indexing_misc.create_entry_data 
				~flipped:false 
				~subst:Subst.empty 
				~literal_position:0
				~term:term
				~clause:[]
				~redundancy_explanation:redundancy_explanation
				()
			in						

				if _productivity then
					(if (add_term 
						~no_duplicates:no_duplicates 
						_data 
						(create_entry ~entry_data:entry_data term data) 
						fterm 
						fterm 
						table)
						then
							(inc_size _tree);
					(match term with
						| Term.Eq eq ->
							(let flipped_term = Term.request_eq ~not_orient:true (eq.Term.right, eq.Term.left)
							in
							let flipped_fterm = flatten_term flipped_term
							in
							let flipped_entry_data = 
								Indexing_misc.create_entry_data 
									~flipped:true 
									~subst:Subst.empty 
									~literal_position:0
									~term:flipped_term
									~clause:[]
									~redundancy_explanation:redundancy_explanation
									()
							in
								if (add_term 
									~no_duplicates:no_duplicates 
									_data 
									(create_entry ~entry_data:flipped_entry_data flipped_term data) 
									flipped_fterm 
									flipped_fterm 
									table) then
									(inc_size _tree))
						| _ -> ()))
				else
					(if (add_term 
						~no_duplicates:no_duplicates 
						_data 
						(create_entry ~entry_data:entry_data term data) 
						[] 
						fterm 
						table) then
						(inc_size _tree);
					(match term with
						| Term.Eq eq ->
							(let flipped_term = Term.request_eq ~not_orient:true (eq.Term.right, eq.Term.left)
							in
							let flipped_fterm = flatten_term flipped_term
							in
							let flipped_entry_data = 
								Indexing_misc.create_entry_data 
									~flipped:true 
									~subst:Subst.empty 
									~literal_position:0
									~term:flipped_term
									~clause:[]
									~redundancy_explanation:redundancy_explanation
									()
							in
								if (add_term 
										~no_duplicates:no_duplicates 
										_data 
										(create_entry ~entry_data:flipped_entry_data flipped_term data)
										[] 
										flipped_fterm 
										table) then
									(inc_size _tree))
						| _ -> ()))
		| _ ->
			raise (Failure "Discrimination_tree.add: Bad index toplevel type.")



  (*** traversal ***)
      
  method iter (func: term -> 'data -> unit) : unit =
    iter func _tree.root
      
  method fold:
    'a. ('a -> term -> 'data -> 'a) -> 'a -> 'a =
    fun func acc ->
      fold func acc _tree.root


  method remove (term: term) (data: 'data option) : bool = 
    if get_size _tree = 0 then
      (false)
    else
      remove_variant ~equal:true _tree term data _data




  (*** variants ***)

  method find_all_variants (term: term) : 'data list =
    if get_size _tree = 0 then
      []
    else
      find_variants
	(create_leaf_check ~find_all:true
	   _tree
	   (fun _ -> true)
	   (Term.are_terms_variants term)
	)
	(flatten_term term)
	_tree.root

  method find (term: term) : 'data option = 
    if get_size _tree = 0 then
      None
    else
      try
	ignore (
	  find_variants
	    (create_leaf_check ~find_all:false
	       _tree
	       (fun _ -> true)
	       (Term.term_equal term)
	    )
	    (flatten_term term)
	    _tree.root
	    : 'data list);
	None
      with FOUND ->
	Some (get_found _tree)



  (*** generalization ***)

  method private find_generalizations' ~(find_all: bool) ~(p_preserving: bool) (term: term) : 'data list =
    find_generalizations ~p_preserving:p_preserving
      (create_leaf_check
	 ~find_all:find_all
	 _tree
	 (fun _ -> true)
	 (Unification.is_term_instance ~p_preserving:p_preserving term)
      )
      []
      (flatten_term term)
      _tree.root
      _tree.root.prefix
	    

  method find_all_generalizations ~(p_preserving: bool) (term: term) : 'data list =  
    if get_size _tree = 0 then
      []
    else
      self#find_generalizations' ~find_all:true ~p_preserving:p_preserving term

  method find_generalization ~(p_preserving: bool) (term: term) : 'data option =  
    if get_size _tree = 0 then
      None
    else
      try
	ignore(
	  self#find_generalizations' ~find_all:false ~p_preserving:p_preserving term
	    : 'data list);
	None
      with FOUND ->
	Some (get_found _tree)
      


  (*** unifiable ***)
	    
  method private find_unifiables' ~(find_all: bool) ~(p_preserving: bool) (term: term) : 'data list = 
    find_unifiables ~p_preserving:p_preserving
      (create_leaf_check
	 ~find_all:find_all
	 _tree
	 (fun _ -> true)
	 (Unification.are_terms_unifiable ~p_preserving:p_preserving term)
      )
      []
      (flatten_term term)
      _tree.root
      _tree.root.prefix

  method find_all_unifiable ~(p_preserving: bool) (term: term) : 'data list = 
    if get_size _tree = 0 then
      ([])
    else
      (self#find_unifiables' ~find_all:true ~p_preserving:p_preserving term)

  
      
  method private find_unifiables_entry_data' ~(find_all: bool) ~(p_preserving: bool) (term: term) : ('data * entry_data) list = 
    find_unifiables_entry_data ~p_preserving:p_preserving
      (create_leaf_check_entry_data
	 ~find_all:find_all
	 _tree
	 (fun _ -> true)
	 (Unification.are_terms_unifiable ~p_preserving:p_preserving term)
      )
      []
      (flatten_term term)
      _tree.root
      _tree.root.prefix      
       
      
           
      
  method find_all_unifiable_entry_data ~(p_preserving: bool) (term: term) : ('data * entry_data) list = 
    if get_size _tree = 0 then
      ([])
    else
      (self#find_unifiables_entry_data' ~find_all:true ~p_preserving:p_preserving term)
      
      
  method find_unifiable ~(p_preserving: bool) (term: term) : 'data option = 
    if _tree.size == 0 then
      None
    else		(
      try
	ignore(
	  self#find_unifiables' ~find_all:false ~p_preserving:p_preserving term
	  : 'data list);
	None
      with FOUND ->
	(Some (get_found _tree)))

  method find_all_unifiable_subst ~(p_preserving: bool) (term: term) : ('data * Subst.subst) list = 
    if get_size _tree = 0 then
      ([])
    else 
    (find_unifiables_subst ~p_preserving:p_preserving
      (create_leaf_check_subst
(*	 ~find_all:true*)
	 _tree
	 (fun _ -> true)
	 (fun tree_term ->
	    (* context literal term first,
	       so that bindings are from context literal vars to clause literal vars.
	       see Subst.remove_universal_context_vars.
	    *)
	    Unification.unify_terms ~recompute:false ~p_preserving:p_preserving
	      term Term_indexing.query_offset tree_term Term_indexing.index_offset)
      )
      []
      (flatten_term term)
      _tree.root
      _tree.root.prefix)

  method find_all_unifiable_entry_data_subst ~(p_preserving: bool) (term: term) : ('data * entry_data * Subst.subst) list = 
    if get_size _tree = 0 then
      ([])
    else 
    (find_unifiables_entry_data_subst ~p_preserving:p_preserving
      (create_leaf_check_entry_data_subst
	 ~find_all:true
	 _tree
	 (fun _ -> true)
	 (fun tree_term ->
	    (* context literal term first,
	       so that bindings are from context literal vars to clause literal vars.
	       see Subst.remove_universal_context_vars.
	    *)
	    Unification.unify_terms ~recompute:false ~p_preserving:p_preserving
	      term Term_indexing.query_offset tree_term Term_indexing.index_offset)
      )
      []
      (flatten_term term)
      _tree.root
      _tree.root.prefix)

  (*** instance ***)

  method private find_instances' ~(find_all: bool) ~(p_preserving: bool) (term: term) : 'data list =
    find_instances ~p_preserving:p_preserving
      (create_leaf_check
	 ~find_all:find_all
	 _tree
	 (fun _ -> true)
	 (fun tree_term ->
	    Unification.is_term_generalization ~p_preserving:p_preserving term tree_term
	 )
      )
      []
      (flatten_term term)
      _tree.root
      _tree.root.prefix


  method find_all_instances ~(p_preserving: bool) (term: term) : 'data list =
    if get_size _tree = 0 then
      []
    else
      self#find_instances' ~find_all:true ~p_preserving:p_preserving term

  method find_instance ~(p_preserving: bool) (term: term) : 'data option = 
    if get_size _tree = 0 then
      None
    else
    try 
      ignore(
	self#find_instances' ~find_all:false ~p_preserving:p_preserving term
	  : 'data list);
      None
    with FOUND ->
      Some (get_found _tree)



  (*** productivity ***)

  (* find an instance of producer that is more general than produced *)
  method find_strongly_shielding (producer: term) (produced: term) : 'data option =
    if not _productivity then
      failwith "Discrimination_tree: productivity not enabled"
    else if get_size _tree = 0 then
      None
    else
      let fterm_produced =
	flatten_term produced
      in
      try 
	ignore(
	  find_instances ~p_preserving:false
	    (create_leaf_check
	       ~find_all:false
	       _tree
	       (fun leaf_fterm -> fterm_equal_generalization ~p_preserving:false leaf_fterm fterm_produced)
	       (fun tree_term ->
		  (Term.term_equal tree_term produced)
		  ||
		    (
		      (* must be a true instance *)
		      not (Term.term_equal producer tree_term)
		      &&
		      Unification.is_term_instance produced tree_term
		      &&
		      Unification.is_term_generalization producer tree_term
		    )
	       )
	    )
	    []
	    (flatten_term producer)
	    _tree.root
	    _tree.root.prefix
	    : 'data list);
	None
      with FOUND ->
	Some (get_found _tree)
	  

  method private shielding_entry_check (producer: term) (produced: term) (tree_term: term) : bool =
    (* produced itself is ok *)
    (Term.term_equal tree_term produced)
    ||
    (
      (* must be a true instance *)
      not (Term.term_equal tree_term producer)
      &&
      (* must be more general then the produced term *)
	Unification.is_term_instance produced tree_term
	(* p-preservingly unifiable from the tree term -
	   this instance of producer and tree term is still more general
	   than produced, as producer and tree term themselves
	   are more general than produced.
	   (at least, I, Alex, think so)
	*)
      &&
      (
	try
	  let subst =
	    Unification.unify_terms ~recompute:false
	      producer Term_indexing.query_offset
	      tree_term Term_indexing.index_offset
	  in
	    Subst.is_p_renaming subst Term_indexing.index_offset
	with
	  | Unification.UNIFICATION_FAIL  ->
	      false
      )
    )

  (* find a term p-preservingly instantiable to an instance of producer,
     which is also more general than produced *)
  method private find_shielding' ~(find_all: bool) (producer: term) (produced: term) : 'data list =
    let fterm_produced =
      flatten_term produced
    in
    find_shielding ~p_preserving:false
      (create_leaf_check
	 ~find_all:find_all
	 _tree
	 (*(fterm_equal_shielding)*)
	 (fun leaf_fterm -> fterm_equal_generalization ~p_preserving:false leaf_fterm fterm_produced)
	 (self#shielding_entry_check producer produced)
      )
      []
      (flatten_term producer)
      _tree.root
      _tree.root.prefix


  method find_all_shielding (producer: term) (produced: term) : 'data list =
    if not _productivity then
      failwith "Discrimination_tree: productivity not enabled"
    else if get_size _tree = 0 then
      []
    else
      self#find_shielding' ~find_all:true producer produced

  method find_shielding (producer: term) (produced: term) : 'data option =
    if not _productivity then
      failwith "Discrimination_tree: productivity not enabled"
    else if get_size _tree = 0 then
      None
    else
      try 
	ignore(
	  self#find_shielding' ~find_all:false producer produced
	    : 'data list);
	None
      with FOUND ->
	Some (get_found _tree)
    



  method get_generalization_iterator ~(p_preserving: bool) (term: term) : 'data iterator =
    let generalizations =
      self#find_all_generalizations ~p_preserving:p_preserving term
    in
      object (_iterator)
	val mutable _elements = generalizations

	method is_empty =
	  match _elements with
	    | [] -> true
	    | _ -> false

	method next =
	  match _elements with
	    | [] ->
		raise ITERATOR_EMPTY

	    | element :: tail ->
		_elements <- tail;
		element
      end


  (* like find_all_shielding, but computing matches one at a time *)
  method get_shielding_iterator (producer: term) (produced: term) : 'data iterator =

    object (iterator)
      val _producer_fterm = flatten_term producer
      val _produced_fterm = flatten_term produced
      
      (* index nodes still to search for *)
      val mutable _nodes: (fterm * 'data node * fterm) list = []
      
      (* currently found potentially shielding entries *)
      val mutable _entries: 'entry list = []

      (* a found shielding entry *)
      val mutable _shielding: 'data option = None

	
      initializer
	(* start with query term and root node node *)
	if not _productivity then
	  failwith "Discrimination_tree: productivity not enabled"
	else if get_size _tree > 0 then
	  iterator#save_node _producer_fterm _tree.root _tree.root.prefix


      (* new node to search forfound *)
      method private save_node (query_ftail: fterm) (node: 'data node) (tree_ftail: fterm) =
	_nodes <- (query_ftail, node, tree_ftail) :: _nodes

      (* matching leaf found, evaluate its entries *)
      method private save_leaf (leaf: 'data leaf) =
	(* nodes more general? *)
	if fterm_equal_generalization ~p_preserving:false leaf.fterm _produced_fterm then begin
	  _entries <- _entries @ leaf.entries;
	end
  
      method is_empty =
	match _shielding with
	  | Some _ ->
	      (* yes, entry found *)
	      false
		
	  | None ->
	      (* don't know, check ... *)
	      begin
		match _entries with
		  | entry :: tail ->
		      (* ...leaves *)
		      _entries <- tail;
		      if self#shielding_entry_check producer produced entry.term then begin
			_shielding <- Some entry.data;
		      end;
		      iterator#is_empty
			
		  | [] ->
		      begin
			match _nodes with
			  | (query_fterm, node, tree_fterm) :: tail ->
			      (* ...nodes *)
			      _nodes <- tail;
			      iterator#advance query_fterm node tree_fterm;
			      iterator#is_empty
			    
			  | [] ->
			      true
		      end
	      end

      method next =
	if iterator#is_empty then
	  raise ITERATOR_EMPTY
	
	else  
	  match _shielding with
	    | None ->
		failwith "Discrimination_tree.get_shielding_iterator.next"
		  
	    | Some shielding ->
		_shielding <- None;
		shielding


      method private advance (query_fterm: fterm) (node: 'data node) (tree_fterm: fterm) : unit =
	(* first handle the tree_term *)
	match query_fterm, tree_fterm, node.variant  with
	  | [], _ :: _, _ ->
	      failwith "Discrimination_tree.get_shielding_iterator.advance 1"

	  (* equal keys *)
	  | query_key :: query_tail, tree_key :: tree_tail, _ when
	      hkey_equal query_key tree_key ->
	      (* variant *)
	      iterator#advance query_tail node tree_tail

	  | Variable :: query_tail, Parameter :: tree_tail, _
	  | Parameter :: query_tail, Variable :: tree_tail, _ ->
	      iterator#advance query_tail node tree_tail

	  | Variable :: query_tail, Symbol _ :: _, _
	  | Parameter :: query_tail, Symbol _ :: _, _ ->
	      let tree_fterm', skip =
		skip_terms_fterm_partial tree_fterm 1
	      in
		if skip = 0 then
		  iterator#advance query_tail node tree_fterm'
		else
		  begin
		    match node.variant with
		      | DtInnerNode inner_node ->
			  iterator#descend_instance query_tail inner_node skip

		      | DtLeafNode _ ->
			  failwith "Discrimination_tree.get_shielding_iterator.advance 2"
		  end

	  (* symbol <-> variable *)
	  | _ :: _, Variable :: tree_tail, _ ->
 	      (* generalization *)
	      iterator#advance (skip_terms_fterm query_fterm 1) node tree_tail
		
		
	  (* mismatch *)
	  | _ :: _, _ :: _, _ ->
	      ()

	  (* query_fterm empty, so check the node itself *)
	  | [], [], DtLeafNode leaf ->
	      iterator#save_leaf leaf
		
	  | Symbol _ as hkey :: query_tail, [], DtInnerNode table ->
	      begin
		try
		  (* for the same symbol don't store it by immediately follow it,
		     for all other cases store the new node to avoid early branching. *)
		  let node' =
		    Table.find table.next hkey
		  in
		    iterator#advance query_tail node' node'.prefix
		with
		  | Not_found ->
		      ()
	      end;
	      begin
		try
		  let node' =
		    Table.find table.next Variable
		  in
		    iterator#save_node (skip_terms_fterm query_fterm 1) node' node'.prefix
		with
		  | Not_found ->
		      ()
	      end
		
	  | Variable :: query_tail, [], DtInnerNode table
	  | Parameter :: query_tail, [], DtInnerNode table ->
	      iterator#descend_instance query_tail table 1
		
	  | [], [], DtInnerNode _ ->
	      failwith "Discrimination_tree.get_shielding_iterator.advance 3"

	  | _, [], DtLeafNode _ ->
	      failwith "Discrimination_tree.get_shielding_iterator.advance 4"
	      
	  | _, [], DtInnerNode _ ->
	      failwith "Discrimination_tree.get_shielding_iterator.advance 5"
	      


      method private descend_instance (query_tail: fterm) (table: 'data inner_node) (skip: int) : unit =
	Table.iter
	  (fun hkey node ->
	     let skip' =
	       skip - 1 + (hkey_arity hkey)
	     in

	     (* now skip also the prefix of this node *)
	     let prefix, skip' =
	       skip_terms_fterm_partial node.prefix skip'
	     in
	       if skip' = 0 then
		 iterator#save_node query_tail node prefix
	       
	       else
		 match node.variant with
		   | DtLeafNode _ ->
		       failwith "shielding_iterator.descend_instance"
			 
		   | DtInnerNode table' ->
		       iterator#descend_instance query_tail table' skip'
	  )
	  table.next
	  
	  
    end
      
      
  (*** to_string ***)

  method to_string : string =
    tree_to_string _tree
      
end

(********************************************************************)
(*	Under construction: perfect indexing
	This may replace the original indexes later on, but for now it will
	only be used for indexing of multi-literal clauses.
	Therefore this class is a modified copy of predicate_index, 
	although the vast majority of its methods and features are not
	used for anything yet.
*)

class ['data] perfect_predicate_index (__productivity: bool) (__data: 'data data) =

object (self)

   
  (**** values ***)

  (* object providing functions on the data type *)
  val _data = __data
  val _productivity = __productivity

  val _tree: 'data tree = create_tree ()

	
  (*** creation ***)
      
  method clear =
    clear _tree
      
  method size =
    get_size _tree
 
      


  (*** add ***)
  

  method add ?(no_duplicates: bool = false) 
	?(redundancy_explanation:State.explanation ref = ref State.invalid_explanation) 
	?(clause: clause = []) 
  	?(literal_position: int = 0) ?(search_flag:(bool ref) = ref false)
  	?(misc_flag:(bool ref) = ref false) (term: term) (data: 'data) : unit = 
    match _tree.root.variant with
      | DtInnerNode table ->
	  let fterm = flatten_term ~perfect:true term
	  and entry_data = 
	  	Indexing_misc.create_entry_data 
	  		~flipped:false 
	  		~subst:Subst.empty 
	  		~literal_position:literal_position 
	  		~term:term
	  		~clause:clause
	  		~search_flag:search_flag
	  		~misc_flag:misc_flag
	  		~redundancy_explanation:redundancy_explanation
	  		()
	  in						
	  let added =
	    if _productivity then
	      ((add_term 
	      		~no_duplicates:no_duplicates 
	      		_data 
	      		(create_entry ~entry_data:entry_data term data) 
	      		fterm 
	      		fterm 
	      		table)
	      	&&
	      (match term with
	      	| Term.Eq eq ->
	      	(let flipped_term = Term.request_eq ~not_orient:true (eq.Term.right, eq.Term.left)
	      	in
	      	let flipped_fterm = flatten_term ~perfect:true flipped_term
	      	in
	      	let flipped_entry_data = 
			  	Indexing_misc.create_entry_data 
			  		~flipped:true 
			  		~subst:Subst.empty 
			  		~literal_position:literal_position 
			  		~term:flipped_term
			  		~clause:clause
			  		~search_flag:search_flag
			  		~misc_flag:misc_flag	
			  		~redundancy_explanation:redundancy_explanation  		
			  		()
	      	in
	      	add_term 
	      		~no_duplicates:no_duplicates 
	      		_data 
	      		(create_entry ~entry_data:flipped_entry_data flipped_term data) 
	      		flipped_fterm 
	      		flipped_fterm 
	      		table)
	       |_ -> true))
	    else
	      ((add_term 
	      		~no_duplicates:no_duplicates 
	      		_data 
	      		(create_entry ~entry_data:entry_data term data) 
	      		[] 
	      		fterm 
	      		table)
	      	&&
	      (match term with
	      	| Term.Eq eq ->
	      	(let flipped_term = Term.request_eq ~not_orient:true (eq.Term.right, eq.Term.left)
	      	in
	      	let flipped_fterm = flatten_term ~perfect:true flipped_term
	      	in
	      	let flipped_entry_data = 
			  	Indexing_misc.create_entry_data 
			  		~flipped:true 
			  		~subst:Subst.empty 
			  		~literal_position:literal_position 
			  		~term:flipped_term
			  		~clause:clause
			  		~search_flag:search_flag
			  		~misc_flag:misc_flag	 
			  		~redundancy_explanation:redundancy_explanation 			
			  		()
	      	in
		      	add_term 
		      		~no_duplicates:no_duplicates 
		      		_data 
		      		(create_entry
		      			~entry_data:flipped_entry_data 
		      			flipped_term 
		      			data) 
		      		[] 
		      		flipped_fterm 
		      		table)
	       |_ -> true))
	  in
	    if added then begin 
	      (inc_size _tree);
	    end;

      | _ ->
	  raise (Failure "discrimination_tree:add: Bad index toplevel type.")


  (*** traversal ***)
      
  method iter (func: term -> 'data -> unit) : unit =
    iter func _tree.root
      
  method fold:
    'a. ('a -> term -> 'data -> 'a) -> 'a -> 'a =
    fun func acc ->
      fold func acc _tree.root


  method remove (term: term) (data: 'data option) : bool =
    if get_size _tree = 0 then
      false
    else
      remove_variant ~perfect:true ~equal:true _tree term data _data

  method remove_clause (term: term) (clause: clause) : bool =
	if get_size _tree = 0 then
		(false)
	else
		(remove_clause ~perfect:true ~equal:true _tree term clause)

  method extract_clause (term: term) (clause: clause) : bool * ('data option) =
	if get_size _tree = 0 then
		(false, None)
	else
		(extract_clause ~perfect:true ~equal:true _tree term clause)

			

  (*** variants ***)

  method find_all_variants ?(subst: Subst.subst = Subst.empty) (term: term) 
  : ('data * entry_data * Subst.subst) list = 
    if get_size _tree = 0 then
      ([])
    else
      perfect_find_variants'
      	subst
	(create_perfect_leaf_check
	   _tree
	   (fun _ -> true)
	   true
	)
	(flatten_term ~perfect:true term)
	_tree.root

  method find ?(subst: Subst.subst = Subst.empty) (term: term) : 'data option = 
    if get_size _tree = 0 then
      None
    else
      try
	ignore (
	  perfect_find_variants'
	  	subst
	    (create_perfect_leaf_check
	       _tree
	       (fun _ -> true)
	       true
	    )
	    (flatten_term ~perfect:true term)
	    _tree.root
	    : ('data * entry_data * Subst.subst) list);
	None
      with FOUND ->
	Some (get_found _tree)


  (*** generalization ***)

  method private find_generalizations' ?(subst: Subst.subst = Subst.empty)
  	~(find_all: bool) ~(p_preserving: bool) (term: term) : 
  	('data * entry_data * Subst.subst) list = 
    perfect_find_generalizations'  
    	subst
    	~p_preserving:p_preserving
      (create_perfect_leaf_check
	 	_tree
	 	(fun _ -> true)
	 	true)
      []
      (flatten_term ~perfect:true term)
      _tree.root
      _tree.root.prefix
	    

  method find_all_generalizations ?(subst: Subst.subst = Subst.empty)
  	~(p_preserving: bool) (term: term) : 
  	('data * entry_data * Subst.subst) list =  
    if get_size _tree = 0 then
      []
    else
      self#find_generalizations'
      	~subst:subst
      	~find_all:true 
      	~p_preserving:p_preserving 
      	term

  method find_generalization  ?(subst: Subst.subst = Subst.empty)
  	~(p_preserving: bool) (term: term) : 'data option =  
    if get_size _tree = 0 then
      None
    else
      try
	ignore(
	  self#find_generalizations' 
	  	~subst:subst
	  	~find_all:false 
	  	~p_preserving:p_preserving 
	  	term
	    : ('data * entry_data * Subst.subst) list);
	None
      with FOUND ->
	Some (get_found _tree)
      


  (*** unifiable ***)
	    
  method private find_unifiables'  ?(subst: Subst.subst = Subst.empty)
  	~(find_all: bool) ~(p_preserving: bool) (term: term) : 
  	('data * entry_data * Subst.subst) list = 
    perfect_find_unifiables'
    	subst
    	~p_preserving:p_preserving
      (create_perfect_leaf_check
	 	_tree
	 	(fun _ -> true)
	 	true)
      []
      (flatten_term ~perfect:true term)
      _tree.root
      _tree.root.prefix

  method find_all_unifiable ?(subst: Subst.subst = Subst.empty) 
  	~(p_preserving: bool) (term: term) : 
  	('data * entry_data * Subst.subst) list = 
    if get_size _tree = 0 then
      ([])
    else
      (self#find_unifiables'
      	~subst:subst
      	~find_all:true 
      	~p_preserving:p_preserving term)

      
  method find_unifiable ?(subst: Subst.subst = Subst.empty) 
  	~(p_preserving: bool) (term: term) : 'data option = 
    if _tree.size == 0 then
      None
    else		(
      try
	ignore(
	  self#find_unifiables'
	  	~subst:subst
	  	~find_all:false 
	  	~p_preserving:p_preserving 
	  	term
	  : ('data * entry_data * Subst.subst) list);
	None
      with FOUND ->
	(Some (get_found _tree)))


  (*** instance ***)

  method private find_instances' ?(subst: Subst.subst = Subst.empty) 
  	~(find_all: bool) ~(p_preserving: bool) (term: term) : 
  	('data * entry_data * Subst.subst) list = 
    perfect_find_instances'
    	subst
    	~p_preserving:p_preserving
      (create_perfect_leaf_check
	 _tree
	 (fun _ -> true)
	 true
      )
      []
      (flatten_term ~perfect:true term)
      _tree.root
      _tree.root.prefix



  method find_all_instances ?(subst: Subst.subst = Subst.empty) 
  	~(p_preserving: bool) (term: term) : 
  	('data * entry_data * Subst.subst) list =
    if get_size _tree = 0 then
      []
    else
      self#find_instances'
      	~subst:subst
      	~find_all:true 
      	~p_preserving:p_preserving 
      	term

  method find_instance ?(subst: Subst.subst = Subst.empty) 
  	~(p_preserving: bool) (term: term) : 'data option = 
    if get_size _tree = 0 then
      None
    else
    try 
      ignore(
	self#find_instances' 
		~subst:subst
		~find_all:false 
		~p_preserving:p_preserving 
		term
	  : ('data * entry_data * Subst.subst) list);
      None
    with FOUND ->
      Some (get_found _tree)


  method get_generalization_iterator ?(subst: Subst.subst = Subst.empty) 
  	~(p_preserving: bool) (term: term) : 'data iterator =
    let generalizations =
      self#find_all_generalizations
      	~subst:subst 
      	~p_preserving:p_preserving 
      	term
    in
      object (_iterator)
	val mutable _elements = 
		List.map
			(fun (data, _, _) -> data)
			generalizations

	method is_empty =
	  match _elements with
	    | [] -> true
	    | _ -> false

	method next =
	  match _elements with
	    | [] ->
		raise ITERATOR_EMPTY

	    | element :: tail ->
		_elements <- tail;
		element
      end
      
  (*** to_string ***)

  method to_string : string =
    tree_to_string _tree
      
end


(********************************************************************)
    
class ['data] index
  (__productivity: bool) (__data: 'data data)
   =
object (_self)
  inherit ['data] Term_indexing.index
    __data
    (new predicate_index __productivity)
end

class ['data] term_index
  (__productivity: bool) (__data: 'data data)
   =
object (_self)
  inherit ['data] Term_indexing.term_index
    __data
    (new predicate_index __productivity)
end

class ['data] symbol_index
  (__productivity: bool) (__data: 'data data)
   =
object (_self)
  inherit ['data] Term_indexing.symbol_index
    __data
    (new predicate_index __productivity)
end

class ['data] perfect_term_index
  (__productivity: bool) (__data: 'data data)
   =
object (_self)
  inherit ['data] Term_indexing.perfect_term_index
    __data
    (new perfect_predicate_index __productivity)
end

class ['data] perfect_symbol_index
  (__productivity: bool) (__data: 'data data)
   =
object (_self)
  inherit ['data] Term_indexing.perfect_symbol_index
    __data
    (new perfect_predicate_index __productivity)
end

let variant_entry_check
	(clause: clause) (entry_data: entry_data) (subst: Subst.subst) : bool = 
	(List.length clause)
	=	
(* A simple length comparison is sufficient, as the variant search has already checked substitutions
in both directions. We only have to check whether the found clause contains any literals that have not
been matched by the search clause. Since the variant search has ensured that each search clause literal
has one unique matching literal in the found clause, we can simply check for unmatched literals in the
found clause by comparing its length to that of the search clause - both lengths must be the same. *)
	(List.length
		(Indexing_misc.get_clause entry_data))
		
let instance_entry_check
	(clause: clause) (entry_data: entry_data) (subst: Subst.subst) : bool =
(* This is more complicated than the variant check, since more than one search clause literal may
have matched a single found clause literal. Even if the lengths of the clauses are equal, the found
clause may still contain unmatched literals. We check this by applying the substitutions to the
respective clauses and then testing for clause equality/identity. *)	
	let candidate_clause = 
		Indexing_misc.get_clause 
			entry_data
	in	(*pe "INSTANCEENTRYCHECK";pc clause;pe "";pc candidate_clause;pe "";psubst subst;pe "";*)
	((candidate_clause == clause)
	||
	(let clause =
		Subst.apply_to_clause 
			subst
			clause 
			0
	in	
		(Term.clause_equal clause candidate_clause)))
	
		
		
let generalization_entry_check			
	(clause: clause) (entry_data: entry_data) (subst: Subst.subst) : bool =
(* Like instance_entry_check, only vice versa. *)	
	let candidate_clause =
		(Indexing_misc.get_clause 
			entry_data)
	in 
	((candidate_clause == clause)
	||
	(let candidate_clause =
		Subst.apply_to_clause 
			subst
			candidate_clause
			1
	in
		(Term.clause_equal clause candidate_clause)))
		
let unifiable_entry_check
	(clause: clause) (entry_data: entry_data) (subst: Subst.subst) : bool =
(* Like instance_entry_check and generalization_entry_check combined. *)	
	let candidate_clause =
		(Indexing_misc.get_clause 
			entry_data)
	in
	((candidate_clause == clause)
	||
	(let candidate_clause =
		Subst.apply_to_clause
			subst
			candidate_clause
			1
	and search_clause =
		Subst.apply_to_clause
			subst
			clause
			0
	in
		(Term.clause_equal search_clause candidate_clause)))
		

		
let find_variants_in_index 
	(index: 'data Term_indexing.perfect_predicate_index) 
	(subst:Subst.subst) 
	(term: Term.term) 
	: ('data * entry_data * Subst.subst) list = 
	index#find_all_variants
		~subst:subst
		term

let find_instances_in_index		
	(index: 'data Term_indexing.perfect_predicate_index) 
	(subst:Subst.subst) 
	(term: Term.term) 
	: ('data * entry_data * Subst.subst) list =
	index#find_all_instances
		~p_preserving:false
		~subst:subst
		term
		
let find_unifiables_in_index		
	(index: 'data Term_indexing.perfect_predicate_index) 
	(subst:Subst.subst) 
	(term: Term.term) 
	: ('data * entry_data * Subst.subst) list = 
	index#find_all_unifiable
		~p_preserving:false
		~subst:subst
		term		
		
let find_generalizations_in_index		
	(index: 'data Term_indexing.perfect_predicate_index) 
	(subst:Subst.subst) 
	(term: Term.term) 
	: ('data * entry_data * Subst.subst) list =
	index#find_all_generalizations
		~p_preserving:false
		~subst:subst
		term		
	
class ['data] perfect_clause_index
	(__productivity: bool) (__data: 'data data) =
object (_self)
	inherit ['data] Term_indexing.perfect_clause_index
		__data
		(new perfect_predicate_index __productivity)
		
(* like above, a workaround to simulate exception FOUND 'data *)		
	val mutable found = None
		
	method private get_found = found
	
	method private set_found (data: 'data option) : unit =
		found <- data
	
	method add_clause (clause: clause) 
		?(redundancy_explanation: State.explanation ref = ref State.invalid_explanation) 
		(data: 'data) : unit =
		let search_flag = ref false
		and misc_flag = ref false
		in
			ignore
				(List.fold_left
					(fun cnt literal ->	
						let index = _self#find literal
						in
							index#add
								~redundancy_explanation:redundancy_explanation 
								~clause:clause 
								~literal_position:cnt 
								~search_flag:search_flag 
								~misc_flag:misc_flag
								literal.Term.atom data;
							cnt + 1)
					0
					clause)

			
	method add_clause_with_subsumption_check (clause: clause) 
		?(redundancy_explanation: State.explanation ref = ref State.invalid_explanation)  
		(data: 'data) : (bool * ('data list)) =
	(* check if there is an indexed clause that subsumes clause *)
		try
			(ignore (_self#find_first_clause_generalization clause);
			(false, []))
		with
			| Not_found ->
	(* clause is not subsumed by indexed clauses *)
				(let subsumed =		
					_self#find_clause_instances clause
				in	(*	ps "ADDED: ";pc clause;pe ".";*)
					_self#add_clause 
						clause
						~redundancy_explanation:redundancy_explanation
						data;
					(true, subsumed))			

					
	method subsumption_check (clause: clause) : (bool * ('data list)) =					
		try
			(ignore (_self#find_first_clause_generalization clause);
			(false, []))
		with
			| Not_found ->			
				(let subsumed =		
					_self#find_clause_instances clause
				in		
					(true, subsumed))			
			
	method remove_clause (clause: clause) : unit =
		ignore
			(List.iter
				(fun literal ->	
					let index = _self#find literal
					in
						ignore (index#remove_clause literal.Term.atom clause);
						match literal.Term.atom with
							| Term.Eq eq ->
								(let flipped_eq =
									Term.request_eq
										~not_orient:true
										(eq.Term.right, eq.Term.left)
								in
									ignore (index#remove_clause flipped_eq clause))
							| _ -> ())
				clause)	
				
	method extract_clause (clause: clause) : 'data option =
		let extracted =
			(List.map
				(fun literal ->	
					let index = _self#find literal
					in
						let _, data =
							(index#extract_clause literal.Term.atom clause)
						in
							data)
				clause)
		in
			match extracted with
				| [] -> None
				| data :: _ -> data	

			
				
	method private find_clauses
		?(find_first_only: bool = false)
		(find_clauses_in_index: 
'data Term_indexing.perfect_predicate_index -> Subst.subst -> term -> ('data * entry_data * Subst.subst) list)
		(entry_check: clause -> entry_data -> Subst.subst -> bool) 
		(clause: clause) 
		: 'data list =		
		
		let rec find_clauses'
			(rest_clause: clause)
			(subst: Subst.subst)
			(previous_results: ('data * entry_data * Subst.subst) list)
			: 'data list =
			
			(match rest_clause with
				| [] ->
					(let results = ref []
					in	
						List.iter
							(fun (data, entry_data, _) -> 
								if (not (List.memq data !results)) && (entry_check clause entry_data subst) then
									(if find_first_only then
										(_self#set_found (Some data);
										raise FOUND)
									else
										(results := data :: !results)))
							previous_results;
						!results)
				| literal :: literals -> (*pl literal;pe ";";List.iter (fun (_, dat, _) -> pc (Indexing_misc.get_clause dat);pe " -<=";) previous_results;*)
					(let previous_subst = ref Subst.empty
					and subst_cluster = ref []
					and subst_clusters = ref []
					in			(* pl literal;pe " <FINDNEXT";psubst subst;pe "";*)
(* find all clauses for this literal *)					
					let literal_results =						
						(let index = _self#find literal
						in
							find_clauses_in_index 
								index
								subst 
								literal.Term.atom)
					in									(*pi (List.length literal_results);pe "<";*)
(* only keep those that have been found for previous search literals *)					
					let literal_results =
						List.find_all
							(fun (_, entry_data, _) ->	  
								Indexing_misc.get_search_flag entry_data)
							literal_results
					in 									 (*pi (List.length literal_results);pe ">"; *)
(* Now start some serious flag messing to ensure the search continues efficiently. The following
makes use of the fact that all entry_datas created for a single clause (one for each literal)
share the flags - for example, set the search_flag in one entry_data to true, and the same flag 
will be true in all the other entry_datas for that clause as well. *)
(* step 1: set the misc_flag of the previous_candidate_clauses, this will be used to ensure 
that we don't flip the same search_flag twice in step 3 *)			
					List.iter
						(fun (_, entry_data, _) -> 
							Indexing_misc.set_misc_flag entry_data true)
						previous_results;			
(* step 2: turn off the search_flags in the literal_results. Thereby the same clauses in the previous_results
can be distinguished from those clauses that only occur in the previous_results: clauses that occur in both
sets have the search_flag set to false, while clauses only in the previous_results still have that flag set
to true. *)
					List.iter
						(fun (_, entry_data, _) -> 	
							Indexing_misc.set_misc_flag entry_data false)
						literal_results;		
(* step 3: for the next round (the next step one node deeper down the search tree) we want to get rid of those clauses
that were not found in this current round. Such clauses have the search_flag set to true, see step 2. Those clauses
that shall stay have the flag set to false. For the next round we want it the other way around, so we flip - true
becomes false, false becomes true. The misc_flag is used to make sure that each search_flag is only flipped once. *)						
					List.iter
						(fun (_, entry_data, _) ->
							if (Indexing_misc.get_misc_flag entry_data) then
								(Indexing_misc.set_search_flag 
									entry_data 
									(not (Indexing_misc.get_search_flag entry_data));
								Indexing_misc.set_misc_flag entry_data false))	
						previous_results;
(* Now only such clauses have true search_flags which have been found for all previous search
literals. Keep in mind that we are moving along the search tree depth-first; when we move upward to a parent node
again to start searching down another child/branch, the search_flags of the parent node may have to be reset to the state
when we started the previous search.
Now we split the result into result clusters according to the substitutions. All result clauses that were
found for the same substitution are put into one result cluster. *)						
						List.iter
							(fun result ->
								match result with
									| (data, entry_data, subst) ->
										(if (subst == !previous_subst) then
											(subst_cluster := result :: !subst_cluster)
										else
											(subst_clusters := 
												(!previous_subst, !subst_cluster) :: !subst_clusters;
											previous_subst := subst;
											subst_cluster := [result])))
							literal_results;
						subst_clusters := 
							(!previous_subst, !subst_cluster) :: !subst_clusters;
(* Finally, we perform a search for each cluster. *)
						let result_datas = ref []
						in				
						List.iter
							(fun (subst, subst_results) -> 
								match subst_results with
									| [] -> ()
									| _ ->
										(try
											(List.iter
												(fun (_, entry_data, _) ->
													Indexing_misc.set_search_flag entry_data true)
												subst_results;
											let subst_result_datas =
												find_clauses'
													literals
													subst
													subst_results
											in
											List.iter
												(fun (_, entry_data, _) ->
													Indexing_misc.set_search_flag entry_data false)
												subst_results;
											List.iter
												(fun data ->
													if (List.memq data !result_datas) then
														()
													else
														(result_datas := data :: !result_datas))
												subst_result_datas)
										with
											| FOUND ->
												(List.iter
													(fun (_, entry_data, _) ->
														Indexing_misc.set_search_flag entry_data false)
													subst_results;
												raise FOUND)))
							!subst_clusters;
						!result_datas))
						
						
	and search_filter (results: ('data * entry_data * Subst.subst) list)
	: ('data * entry_data * Subst.subst) list =
(* dumb stub, for now *)
		results
		
	in
(* now the initial call, similar, but taking into account that nothing has been searched yet *)
		match clause with
			| [] -> 
				([])
			| literal :: literals ->
				(let previous_subst = ref Subst.empty
				and subst_cluster = ref []
				and subst_clusters = ref []
				in					(*pl literal;pe " <FINDFIRST ";*)
				(* find all clauses for this literal *)					
				let literal_results =						
					(let index = _self#find literal
					in
						find_clauses_in_index 
							index
							Subst.empty 
							literal.Term.atom)
				in	
(* preliminary filtering: no use to keep results in the loop that are guaranteed to fail the final tests *)				
				let literal_results =
					search_filter literal_results
				in																															
					List.iter
						(fun result -> 
							match result with
								| (data, entry_data, subst) ->
									(	(*pc (Indexing_misc.get_clause entry_data);pe "";psubst subst;pe "";*)
									if (subst == !previous_subst) then
										(subst_cluster := result :: !subst_cluster)
									else
										(subst_clusters := 
											(!previous_subst, !subst_cluster) :: !subst_clusters;
										previous_subst := subst;
										subst_cluster := [result])))
						literal_results;
(* now search for the next literals *)						
					subst_clusters := 
						(!previous_subst, !subst_cluster) :: !subst_clusters;						
					let result_datas = ref []
					in	
					(try			
						(List.iter
							(fun (subst, subst_results) -> 
								match subst_results with
									| [] -> ()
									| _ ->
										(try
											(List.iter
												(fun (_, entry_data, _) ->	(* pc (Indexing_misc.get_clause entry_data);pe " ---"; *)
													Indexing_misc.set_search_flag entry_data true)
												subst_results;
											let subst_result_datas =
												find_clauses'
													literals
													subst
													subst_results
											in
											List.iter
												(fun (_, entry_data, _) ->
													Indexing_misc.set_search_flag entry_data false)
												subst_results;
											List.iter
												(fun data ->
													if (List.memq data !result_datas) then
														()
													else
														(result_datas := data :: !result_datas))
												subst_result_datas)
										with
											| FOUND ->
												(List.iter
													(fun (_, entry_data, _) ->
														Indexing_misc.set_search_flag entry_data false)
													subst_results;
												raise FOUND)))
							!subst_clusters)
						with
							| FOUND ->
								(let result =
									match _self#get_found with
										| Some data -> data
										| None -> failwith "Discrimination_tree.find_clauses: FOUND exception with no return data."
								in
									_self#set_found None;
									result_datas := [result]));
					!result_datas)						
						
									
							
										
	method find_clause_variants (clause: clause) : 'data list =
		_self#find_clauses
			find_variants_in_index
			variant_entry_check
			clause
			
	method find_clause_instances (clause: clause) : 'data list = 
		_self#find_clauses
			find_instances_in_index
			instance_entry_check
			clause
			
	method find_clause_generalizations (clause: clause) : 'data list =
		_self#find_clauses
			find_generalizations_in_index
			generalization_entry_check
			clause			
			
	method find_clause_unifiables (clause: clause) : 'data list =
		_self#find_clauses
			find_unifiables_in_index
			unifiable_entry_check
			clause
			
	method find_first_clause_variant (clause: clause) : 'data =
		let results =
			_self#find_clauses
				~find_first_only:true
				find_variants_in_index
				variant_entry_check
				clause
		in
			(match results with
				| [] -> raise Not_found
				| [data] -> data
				| _ -> (failwith "Discrimination_tree.find_first_clause_variant: multiple results."))
	
	method find_first_clause_instance (clause: clause) : 'data =
		let results =
			_self#find_clauses
				~find_first_only:true
				find_instances_in_index
				instance_entry_check
				clause
		in
			(match results with
				| [] -> raise Not_found
				| [data] -> data
				| _ -> failwith "Discrimination_tree.find_first_clause_instance: multiple results.")	
				
	method find_first_clause_generalization (clause: clause) : 'data =
		let results =
			_self#find_clauses
				~find_first_only:true
				find_generalizations_in_index
				generalization_entry_check
				clause
		in
			(match results with
				| [] -> raise Not_found
				| [data] -> data
				| _ -> failwith "Discrimination_tree.find_first_clause_generalization: multiple results.")							
	
	method find_first_clause_unifiable (clause: clause) : 'data =
		let results =
			_self#find_clauses
				~find_first_only:true
				find_unifiables_in_index
				unifiable_entry_check
				clause
		in
			(match results with
				| [] -> raise Not_found
				| [data] -> data
				| _ -> failwith "Discrimination_tree.find_first_clause_unifiable: multiple results.")					
							
							
end


(*** public interface ***)

let create_predicate_index (productivity: bool) (data: 'data data)  : 'data Term_indexing.predicate_index =
  new predicate_index productivity data

let create_index (productivity: bool) (data: 'data data) : 'data Term_indexing.index = 
  new index productivity data

let create_subterm_index (productivity: bool) (data: 'data data) : 'data Term_indexing.symbol_index = 
  new symbol_index productivity data

(*  
let create_perfect_clause_index (data: 'data data) : 'data Term_indexing.perfect_term_index =
  new perfect_term_index false data
*)
let create_perfect_predicate_index (productivity: bool) (data: 'data data)  : 'data Term_indexing.perfect_predicate_index =
  new perfect_predicate_index productivity data
    
let create_perfect_clause_index (data: 'data data) : 'data Term_indexing.perfect_clause_index =
  new perfect_clause_index false data
  
(*** specialized term indexes ***)

(* int index *)
class int_data =
object
  method is_equal (x: int) (y: int) = x == y
  method to_string x = string_of_int x
end

let create_int_index (productivity: bool) : int index =
  new index productivity (new int_data)


(* term index *)
class term_data =
object
  method is_equal x y = Term.term_equal x y
  method to_string x = Term.term_to_string x
end

let create_term_index (productivity: bool) : term index =
  new index productivity (new term_data)


(* literal index *)
class literal_data =
object
  method is_equal x y = Term.literal_equal x y
  method to_string x = Term.literal_to_string x
end

let create_literal_index (productivity: bool) : literal index =
  new index productivity (new literal_data)


(* clause index *)
class clause_data =
object
  method is_equal x y = Term.clause_equal x y
  method to_string x = Term.clause_to_string x
end

let create_clause_index (productivity: bool) : clause index =
  new index productivity (new clause_data)



(* literal and clause index *)
class literal_clause_data =
object
  method is_equal (l1, c1) (l2, c2) =
    Term.literal_equal l1 l2 && Term.clause_equal c1 c2

  method to_string (l, c) =
    Term.literal_to_string l ^ " : " ^ Term.clause_to_string c
end

let create_literal_clause_index (productivity: bool) : (literal * clause) index =
  new index productivity (new literal_clause_data)

