(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005, 2006
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Bp

type term = Term.term
type symbol = Symbol.symbol 
type clause = Term.clause
type literal = Term.literal
type problem_literals = Problem_literals.problem_literals
type position = Term.position
type input_partner = Context_unifier.input_partner
type rewrite_partner = Context_unifier.rewrite_partner
type context_rewrite_partner = Context_unifier.context_rewrite_partner
type context_unifier_space = Context_unifier.context_unifier_space
type 'data stack = 'data Stack.stack
type context = Context.context
type state = State.state
type config = Config.config
type subst = Subst.subst


type rewriting =
	{rw_space: context_unifier_space;
	rw_element: Context.element;
	rw_matcher: term;
	rw_rewriter: term;
	rw_position: position;
	rw_literal_index: int;
	rw_result: clause}

let null_rewriting =
	{rw_space = Context_unifier.null_space;
	rw_element = Context.null_element;
	rw_matcher = Term.null_term;
	rw_rewriter = Term.null_term;
	rw_position = Term.top_position;
	rw_literal_index = (-1);
	rw_result = Term.true_clause}

(* the clauses using this rewrite partner,
   i.e. containing this subterm at the given position *)
type element = {
  (* the clause subterm with its position *)
  rewrite_partner: rewrite_partner;

  (* the clauses containing this rewrite partner *)
  mutable spaces: (context_unifier_space * int) list;
}

(* map from offset to element *)
module Elements =
  Map.Make (
    struct
      type t = int

      let compare (first: int) (second: int) : int =
		Tools.compare_int first second
    end
  )

(* accessing clauses by subterm position *)  
module LiteralPositionTable =
  Hashtbl.Make (
    struct
      type t = literal * position
	  
      let equal x y = (x = y)
 (*     let equal (x_lit, x_pos) (y_lit, y_pos) =
      	(x_lit == y_lit) && (x_pos == y_pos)	(* too strict? *)*)
      let hash = Hashtbl.hash
    end
  )  
  
  
(* a subterm - this is just a storage record used during creation, wrapping the rewrite partner
for quick access *)
type problem_subterm = {
  (* the problem / clause subterm *)
  term: term;
  
  (* the leading symbol of the term *)
  symbol: symbol;

  (* each element represents a literal with the [term] as a subterm at a specific position *)
  mutable elements: element LiteralPositionTable.t;
  
  (* provides access to rewriting equations in the context *)
(*  rewrite_partner: rewrite_partner;	*)
}

(* problem subterm accessed by subterm. *)
module SubtermTable = Term.TermTable

(* problem literal accessed by literal. *)
module LiteralTable = Term.LiteralTable





	


(* (bunch of) problem subterms - predicate indexing is reused here as subterm indexing,
with accessing by the top subterm symbol (predicate symbols are essentially just term symbols,
after all, so there is no difference in the internal handling) *)

class problem_subterm_data =
object
  method is_equal (first: problem_subterm) (second: problem_subterm) : bool =
    Term.term_equal first.term second.term
      
  method to_string (data: problem_subterm) : string =
    "problem_subterm: " ^ Term.term_to_string data.term
end

type subterm_index = problem_subterm Term_indexing.term_index


type problem_subterms = {
  (* environment *)
  config: config;
  state: state;
  context: context;
  
  (* access to problem subterms by term - used during creation *)
  problem_subterms: problem_subterm SubtermTable.t;

  (* access to problem subterms by top symbol and term *)
  index: problem_subterm Term_indexing.symbol_index;
  
  
}

let get_config (problem_subterms: problem_subterms) : config =
	problem_subterms.config

let create_rewrite_partner 
	(problem_subterms: problem_subterms) 
	(index: int) 
	(literal: literal) 
	(subterm: term) 
	(position: position)
	(smaller: bool) : rewrite_partner =
  let vars =
    List.map
      (fun var ->
	 Subst.make_var var Subst.input_literal_offset
      )
      (Term.vars_of_literal literal)
  in
  let rewrite_partner =
    Context_unifier.create_rewrite_partner
      index literal subterm position smaller vars
  in
(*    create_initial_context_rewrite_partners problem_subterms rewrite_partner;*)
    rewrite_partner




(* create a problem_subterms record *)
let create (config: config) (context: context) (state: state) : problem_subterms =
	{
		config = config;
		state = state;
		context = context;
		problem_subterms = SubtermTable.create 16;
		index = Discrimination_tree.create_subterm_index false (new problem_subterm_data);
	}

(* add the problem literal's subterms to the index *)

let register_to_index (problem_subterms: problem_subterms) (problem_subterm: problem_subterm) : unit =
  let index =
    problem_subterms.index#find problem_subterm.symbol
  in	
    index#add problem_subterm.term problem_subterm

   
(* combination of several preprocessing steps for a clause *)    
exception TAUTOLOGY  
exception EMPTY_CLAUSE  
exception FAILED_PARAMODULATION
    
let optimize_clause (clause: clause) : clause =
	if Term.is_tautology clause then
		(raise TAUTOLOGY)
	else
		(if (List.exists
				(fun literal ->
					literal.Term.is_constraint
					&&
					(Term.is_trivial_eq literal.Term.atom))
				clause) then
			(raise TAUTOLOGY); (* not taut in the strictest sense, but can be dismissed like one *)
		let clause =
			Term.remove_duplicates clause
		in
		let clause =
			Term.sort_clause clause
		in
		let clause =
			Subst.normalize_clause clause
		in 
			clause)	    


		
		
let rec reflexivity_processing (clause: clause) : clause list =
	let rec reflexivity_processing' (done_literals: literal list) (to_do_literals: literal list) 
	: clause list =
		(match to_do_literals with
			| literal :: remaining_literals ->
				(if literal.Term.is_constraint || literal.Term.sign then
					(reflexivity_processing' (literal :: done_literals) remaining_literals)
				else
					(match literal.Term.atom with
						| Term.Eq eq ->
							(try
								(let unifier =
									Unification.unify_terms 
									~recompute:false
									eq.Term.left
									0
									eq.Term.right
									0
								in
								let shortened_clause =
									List.rev_append
										done_literals
										remaining_literals
								in
								match shortened_clause with
									| [] -> raise EMPTY_CLAUSE
									| _ ->
										(try
											(let reflexivity_clause =
												optimize_clause
													(Subst.apply_to_clause
														unifier
														shortened_clause
														0)
											in
												List.rev_append
													(reflexivity_clause :: (reflexivity_processing reflexivity_clause))
													(reflexivity_processing' 
														(literal :: done_literals) 
														remaining_literals))
										with
											| TAUTOLOGY ->
												(reflexivity_processing' 
													(literal :: done_literals) 
													remaining_literals)))
							with
								| FAILED_PARAMODULATION
								| Unification.UNIFICATION_FAIL ->
					(* the equation cannot be unified away, so continue with the other literals *) 
									(reflexivity_processing' (literal :: done_literals) remaining_literals))
						| _ ->
							(reflexivity_processing' (literal :: done_literals) remaining_literals)))
			| [] ->
				[])
	in
		reflexivity_processing' [] clause
			
			
			
					

let compute_initial_context_rewrite_partners 
	(problem_subterms: problem_subterms) (rewrite_partner: rewrite_partner) : unit = 
	let candidate_equations =
		(let unify_wrapped_term =
			(Term.request_eq 
				~insert_db:false 
				(rewrite_partner.Context_unifier.rwp_subterm, (Term.request_var (Var.max_universal))))
		in		
			Context.find_all_unifiable_entry_data
				problem_subterms.context
				(Term.request_literal
					~insert_db:false
					true
					unify_wrapped_term))
	and offset = 
		Config.variable_renaming_offset_for_paramodulation problem_subterms.config
	in	
		List.iter
			(fun (element, entry_data) -> 
				try
					(let matcher, rewriter =
						(let left, right, oriented =
							(match element.Context.el_literal.Term.atom with
								Term.Eq eq ->
									eq.Term.left,
									eq.Term.right,
									eq.Term.oriented
								|_ -> failwith "Problem_subterms.compute_initial_context_rewrite_partners: bad equation in index.")
						in
							if (Indexing_misc.get_flipped entry_data) then
								(if oriented then
									(raise FAILED_PARAMODULATION)
								else
									(right, left))
							else
								(left, right))
					in
(* ensure matcher and rewriter are parameter-free and unlikely to share variables with other terms *)					
					let matcher, mapping =
						Term.request_universal_term_with_mapping 
							~offset:offset 
							matcher
					in
					let rewriter, _ = 
						Term.request_universal_term_with_mapping 
							~offset:offset 
							~mapping:mapping
							rewriter
					in	
						let sigma =
(* no need to catch a unification failure here - if these terms do not unify, then there is a bug in the index,
so we want to know about the failure
No, wait, that is not true. This is not a perfect index, so the results will not necessarily unify. *)						
							Unification.unify_terms 
								~recompute:false
								matcher
								0
								rewrite_partner.Context_unifier.rwp_subterm
								1
						in	
							let matcher_sigma, rewriter_sigma =
								(Subst.apply_to_term ~insert_db:true ~normalize:true sigma matcher 0),
								(Subst.apply_to_term ~insert_db:true ~normalize:true sigma rewriter 0)
							in
								if (Term.term_equal rewriter_sigma matcher_sigma)
									||
									(Term.term_ordering_greater rewriter_sigma matcher_sigma)
									then (raise FAILED_PARAMODULATION);
								let context_rewrite_partner = 
									Context_unifier.create_context_rewrite_partner
										element
										matcher
										rewriter
										sigma
										(not (Indexing_misc.get_flipped entry_data))
								in	
									Stack.push 
										rewrite_partner.Context_unifier.rwp_context_rewrite_partners
										context_rewrite_partner
										
										
				)
				with	
					| FAILED_PARAMODULATION 
					| Unification.UNIFICATION_FAIL -> ()
					|_ -> failwith "Problem_subterms.compute_initial_context_rewrite_partners"			
			)				
			candidate_equations
		
	

			
let set_initial_context_rewrite_partners 
	(problem_subterms: problem_subterms) 
	(rewrite_partner: rewrite_partner) : unit =			
	compute_initial_context_rewrite_partners problem_subterms rewrite_partner
	

	
let get_problem_subterm (problem_subterms: problem_subterms) (subterm: term) : problem_subterm =
  try
    (SubtermTable.find problem_subterms.problem_subterms subterm)
  with Not_found ->
	(* create a new problem subterm
	   and update index structures to refer to it *)
	let new_problem_subterm = {
	  term = subterm;
	  symbol = Term.top_symbol_term subterm;
	  elements = LiteralPositionTable.create 16;
	}
	in

	  (* add new problem subterm to indexes *)
	  SubtermTable.add problem_subterms.problem_subterms subterm new_problem_subterm;
	  register_to_index problem_subterms new_problem_subterm;
	  new_problem_subterm		


				

let register_rewrite_partner
	(problem_subterms: problem_subterms)
	(rewrite_partner: rewrite_partner)
	(index: int)
	(position: position)
	(space: context_unifier_space) : unit = 
	let subterm =
		rewrite_partner.Context_unifier.rwp_subterm
	and
		literal =
		rewrite_partner.Context_unifier.rwp_literal
	in		
	let problem_subterm =
		(try 
			(SubtermTable.find problem_subterms.problem_subterms subterm)
		with
			| Not_found ->
				failwith "Problem_subterms.register_rewrite_partner: problem_subterm")
	in
	let element =
		(try
			(LiteralPositionTable.find problem_subterm.elements (literal, position))
		with
			| Not_found ->
				failwith "Problem_subterms.register_rewrite_partner: problem_subterm.element")
	in	
		element.spaces <- (space, index) :: element.spaces
		
		


let register_literal_subterms
    (problem_subterms: problem_subterms) (literal: literal) (index: int) : (rewrite_partner * int * position) list =
    if literal.Term.is_constraint then 
    	([])	(* no paramodulation in constraints *)
    else
	    (let registered_problem_subterms = ref []
	    and get_element problem_subterm term position smaller =
	    	  try
				(LiteralPositionTable.find problem_subterm.elements (literal, position))
			with Not_found ->
				(let new_rewrite_partner = create_rewrite_partner problem_subterms index literal term position smaller
				in
				let new_element =
					{rewrite_partner = new_rewrite_partner;
					spaces = []}
				in
					set_initial_context_rewrite_partners 
						problem_subterms 
						new_rewrite_partner;
					LiteralPositionTable.add problem_subterm.elements (literal, position) new_element;
					new_element)			
	    in
		let rec register_subterms (term: term) (smaller: bool) (index_list: int list) : unit =
			(match term with
				Term.Var _ -> ()
				| Term.Const _ ->
					let new_problem_subterm = get_problem_subterm problem_subterms term
					and position = Term.make_position (List.rev index_list)
					in
						let element =
							get_element new_problem_subterm term position smaller
						in
							registered_problem_subterms := 
								(element.rewrite_partner, index, position) 
								:: !registered_problem_subterms
				| Term.Func func ->
					let new_problem_subterm = get_problem_subterm problem_subterms term
					and position = Term.make_position (List.rev index_list)
					in
						let element =
							get_element new_problem_subterm term position smaller
						in
							registered_problem_subterms := 
								(element.rewrite_partner, index, position) 
								:: !registered_problem_subterms;
							Array.iteri
								(fun subindex subterm -> 
									register_subterms subterm smaller (subindex :: index_list))
								func.Term.subterms
				| Term.Eq eq ->
					failwith "Problem_subterms.register_literal_subterms: bad literal")
		 in
		 	(match literal.Term.atom with
		 		Term.Var _ -> failwith "Problem_subterms.register_literal_subterms: bad literal"
		 		| Term.Const _ -> ()	(* no rewriting in nullary predicate literals *)
		 		| Term.Func func -> 
		 				Array.iteri
							(fun subindex subterm -> 
								register_subterms subterm false [subindex])
							func.Term.subterms
				| Term.Eq eq ->
						register_subterms eq.Term.left false [0];
(* do not rewrite right side if it is smaller and Superposition is used *)						
						if (not eq.Term.oriented)
							|| 
							(match (Config.rewriting problem_subterms.config) with
								| Flags.RW_Paramodulation -> true
								| _ -> false) then 
							(register_subterms eq.Term.right eq.Term.oriented [1]));
			!registered_problem_subterms)
			
let paramodulate_clause 
		(space: context_unifier_space) 
		(index: int)
		(rewritten_literal: literal)
		?(constraint_literal: literal = Term.null_literal)
		(unifier: subst)
		: clause =	
	let i = ref (-1)
	in	
	let rewritten_cus_clause =
		(List.map 
			(fun literal ->
				i := !i+1;
				if (!i = index) then
					(rewritten_literal)
				else
					(Subst.apply_to_literal
						~insert_db:true
						~normalize:false
						unifier
						literal
						0))
			space.Context_unifier.cus_clause)
	in
	let rewritten_cus_clause =
		if (Term.is_trivial_eq constraint_literal.Term.atom)
			|| (constraint_literal == Term.null_literal) then
			(rewritten_cus_clause)
		else	
			(constraint_literal :: rewritten_cus_clause)
	in 	 (*pc space.Context_unifier.cus_clause; ps " INTO ";pc rewritten_cus_clause;pe "";*)
		optimize_clause rewritten_cus_clause
		
		
let get_paramodulation_unifier 
	(matcher: term)
	(rewriter: term) 
	(term_to_rewrite: term) : Subst.subst = 
	let unifier = 
		(Unification.unify_terms 
			~recompute:false
			matcher
			0
			term_to_rewrite
			0)
	in 
	let unified_matcher = 
		(Subst.apply_to_term
			~insert_db:true
			~normalize:false
			unifier
			matcher
			0)
	and unified_rewriter =
		(Subst.apply_to_term
			~insert_db:true
			~normalize:false
			unifier
			rewriter
			0)
	in		(* pt unified_matcher;pe " <UNIFIEDMATCHER";pt unified_rewriter;pe " <UNIFIEDREWRITER"; *)
		if (Term.term_equal unified_matcher unified_rewriter)
		|| (Term.term_ordering_greater_var_checked unified_rewriter unified_matcher)
		then 
			(raise FAILED_PARAMODULATION)
		else
			(unifier)	
		
(* creation of a constraint literal *)					
let create_constraint_literal (matcher: term) (rewriter: term) (sigma: Subst.subst) : Term.literal =
	let matcher_sigma =
		Subst.apply_to_term
			~insert_db:true
			~normalize:false
			sigma
			matcher
			0
	and rewriter_sigma =
		Subst.apply_to_term
			~insert_db:true
			~normalize:false
			sigma
			rewriter
			0	
	in		
	let context_equation_sigma =
		Term.request_eq
			~insert_db:true
			(matcher_sigma, rewriter_sigma)
	in
		Term.request_literal
			~insert_db:true
			~is_constraint:true
			false
			context_equation_sigma
				
		
let add_to_problem_subterm
		?(recomputing: bool = false)
(* if true, then we are only recomputing the rewritten clauses result, but the
context_rewrite_partner is not added to the subterm *)		
		(problem_subterms: problem_subterms)
		(problem_subterm: problem_subterm)
		(context_element: Context.element)
		(context_rewrite_partner: context_rewrite_partner)
		: (rewriting list) =	(*pl context_rewrite_partner.Context_unifier.crwp_element.Context.el_literal;pe "f!";*)
	let matcher =
		ref context_rewrite_partner.Context_unifier.crwp_matcher
	and rewriter =
		ref context_rewrite_partner.Context_unifier.crwp_rewriter
	and rewritten = ref []
	in	
	(try	
		(let unifier = 
			ref
				(get_paramodulation_unifier
					!matcher
					!rewriter
					problem_subterm.term)
		and min_var_num =
			ref context_rewrite_partner.Context_unifier.crwp_min_var_num
	
		in	(*pl context_rewrite_partner.Context_unifier.crwp_element.Context.el_literal;pe "!";*)
(* create the constraint literal *)		
		let constraint_literal =
			if (Config.resolve problem_subterms.config)
			&& (context_rewrite_partner.Context_unifier.crwp_universal) then
				(ref Term.null_literal)
			else
				(ref
					(create_constraint_literal
						!matcher
						!rewriter
						!unifier))
		in	
	(*	pl !constraint_literal;ps " FROM ";pl context_rewrite_partner.Context_unifier.crwp_element.Context.el_literal;pe "";
		pt context_rewrite_partner.Context_unifier.crwp_matcher;pe " <MATCHER";
		pt context_rewrite_partner.Context_unifier.crwp_rewriter;pe " <REWRITER";	*)
(* paramodulation inferences *)		
		LiteralPositionTable.iter
			(fun (literal, position) element ->	
				if (match (Config.rewriting problem_subterms.config) with
						| Flags.RW_Limited -> true
						| _ -> false)
				&& context_rewrite_partner.Context_unifier.crwp_universal 
				&& element.rewrite_partner.Context_unifier.rwp_smaller then 
					()
				else
					(if not recomputing then
						(Stack.push 
							element.rewrite_partner.Context_unifier.rwp_context_rewrite_partners 
							context_rewrite_partner);
					let rewritten_term = Term.set_position literal.Term.atom position !rewriter
					in
					let rewritten_term = Term.recursive_request_term rewritten_term 
					in
					let proto_rewritten_literal =
						(Term.request_literal
							~insert_db:true
							literal.Term.sign
							rewritten_term)
					in
					let rewritten_literal =
						ref
							(Subst.apply_to_literal
								~insert_db:true
								~normalize:false
								!unifier
								proto_rewritten_literal
								0)
					in		
						(List.iter
							(fun (space, index) -> 
(* spaces may become redundant during this loop, so check for each if it is actually necessary to
still consider it for paramodulation *)						
								if (State.is_explanation_invalid
									!(space.Context_unifier.cus_redundancy_explanation)) then
								(try
									(if !min_var_num <= space.Context_unifier.cus_max_var_num then
(* Uh-oh, the space has more variables than we are prepared for, this could lead to
variable capturing by paramodulation. The paramodulating terms will have to have their variables
renamed into a safe higher namespace, and as a result the literal and the unifier will have to 
change, too. All this should rarely happen, see Config.variable_renaming_offset_for_paramodulation . *)						
										(let offset = space.Context_unifier.cus_max_var_num - !min_var_num + 1
										in
										min_var_num := space.Context_unifier.cus_max_var_num + 1;
										let renamed_matcher, mapping =
											Term.request_universal_term_with_mapping 
												~offset:offset 
												!matcher
										in
										let renamed_rewriter, _ =
											Term.request_universal_term_with_mapping 
												~offset:offset 
												~mapping:mapping 
												!rewriter
										in
										matcher := renamed_matcher;
										rewriter := renamed_rewriter;
										unifier :=
											(get_paramodulation_unifier
												!matcher
												!rewriter
												problem_subterm.term);
										if !constraint_literal == Term.null_literal then
											()
										else
											(constraint_literal :=
												create_constraint_literal
													!matcher
													!rewriter
													!unifier);
										let proto_rewritten_term = 
											Term.set_position literal.Term.atom position !rewriter
										in
										let rewritten_term = 
											Term.recursive_request_term proto_rewritten_term 
										in
										let proto_rewritten_literal =
											(Term.request_literal
												~insert_db:true
												literal.Term.sign
												rewritten_term)
										in
										rewritten_literal :=
											(Subst.apply_to_literal
												~insert_db:true
												~normalize:false
												!unifier
												proto_rewritten_literal
												0));
(* now it is safe to paramodulate *)

									let rewritten_clause =																	
										paramodulate_clause 
											space 
											index
											!rewritten_literal
											~constraint_literal:(!constraint_literal)
											!unifier
										in
										let rewriting =
											{rw_space = space;
											rw_element = context_element;
											rw_matcher = context_rewrite_partner.Context_unifier.crwp_matcher;
											rw_rewriter = context_rewrite_partner.Context_unifier.crwp_rewriter;
											rw_position = position;
											rw_literal_index = index;
											rw_result = rewritten_clause}
										in
										rewritten :=
												rewriting :: !rewritten)
								with
									| FAILED_PARAMODULATION
									| TAUTOLOGY -> ()))
						element.spaces)))
			problem_subterm.elements)
	with
		| FAILED_PARAMODULATION -> ());
	!rewritten
		
	
	
let compute_for_element (problem_subterms: problem_subterms) (context_element: Context.element)
	: (rewriting list) =
	let rewritten = ref []
	in
	let compute_for_element_side side =
		let unifying_candidates =
			let index =
				problem_subterms.index#find (Term.top_symbol_term side)
			in	
				index#find_all_unifiable_subst ~p_preserving:false side
		in		
			List.iter
				(fun (problem_subterm, _) ->
					LiteralPositionTable.iter
						(fun _ element ->
							try
								(let context_rewrite_partner =
									Stack.find
										element.rewrite_partner.Context_unifier.rwp_context_rewrite_partners
										(fun x ->
											compare
											context_element.Context.el_id
											x.Context_unifier.crwp_element.Context.el_id)
								in
								let results =
									add_to_problem_subterm
										~recomputing:true
										problem_subterms
										problem_subterm
										context_element
										context_rewrite_partner
								in
									rewritten :=
										List.rev_append
											results
											!rewritten)	
							with
								| Not_found -> 
									())
						problem_subterm.elements)
				unifying_candidates	
	in 
	(try
		(let equation =
			(match context_element.Context.el_literal.Term.atom with
				| Term.Eq eq -> eq
				| _ -> raise FAILED_PARAMODULATION)
		in
			compute_for_element_side equation.Term.left;
			if (not (equation.Term.oriented)) then
				(compute_for_element_side equation.Term.right))
	with
		| FAILED_PARAMODULATION -> ());
	!rewritten
			
		
(* add a new equational context literal and compute all superposition unifiers *)
let add 
	(problem_subterms: problem_subterms)
	(context_element: Context.element)
	(candidate_type: Context_unifier.candidate_type) : (rewriting list) = 
	let rewritten = ref []
	in 		
	(try
		(let equation =
(* only positive non-trivial equational units may paramodulate *)		
			if context_element.Context.el_literal.Term.sign then
				(match context_element.Context.el_literal.Term.atom with
					| Term.Eq eq -> 
						(if (Term.term_equal eq.Term.left eq.Term.right) then
(* tautological cases, useless for paramodulation *)
							(raise FAILED_PARAMODULATION)
						else
							(eq))
					| _ -> raise FAILED_PARAMODULATION)
			else
				(raise FAILED_PARAMODULATION)
		in	
		let left =
			equation.Term.left
		and right =
			equation.Term.right
		and offset = Config.variable_renaming_offset_for_paramodulation problem_subterms.config
		in		
(* ensure left and right sides are parameter-free for paramodulation, and that they are unlikely
to share variables with any subterms from clauses by adding an offset to the variables - in specific
cases when very large clauses are about to be rewritten this might be insufficient, but we check
for this and do another renaming if it is absolutely necessary *)		
		let left_for_paramodulation, mapping = 
			Term.request_universal_term_with_mapping 
				~offset:offset 
				left
		in
		let right_for_paramodulation, _ =
			Term.request_universal_term_with_mapping 
				~mapping:mapping 
				~offset:offset 
				right
		in

		if not (Term.is_term_var left) then
(* Only bother with paramodulations from left to right when left is no variable.
If left is a variable, then so is right (due to the orientation algorithm which 
moves variables to the right first) - in that case it is sufficient to do the
paramodulation from right to left, since it is actually the same (X=Y is Y=X after
renaming, the direction does not matter). *)		
			(let context_rewrite_partner =
				Context_unifier.create_context_rewrite_partner
					context_element
					left_for_paramodulation
					right_for_paramodulation
					Subst.empty
					false
			in	
			
			let unifying_candidates =
				(let index =
					problem_subterms.index#find (Term.top_symbol_term left)
				in
					index#find_all_unifiable_subst ~p_preserving:false left)
			in		
				List.iter
					(fun (problem_subterm, subst) -> 
						rewritten :=
							List.rev_append 
								(add_to_problem_subterm
									problem_subterms
									problem_subterm
									context_element
									context_rewrite_partner)
								!rewritten)
					unifying_candidates);

	(* if the context equation is not orientable, then the other direction has to be
	processed as well *)						
			if (not (equation.Term.oriented)) then
				(let context_rewrite_partner_rl =
					Context_unifier.create_context_rewrite_partner
						context_element
						right_for_paramodulation
						left_for_paramodulation
						Subst.empty
						true
				in	
				let unifying_candidates =
					(match right with
						| Term.Var var ->
							(let unifying_candidates = ref []
							in
								SubtermTable.iter
									(fun term problem_subterm ->
										unifying_candidates :=
											(problem_subterm, Subst.empty) :: !unifying_candidates)
									problem_subterms.problem_subterms;
								!unifying_candidates)
						|_ ->
							(let index =
									problem_subterms.index#find (Term.top_symbol_term right)
								in	
									index#find_all_unifiable_subst ~p_preserving:false right))
				in 
					List.iter
						(fun (problem_subterm, subst) -> 
							rewritten :=
								List.rev_append 
									(add_to_problem_subterm
										problem_subterms
										problem_subterm
										context_element
										context_rewrite_partner_rl)
									!rewritten)
						unifying_candidates))
	with FAILED_PARAMODULATION -> ());
	!rewritten
		
		
(*** backtrack ***)

let backtrack (problem_subterms: problem_subterms) : unit =
	(* backtrack each problem subterm *)
	problem_subterms.index#iter
    	(fun _ index ->
			index#iter
	   			(fun _ problem_subterm ->
	   				LiteralPositionTable.iter
	   					(fun _ element ->
	   						let rec remove_invalid_context_rewrite_partner () =
		   						if Stack.size element.rewrite_partner.Context_unifier.rwp_context_rewrite_partners = 0 then
		   							()
		   						else
		   							(let context_rewrite_partner =
		   								Stack.top element.rewrite_partner.Context_unifier.rwp_context_rewrite_partners
		   							in
		   								if (State.is_choice_point_invalid 
		   									context_rewrite_partner.Context_unifier.crwp_element.Context.el_choice_point) then
		   									(Stack.remove_top element.rewrite_partner.Context_unifier.rwp_context_rewrite_partners;
		   									remove_invalid_context_rewrite_partner ()))
		   					in
		   						remove_invalid_context_rewrite_partner ())
	   					problem_subterm.elements))
	   					
	   					
   					
	   					
	   					
let unregister_space (problem_subterms: problem_subterms) (space: context_unifier_space) : unit =
	Array.iter
		(fun (rewrite_partner, index, position) ->
				let problem_subterm =
					try
						(SubtermTable.find 
							problem_subterms.problem_subterms 
							rewrite_partner.Context_unifier.rwp_subterm)
					with
						| Not_found ->
							(failwith "Problem_subterms.unregister_space: missing problem_subterm.")
				and literal = rewrite_partner.Context_unifier.rwp_literal
				in		(*pl literal;ps " <-- ";pt (Term.get_position position);ps " ";pi (Term.get_position_index position);pe "";*)
				let element =
					try
						(LiteralPositionTable.find
							problem_subterm.elements
							(literal, position))
					with
						| Not_found ->
							(failwith "Problem_subterms.unregister_space: missing element.")
				in
	 (* remove space from rewrite partner *)			
					element.spaces <-
						List.find_all
							(fun (space', _) ->
								space' != space)
							element.spaces;			
	 (* remove rewrite_partner/element if no space left *)							
					(match element.spaces with
						| [] ->
							LiteralPositionTable.remove
								problem_subterm.elements
								(literal, position)
						| _ -> ());
					if (LiteralPositionTable.length problem_subterm.elements) = 0 then
						(SubtermTable.remove 
							problem_subterms.problem_subterms 
							rewrite_partner.Context_unifier.rwp_subterm;
						let index =
		     				problem_subterms.index#find (Term.top_symbol_term problem_subterm.term)
		   				in 
		     				ignore (index#remove problem_subterm.term (Some problem_subterm))))
		space.Context_unifier.cus_rewrite_partner_triples
		