(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005, 2006
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


type term = Term.term
type literal = Term.literal
type position = Term.position
type clause = Term.clause
type rewrite_partner = Context_unifier.rewrite_partner
type context_rewrite_partner = Context_unifier.context_rewrite_partner
type context_unifier_space = Context_unifier.context_unifier_space
type 'data stack = 'data Stack.stack
type config = Config.config
type context = Context.context
type state = State.state

type problem_subterms

val get_config : problem_subterms -> config

(** a container type for the results of rewriting inferences *)
type rewriting =
	{rw_space: context_unifier_space;
	(* the parent clause being rewritten *)
	
	rw_element: Context.element;
	(* the rewriting context element *)
	
	rw_matcher: term;
	(* the side of the rewriting context element that unified with a subterm in the clause *)
	
	rw_rewriter: term;
	(* the other side of the rewriting context element, which then replaced the subterm in the clause *)
	
	rw_position: position;
	(* the position of the rewritten subterm *)
	
	rw_literal_index: int;
	(* the index of the rewritten literal within the clause's list representation *)
	
	rw_result: clause
	(* the resulting rewritten clause *)
	}

(** a dummy [rewriting] *)
val null_rewriting: rewriting

type element = {
	rewrite_partner: rewrite_partner;
	mutable spaces: (context_unifier_space * int) list;
}

exception FAILED_PARAMODULATION
exception EMPTY_CLAUSE
exception TAUTOLOGY

(** {6 Functions} *)


(** [register_rewrite_partner problem_subterms rewrite_partner index position context_unifier_space]
    registers the [context_unifier_space] as containing the subterm [rewrite_partner],
    so that if a side of a context equation unifies with this subterm, then
    [context_unifier_space] may be rewritten by paramodulation.

    [input_partner] must have been previously retrieved
    with {!Problem_subterms.register_literal_subterms}.
*)
val register_rewrite_partner: problem_subterms -> rewrite_partner -> 
	int -> position -> context_unifier_space -> unit


(** registers all subterms of a literal to the index, and returns a list of triples: a [rewrite_partner] that
informs the clause when a subterm can be rewritten, and an integer and a [position] indicating the literal and
term position of the specific subterm occurrence within the clause. *)
val register_literal_subterms: problem_subterms -> literal -> int -> (rewrite_partner * int * position) list

(** {2 Creation} *)

(** creation of the [problem_subterms] data structure within a given environment. *)
val create: config -> context -> state -> problem_subterms


(** {2 Paramodulation} *)

(** adds a context element and computes all paramodulations *)
val add: problem_subterms -> Context.element -> Context_unifier.candidate_type 
-> (rewriting list)

(** computes the rewritten clauses for a context element (without adding the element) *)
val compute_for_element: problem_subterms -> Context.element -> rewriting list

(** tries to unify the matcher of the [context_rewrite_partner] with the [term]
and returns the unifier; raises FAILED_PARAMODULATION if unification is impossible *)
val get_paramodulation_unifier: term -> term -> term -> Subst.subst

(** [create_constraint_literal left right sigma] creates a constraint literal with the
equational sides [left] and [right] and the unifier [sigma] applied. *)
val create_constraint_literal: term -> term -> Subst.subst -> literal

(** [paramodulate_clause space index rewritten_literal constraint_literal unifier] takes the clause
represented by [space], replaces the literal at position [index] with [rewritten_literal],
applies [unifier] to the other literals, appends [constraint_literal] (unless it is 
[Term.null_literal]) and returns the resulting clause. *)
val paramodulate_clause: context_unifier_space -> int -> literal -> ?constraint_literal:literal -> Subst.subst -> clause

(** performs all Reflexivity inferences possible on a clause, and recursively for the Reflexivity
results as well. Returns a list of all the resulting clauses.
Raises [EMPTY_CLAUSE] when the empty clause is derived.
Some weaknesses that could/should be addressed:
 - there is no check whether the parent clause is made redundant by the result,
 - the resulting clauses are neither ordered nor normalized,
 - the result list may contain variants of the same clause.
  *)
val reflexivity_processing: clause -> clause list

(** {2 Backtracking} *)

(** removes all information based on retracted context literals. 

    {b DEPENDENCY}: {!Context.backtrack} must have been executed before.
*)
val backtrack: problem_subterms -> unit


(** removes a context_unifier_space *)
val unregister_space: problem_subterms -> context_unifier_space -> unit