(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005, 2006
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

open Bp

type config = Config.config
type bound = Bound.bound
type statistic = Statistic.statistic
type symbol = Symbol.symbol
type literal = Term.literal
type choice_point = State.choice_point
type explanation = State.explanation
type state = State.state
type raw_context_unifier = Context_unifier.raw_context_unifier
type problem_literals = Problem_literals.problem_literals
type problem_subterms = Problem_subterms.problem_subterms
type selected = Selection_types.selected
type selection_lookahead = Selection_lookahead.selection_lookahead
type context = Context.context
type clause = Term.clause
type element = Context.element
type candidate_type = State.literal_type
type context_unifier_space = Context_unifier.context_unifier_space
type rewriting = Problem_subterms.rewriting

type candidate = {
	ca_id: int;
	ca_type: candidate_type;
	ca_clause: clause;
	ca_generation: int;
	ca_element: element;
	ca_space: context_unifier_space;
	ca_rewriting: rewriting;
	ca_explanation: explanation;	
}

let candidate_to_string title candidate =
	let type_string =
		(match candidate.ca_type with
			| State.Paramodulation -> "Paramodulation"
			| State.Reflexivity -> "Reflexivity"
			| _ -> "Error")
	and generation_string = string_of_int candidate.ca_generation
	in
    (title ^ (Term.clause_to_string candidate.ca_clause) ^ "   " ^ type_string ^ "   " ^ generation_string)

class clause_data =
object
	method is_equal (first: clause) (second: clause) : bool =
		Term.clause_equal first second
			
	method to_string (data: clause) : string =
		"clause: " ^ Term.clause_to_string data
end    
	



let candidate_counter = ref (-1)

let null_candidate = {
	ca_clause = [];
	ca_type = State.Paramodulation; (* doesn't really matter which type... *)
	ca_id = (-2);
	ca_generation = (-1);
	ca_element = Context.null_element;
	ca_space = Context_unifier.null_space;
	ca_rewriting = Problem_subterms.null_rewriting;
	ca_explanation = State.valid_empty_explanation;
}


		

let compare_candidates (first: candidate) (second: candidate) : int =
	let first_clause = first.ca_clause
	and second_clause = second.ca_clause
	in
	if (Term.is_negative_clause first_clause) && (not (Term.is_negative_clause second_clause)) then
		(-1)
	else
		(if (Term.is_negative_clause second_clause) && (not (Term.is_negative_clause first_clause)) then
			(1)
		else
			(if (List.length first_clause) < (List.length second_clause) then
				(-1)
			else
				(if (List.length second_clause) < (List.length first_clause) then
					(1)
				else
					(if first.ca_generation < second.ca_generation then
	      				(-1)
	    			else 
	    				(if first.ca_generation > second.ca_generation then
	      					(1)
	      				else
							(if (first.ca_id < second.ca_id) then
								(-1)
							else
								(1)))))))
							
	
(* cache for potentially applicable candidates *)
module Valid =
  Heap.MinMaxHeap (
    struct
      type t = candidate

      let compare = compare_candidates
      let to_string = candidate_to_string ""
    end
  )	
		
(* for potentially applicable but forgotten candidates *)
(* access by candidate ordering *)
module ForgottenOrder =
  Set.Make (
    struct
      type t = (Context.element * candidate)

      let compare (_, candidate0) (_, candidate1) =
	  compare_candidates candidate0 candidate1
    end
  )

(* access by context element *)
module ForgottenBarrier =
  Hashtbl.Make (
    struct
      type t = Context.element

      let equal = Context.element_equal

      let hash element =
	element.Context.el_id
    end
  )  
  
  
  
type candidates = {
  	cd_state: state;
  	cd_statistic: statistic;
  	cd_config: config;
  	cd_bound: bound;
  	cd_context: context;
  	cd_problem_literals: problem_literals;
  	cd_problem_subterms: problem_subterms;
  	mutable cd_preselected: candidate option; 
	
(* Clauses that have not yet been turned into valid candidates. *)	
	cd_unprocessed: (clause * element * context_unifier_space * candidate_type * rewriting) array;
(* number of valid entries *)
    mutable cd_unprocessed_size: int;	  	
  	
(* The other candidate types use a heap for storage, but that requires a proper
comparison function, including an ordering of the candidates. But how to order
clauses? Maybe this can be changed to something more efficient once the candidate
type has become more complex and contains some quality/utility data that can actually be
ordered.  *)
	cd_valid: Valid.t;
	
(* candidates which are not applicable because of a context literal
added to the context at the stored choice point. *)
	mutable cd_invalid: (candidate * explanation) list;	
	
(* The index serves a similar purpose for clauses as the term table does for terms.
It does NOT keep track of which clauses are valid candidates, it only ensures object
identity of variant clauses. More specifically, if a new clause is about to be added
to the candidates, the index is checked whether a variant is already 'known' to the
system, and if so, then that variant is used instead. *)	
	cd_index: Term.clause Term_indexing.perfect_clause_index;
	
(* This index is shared with the Selection.selection that this candidates set belongs to. *)	
	cd_space_index: context_unifier_space Term_indexing.perfect_clause_index;
	
(*** forgotten & exceeding - analogous to Selection_assert.candidates ***)

  (* ordering over the barriers *)
	mutable cd_forgotten_order: ForgottenOrder.t;

  (* mapping from context element to its barrier *)
	cd_forgotten_barrier: candidate ForgottenBarrier.t;

  (* if the forgotten candidates of an element are to be recomputed,
     its former barrier is stored here while the recomputation takes place.
     i.e., all recomputed candidates below this barrier must be ignored. *)
	mutable cd_forgotten_recomputation_barrier: (Context.element * candidate) option;

  (* hash table for context elements by their id.

     for these context literals exceeding applicable candidates
     have been found and dropped during their initial computation.
     that is, when checking for a complete exhausted branch
     these have to be recomputed and rechecked. *)
	cd_exceeding_elements: (int, Context.element) Hashtbl.t;

  (* flag to denote that currently candidates are recomputed
     just to be checked for finding applicable exceeding ones. *)
	mutable cd_check_exceeding: bool;
	
}  



(*** exceeding candidates ***)

(* raised if the current checking of exceeding candidates
   already implies a restart with the next higher deepening bound.
   that is, further evaluation of exceeding candidates
   yields no further information and can be omitted. *)
exception EXCEEDING_EXHAUSTED


(* have exceeding applicable candidates been dropped for this context literal? *)
let is_marked_exceeding (candidates: candidates) (context_element: Context.element) : bool =
  try
    ignore (Hashtbl.find candidates.cd_exceeding_elements context_element.Context.el_id: Context.element);
    true
  with
    | Not_found ->
	false
	
	
(* an exceeding applicable candidate has been dropped for this context literal *)
let mark_exceeding (candidates: candidates) (context_element: Context.element) : unit =
  Hashtbl.add candidates.cd_exceeding_elements context_element.Context.el_id context_element
	
	
	
(* filler for the cd_unprocessed_array *)
let null_unprocessed =
	([], Context.null_element, Context_unifier.null_space, State.Reflexivity, Problem_subterms.null_rewriting)
	
	
let clear_unprocessed (candidates: candidates) : unit =
	for i = 0 to (candidates.cd_unprocessed_size - 1) do
		(candidates.cd_unprocessed.(i) <- null_unprocessed)
	done;
	candidates.cd_unprocessed_size <- 0
	
	
(* compute the quality of a clause based on the deepening bound criterion. *)
let get_clause_quality (candidates: candidates) (clause: clause) : int =
  if Config.iterative_deepening candidates.cd_config = Flags.IT_TermDepth then
    Term_attributes.weight_of_clause clause
  else
    Term_attributes.depth_of_clause clause	
  
let create (config: config) (bound: bound) (state: state) (context: context) (statistic: statistic)
	(problem_literals: problem_literals) (problem_subterms: problem_subterms) 
	(space_index: context_unifier_space Term_indexing.perfect_clause_index) 
	: candidates = 
	let unprocessed =
    	Array.make (Tools.max_int 1 Const.max_unprocessed_clause_candidates) null_unprocessed
  	in
{
	cd_state = state;
	cd_statistic = statistic;
	cd_config = config;
	cd_bound = bound;
	cd_context = context;
	cd_problem_literals = problem_literals;
	cd_problem_subterms = problem_subterms;
	cd_preselected = None;
	cd_unprocessed = unprocessed;
	cd_unprocessed_size = 0;
	cd_valid = Valid.create null_candidate;
	cd_invalid = [];
	cd_index = Discrimination_tree.create_perfect_clause_index (new clause_data);
	cd_space_index = space_index;
	cd_forgotten_order = ForgottenOrder.empty;
    cd_forgotten_barrier = ForgottenBarrier.create 64;
    cd_forgotten_recomputation_barrier = None;
    cd_exceeding_elements = Hashtbl.create 64;
    cd_check_exceeding = false;
}	  
  		

let check_subsumed_clause (candidates: candidates) (clause: clause) : explanation option =
	match (Context.check_subsumed_clause candidates.cd_context clause) with
		| Some element ->
			Some [element.Context.el_choice_point]
		| None ->
			(try
				(let space =
					(candidates.cd_space_index#find_first_clause_generalization clause)
				in
					Some space.Context_unifier.cus_explanation)
			with
				| Not_found ->
					None)

let build_candidate (candidates: candidates) (candidate_type: State.literal_type) 
	(element: element) (space: context_unifier_space) (clause: Term.clause) (rewriting: rewriting)
	: unit = 
	if element == Context.null_element then 
		()
	else
	(let id = !candidate_counter
	and generation = element.Context.el_generation
	in	
	let candidate = 
		{
			ca_clause = clause;
			ca_type = candidate_type;
			ca_id = id;
			ca_generation = generation;
			ca_element = element;
			ca_space = space;
			ca_rewriting = rewriting;
			ca_explanation = 
				State.merge_explanations
					[element.Context.el_choice_point]
					space.Context_unifier.cus_explanation;			
		}	
	and complexity =
		(let (literal_complexity, constraint_complexity) =
			List.fold_left
				(fun (clause_complexity, constraint_count) literal ->
					let literal_complexity =
						candidates.cd_bound#get_complexity literal
					and constraint_count =
						if literal.Term.is_constraint then
							(constraint_count + 1)
						else
							(constraint_count)
					in
						if literal_complexity > clause_complexity then
							(literal_complexity, constraint_count)
						else
							(clause_complexity, constraint_count))
				(0, 0)
				clause
		in
			max literal_complexity constraint_complexity)
	in
	match clause with
		| [] -> 
(* empty clause - no need to check anything, just build and store for rapid closing *)		
			(candidates.cd_preselected <- Some candidate)
		| _ ->
(* are we recomputing forgotten candidates? *)
		(match candidates.cd_forgotten_recomputation_barrier with
			| Some (_, forgotten_barrier) when
(* ignore candidates below the recomputation threshhold *)
				compare_candidates candidate forgotten_barrier < 0
				||
(* ignore candidates exceeding the deepending bound - they have been stored before *)
				candidates.cd_bound#exceeds complexity
				->
					()
			| _ ->	  
(* are we checking exceeding candidates for applicability? *)
				(if candidates.cd_check_exceeding then
					(if (candidates.cd_bound#exceeds_current 
						complexity 
						element.Context.el_choice_point) then 
(* is the candidate subsumed by the current clauses? *)						
							(match check_subsumed_clause candidates clause with
								| Some explanation ->
									candidates.cd_invalid <- (candidate, explanation) :: candidates.cd_invalid
								| None ->
(* candidate is not subsumed *)								
									(ignore (candidates.cd_bound#register complexity element.Context.el_choice_point);
									if candidates.cd_bound#has_min_exceeding then
			  							(raise EXCEEDING_EXHAUSTED))))
			  	else
(* is this candidate exceeding the current bound? *)			  	
			  		(if candidates.cd_bound#exceeds complexity then
			  			(match Config.restart candidates.cd_config with
							| Flags.RS_Eager
							| Flags.RS_Lazy ->
							(* no, drop the candidate *)
								(ignore 
									((candidates.cd_bound#register complexity element.Context.el_choice_point) : bool))
							| Flags.RS_Delayed ->
(* for this choice point / context literal exceeding applicable candidates have already been dropped. *)
								(if (is_marked_exceeding candidates element) then
									()
								else
(* is the candidate subsumed by the current clauses? *)						
									(match check_subsumed_clause candidates clause with
										| Some choice_point ->
											(candidates.cd_invalid <-
												(candidate, choice_point) :: candidates.cd_invalid)
										| None ->
(* candidate is not subsumed *)
											(mark_exceeding candidates element))))
					else
(* a regular candidate (obeying the depth bound) *)
(* is the candidate subsumed? *)						
						(match check_subsumed_clause candidates clause with
							| Some choice_point ->
								(candidates.cd_invalid <-
									(candidate, choice_point) :: candidates.cd_invalid)
							| None ->
(* candidate is not subsumed *)
								(candidate_counter := !candidate_counter + 1;
								Valid.add candidates.cd_valid candidate))))))
	
						
let build_unprocessed ~(force: bool) (candidates: candidates) : unit = 
(* buffer not full, do nothing *)
	if (not force) && (candidates.cd_unprocessed_size < Array.length candidates.cd_unprocessed - 1) then
  		()
  	else
(* no need to check for closing_found, this is handled by Selection_assert. *)  	
  		(for i = 0 to (candidates.cd_unprocessed_size - 1) do
  			(match candidates.cd_unprocessed.(i) with
  				| (clause, element, space, candidate_type, rewriting) ->
  					(build_candidate candidates candidate_type element space clause rewriting))
  		done;
  		clear_unprocessed candidates)

let add ?(rewriting: Problem_subterms.rewriting = Problem_subterms.null_rewriting) (candidates: candidates) 
	(candidate_type: State.literal_type) (element: element) 
	(space: context_unifier_space) (clause: clause) : unit =
	if Term.is_tautology clause then 
		()
	else
		(let clause =
			(try
				((*let found = candidates.cd_index#find_clause_variants clause
				in
				pc clause;pe ":";
				List.iter (fun c -> pc c;pe "") found;*)
				candidates.cd_index#find_first_clause_variant clause)
			with
				| Not_found -> 
					(candidates.cd_index#add_clause clause clause;
					clause))
		in	
			match clause with
				| [] ->
(* empty clause - don't store it in the temporary buffer, instead turn it into a proper
candidate right away, it will be used for closing soon *)				
					(build_candidate candidates candidate_type element space clause rewriting)
				| _ ->	
					(candidates.cd_unprocessed.(candidates.cd_unprocessed_size) <- (clause, element, space, candidate_type, rewriting);
					candidates.cd_unprocessed_size <- candidates.cd_unprocessed_size + 1;
					(match candidate_type with
						| State.Paramodulation ->
							Statistic.inc_computed_paramodulation_candidates candidates.cd_statistic
						| State.Reflexivity ->
							Statistic.inc_computed_reflexivity_candidates candidates.cd_statistic
						| State.Simplification ->
							Statistic.inc_computed_simplification_candidates candidates.cd_statistic
						| _ ->
							failwith "Selection_clauses.add: bad literal_type.");
					build_unprocessed ~force:false candidates))

			
(* remove all retracted candidates at the beginning of the valid candidate priority queue *)
let rec retract_valid_min (candidates: candidates) : unit =
  	if Valid.is_empty candidates.cd_valid then 
  		()
	else
	(let candidate =
		Valid.min candidates.cd_valid
	in
(* candidate is no longer valid (due to backtracking) *)
		if (State.is_explanation_invalid candidate.ca_explanation) then
			(	(*pe (candidate_to_string "" candidate);*)
			ignore (Valid.remove_min candidates.cd_valid);
			retract_valid_min candidates))

(* remove all retracted candidates at the beginning of the valid candidate priority queue *)
let rec retract_valid_max (candidates: candidates) : unit =
  	if Valid.is_empty candidates.cd_valid then 
  		()
	else
	(let candidate =
		Valid.max candidates.cd_valid
	in
(* candidate is no longer valid (due to backtracking) *)
		if (State.is_explanation_invalid candidate.ca_explanation) then
			(ignore (Valid.remove_max candidates.cd_valid);
			retract_valid_max candidates))    		
			
			
			
(* if there is a forgotten candidate which is better than the best valid one,
   rebuild all its elements forgotten candidates. *)
let rec rebuild_forgotten (candidates: candidates) : unit =
	retract_valid_min candidates;

(* no forgotten candidates *)
	if ForgottenOrder.is_empty candidates.cd_forgotten_order then
    	()
	else
		(let (forgotten_element, forgotten_barrier) =
			ForgottenOrder.min_elt candidates.cd_forgotten_order
		in
(* no cached elements left *)
			if Valid.is_empty candidates.cd_valid
				||
(* or best forgotten element is better than best cached element *)
				compare_candidates forgotten_barrier (Valid.min candidates.cd_valid) < 0
			then
(* rebuild candidates for the forgotten element *)
(* no candidate is forgotten anymore *)
				(ForgottenBarrier.remove candidates.cd_forgotten_barrier forgotten_element;
				candidates.cd_forgotten_order <- 
					ForgottenOrder.remove 
						(forgotten_element, forgotten_barrier) 
						candidates.cd_forgotten_order;
				
(* ... and recompute them *)

(* ... but candidates below the previous barrier should be ignored upon recomputation,
so enter guard ... *)
				candidates.cd_forgotten_recomputation_barrier <- 
					Some (forgotten_element, forgotten_barrier);
				
				Problem_literals.compute_for_element 
					candidates.cd_problem_literals 
					forgotten_element 
					Context_unifier.Assert;
				let rewritten =
					Problem_subterms.compute_for_element
						candidates.cd_problem_subterms
						forgotten_element
				in
				List.iter
					(fun rewriting ->
						add
							candidates
							State.Paramodulation
							forgotten_element
							rewriting.Problem_subterms.rw_space
							rewriting.Problem_subterms.rw_result)
					rewritten;
				build_unprocessed ~force:true candidates;
				
(* ... and exit guard *)
				candidates.cd_forgotten_recomputation_barrier <- None;
				
				rebuild_forgotten candidates))

			

			
			
let rec get_best_candidate (candidates: candidates) : candidate option =
	rebuild_forgotten candidates;
	if (Valid.is_empty candidates.cd_valid) then
		(None)
	else
		(let candidate =
      		Valid.remove_min candidates.cd_valid
      	in
(* eagerly drop retracted candidates, if any *)
      		retract_valid_min candidates;      
      		if (State.is_choice_point_invalid candidate.ca_element.Context.el_choice_point) then
      			(get_best_candidate candidates)
      		else
(* check for various subsumptions *)      			
				(match check_subsumed_clause candidates candidate.ca_clause with
					| Some choice_point ->
						(candidates.cd_invalid <-
							(candidate, choice_point) :: candidates.cd_invalid;
						get_best_candidate candidates)
					| None ->
      			     	(Some candidate)))
      			

      			
let select (candidates: candidates) : selected option = 
	build_unprocessed ~force:true candidates;
	let candidate =
		match candidates.cd_preselected with
			| Some empty_clause ->
				(candidates.cd_preselected <- None;
				Some empty_clause)
			| None ->
				(get_best_candidate candidates)
	in
	match candidate with
		| None ->
			(None)
		| Some candidate -> 
			let selected_clause =
				{
	  				Selection_types.candidate_type = candidate.ca_type;
	  				Selection_types.literal = Term.null_literal;
	  				Selection_types.clause = candidate.ca_clause;
	  				Selection_types.element = candidate.ca_element;
					Selection_types.raw_context_unifier = Context_unifier.null_context_unifier;
					Selection_types.context_unifier_space = candidate.ca_space;
					Selection_types.rewriting = candidate.ca_rewriting;
	      			Selection_types.explanation = candidate.ca_explanation;					
				}
			in	
				Some selected_clause
			
				
				
				
(* check all exceeding candidates for applicability. *)
let check_exceeding (candidates: candidates) : unit = 
 (* if Config.is_horn candidates.cd_config then begin*)
  if not (candidates.cd_unprocessed_size = 0) then
    failwith "Selection_clauses.check_exceeding";

  (* enter exceeding checking mode *)
  candidates.cd_check_exceeding <- true;

  (* recompute for each context literal all candidates in chronological order *)
  begin
    try
      (Context.iter
	(fun element ->
	   if is_marked_exceeding candidates element then
	     (let rewritten =
			Problem_subterms.compute_for_element
				candidates.cd_problem_subterms
				element
		in
		List.iter
			(fun rewriting -> 
				add
					candidates
					State.Paramodulation
					element
					rewriting.Problem_subterms.rw_space
					rewriting.Problem_subterms.rw_result)
			rewritten))
	candidates.cd_context;

      (* build all unprocessed exceeding candidates *)
      build_unprocessed ~force:true candidates)
    with
      | EXCEEDING_EXHAUSTED ->
	  ()
  end;

  (* unprocessed candidates might be in an inconsistent state due to EXCEEDING_EXHAUSTED.
     as it can only contain recomputed candidates, it can be savely emptied. *)
  clear_unprocessed candidates;

  (* leave exceeding checking mode *)
  candidates.cd_check_exceeding <- false


				
				
				
(*** backtrack ***)

let backtrack_forgotten (candidates: candidates) : unit =
  (* if recomputation was done, it is finished by now. *)
  candidates.cd_forgotten_recomputation_barrier <- None;

  (* remove barriers of retracted context elements *)
  let invalid_elements =
    ForgottenBarrier.fold
      (fun element candidate acc ->
	 if State.is_choice_point_invalid element.Context.el_choice_point then
	   (element, candidate) :: acc

	 else
	   acc
      )
      candidates.cd_forgotten_barrier
      []
  in
    List.iter
      (fun (element, _) ->
	 ForgottenBarrier.remove candidates.cd_forgotten_barrier element;
      )
      invalid_elements;

    candidates.cd_forgotten_order <-
      List.fold_left
        (fun forgotten (element, candidate) ->
	   ForgottenOrder.remove (element, candidate) forgotten 
	)
        candidates.cd_forgotten_order
        invalid_elements
        
        
let backtrack_exceeding (candidates: candidates) : unit =
  (* remove candidates exceeding the depth bound. *)
  let invalid_elements =
    Hashtbl.fold
      (fun id element acc ->
	 if State.is_choice_point_invalid element.Context.el_choice_point then
	   id :: acc

	 else
	   acc
      )
      candidates.cd_exceeding_elements
      []
  in
    List.iter
      (fun id ->
	 Hashtbl.remove candidates.cd_exceeding_elements id
      )
      invalid_elements        	
      
      
let add_unprocessed (candidates: candidates) (candidate: candidate) : unit =
	candidates.cd_unprocessed.(candidates.cd_unprocessed_size) <- 
		(candidate.ca_clause, candidate.ca_element, candidate.ca_space, candidate.ca_type, candidate.ca_rewriting);
	candidates.cd_unprocessed_size <- candidates.cd_unprocessed_size + 1;
	build_unprocessed ~force:false candidates
      
(* backtrack all invalid candidates *)
let backtrack_invalid (candidates: candidates) : unit =
	let rec backtrack_invalid still_invalid now_unprocessed invalid =
		match invalid with
			| [] ->
				candidates.cd_invalid <- still_invalid;
				List.iter 
					(add_unprocessed candidates) 
					now_unprocessed
	
			| (candidate, conflicting_explanation) as invalid :: tail ->
	(* candidate does no longer exist *)
	if State.is_explanation_invalid candidate.ca_explanation then
		backtrack_invalid still_invalid now_unprocessed tail
	
	(* candidate is no longer invalidated by the context *)
	else if State.is_explanation_invalid conflicting_explanation then
		backtrack_invalid still_invalid (candidate :: now_unprocessed) tail
	
	(* candidate still not applicable *)
	else
		backtrack_invalid (invalid :: still_invalid) now_unprocessed tail
	in
		backtrack_invalid [] [] candidates.cd_invalid			
				
				
let backtrack (candidates: candidates) : unit =
	(match candidates.cd_preselected with
		| Some empty_clause ->
			if (State.is_explanation_invalid empty_clause.ca_explanation) then 
				(candidates.cd_preselected <- None)
			else
				(failwith "Selection_clauses.backtrack: preselected")
		| None ->
			());
	retract_valid_min candidates;
	retract_valid_max candidates;
	
	backtrack_invalid candidates;
	backtrack_forgotten candidates;
	backtrack_exceeding candidates;	
	
	clear_unprocessed candidates
		
		
		
let has_closing_candidate (candidates: candidates) : bool =
  match candidates.cd_preselected with
    | None -> false
    | _ -> true		