(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005, 2006
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(* the term database consists basically of hash sets containing the terms.

   the hash values are for a
   - variable term: its var's id
   - constant term: its symbol's id
   - function term: its hash value
   - literal: its enclosed term's hash value

   terms are simply stored in a weak set.
   as long as a term is referenced outside of the set
   it is kept alive and subsequent creation requests return the same term.
   as soon as the term is not referenced outside of the set
   it may be automatically garbage collected.
*)

(*** global term ordering ***)

type term_ordering =
	| LRPO
	| KBO

let selected_term_ordering = ref LRPO	


(*** types ***)

type var = Var.var
type symbol = Symbol.symbol


type func = {
  symbol: symbol;
  subterms: term array;
  ground: bool;
  hash: int;
  in_db: bool;
  kbo_var_occs: (var * int) list;
  kbo_weight: int;
}

and equation = {
	term: term;
	left: term;
	right: term;
	oriented: bool;
}

and term =
  | Var of var
  | Const of symbol
  | Func of func
  | Eq of equation

let rec is_ground term =
	match term with
		| Var _ -> false
		| Const _ -> true
		| Func func -> func.ground
		| Eq eq -> is_ground eq.term
	
let is_ground_eq eq =
	match eq.term with
		| Func func ->
			func.ground
		| _ -> false
	  
let is_ground_eq_term term =
	match term with
		| Eq eq ->
			(is_ground_eq eq)
		| _ -> false	
	
		
		
type literal = {
  sign: bool;
  atom: term;
  literal_in_db: bool;
  is_constraint: bool;
  id: int;
}

type clause =
    literal list

let is_positive_clause clause =
	List.for_all
		(fun literal ->
			literal.sign)
		clause    
    
let is_negative_clause clause =
	List.for_all
		(fun literal ->
			not literal.sign)
		clause    
	
			
let is_eq term =
	match term with
		| Eq eq -> true
		| _ -> false

		
let rec term_copy (term: term) : term =
	match term with
		| Var var -> 
			Var
				(Var.clone_renumbered
					var
					(Var.id_of_var var))
		| Const symbol ->
			Const symbol
		| Func func ->
			Func
				({symbol = func.symbol;
				subterms = 
					(Array.map
						(fun subterm ->
							term_copy subterm)
						func.subterms);
				ground = func.ground;
				hash = func.hash;
				in_db = false;
				kbo_var_occs = func.kbo_var_occs;
				kbo_weight = func.kbo_weight})
		| Eq eq ->
			let copied_term = term_copy eq.term
			in
			let left, right =
				match copied_term with
					| Func func ->
						(if eq.oriented then
							(func.subterms.(0), func.subterms.(1))
						else
							(func.subterms.(1), func.subterms.(0)))
					| _ -> failwith "Term.term_copy: bad equation"
			in		
				Eq
					({term = copied_term;
					left = left;
					right = right;
					oriented = eq.oriented})		


					
(*** hashing ***)

(* the hash value of a term. *)
let rec hash_of_term (term: term) : int =
  match term with
    | Var var ->
	Var.hash_of_var var
	  
    | Const symbol ->
	Symbol.id symbol
	  
    | Func func ->
	func.hash

	| Eq eq ->
	hash_of_term eq.term

let hash_of_literal (literal: literal) : int =
  hash_of_term literal.atom


let hash_of_clause (clause: clause) : int =
  List.fold_left
    (fun acc literal ->
      (acc + (hash_of_literal literal)) * 131
    )
    0
    clause


(* computation of the hash value of a Func term.

   131 is known to work quite well by ... someone Cesare Tinelli knows.
   tests seem to support it. *)
let compute_hash (predicate_symbol: symbol) (arguments: term array) : int =

  let compute_hash' (old_value: int) (new_value: int) =
    (old_value + new_value) * 131
  in

    Array.fold_left 
      (fun old_hash term ->
	 compute_hash' old_hash (hash_of_term term)
      )
      (compute_hash' 0 (Symbol.id predicate_symbol))
      arguments


(*** kbo values ***)

let rec kbo_weight_of_term (term: term) : int =
	match term with
		| Var var ->
			1
		
		| Const symbol ->
			2
		
		| Func func ->
			func.kbo_weight
		
		| Eq eq ->
			hash_of_term eq.term      
      
let kbo_weight_of_literal (literal: literal) : int =
	kbo_weight_of_term literal.atom		
		
let rec kbo_var_occs_of_term (term: term) : (var * int) list =
	match term with
		| Var var ->
			[(var, 1)]
		| Const _ ->
			[]
		| Func func ->
			func.kbo_var_occs
		| Eq eq ->
			kbo_var_occs_of_term eq.term
			
let kbo_var_occs_of_literal (literal: literal) : (var * int) list =
	kbo_var_occs_of_term literal.atom				
	
	
(* combined computation of hash, kbo_weight and kbo_var_occs *)      
let compute_func_values (predicate_symbol: symbol) (arguments: term array) : int * int * ((var * int) list)=
	let rec hash = ref (Symbol.id predicate_symbol)
	and kbo_weight = ref 0
	and kbo_var_occs = ref []
	and merge_kbo_var_occs merged occs_1 occs_2 =
		match occs_1, occs_2 with
			| [], [] ->
				List.rev merged
			| [], _ ->
				List.rev_append merged occs_2
			| _, [] ->
				List.rev_append merged occs_1
			| (var_1, occ_1) :: rest_occs_1, (var_2, occ_2) :: rest_occs_2 ->
				(match (Var.compare var_1 var_2) with
					| 0 ->
						merge_kbo_var_occs
							((var_1, (occ_1 + occ_2)) :: merged)
							rest_occs_1 rest_occs_2
					| (-1) ->
						merge_kbo_var_occs
							((var_1, occ_1) :: merged)
							rest_occs_1
							occs_2
					| _ ->
						merge_kbo_var_occs
							((var_2, occ_2) :: merged)
							occs_1
							rest_occs_2)					
	in
	(match !selected_term_ordering with
		| LRPO ->
			Array.iter
				(fun term ->
					hash := (!hash + (hash_of_term term)) * 131;
					kbo_weight := 0;
					kbo_var_occs := [])
				arguments	
		| KBO ->		
			Array.iter
				(fun term ->
					hash := (!hash + (hash_of_term term)) * 131;
					kbo_weight := !kbo_weight + (kbo_weight_of_term term);
					kbo_var_occs := merge_kbo_var_occs [] !kbo_var_occs (kbo_var_occs_of_term term))
				arguments);
	!hash, !kbo_weight, !kbo_var_occs

	
			

let rec is_term_ground (term: term) : bool =
  match term with
    | Var _ ->
	false

    | Const _ ->
	true

    | Func func ->
	func.ground

	| Eq eq ->
	is_term_ground eq.term
	




(*** comparison ***)

let rec term_equal (term1: term) (term2: term) : bool =
  term1 == term2
  ||
  begin
    match term1, term2 with
      | Func func1, Func func2 ->
	  not (func1.in_db && func2.in_db)
	  &&
	  Symbol.equal func1.symbol func2.symbol
	  &&
	  Tools.array_for_all2
	    (fun term1 term2 ->
	       term_equal term1 term2
	    )
	    func1.subterms
	    func2.subterms
	  | Eq eq1, Eq eq2 ->
	  	if (eq1.oriented && eq2.oriented) then
	  		((eq1.left == eq2.left)&&(eq1.right == eq2.right))	
	  	else	
		  	(((eq1.left == eq2.left)&&(eq1.right == eq2.right))
		  	||((eq1.left == eq2.right)&&(eq1.right == eq2.left)))
	    
      | _ ->
	  false
  end


let literal_equal (literal1: literal) (literal2: literal) : bool =
  literal1 == literal2
  ||
  begin
    not (literal1.literal_in_db && literal2.literal_in_db)
    &&
    (literal1.sign == literal2.sign)
    &&
    (literal1.is_constraint == literal2.is_constraint)
    &&
    (term_equal literal1.atom literal2.atom)
  end


let clause_equal (clause1: clause) (clause2: clause) : bool =
  Tools.lists_unordered_equal literal_equal clause1 clause2

let clause_approx_equal (clause1: clause) (clause2: clause) : bool =
  Tools.lists_equal literal_equal clause1 clause2



(*** positions ***)					
					


type position =
	| Top
	| Position of subterm_position
	
and subterm_position =
	{index: int;
	position: position;
	pos_hash: int;
	pos_in_db: bool}
	
let rec position_to_string position =				
	match position with
		| Top ->
			""
		| Position pos ->
			let trailing_string =
				position_to_string pos.position
			in
			let trailing_string' =
				(if trailing_string = "" then
					(trailing_string)
				else
					("."^trailing_string))
			in
			(string_of_int (pos.index + 1))
			^
			trailing_string'
			
(*** database types ***)



(* term database *)
module TermTableBase =
  Weak.Make (
    struct
      type t = term

      let rec are_terms_equal_at (index: int) (terms1: term array) (terms2: term array) : bool =
	if index >= Array.length terms1 then
	  true
	else begin
	  (* subterms can be compared using pointer equality,
	     as a term is constructed bottom up with this database *)
	  if terms1.(index) == terms2.(index) then
	    are_terms_equal_at (index + 1) terms1 terms2
	  else
	    false
	end


      (* term_equal can not be used, as it uses object identity 
	 which is assured exactly by this module *)
      let rec equal (term1: term) (term2: term) : bool =
	match term1, term2 with
	  | Var var1, Var var2 ->
	      Var.equal var1 var2

	  | Const const1, Const const2 ->
	      Symbol.equal const1 const2

	  | Func func1, Func func2 ->
	      (Symbol.equal func1.symbol func2.symbol)
	      &&
	      (are_terms_equal_at 0 func1.subterms func2.subterms)
	      
	  | Eq eq1, Eq eq2 ->
	  		if (eq1.oriented && eq2.oriented) then
	  			(equal eq1.left eq2.left) && (equal eq1.right eq2.right)
	  		else
		  	 (((equal eq1.left eq2.left) && (equal eq1.right eq2.right))
		  	 ||((equal eq1.left eq2.right) && (equal eq1.right eq2.left)))

	  | _ ->
	      false

      let hash (term: term) : int =
	hash_of_term term
    end
  )


(* literal database *)
module LiteralTableBase =
  Weak.Make (
    struct
      type t = literal

      let equal (literal1: literal) (literal2: literal) : bool =
	(literal1.sign == literal2.sign)
	&&
	(literal1.is_constraint = literal2.is_constraint)
	&&
	(literal1.atom == literal2.atom)
	
      let hash (literal: literal) : int =
	hash_of_literal literal

    end
  )

(* position database *)
module PositionTableBase =
	Weak.Make (
		struct
			type t = position
			let rec equal (position1: position) (position2: position) : bool =
				match position1, position2 with
					| Top, Top -> true
					| Top, _ -> false
					| _, Top -> false
					| Position p1, Position p2 ->
						p1.index = p2.index
						&&
						equal p1.position p2.position
			let rec hash (position:position) : int =
				match position with
					| Top -> 0
					| Position pos ->
						pos.index + (hash pos.position)
		end
	)


type database = {
  (* contains the terms: consts, vars, functions *)
  terms: TermTableBase.t;
  (* contains the literals *)
  literals: LiteralTableBase.t;
  (* contains the positions *)
  positions: PositionTableBase.t;
}



(*** globals ***)

(* this global database contains the global terms *)
let global: database =
  {
    terms = TermTableBase.create 2048;
    literals = LiteralTableBase.create 512;
    positions = PositionTableBase.create 2048;
  }




(*** decomposition ***)

let rec top_symbol_term (term: term) : symbol =
  match term with
    | Var _ -> raise Not_found
    | Const symbol -> symbol
    | Func func -> func.symbol
    | Eq eq -> top_symbol_term eq.term

let top_symbol_literal (literal: literal) : symbol =
  top_symbol_term literal.atom


let is_term_var (term: term) : bool =
  match term with
    | Var _ -> true
    | _ -> false

let is_literal_var (literal: literal) : bool =
  is_term_var literal.atom



let rec do_vars_of_term (acc: var list) (term: term) : var list =
  match term with
    | Var var ->
	(* store each variable only once *)
	if List.exists
	  (fun acc_var ->
	     Var.equal acc_var var
	  )
	  acc
	then
	  acc
	else
	  var :: acc

    | Const _ ->
	acc

    | Func func ->
	if func.ground then
	  acc
	else
	  Array.fold_left do_vars_of_term acc func.subterms
	
	| Eq eq ->
		do_vars_of_term acc eq.term

let vars_of_term (term: term) : var list =
  do_vars_of_term [] term

let vars_of_literal (literal: literal) : var list =
  vars_of_term literal.atom

let vars_of_clause (clause: clause) : var list =
  List.fold_left
    (fun acc literal ->
      do_vars_of_term acc literal.atom
    )
    []
    clause

exception FOUND
    
let is_term_universal (term: term) : bool =
	let rec check_term_universal (term: term) : unit =
		(match term with
			| Var var ->
				if (Var.is_parametric var) then
					(raise FOUND)	    
	    	| Const _ ->
	    		()
	    	| Func func ->
	    		Array.iter check_term_universal func.subterms
	    	| Eq eq ->
	    		check_term_universal eq.left;
	    		check_term_universal eq.right)
	in
		try
			(check_term_universal term;
			true)
		with
			| FOUND ->
				false
				
let is_literal_universal (literal: literal) : bool =
	is_term_universal literal.atom			
					
let rec term_contains_var (term: term) (var: var) : bool =
  match term with
    | Var var' ->
	Var.equal var var'
	  
    | Const _ ->
	false
	  
    | Func func ->
	Tools.array_exists
	  (fun term' -> term_contains_var term' var)
	  func.subterms

	| Eq eq ->
		term_contains_var eq.term var
	  
let rec iter_vars f (term: term) : unit =
	match term with
		Var var -> f var
		|Func func ->
			if (func.ground) then ()
			else
				(Array.iter 
					(fun sub ->
						iter_vars f sub)
					func.subterms)
		|Const const -> ()
		|Eq eq ->
			iter_vars f eq.term
	  
let term_contains_vars_of_term (t1: term) (t2: term) : bool =
	try
		(iter_vars
			(fun v1 ->
				if (List.exists 
						(fun v2 -> v1==v2) 
						(vars_of_term t1)) then ()
				else raise Exit)
			t2;
		true)
	with Exit -> false
	
let literal_contains_var literal var =
  term_contains_var literal.atom var



let rec term_contains_term (term: term) (contained: term) : bool =
  term_equal contained term
  ||
  match contained with
  	Eq _ -> false
  	|_ ->
  match term with
    | Var _
    | Const _ ->
	false
	  
    | Func func ->
	Tools.array_exists
	  (fun term' -> term_contains_term term' contained)
	  func.subterms
	| Eq eq ->
	term_contains_term
		eq.term contained

let literal_contains_term literal term =
  term_contains_term literal.atom term




let is_literal_ground literal =
  is_term_ground literal.atom









(*** string representation ***)

let rec term_to_string ?(pretty:bool = true) ?(debug:bool = false) (term: term) : string =
  match term with
    | Var var ->
	Var.to_string var

    | Const const ->
	(if Symbol.is_fd_element const then
	  "'" ^ Symbol.to_string ~pretty:pretty const
	else
	  Symbol.to_string ~pretty:pretty const)

    | Func func when Array.length func.subterms = 0 ->
	(* a constant in disguise *)
	let const = func.symbol in
	if Symbol.is_fd_element const then
	  "'" ^ Symbol.to_string ~pretty:pretty const
	else
	  Symbol.to_string ~pretty:pretty const

    | Func func ->
	(if pretty && Symbol.equal Symbol.equality func.symbol then begin
	  "(" ^ term_to_string ~pretty:pretty ~debug:debug func.subterms.(0)
	    ^ " = "
	    ^ term_to_string ~pretty:pretty ~debug:debug func.subterms.(1) ^ ")"
	end

	(* finite model: r(a, ...) ---> r_a(...) *)
	else if pretty
	  &&
	  Symbol.is_fd_relation func.symbol
	then begin
	  let constant, subterms =
	    match Array.to_list func.subterms with
	      | []
	      | _ :: [] ->
		  failwith ("Term.term_to_string: fd relation of arity < 2: " ^ Symbol.to_string func.symbol);

	      | head :: tail ->
		  head, tail
	  in
	  let index =
	    if is_term_var constant then
	      "r_" ^ string_of_int (Symbol.arity func.symbol - 1)
	    else
	      "r_" ^ Symbol.to_string ~pretty:pretty (top_symbol_term constant)
	  in
	    index
	    ^ "("
	    ^ String.concat ", " (List.map (term_to_string ~pretty:pretty ~debug:debug) subterms)
	    ^ ")"
	end

	else begin
	  Symbol.to_string ~pretty:pretty func.symbol
	  ^ "("
	  ^ String.concat ", " (List.map (term_to_string ~pretty:pretty) (Array.to_list func.subterms))
	  ^ ")"
	  ^ (if debug then
	  		(let kbo_occ_list =
	  			(let tmp_string = ref ""
	  			in
	  			List.iter
	  				(fun (var, occ) ->
	  					tmp_string := " " ^ !tmp_string ^ (Var.to_string var) ^ "(" ^ (string_of_int occ) ^ ")")
	  				func.kbo_var_occs;
	  			!tmp_string)
	  		in
	  		" [KBO-weight: " ^ (string_of_int func.kbo_weight) ^ " KBO-variables:" ^ kbo_occ_list ^ "]")
	  	else
	  		(""))
	end)
	|Eq eq ->
		term_to_string ~pretty:pretty ~debug:debug eq.term





(*** term sort ***)

let rec get_term_sort (term: term) : Symbol.sort =
  match term with
    | Const symbol
    | Func { symbol = symbol } ->
	Symbol.sort symbol
	| Eq eq ->
		get_term_sort eq.term

    | _ ->
	failwith "Term.get_term_sort"

let get_literal_sort (literal: literal) : Symbol.sort =
  get_term_sort literal.atom


let is_connection_literal (literal: literal) : bool =
  try
    Symbol.is_connection (top_symbol_literal literal)
  with
    | Not_found ->
	false

let is_fd_relation (literal: literal) : bool =
  try
    Symbol.is_fd_relation (top_symbol_literal literal)
  with
    | Not_found ->
	false

let is_fd_size_marker (literal: literal) : bool =
  try
    Symbol.is_fd_size_marker (top_symbol_literal literal)
  with
    | Not_found ->
	false

let is_fd_element (term: term) : bool =
  try
    Symbol.is_fd_element (top_symbol_term term)
  with
    | Not_found ->
	false


let rec is_input_term (term: term) : bool =
  match term with
    | Var _ ->
	true

    | Const symbol ->
	Symbol.is_input symbol

    | Func func ->
	Symbol.is_input func.symbol
	&&
	Tools.array_for_all is_input_term func.subterms
	
	| Eq eq ->
	is_input_term eq.term

let is_input_literal (literal: literal) : bool =
  is_input_term literal.atom


let rec is_fd_term (term: term) : bool =
  match term with
    | Var _ ->
	true

    | Const symbol ->
	Symbol.is_input symbol
	||
	Symbol.is_fd_symbol symbol
	||
	Symbol.is_fd_element symbol

    | Func func ->
	(
	  Symbol.is_input func.symbol
	  ||
	  Symbol.is_fd_relation func.symbol
	  ||
	  (Symbol.equal Symbol.diff func.symbol)
	)
	&&
	Tools.array_for_all is_fd_term func.subterms
	
	| Eq eq ->
		is_fd_term eq.term



let is_fd_literal (literal: literal) : bool =
  is_fd_term literal.atom


let rec is_skolem_free_term (term: term) : bool =
  match term with
    | Var _ ->
	true

    | Const const ->
	not (Symbol.is_skolem const)

    | Func func ->
	not (Symbol.is_skolem func.symbol)
	&&
	Tools.array_for_all is_skolem_free_term func.subterms
	
	| Eq eq ->
		is_skolem_free_term eq.term

let rec is_skolem_free_literal (literal: literal) : bool =
  is_skolem_free_term literal.atom





(*** Hash tables ***)


(* now export tables for further use.
   here equality can be checked by pointer identity. *)

module TermTable =
  Hashtbl.Make (
    struct
      type t = term

      let equal = term_equal
      let hash = hash_of_term
    end
  )


(* access by literal *)
module LiteralTable =
  Hashtbl.Make (
    struct
      type t = literal
	  
      let equal = literal_equal
      let hash = hash_of_literal
    end
  )

module ClauseTable =
  Hashtbl.Make (
    struct
      type t = clause
	  
      let equal = clause_equal
      let hash = hash_of_clause
    end
  )


module ClauseApproxTable =
  Hashtbl.Make (
    struct
      type t = clause
	  
      let equal = clause_approx_equal
      let hash = hash_of_clause
    end
  )

(* access by literal type, i.e. sign and predicate symbol
   but not the concrete literal. *)
module LiteralTypeTable =
  Hashtbl.Make (
    struct
      type t = literal

      let equal (literal1: literal) (literal2: literal) : bool =
	(literal1.sign == literal2.sign)
	&&
	match literal1.atom, literal2.atom with
	  | Func func1(*(symbol1, _, _)*), Func func2(*(symbol2, _, _)*) ->
	      (* same predicate symbol *)
	      Symbol.equal func1.symbol func2.symbol
		
	  | Eq _, Eq _ ->
	  	true    
	      
	  | _ ->
	      (* or for var/const same term *)
	      literal1.atom == literal2.atom

      let hash (literal: literal) : int =
	match literal.atom with
	  | Const symbol
	  | Func { symbol = symbol } (*(symbol, _, _)*) ->
	      Symbol.id symbol
	  | Eq _ ->
	  	  Symbol.id Symbol.equality
	  | Var var ->
	      Var.hash_of_var var
    end
  )


(*** lexicographic/recursive path ordering of terms ***)

exception Found
exception Failed


	
let symbol_precedence_compare (sym1:Symbol.symbol) (sym2:Symbol.symbol) : int =
	let arity_comp = Pervasives.compare (Symbol.arity sym1) (Symbol.arity sym2)
	in
		if (arity_comp>0) then 1
		else
			(if (arity_comp<0) then (-1)
			else
				(let name_comp =
					String.compare (Symbol.name sym1) (Symbol.name sym2)
				in
					if (name_comp>0) then 1
					else
						(if (name_comp<0) then (-1)
						else 0)))


(* tests whether t1 is greater than t2 in the RPO; can compare equations, but
care should be taken - if one input term is an equation and the other is not,
then an exception will be thrown *)
let rec lrpo_greater (t1:term) (t2:term) : bool =
	if (term_equal t1 t2) then false
	else
		(match t1, t2 with
			Var _, _ -> false
			|_, Var var -> (term_contains_var t1 var)
			|Const c1, Const c2 ->
				if (symbol_precedence_compare c1 c2)>0 then true
				else false
			|Const _, Func _ -> false
			|Func _, Const _ -> true
			|Eq eq1, Eq eq2 ->
				lrpo_greater eq1.term eq2.term
			|Func f1, Func f2 ->	
				(try
					(Array.iter
						(fun s ->
							if (term_equal s t2)
								||(lrpo_greater s t2)
								then (raise Found))
						f1.subterms;
					(match (symbol_precedence_compare f1.symbol f2.symbol) with
						(-1) -> false
						|0 ->
							(let exclusive_1 = ref []
							and exclusive_2 = ref []
							and check_1 = Array.make (Array.length f1.subterms) true
							and check_2 = Array.make (Array.length f1.subterms) true
							in
								Array.iteri
									(fun i1 s1 ->
										if check_1.(i1) then (
											try
												(Array.iteri
													(fun i2 s2 ->
														if term_equal s1 s2 then
															(check_1.(i1)<-false;
															check_2.(i2)<-false;
															raise Failed))
													f2.subterms;
												exclusive_1 := s1 :: !exclusive_1)
											with Failed -> ()))
									f1.subterms;
								Array.iteri
									(fun i2 s2 ->
										if check_2.(i2) then
											(exclusive_2 := s2 :: !exclusive_2))
									f2.subterms;
								if ((List.length !exclusive_1) > 0) then
									(List.for_all
										(fun s2 ->
											List.exists
												(fun s1 ->
													lrpo_greater s1 s2)
												!exclusive_1)
										!exclusive_2)
								else (false))
						|1 -> 
							Array.iter
								(fun s ->
									if (lrpo_greater t1 s) then ()
									else (raise Failed))
								f2.subterms;
							true
						|_ -> raise (Failure "Term.lrpo_greater: bad comparison value")))
				with 
					Found -> true	
					|Failed -> false)
		|_, _ -> failwith "Term.lrpo_greater: bad comparison input")
				
let lrpo_greater_var_checked (t1:term) (t2:term) : bool =
	if (term_contains_vars_of_term t1 t2) then
		(lrpo_greater t1 t2)
	else 
		(false)
		
let compare_lrpo_greater_var_checked (t1:term) (t2:term) : int =
	if (term_equal t1 t2) then 0
	else
		(if (lrpo_greater_var_checked t1 t2) then 1
		else
			(if (lrpo_greater_var_checked t2 t1) then (-1)
			else 0))
		
let rec lrpo_greater_ground_like (t1:term) (t2:term) : bool =
	if (term_equal t1 t2) then 
		(false)
	else
		(match t1, t2 with
			Var v1, Var v2 ->
				((Var.compare v1 v2)=1)
			|Var _, _ -> false
			|_, Var var -> true
			|Const c1, Const c2 ->
				if (symbol_precedence_compare c1 c2)>0 then true
				else false
			|Const _, Func _ -> false
			|Func _, Const _ -> true
			|Eq eq1, Eq eq2 ->
				lrpo_greater_ground_like eq1.term eq2.term			
			|Func f1, Func f2 ->	
				(try
					(Array.iter
						(fun s ->
							if (term_equal s t2)
								||(lrpo_greater_ground_like s t2)
								then (raise Found))
						f1.subterms;
					match (symbol_precedence_compare f1.symbol f2.symbol) with
						(-1) -> false
						|0 ->
							(let exclusive_1 = ref []
							and exclusive_2 = ref []
							and check_1 = Array.make (Array.length f1.subterms) true
							and check_2 = Array.make (Array.length f1.subterms) true
							in
								Array.iteri
									(fun i1 s1 ->
										if check_1.(i1) then (
											try
												(Array.iteri
													(fun i2 s2 ->
														if term_equal s1 s2 then
															(check_1.(i1)<-false;
															check_2.(i2)<-false;
															raise Failed))
													f2.subterms;
												exclusive_1 := s1 :: !exclusive_1)
											with Failed -> ()))
									f1.subterms;
								Array.iteri
									(fun i2 s2 ->
										if check_2.(i2) then
											(exclusive_2 := s2 :: !exclusive_2))
									f2.subterms;
								if ((List.length !exclusive_1) > 0) then
									(List.for_all
										(fun s2 ->
											List.exists
												(fun s1 ->
													lrpo_greater_ground_like s1 s2)
												!exclusive_1)
										!exclusive_2)
								else (false))
						|1 -> 
							Array.iter
								(fun s ->
									if (lrpo_greater_ground_like t1 s) then ()
									else (raise Failed))
								f2.subterms;
							true
						|_ -> raise (Failure "Bad comparison value in Term.lrpo_greater_ground_like."))
				with 
					Found -> true	
					|Failed -> false)
		|_, _ -> failwith "Term.lrpo_greater_ground_like: bad comparison input")					
					
let compare_lrpo_greater_ground_like (t1:term) (t2:term) : int =
	if (term_equal t1 t2) then 
		(0)
	else
		(if (lrpo_greater_ground_like t1 t2) then 
			(1)
		else 
			(if (lrpo_greater_ground_like t2 t1) then
				(-1)
			else
				(0)))


let rec kbo_greater (t1:term) (t2:term) : bool =
	let rec var_occs_greater occs1 occs2 =
		(match occs1, occs2 with
			| _, [] -> true
			| [], _ -> false
			| (var1, occ1) :: rest_occs1, (var2, occ2) :: rest_occs2 ->
				(match (Var.compare var1 var2) with
					| (-1) ->
						var_occs_greater rest_occs1 occs2
					| 0 ->
						if occ1 < occ2 then
							(false)
						else
							(var_occs_greater rest_occs1 rest_occs2)
					| _ ->
						false))
	in
	if (term_equal t1 t2) then false
	else
		(match t1, t2 with
			| Var _, _ -> false
			| Const s1, Const s2 ->
				(symbol_precedence_compare s1 s2) = 1
			| Const _, _ -> false
			| Func func, Var var ->
				List.exists
					(fun (var', _) ->
						var = var')
					func.kbo_var_occs
			| Func func, Const symbol -> true
			| Func func1, Func func2 ->
				if var_occs_greater func1.kbo_var_occs func2.kbo_var_occs then
					(match (Tools.compare_int func1.kbo_weight func2.kbo_weight) with
						| (-1) -> false
						| 0 -> 
							(match (symbol_precedence_compare func1.symbol func2.symbol) with
								| (-1) -> false
								| 0 ->
									(try
										(Array.iteri
											(fun i subterm ->
												if (kbo_greater subterm func2.subterms.(i)) then
													(raise Found)
												else
													(if (term_equal subterm func2.subterms.(i)) then
														()
													else
														(raise Failed)))
											func1.subterms;
										true)
									with
										| Found -> true
										| Failed -> false)
								| _ -> true)
						| _ -> true)					
				else
					(false)
			| Eq eq1, Eq eq2 ->
				kbo_greater eq1.term eq2.term
			| Eq _, _ -> false
			| _, Eq _ -> false)

let rec kbo_greater_ground_like (t1:term) (t2:term) : bool =
	let rec var_occs_greater occs1 occs2 =
		(match occs1, occs2 with
			| _, [] -> true
			| [], _ -> false
			| (var1, occ1) :: rest_occs1, (var2, occ2) :: rest_occs2 ->
				(match (Var.compare var1 var2) with
					| (-1) ->
						var_occs_greater rest_occs1 occs2
					| 0 ->
						if occ1 < occ2 then
							(false)
						else
							(var_occs_greater rest_occs1 rest_occs2)
					| _ ->
						false))
	in
	if (term_equal t1 t2) then false
	else
		(match t1, t2 with
			| Var var1, Var var2 -> var1 > var2
			| Var _, _ -> false
			| Const s1, Const s2 ->
				(symbol_precedence_compare s1 s2) = 1
			| Const _, _ -> false
			| Func func, Var var ->
				List.exists
					(fun (var', _) ->
						var = var')
					func.kbo_var_occs
			| Func func, Const symbol -> true
			| Func func1, Func func2 ->
				if var_occs_greater func1.kbo_var_occs func2.kbo_var_occs then
					(match (Tools.compare_int func1.kbo_weight func2.kbo_weight) with
						| (-1) -> false
						| 0 -> 
							(match (symbol_precedence_compare func1.symbol func2.symbol) with
								| (-1) -> false
								| 0 ->
									(try
										(Array.iteri
											(fun i subterm ->
												if (kbo_greater_ground_like subterm func2.subterms.(i)) then
													(raise Found)
												else
													(if (term_equal subterm func2.subterms.(i)) then
														()
													else
														(raise Failed)))
											func1.subterms;
										true)
									with
										| Found -> true
										| Failed -> false)
								| _ -> true)
						| _ -> true)					
				else
					(false)
			| Eq eq1, Eq eq2 ->
				kbo_greater_ground_like eq1.term eq2.term
			| Eq _, _ -> false
			| _, Eq _ -> false)
			
let compare_kbo_greater	(t1:term) (t2:term) : int =
	if (term_equal t1 t2) then 
		(0)
	else
		(if (kbo_greater t1 t2) then 
			(1)
		else 
			(if (kbo_greater t2 t1) then
				(-1)
			else
				(0)))			
			
				
let compare_kbo_greater_ground_like	(t1:term) (t2:term) : int =
	if (term_equal t1 t2) then 
		(0)
	else
		(if (kbo_greater_ground_like t1 t2) then 
			(1)
		else 
			(if (kbo_greater_ground_like t2 t1) then
				(-1)
			else
				(0)))		
			
let term_ordering_greater (term1: term) (term2: term) : bool =
	match !selected_term_ordering with
		| LRPO -> lrpo_greater term1 term2
		| KBO -> kbo_greater term1 term2
	
let term_ordering_greater_var_checked (term1: term) (term2: term) : bool =
	match !selected_term_ordering with
		| LRPO -> lrpo_greater_var_checked term1 term2
		| KBO -> kbo_greater term1 term2
	
let term_ordering_greater_ground_like (term1: term) (term2: term) : bool =
	match !selected_term_ordering with
		| LRPO -> lrpo_greater_ground_like term1 term2
		| KBO -> kbo_greater_ground_like term1 term2

let compare_term_ordering_greater_var_checked (term1: term) (term2: term) : int =
	match !selected_term_ordering with
		| LRPO -> compare_lrpo_greater_var_checked term1 term2
		| KBO -> compare_kbo_greater term1 term2
		
let compare_term_ordering_greater_ground_like (term1: term) (term2: term) : int =
	match !selected_term_ordering with
		| LRPO -> compare_lrpo_greater_ground_like term1 term2
		| KBO -> compare_kbo_greater_ground_like term1 term2		
	
		
				
(*** creation ***)



(* these functions try to find a requested term in the database
   and to return the old instance.

   otherwise a fresh instance is created, stored, and returned. *)


let request_term ?(insert_db:bool = true) (term: term) : term =
  try
    TermTableBase.find global.terms term
  with
    | Not_found ->	
	if insert_db then begin
	  TermTableBase.add global.terms term;
	end;
 	term


let request_var (var: var) : term =
  let new_term =
    Var var
  in
    request_term new_term

let request_const (symbol: symbol) : term =
  if Const.debug then begin
    if Symbol.arity symbol != 0 then begin
      failwith ("Term.request_const on non-constant symbol: " ^ Symbol.to_string symbol);
    end
  end;

  let new_term =
    Const symbol
  in
    request_term new_term


let request_func ?(insert_db:bool = true) ((symbol, terms): symbol * term array) : term =
  if Const.debug then begin
    if (Symbol.arity symbol != Array.length terms) then begin
      print_newline ();
      print_endline (Symbol.to_string symbol);
      print_endline (string_of_int (Symbol.arity symbol));
      Array.iter (fun term -> print_endline (term_to_string term)) terms;
      failwith "request_func: arity mismatch";
    end;
  end;

  let new_term =
  	let hash, kbo_weight, kbo_var_occs =
  		compute_func_values symbol terms
  	in
    Func (*(symbol, terms, compute_hash symbol terms)*)
      {
	symbol = symbol;
	subterms = terms;
	ground = Tools.array_for_all is_term_ground terms;
	hash = hash;
	in_db = insert_db;
	kbo_weight = kbo_weight;
	kbo_var_occs = kbo_var_occs;
      }
  in
   request_term ~insert_db:insert_db new_term

let request_eq ?(insert_db:bool = true) ?(force_orient:bool = true) 
	?(not_orient:bool = false) ((left, right): term * term) : term =
	match left, right with
		Eq _, _
		|_, Eq _ -> failwith "Term.request_eq: attempt to nest equations."
		|_, _ ->
			let new_left, new_right, oriented =
				(match (compare_term_ordering_greater_var_checked left right) with
					| (-1) -> 
						(if not_orient then
							(left, right, false)
						else
							(right, left, true))
					| 0 -> 
						(if not_orient then
							(left, right, false)
						else
							(if force_orient then
								(match (compare_term_ordering_greater_ground_like left right) with
									| (-1) -> (right, left, false)
									| 0 -> (left, right, false)
									| 1 -> (left, right, false)
									| _ -> (failwith "Term.request_eq: bad equational comparison. I"))
							else
								(left, right, false)))
					| 1 -> (left, right, true)
					| _ -> (failwith "Term.request_eq: bad equational comparison. II"))
			in
			let eq_term = request_func ~insert_db:insert_db (Symbol.equality, [|new_left; new_right|])
			in 
			let eq_left, eq_right =
				(match eq_term with
					Func eq_func -> eq_func.subterms.(0), eq_func.subterms.(1)
					|_ -> failwith "Term.request_eq: call to request_func delivered unexpected result.")
			in
			let new_eq =
				Eq 
					{
				term = eq_term;
				left = eq_left;
				right = eq_right;
				oriented = oriented;
					}
			in
				request_term ~insert_db:insert_db new_eq
				
				
let rec recursive_request_term (term: term) : term =
	match term with
		| Var var -> request_var var
		| Const const -> request_const const
		| Func func ->
			let new_subterms =
				Array.map
					(fun subterm ->
						recursive_request_term subterm)
					func.subterms
			in
				request_func (func.symbol, new_subterms)
		| Eq eq ->
			(let term =
				recursive_request_term eq.term
			in
				match term with
					| Func func ->
						request_eq (func.subterms.(0), func.subterms.(1))		
					| _ -> failwith "Term.recursive_request_term: bad equation")
					
let top_position = Top

let get_position_hash (position: position) : int =
	match position with
		| Top -> 
			0
		| Position pos ->
			pos.pos_hash

let rec make_position (numbers: int list) : position =
	match numbers with
		| [] -> 
			Top
		| number :: rest ->
			let sub_position =
				make_position rest
			in
			let new_position =
				Position
					{index = number;
					position = sub_position;
					pos_hash = number + (get_position_hash sub_position);
					pos_in_db = true}
			in
				try
					(PositionTableBase.find
						global.positions
						new_position)
				with
					| Not_found ->
						(PositionTableBase.add
							global.positions
							new_position;
						new_position)


		
let rec get_position (term: term) (position: position) : term =
	match position with
		| Top ->
			(term)
		| Position subterm_position ->
			match term with
				| Func func ->
					(get_position func.subterms.(subterm_position.index) subterm_position.position)
				| Eq eq ->
					(get_position eq.term position)
				| _ ->
					(failwith "Term.get_position: nonexistent position in term.")
			
										
let rec set_position (term: term) (position: position) (to_insert: term) : term =
	match position with 
		| Top ->
			(to_insert)
		| Position subterm_position ->
			match term with
				| Func func ->
					let new_subterms = Array.copy func.subterms
					in
					new_subterms.(subterm_position.index) <- 
						(set_position 
							new_subterms.(subterm_position.index) 
							subterm_position.position 
							to_insert);
					let hash, kbo_weight, kbo_var_occs =
						compute_func_values func.symbol new_subterms
					in
					(Func
						{symbol = func.symbol;
						subterms = new_subterms;
						ground = func.ground && (is_term_ground to_insert);
						hash = hash; 
						in_db = false;
						kbo_weight = kbo_weight;
						kbo_var_occs = kbo_var_occs})
				| Eq eq ->
					(let new_eq_term =
						set_position eq.term position to_insert
					in
					match new_eq_term with
						| Func func ->
							(request_eq (func.subterms.(0), func.subterms.(1)))
						| _ ->
							(failwith "Term.set_position: bad equation."))
				| _ ->
					(failwith "Term.set_position: nonexistent position in term.")
				
					
					
			
let lit_id = ref 0
						
let gen_lit_id () =
	lit_id := !lit_id + 1;
	!lit_id
   
let request_literal ?(insert_db:bool = true) ?(is_constraint:bool = false) (sign: bool) (term: term) : literal =
  let new_literal = {
    sign = sign;
    atom = term;
    is_constraint = is_constraint;
    literal_in_db = insert_db;
    id = gen_lit_id ();
  }
  in
    try
      LiteralTableBase.find global.literals new_literal
    with
      | Not_found ->
	  if insert_db then begin
	    LiteralTableBase.add global.literals new_literal;
	  end;
 	  new_literal

let request_constraint_literal ?(insert_db:bool = true) (sign: bool) (term: term) : literal =
  let new_literal = {
    sign = sign;
    atom = term;
    is_constraint = true;
    literal_in_db = insert_db;
    id = gen_lit_id ();
  }
  in
    try
      LiteralTableBase.find global.literals new_literal
    with
      | Not_found ->
	  if insert_db then begin
	    LiteralTableBase.add global.literals new_literal;
	  end;
 	  new_literal

 	  
let request_constraint_counterpart ?(insert_db:bool = true) (literal: literal) : literal =
	request_literal
		~insert_db:insert_db
		~is_constraint:(not literal.is_constraint)
		literal.sign
		literal.atom
		 	  
 	  
let rec insert_term term =
  match term with
    | Func func ->
	if func.in_db then
	  term

	else begin	
	  let subterms =
	    Array.map insert_term func.subterms
	  in
	    request_func (func.symbol, subterms)
	end

    | _ ->
	term

let insert_literal literal =
  if literal.literal_in_db then
    literal

  else begin
    let term =
      insert_term literal.atom
    in
      request_literal literal.sign term
  end



		


(*** string representation ***)





let literal_to_string ?(pretty:bool = true) ?(constraints:bool = true) (literal: literal) : string =
  (* replace diff by -equal *)
  let literal =
    match literal.atom with
      | Func func when pretty && Symbol.equal Symbol.diff func.symbol ->
	  request_literal
	    (not literal.sign)
	    (request_func (Symbol.equality, func.subterms))
      | _ ->
	  literal
  in
	if literal.is_constraint then
		(match literal.atom with
			| Eq eq ->
				("(" ^ (term_to_string ~pretty:pretty eq.left) ^ " -> " ^ (term_to_string ~pretty:pretty eq.right) ^ ")")
			| _ ->
				(failwith "Term.literal_to_string: bad constraint literal."))
	else
  (let prefix =
    if literal.sign then
      "+" (* "" *)
    else
      "-" (* "~" *)
  in
  let term =
    term_to_string ~pretty:pretty literal.atom
  in
    prefix ^ term)



let clause_to_string ?(pretty:bool = true) ?(constraints:bool = true) (clause: clause) : string =
	if constraints then
		(let constraint_literals, literals =
			List.partition
				(fun lit -> lit.is_constraint)
				clause
		in
			"[" ^ String.concat ", " (List.map (literal_to_string ~pretty:pretty) literals) ^ "].["
			^ String.concat ", " (List.map (literal_to_string ~pretty:pretty ~constraints:true) constraint_literals) ^ "]")
	else
		("[" ^ String.concat ", " (List.map (literal_to_string ~pretty:pretty ~constraints:false) clause) ^ "]")









(*** Term Modification ***)


let request_negated_literal ?(insert_db:bool = true) (literal: literal) : literal = 
  request_literal ~insert_db:insert_db (not literal.sign) literal.atom


let replace_term_in_term (term: term) (old_term: term) (new_term: term) : term =

  let rec replace_in_term' (term: term) : term =
    if term_equal term old_term then
      new_term
    else
      match term with
	| Var _
	| Const _ ->
	    term
	    
	| Func func ->
	    let new_terms =
	      Array.map
		(fun term ->
		   replace_in_term' term
		)
		func.subterms
	    in
	      request_func (func.symbol, new_terms)
	| Eq eq ->
		let new_left, new_right =
			(replace_in_term' eq.left),
			(replace_in_term' eq.right)
		in
			request_eq (new_left, new_right)
  in
    if term_equal old_term new_term then
      term
    else
      replace_in_term' term
		

let replace_term_in_literal literal old_term new_term =
  request_literal
    literal.sign
    (replace_term_in_term literal.atom old_term new_term)




let rec replace_terms_in_term (term: term) (mapping: (term * term) list) : term =

  let rec replace_in_term' (term: term) : term =
    try
      let (_, new_term) =
	List.find
	  (fun (old_term, _) -> term_equal term old_term)
	  mapping
      in
	new_term
    with
      | Not_found ->
	  begin
	    match term with
	      | Var _
	      | Const _ ->
		  term
		    
	      | Func func ->
		  let new_terms =
		    Array.map
		      (fun term ->
			 replace_in_term' term
		      )
		      func.subterms
		  in
		    request_func (func.symbol, new_terms)
		  | Eq eq ->
		  let new_left, new_right =
		  	(replace_in_term' eq.left),
		  	(replace_in_term' eq.right)
		  in
		  	request_eq (new_left, new_right)
	  end
  in
    replace_in_term' term

let replace_terms_in_literal literal map =
  request_literal
    literal.sign
    (replace_terms_in_term literal.atom map)



let rec replace_vars_in_term (term: term) (func: var -> term -> term) : term =
  match term with
    | Var var ->
	func var term
	  
    | Const _ ->
	term
	  
    | Func func' ->
	let new_terms =
	  Array.map
	    (fun term ->
	       replace_vars_in_term term func
	    )
	    func'.subterms
	in
	  request_func (func'.symbol, new_terms)

	| Eq eq ->
		let new_left, new_right =
			(replace_vars_in_term eq.left func),
			(replace_vars_in_term eq.right func)
	in
		request_eq (new_left, new_right)		
		

let replace_vars_in_literal literal func =
  request_literal
    literal.sign
    (replace_vars_in_term literal.atom func)




let rec remove_duplicates clause =
  match clause with
    | [] ->
	[]

    | literal :: tail ->
	if
	  List.exists
	    (fun literal' -> literal_equal literal literal')
	    tail
	then
	  tail
	else
	  literal :: remove_duplicates tail

	  

let is_trivial_eq term =
	match term with
		| Eq eq ->
			term_equal eq.left eq.right
		| _ -> false
			  
	  
let rec remove_trivial_negative_equations clause =
	List.filter 
		(fun literal ->
			literal.sign
			||
			(not (is_trivial_eq literal.atom)))
		clause






(* does also normalize the remaining variables,
   i.e. removes 'gaps' created by skolemizing some variables *)
let rec request_skolemized_term' (term: term)
    ((var_skolemization, par_normalization) as acc: ((var * term) list) * ((var * term) list))
    : ((var * term) list) * ((var * term) list) =

    match term with
      | Var var ->
	  if Var.is_universal var then begin
	    if List.exists (fun (var', _) -> Var.equal var var') var_skolemization then
	      acc
	    else
	      let fresh_const =
		request_const (Symbol.create_skolem ())
	      in
		((var, fresh_const) :: var_skolemization), par_normalization
	  end

	  else begin
	    if List.exists (fun (var', _) -> Var.equal var var') par_normalization then
	      acc
	    else
	      let fresh_par =
		(* the next id can be deduced from the number of normalized pars *)
		Var.clone_renumbered var (List.length par_normalization)
	      in
	      let fresh_term =
		request_var fresh_par
	      in
		var_skolemization, ((var, fresh_term) :: par_normalization)
	  end

      | Const _ ->
	  acc

      | Func func ->
	  Array.fold_left
	    (fun acc' term' ->
	      request_skolemized_term' term' acc'
	    )
	    acc
	    func.subterms

	  | Eq eq ->
	  	request_skolemized_term' eq.term acc

	    
let request_skolemized_term (term: term) : term =
  let var_skolemization, par_normalization =
    request_skolemized_term' term ([], [])
  in
    match var_skolemization with
      | [] ->
	  (* no universal variables in term *)
	  term

      | _ ->
	  let mapping =
	    List.map
	      (fun (var, term) ->
		request_var var, term
	      )
	      (par_normalization @ var_skolemization)
	  in
	    replace_terms_in_term term mapping


let request_skolemized_literal (literal: literal) : literal =
  request_literal literal.sign (request_skolemized_term literal.atom)


let request_skolemized_clause (clause: clause) : clause =
  let var_skolemization, par_normalization =
    List.fold_left
      (fun acc literal ->
	request_skolemized_term' literal.atom acc
      )
      ([], [])
      clause
  in
    match var_skolemization with
      | [] ->
	  (* no universal variables in term *)
	  clause

      | _ ->
	  let mapping =
	    List.map
	      (fun (var, term) ->
		request_var var, term
	      )
	      (par_normalization @ var_skolemization)
	  in
	    List.map
	      (fun literal ->
		replace_terms_in_literal literal mapping
	      )
	      clause




(* original version,no offset
let rec request_universal_term' (term: term) (mapping: (var * term) list) : (var * term) list  =
  match term with
    | Var var ->
	if Var.is_parametric var then begin
	  if List.exists (fun (var', _) -> Var.equal var var') mapping then
	    mapping
	  else
	    let fresh_var =
	      Var.create_universal (List.length mapping)
	    in
	    let fresh_term =
	      request_var fresh_var
	    in
	      (var, fresh_term) :: mapping
	end

	else
	  (* ensure that no parameter is mapped to this variable
	     by creating a fresh variable with the same id *)
	  (var, term) :: mapping

      | Const _ ->
	  mapping

      | Func func ->
	  Array.fold_left
	    (fun mapping' term' ->
	      request_universal_term' term' mapping'
	    )
	    mapping
	    func.subterms
	  | Eq eq ->
	  	request_universal_term' eq.term mapping
*)	    
	  	
(* original version, no offset 
let request_universal_term (term: term) : term =
  let mapping =
    request_universal_term' term []
  in
    match mapping with
      | [] ->
	  (* no parametric variables in term *)
	  term

      | _ ->
	  let mapping =
	    List.map
	      (fun (var, term) ->
		request_var var, term
	      )
	      mapping
	  in
	    replace_terms_in_term term mapping
*)
	   
let rec request_universal_term' (offset: int) (term: term) (mapping: (var * term) list) : (var * term) list  =
  match term with
    | Var var ->
	if (Var.is_parametric var) then
		(if List.exists (fun (var', _) -> Var.equal var var') mapping then
			(mapping)
		else
			(let fresh_var =
				Var.create_universal ((List.length mapping) + offset)
			in
			let fresh_term =
				request_var fresh_var
			in
				(var, fresh_term) :: mapping))
	else
(* ensure that no parameter is mapped to this variable
by creating a fresh variable with the same id *)
		(if (offset = 0) then
			((var, term) :: mapping)
		else
			(if List.exists (fun (var', _) -> Var.equal var var') mapping then
				(mapping)
			else
				(let fresh_var =
					Var.create_universal ((Var.id_of_var var) + offset)
				in
				let fresh_term =
					request_var fresh_var
				in
					(var, fresh_term) :: mapping)))

      | Const _ ->
	  	mapping

      | Func func ->
		  Array.fold_left
		    (fun mapping' term' ->
		      request_universal_term' offset term' mapping'
		    )
		    mapping
		    func.subterms
	  | Eq eq ->
		  	request_universal_term' offset eq.term mapping


let request_universal_term_with_mapping 
	?(offset: int = 0) ?(mapping: (var * term) list = []) (term: term) : term * ((var * term) list) =
  let mapping' =
    request_universal_term' offset term mapping
  in
    match mapping' with
      | [] ->
	  (* no parametric variables in term *)
	  	(term, mapping)

      | _ ->
	  let term_mapping =
	    List.map
	      (fun (var, term) ->
		request_var var, term
	      )
	      mapping'
	  in
	    ((replace_terms_in_term term term_mapping), mapping')		  	
		  	
 
let request_universal_term ?(offset: int = 0) (term: term) : term =
	let term', _ = request_universal_term_with_mapping ~offset:offset term
	in
		term'
		

	    
		    
	    
let request_universal_literal (literal: literal) : literal =
  request_literal literal.sign (request_universal_term literal.atom)

let request_universal_clause (clause: clause) : clause =
  let mapping =
    List.fold_left
      (fun acc literal ->
	request_universal_term' 0 literal.atom acc
      )
      []
      clause
  in
    match mapping with
      | [] ->
	  clause

      | _ ->
	  let mapping =
	    List.map
	      (fun (var, term) ->
		request_var var, term
	      )
	      mapping
	  in
	    List.map
	      (fun literal ->
		replace_terms_in_literal literal mapping
	      )
	      clause











(*** special literals ***)


(*** creation ***)

let true_literal : literal =
  request_literal true (request_const (Symbol.create_predicate "true" 0))

let false_literal : literal =
  request_literal true (request_const (Symbol.create_predicate "false" 0))

let assert_literal : literal =
  request_literal true (request_const (Symbol.create_predicate "__assert__" 0))

let init_literal : literal =
  request_literal true (request_const (Symbol.create_predicate "__init__" 0))

let null_term : term =
  request_const (Symbol.create_function "__null_term__" 0)

let null_literal : literal =
  request_literal true null_term

let v_par =
  Var.create_parametric 0

let v_uni =  
  Var.create_universal 0
    
let v_term =
  let v_name =
    "__default_v_constant__"
  in
    request_const (Symbol.create_function v_name 0)

let minus_v : literal =
  let term =
    request_func (
      Symbol.create_predicate
		    "__-v__" 1,
		    [| request_var v_par |]
    )
  in
    request_literal false term

let plus_v : literal =
  let term =
    request_func (
      Symbol.create_predicate
			   "__+v__" 1,
			   [| request_var v_par |]
    )
  in
    request_literal true term




(*** special clauses ***)
let true_clause =
	[true_literal]


(*** schema terms ***)

(* is this a schema term? *)
let is_schema_term (term: term) : bool =
  let terms =
    match term with
      | Func func -> func.subterms
      | _ -> [| |]
  in
    try
      ignore (
	Array.fold_left
	  (fun acc term ->
	     match term with
	       | Var var ->
		   (* check that this variable has not been used before.*)
		   if List.exists (fun old_var -> Var.equal var old_var) acc then
		     raise Exit
		   else
		     var :: acc
			 
	       | _ ->
		   (* must be a variable *)
		   raise Exit
	  )
	  []
	  terms
      );
      true
    with
      | Exit ->
	  false

(* create the schema term for a symbol. *)
let create_schema_term (predicate: symbol) : term =
  if Symbol.arity predicate = 0 then
    request_const predicate

  else
    (* create the most general term,
       i.e. a term with a fresh parameter at each position *)
    let terms =
      Array.make (Symbol.arity predicate) null_term
    in
      for i = 0 to Symbol.arity predicate - 1 do
      terms.(i) <- request_var (Var.create_universal i)
      done;
      request_func (predicate, terms)


(* create the schema term for a term. *)
let rec create_schema_term_from_term (term: term) : term =
  match term with
    | Const _ ->
	(* consts are most general *)
	term

    | Func func ->
	create_schema_term func.symbol

    | Var _ ->
	failwith "Context.create_schema"
	
	| Eq eq ->
	create_schema_term Symbol.equality





(*** clause properties ***)


let is_definit (clause: literal list) : bool =

  let rec is_definit' (clause: literal list) (pos_count: bool) : bool =
    match clause with
      | [] ->
	  pos_count
	    
      | literal :: tail ->
	  if literal.sign then begin
	    if pos_count then
	      false
	    else
	      is_definit' tail true
	  end
	  else
	    is_definit' tail pos_count
  in
    is_definit' clause false

let is_Horn (clause: literal list) : bool =

  let rec is_horn' (clause: literal list) (pos_count: bool) : bool =
    match clause with
      | [] ->
	  true
	    
      | literal :: tail ->
	  if literal.sign then begin
	    if pos_count then
	      false
	    else
	      is_horn' tail true
	  end
	  else
	    is_horn' tail pos_count
  in
    is_horn' clause false


let are_Horn (clauses: clause list) : bool =
  List.for_all
    (fun clause ->
       is_Horn clause
    )
    clauses




let is_BS (clause: clause) : bool =
 	let is_BS' term =
 		match term with
 			| Var _
 			| Const _ ->
 			true
 			| Func _
 			| Eq _ ->
 			false
 	in
  List.for_all
    (fun literal ->
       match literal.atom with
	 | Var _ ->
	     failwith "Term.is_BS"
	       
	 | Const _ ->
	     true
	       
	 | Func func ->
	     Tools.array_for_all
	       is_BS'
	       func.subterms
	 | Eq eq ->
		 (is_BS' eq.left)
		 &&
		 (is_BS' eq.right)
    )
    clause

let are_BS (clauses: clause list) : bool =
  List.for_all is_BS clauses



let contains_empty_clause (clauses: clause list) : bool =
  List.exists
    (List.for_all (literal_equal false_literal))
    clauses


let contains_equality (clauses: clause list) : bool =
  List.exists
    (fun clause ->
       List.exists
	 (fun literal ->
	    match literal.atom with
		  | Eq _ ->
		  true
	    
	      | Func func ->
		  Symbol.equal Symbol.equality func.symbol
		    
	      | _ ->
		  false
       )
       clause
    )
    clauses







(*** Variants ***)



let are_terms_variants (_term1: term) (_term2: term) : bool =

  (* acc keeps track of the bijective mapping
     from the variables of _term1 to the variables of _term2 *)
  let rec do_variants (acc: (var * var) list) (term1: term) (term2: term) :
    (var * var) list =
    
    match term1, term2 with
      | Var var1, Var var2 when
	  Var.is_universal var1 == Var.is_universal var2 ->
	  Tools.mapping_extend acc Var.equal Var.equal var1 var2
	      
      | Const symbol1, Const symbol2 when
	  Symbol.equal symbol1 symbol2 ->
	  acc
	      
      | Func func1, Func func2 when
	  Symbol.equal func1.symbol func2.symbol ->
	  Tools.array_fold2 do_variants acc func1.subterms func2.subterms
	      
	  | Eq eq1, Eq eq2 ->
	  if (eq1.oriented && eq2.oriented) then
	  	(match eq1.term, eq2.term with
	  		Func func1, Func func2 ->
	  			Tools.array_fold2 do_variants acc func1.subterms func2.subterms
	  		|_ -> failwith "Bad equation in Term.are_terms_variants.")
	  else
	  	(try
	  		(match eq1.term, eq2.term with
	  		Func func1, Func func2 ->
	  			Tools.array_fold2 do_variants acc func1.subterms func2.subterms
	  		|_ -> failwith "Bad equation in Term.are_terms_variants.")
	  	with Exit ->
	  		(let acc = do_variants acc eq1.left eq2.right
	  		in
	  			do_variants acc eq1.right eq2.left))
	  
      | _ ->
	  raise Exit
  in	    
    (term_equal _term1 _term2)
    ||
    (
      try
	ignore (do_variants [] _term1 _term2);
	true
      with
	| Exit ->
	    false
    )


let are_literals_variants (literal1: literal) (literal2: literal) : bool =
  (literal1.sign == literal2.sign)
  &&
  (are_terms_variants literal1.atom literal2.atom)



let are_terms_skolem_variants (_term1: term) (_term2: term) : bool =

  let var_mapping =
    ref []
  in
  let skolem_mapping =
    ref []
  in

  (* acc keeps track of the bijective mapping
     from the variables of _term1 to the variables of _term2 *)
  let rec do_variants (term1: term) (term2: term) : unit =    
    match term1, term2 with
      | Var var1, Var var2 when
	  Var.is_universal var1 == Var.is_universal var2 ->
	  var_mapping := Tools.mapping_extend !var_mapping Var.equal Var.equal var1 var2
	      
      | Const symbol1, Const symbol2 when
	  Symbol.equal symbol1 symbol2 ->
	  ()
	      
      | Const symbol1, Const symbol2 when
	  Symbol.is_skolem symbol1
	  &&
	  Symbol.is_skolem symbol2
	  ->
	  skolem_mapping := Tools.mapping_extend !skolem_mapping Symbol.equal Symbol.equal symbol1 symbol2

      | Func func1, Func func2 when
	  Symbol.equal func1.symbol func2.symbol ->
	  Tools.array_iter2 do_variants func1.subterms func2.subterms
	      
	  | Eq eq1, Eq eq2 ->
	  if (eq1.oriented && eq2.oriented) then
	  	(match eq1.term, eq2.term with
	  		Func func1, Func func2 ->
	  			Tools.array_iter2 do_variants func1.subterms func2.subterms
	  		|_ -> failwith "Bad equation in Term.do_variants.")
	  else
	  	(try
	  		(match eq1.term, eq2.term with
	  		Func func1, Func func2 ->
	  			Tools.array_iter2 do_variants func1.subterms func2.subterms
	  		|_ -> failwith "Bad equation in Term.do_variants.")
	  	with Exit ->
	  		(do_variants eq1.left eq2.right;
	  		do_variants eq1.right eq2.left))
	  
      | _ ->
	  raise Exit
  in	    
    (term_equal _term1 _term2)
    ||
    (
      try
	ignore (do_variants _term1 _term2);
	true
      with
	| Exit ->
	    (*
	    List.iter
	      (fun (x, y) ->
		 print_endline (Symbol.to_string x ^ " - " ^ Symbol.to_string y)
	      )
	      !skolem_mapping;
	    *)
	    false
    )


let are_literals_skolem_variants (literal1: literal) (literal2: literal) : bool =
  (literal1.sign == literal2.sign)
  &&
  (are_terms_skolem_variants literal1.atom literal2.atom)












(* remove variants from a literal set *)
let remove_variants (literals: literal list) =
  (* extend an existing variant mapping to two more terms.
     invariant: variable with lower id are bound to variables with higher id.
     avoids cycles in renaming.
  *)
  let rec are_terms_variants mapping (term1: term) (term2: term) =
    match term1, term2 with
      | Var var1, Var var2 ->
	  let cmp =
	    Var.compare var1 var2
	  in
	    if cmp = 0 then
	      mapping
	    else if cmp < 0 then
	      (var1, var2) :: mapping
	    else
	      (var2, var1) :: mapping
	      
      | Const symbol1, Const symbol2 when
	  Symbol.equal symbol1 symbol2 ->

	  mapping
	    
      | Func func1, Func func2 when
	  Symbol.equal func1.symbol func2.symbol ->
	  
	  Tools.array_fold2 are_terms_variants mapping func1.subterms func2.subterms
	  
	  | Eq eq1, Eq eq2 ->
	  	if (eq1.oriented && eq2.oriented) then
	  		(let mapping' = (are_terms_variants mapping eq1.left eq2.left)
	  		in
	    		(are_terms_variants mapping' eq1.right eq2.right))
	    else
	    	(try
		  		(let mapping' = (are_terms_variants mapping eq1.left eq2.left)
		  		in
		    		(are_terms_variants mapping' eq1.right eq2.right))
	    	with Exit ->
		  		(let mapping' = (are_terms_variants mapping eq1.left eq2.right)
		  		in
		    		(are_terms_variants mapping' eq1.right eq2.left)))
		    			    	
      | _ ->
	  raise Exit
  in

  (* find a renaming making all variants identical *)
  let rec find_renaming mapping literals =
    match literals with
      | [] ->
	  failwith "Lemma: find_renaming"
	  
      | _ :: [] ->
	  mapping
	  
      | l1 :: tail ->
	  (* search for a variant for l1 in tail *)
	  begin
	    let rec find_renaming' mapping tail' =
	      match tail' with
		| [] ->
		    (* no variant found *)
		    find_renaming mapping tail
		    
		| l2 :: tail' ->
		    let mapping' =
		      try
			Some (are_terms_variants mapping l1.atom l2.atom)
		      with
			| Exit ->
			    None
		    in
		      (* variant found *)
		      match mapping' with
			| Some mapping ->
			    find_renaming mapping tail

			| None ->
			    find_renaming' mapping tail'
	    in
	      find_renaming' mapping tail	    

	  end
  in
  let mapping =
    find_renaming [] literals
  in

  let rec dereference mapping var =
    try
      let var' =
	Var.VarTable.find mapping var
      in
	dereference mapping var'
    with
      | Not_found ->
	  var
  in

    match mapping with
      | [] ->
	  (* no variants *)
	  literals

      | _ ->
	  begin
	    (* apply the renaming *)
	    let mapping' =
	      Var.VarTable.create 61
	    in
	      List.iter
		(fun (var1, var2) ->
		   Var.VarTable.add mapping' var1 var2
		)
		mapping;

	      let mapping =
		Var.VarTable.fold
		  (fun var1 var2 acc ->
		     let var2' =
		       dereference mapping' var2
		     in
		       (request_var var1, request_var var2') :: acc
		  )
		  mapping'
		  []
	      in
		
		(* apply unifier - have to use replace to avoid independent normalization *)
		List.map
		  (fun literal ->
		     replace_terms_in_literal literal mapping
		  )
		  literals
	  end
			     


      
(* remove duplicates *)
let remove_duplicates (literals: literal list) : literal list =
  let seen =
    LiteralTable.create 63
  in
    List.find_all
      (fun literal ->
	 if LiteralTable.mem seen literal then
	   false
	     
	 else begin
	   LiteralTable.add seen literal ();
	   true
	 end
      )
      literals









(* t = t \/ C *)
let is_tautology_1 (clause: clause) : bool =
  List.exists
    (fun literal ->
       literal.sign
       &&
       match literal.atom with
	 | Func func when
	     Symbol.equal Symbol.equality func.symbol
	     &&
	     term_equal func.subterms.(0) func.subterms.(1)
	     ->
	     true
	 | Eq eq when
	 	 term_equal eq.left eq.right
	 	 ->
	 	 true
	 | _ ->
	     false
	    
    )
    clause

(* p \/ -p \/ C *)
let is_tautology_2 (clause: clause) : bool =
  List.exists
    (fun literal ->
      List.exists
	(fun literal' ->
	  literal.sign != literal'.sign
	  &&
	  term_equal literal.atom literal'.atom
	)
	clause
    )
    clause

let is_tautology (clause: clause) : bool =
  is_tautology_1 clause
  ||
  is_tautology_2 clause










(*** Ordering ***)

let rec compare_terms x y =
  if term_equal x y then
    0
  else
    match x, y with
	(* var first *)
      | Var u, Var v ->
	  Var.compare u v
	    
      | Var _, _ ->
	  -1
	    
      | _, Var _ ->
	  1

      (* const second *)
      | Const a, Const b ->
	  Symbol.compare a b
	    
      | Const _, _ ->
	  -1
	    
      | _, Const _ ->
	  1
	    
	  | Eq eq1, Eq eq2 ->
	  compare_terms eq1.term eq2.term
	  (* This might be too simple, need to check what this function
	  is actually used for, as a multiset extension might be necessary. *)	
	  | Eq _, _ ->
	  -1
	  
	  | _, Eq _ ->
	  1
	  
      (* function term last *)
      | Func func1, Func func2 ->
	  let cmp =
	    Symbol.compare func1.symbol func2.symbol
	  in
	    if cmp <> 0 then
	      cmp
	    else
	      let rec compare_terms' index =
		if index >= Array.length func1.subterms then
		  0
		else
		  let cmp' =
		    compare_terms func1.subterms.(index) func2.subterms.(index)
		  in
		    if cmp' <> 0 then
		      cmp'
		    else
		      compare_terms' (index + 1)
	      in
		compare_terms' 0

	

(* pos before neg *)
let compare_literals x y =
  let cmp =
    compare
      (Symbol.name (top_symbol_literal x))
      (Symbol.name (top_symbol_literal y))
  in
    if cmp <> 0 then
      cmp
    else if x.sign == y.sign then
      compare_terms x.atom y.atom
    else if x.sign then
      -1
    else
      1

let sort_clause clause =
  List.sort compare_literals clause




(*** tptp ***)

(* get the i.th finite domain element in tptp format: "1" *)
let tptp_get_domain_element (i: int) : term =
(*  request_const (Symbol.create_skolem ~name:("\"" ^ string_of_int i ^ "\"") 0)*)
  request_const
    (Symbol.create_skolem
      ~name:(Some ("e" ^ string_of_int i)) ()
    )

let rec tptp_replace_domain_elements (term: term) : term =
  match term with
    | Var _ ->
	term

    | Const const ->
	if Symbol.is_fd_element const then
	  tptp_get_domain_element (int_of_string (Symbol.to_string ~pretty:true const))
	else
	  term

    | Func func ->
	let subterms =
	  Array.map tptp_replace_domain_elements func.subterms
	in
	  request_func (func.symbol, subterms)

	| Eq eq ->
		let left = tptp_replace_domain_elements eq.left
		and right = tptp_replace_domain_elements eq.right
		in
			request_eq (left, right)


(* replaces all variables in a term by constants,
   so that instead of Darwin's _0 the output is a tptp conform X0. *)
let tptp_replace_var (var: var) : term =
  let const =
    Symbol.create_function ("X" ^ string_of_int (Var.id_of_var var)) 0
  in
    request_const const

let rec tptp_replace_vars (term: term) : term =
  match term with
    | Var var ->
	tptp_replace_var var

    | Const _ ->
	term

    | Func func ->
	let subterms =
	  Array.map tptp_replace_vars func.subterms
	in
	  request_func (func.symbol, subterms)
	  
	| Eq eq ->
		let left = tptp_replace_vars eq.left
		and right = tptp_replace_vars eq.right
		in
			request_eq (left, right)	  

(* tptp can't take symbols starting with '_' *)
let rec tptp_replace_internal_symbols (term: term) : term =
  match term with
    | Var _ ->
	term

    | Const symbol ->
	if (String.get (Symbol.to_string ~pretty:false symbol) 0) = '_' then
	  request_const (
	      Symbol.create_skolem
		~arity:(Symbol.arity symbol)
		~name:(Some ("int" ^ (Symbol.name symbol)))
		()
	  )
	else
	  term

    | Func func ->
	let symbol =
	  if (String.get (Symbol.to_string ~pretty:false func.symbol) 0) = '_' then
	    Symbol.create_skolem
	      ~arity:(Symbol.arity func.symbol)
	      ~name:(Some ("int" ^ (Symbol.name func.symbol)))
	      ()
	  else
	    func.symbol
	in
	let subterms =
	  Array.map tptp_replace_internal_symbols func.subterms
	in
	  request_func (symbol, subterms)
	| Eq eq ->
	let tptp_eq_term = tptp_replace_internal_symbols eq.term
	in
		match tptp_eq_term with
			Func func ->
				request_eq (func.subterms.(0), func.subterms.(1))
			|_ -> (failwith "Term.tptp_replace_internal_symbols: bad equation conversion")


(* replaces a finite domain relation with the original function symbol *)
let tptp_replace_relation (term: term) =
  match term with
    | Func func when Symbol.is_fd_relation func.symbol ->
	let symbol =
	  match func.subterms.(0) with
	    | Const symbol ->
		Symbol.get_symbol_from_fd_symbol symbol
	    | _ ->
		failwith ("Context.tptp_replace_relation : " ^ term_to_string term);
	in
	let result =
	  func.subterms.(Array.length func.subterms - 1)
	in
	let arguments =
	  Array.sub func.subterms 0 (Array.length func.subterms - 2)
	in
	let function_term =
	  request_func (symbol, arguments)
	in
	  request_func (Symbol.equality, [| function_term; result |])

    | _ ->
	term


(* performs all transformations necessary to output a term as a tptp term *)
let to_tptp_term (term: term) : term =
  tptp_replace_relation (tptp_replace_domain_elements (tptp_replace_vars term))

let tptp_clause_to_tptp_string (label: string) (clause: clause) : string =
  let clause =
    List.map
      (fun literal ->
	let sign =
	  if literal.sign then
	    "  "
	  else
	    "~ "
	in
	let term =
	  tptp_replace_internal_symbols (tptp_replace_domain_elements (tptp_replace_vars literal.atom))
	in

	(* have to treat '=' special *)
	let term_repr =
	  match term with
	    | Func func when Symbol.equal Symbol.equality func.symbol ->
		term_to_string ~pretty:false func.subterms.(0)
		^ " = "
		^ term_to_string ~pretty:false  func.subterms.(1)

	    | _ ->
		term_to_string ~pretty:false term
	in
	  sign ^ term_repr ^ "\n"
      )
      clause;
  in
    "cnf(" ^ label ^ ",axiom,(\n"
    ^ "      " ^ String.concat "    | " clause
    ^ ")).\n"
  
(* misc *)
		
let eq_funcs_to_equations clause =
	List.map
		(fun lit ->
			match lit.atom with
				| Func func when func.symbol == Symbol.equality ->
					(let eq = request_eq (func.subterms.(0), func.subterms.(1))
					in
						request_literal
							~is_constraint:lit.is_constraint
							lit.sign
							eq)
				| _ ->
					(lit))
		clause