(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005, 2006
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** terms, literals, clauses

  All persistent terms are stored in a global database,
  so that each term exists only once in the system (hash-consing).

  Fast access is based on a hash value for every term {!Term.hash_of_term}.
  For variables and constants that's their id,
  for terms it is (for efficiency) stored in the term data structure.

  Benefits: low memory usage by term sharing, fast term comparison by pointer equality.

  Drawbacks: term construction implies a term database access, terms store a hash value.

  Only temporarily needed terms may not be entered into the database,
  if the argument ?insert_db is false in the request functions below.
  By default ?insert_db is true.
*)


(** {6 Types} *)

type var = Var.var
type symbol = Symbol.symbol


(** a functional term, e.g. f(a, b) *)
type func = private {
    symbol: symbol;
    (** the top symbol, e.g. f *)
    
    subterms: term array;
    (** the subterms, e.g. a, b *)

    ground: bool;
    (** is the term ground? *)

    hash: int;
    (** the hash value of this term used in the term database. *)

    in_db: bool;
    (** is this term in the database? *)
    
    kbo_var_occs: (var * int) list;
    (** a list of the variables occurring in the term, sorted from lowest id to largest,
    each with an associated integer representing the number of occurrences of that variable
    within the term *)
    
    kbo_weight: int;
    (** weight of the term for the Knuth-Bendix-Ordering *)
}

(** an equation, e.g. f(a, b) = g(b), with orientation information
and direct side access *)
and equation = private {
	term: term;
	(** the equation as a whole, e.g. f(a, b) = g(b) *)
	
	left: term;
	(** the left side of the equation, e.g. f(a, b) *)
	
	right: term;
	(** the right side of the equation, e.g. g(b) *)
	
	oriented: bool;
	(** is true iff left is larger than right in reduction ordering *)
}

(** a term can be a variable, a constant or a function. *)
and term = private
  | Var of var
      (** a variable *)

  | Const of symbol
      (** a constant *)

  | Func of func
      (** a function *)

  | Eq of equation
  	  (** an equation *)    

(** is this term ground? *)  	  
val is_ground: term -> bool  	  
  	  
(** is this term an equation? *)  	  
val is_eq: term -> bool  	  

(** is this equation ground? *)
val is_ground_eq: equation -> bool

(** is this term a ground equation? *)
val is_ground_eq_term: term -> bool
  	  
(** is this term a trivial equation, i.e. does the term have the form s=s? *)  	  
val is_trivial_eq: term -> bool  


(** position: a pointer system for accessing subterms *)
type position

(** creates a position from a list of integers *)
val make_position: (int list) -> position

(** the position pointing at the term itself *)
val top_position: position

(** returns the subterm within [term] pointed at by [position] *)
val get_position: term -> position -> term

(** replaces the subterm at the position within [term] pointed at by [position] by the term [term] *)
val set_position: term -> position -> term -> term

(*
(** returns the index of the position, mostly intended for debugging *)
val get_position_index: position -> int
*)

(** [term_copy term] returns a copy of [term] that can be manipulated without
affecting [term]. *)
val term_copy: term -> term

(** a literal is a signed term - the term is interpreted as a predicate. *)
type literal = private {
  sign: bool;
  (** the sign/polarity *)

  atom: term;
  (** the atom *)

  literal_in_db: bool;
  (** is this term in the database? *)
  
  is_constraint: bool;
  (** is this a constraint literal? *)  
  
  id: int;
  (** unique identification number *)
}

(** a clause: a list, not a set as in theory. *)
type clause =
    literal list

(** is this clause positive? *)
val is_positive_clause: clause -> bool
    
(** is this clause negative? *)
val is_negative_clause: clause -> bool
    

(** a specialized [Hashtbl] with terms as keys. *)
module TermTable : Hashtbl.S with type key = term

(** a specialized [Hashtbl] with literals as keys. *)
module LiteralTable : Hashtbl.S with type key = literal

(** a specialized [Hashtbl] with clauses as keys.
    order of literals in a clause does not matter.  *)
module ClauseTable : Hashtbl.S with type key = clause

(** a specialized [Hashtbl] with clauses as keys.
    order of literals in a clause does matter. *)
module ClauseApproxTable : Hashtbl.S with type key = clause

(** a specialized [Hashtbl] with literal types as keys,
  that is not the concrete literal but it's sign and predicate symbol. *)
module LiteralTypeTable : Hashtbl.S with type key = literal




(** {6 Functions} *)

(** {2 Special Terms} *)


(** the {e true} literal (sign is true). *)
val true_literal: literal

(** the {e false} literal (sign is true). *)
val false_literal: literal

(** to be used for default initialization. *)
val null_term: term

(** to be used for default initialization. *)
val null_literal: literal

(** marks an assert gap in a context unifier ({!Context_unifier}). *)
val assert_literal: literal

(** marks an assert gap for a literal of the initial interpretation
    of a context. *)
val init_literal: literal

(** the {e +v} context literal. *)
val plus_v: literal

(** the {e -v} context literal. *)
val minus_v: literal

(** dummy parameter contained in plus_v and minus_v,
    to be used together with [v_term]. *)
val v_par: var

(** dummy universal - is not actually used with v, but named
	as such for consistency. *)
val v_uni: var

(** dummy term to bind plus_v oder minus_v in a context unifier,
    see implementation of {!Context_unifier.extend_partial_context_unifier}. *)
val v_term: term

(** a dummy tautological clause, containing only the [true_literal] *)
val true_clause: clause

(** {2 Lexicographic/Recursive Path Ordering} *)

(** used to select a term ordering method *)
type term_ordering =
	| LRPO
	| KBO

(** used as a flag to indicate the selected term ordering method. The proper [Flags]
and [Config] modules are inaccessible from here, and the use of the term ordering functions 
is ubiquitous, so the use of a proper flag would require excessive transferring of structures
and values throughout the program. A single global reference like this one, while certainly
clumsy, is still far more elegant than the 'proper' alternative. *) 	
val selected_term_ordering: term_ordering ref	
	
(** [lrpo_greater t1 t2] returns true if [t1] is greater than
[t2] and false otherwise. The formal definition of the RPO is
implemented, and thus variables will be compared. This means
that the function may be unsuitable for comparisons of terms
from different clauses, as their variables are usually 
uncomparable.  *)
val lrpo_greater: term -> term -> bool

(** [lrpo_greater_var_checked t1 t2] works like [lrpo_greater],
but performs a preliminary test whether all variables of [t2]
are contained in t1, a necessary condition for [t2] being smaller
than [t1] if [t2] is not ground.
This is the recommended function for terms that may be non-ground. *)
val lrpo_greater_var_checked:  term -> term -> bool

(** [compare_lrpo_greater_var_checked t1 t2] returns 1 if [t1] is
greater than [t2], -1 if [t2] is greater than [t1], and 0 otherwise.
Comparison is done using [lrpo_greater_var_checked]. *)
val compare_lrpo_greater_var_checked:  term -> term -> int

(** [lrpo_greater_ground_like t1 t2] works like [lrpo_greater],
but treats variables as lower order constants ordered by their 
variable-id number. It is thus total also on non-ground terms, in
the limited sense that it can be used for term sorting within a clause.
It should not be used for comparisons between clauses, as variable 
numbers can change during processing. *)
val lrpo_greater_ground_like: term -> term -> bool

(** Works like [compare_lrpo_greater_var_checked], but uses the
ordering of [lrpo_greater_ground_like]. *) 
val compare_lrpo_greater_ground_like:  term -> term -> int


val term_ordering_greater: term -> term -> bool

val term_ordering_greater_var_checked: term -> term -> bool

val term_ordering_greater_ground_like: term -> term -> bool

val compare_term_ordering_greater_var_checked: term -> term -> int

val compare_term_ordering_greater_ground_like: term -> term -> int



(** {2 Construction} *)

val request_term: ?insert_db:bool -> term -> term

(** requests a {!Term.term}[.Var] term from a database.

    there might be different instances of the same variable {!Var.var} in the system.
    Still, as the first request to {!Term.request_var} creates a database entry
    for this specific instance of the variable,
    and the following requests return this first stored instance,
    all {!Term.term}[.Var]s built on this variable are identical. *)
val request_var: var -> term

(** requests a {!Term.term}[.Const] term from a database. *)
val request_const: symbol -> term

(** requests a {!Term.term}[.Func] term.
    if [insert_db] is true (default: true) the term is inserted into the database,
    if it is not already contained. *)
val request_func: ?insert_db:bool -> symbol * term array -> term

(** requests a {!Term.term}[.Eq] term.
    if [insert_db] is true (default: true) the term is inserted into the database,
    if it is not already contained.
    [term_ordering] is used to select the term ordering used for orientation,
    default is LRPO. 
    if [force_orient] is true (default: true) then the equation will be oriented
    as if its variables are lower order constants, provided that it cannot be
    oriented using the RPO. This method at least ensures that multiple occurrences
    of the same equation within a clause are oriented the same way.
    if [not_orient] is true (default: false) then the equation will be oriented
    in the order of the arguments, i.e. this allows to override the RPO, for
    example to get a flipped version of an equation that actually is orientable.
    [not_orient] set to true overrides [force_orient].
    if [not_orient] is true and the input happens to be oriented, then the [oriented]
    field of the resulting equation will be true. *)
val request_eq: ?insert_db:bool -> ?force_orient:bool 
-> ?not_orient:bool -> term * term -> term

(** requests a {!Term.literal}.
    if [insert_db] is true (default: true) the term is inserted into the database,
    if it is not already contained. *)
val request_literal: ?insert_db:bool -> ?is_constraint:bool -> bool -> term -> literal

(** requests the constraint counterpart literal to [literal].
	if [literal] is a constraint, then the result is a non-constraint.
	if [literal] is not a constraint, then the result is a constraint.
	[literal] and the result are identical in all other aspects. *)
val request_constraint_counterpart: ?insert_db:bool -> literal -> literal

(** [request_term] is insufficient for paramodulation purposes, as the replacement
at subterm positions messes up the term structure sufficiently for the pointer based
term equality to fail; a generated term may thus not be recognized as already existing 
in the term database. [recursive_request_term] reconstructs [term] bottom-up from the
database and will thereby produce the desired result. *)
val recursive_request_term: term -> term


(** {2 Term Modification} *)

(** inserts the literal into the database if it is not already contained. *)
val insert_term: term -> term

(** see {!Term.insert_term}. *)
val insert_literal: literal -> literal


(** requests a negation of a literal. *)
val request_negated_literal: ?insert_db:bool -> literal -> literal

(** requests the skolemization of a literal from the database.
    this is the skolemization in the Model Evolution sense,
    i.e each universal variable is replaced by a fresh constant,
    while parameters are normalized
    (each occurrence of the same variable is replaced by the same constant).
*)
val request_skolemized_literal: literal -> literal

(** see {!Term.request_skolemized_literal}. *)
val request_skolemized_clause: clause -> clause

(** requests the universal version of a term from the database,
    i.e. each parameter occuring in the term is replaced
    by a fresh universal variable.
    (each occurrence of the same parameter is replaced by the same variable).
    [offset] can be used to rename/increase the variables - this will affect
    all variables, including non-parameters.
*)
val request_universal_term: ?offset:int -> term -> term


(** see {!Term.request_universal_term}, also returns the mapping used. *)
val request_universal_term_with_mapping: ?offset:int -> ?mapping: (var * term) list 
	-> term -> term * ((var * term) list)

(** see {!Term.request_universal_term}. *)
val request_universal_literal: literal -> literal

(** see {!Term.request_universal_clause}. *)
val request_universal_clause: clause -> clause

(** [replace_in_term_term term sub_term replace_term]
    builds a new term from [term] by replacing
    every occurrence of [sub_term] in [term] by [replace_term]. *)
val replace_term_in_term: term -> term ->  term -> term

(** see {!Term.replace_term_in_term}. *)
val replace_term_in_literal: literal -> term -> term -> literal

(** like {!Term.replace_term_in_term}, but for a number of terms simultaneously. *)
val replace_terms_in_term: term -> (term * term) list -> term

(** see {!Term.replace_terms_in_term}. *)
val replace_terms_in_literal: literal -> (term * term) list -> literal

(** [replace_vars_in_term term map]
    like {!Term.replace_terms_in_term},
    but a variable is replaced by a term by the function map.  *)
val replace_vars_in_term: term -> (var -> term -> term) -> term

(** see {!Term.replace_vars_in_term}. *)
val replace_vars_in_literal: literal -> (var -> term -> term) -> literal

(** removes duplicate literals. *)
val remove_duplicates: clause -> clause

(** removes trivial negative equational literals. *)
val remove_trivial_negative_equations: clause -> clause

(** replace literals that use the equation symbol as predicate (i.e. a [func] type term) 
with literals that use true equations ([equation] type term). This form is necessary
when using the 'Builtin' equality handling.  *)
val eq_funcs_to_equations: clause -> clause

(** {2 Term Sorts} *)

(** the sort of the top symbol *)
val get_term_sort: term -> Symbol.sort

(** the sort of the top symbol *)
val get_literal_sort: literal -> Symbol.sort

(** a connection predicate? *)
val is_connection_literal: literal -> bool

(** a finite domain relation predicate? *)
val is_fd_relation: literal -> bool

(** a finite domain marker predicate? *)
val is_fd_size_marker: literal -> bool

(** a finite domain element? *)
val is_fd_element: term -> bool


(** a term built over the input signature? *)
val is_input_term: term -> bool

(** see {!Term.is_input_term} *)
val is_input_literal: literal -> bool

(** a term built over the input signature
    and additional finite domain symbols?
    I.e. {!Symbol.sort} in FD_Relation, FD_Element, FD_Symbol *)
val is_fd_term: term -> bool

(** see {!Term.is_fd_term} *)
val is_fd_literal: literal -> bool

(** a skolem free term? *)
val is_skolem_free_term: term -> bool

(** a skolem free literal? *)
val is_skolem_free_literal: literal -> bool


(** is this term parameter-free? *)
val is_term_universal: term -> bool

(** is this literal parameter-free? *)
val is_literal_universal: literal -> bool


(** {2 Decomposition} *)


(** the hash value of a term. *)
val hash_of_term: term -> int

(** see {!Term.hash_of_term}. *)
val hash_of_literal: literal -> int

(** see {!Term.hash_of_clause}. *)
val hash_of_clause: clause -> int


(** returns the top symbol of the term.
    @raise Not_found if term is of type [Var]. *)
val top_symbol_term: term -> symbol

(** see {!Term.top_symbol_term}. *)
val top_symbol_literal: literal -> symbol


(** is the term of type [Var]? *)
val is_term_var: term -> bool

(** see {!Term.is_term_var}. *)
val is_literal_var: literal -> bool


(** returns the variables of the term (without duplicates). *)
val vars_of_term: term -> var list

(** see {!Term.vars_of_term}. *)
val vars_of_literal: literal -> var list

(** see {!Term.vars_of_literal}. *)
val vars_of_clause: clause -> var list


(** is the variable contained in the term? *)
val term_contains_var: term -> var -> bool

(** does term_1 contain all variables of term_2? *)
val term_contains_vars_of_term: term -> term -> bool

(** see {!Term.term_contains_var}. *)
val literal_contains_var: literal -> var -> bool

(** [term_contains_term term contained] is the term [contained] contained in the term [term]? *)
val term_contains_term: term -> term -> bool

(** see {!Term.term_contains_var}. *)
val literal_contains_term: literal -> term -> bool


(** is the term ground? *)
val is_term_ground: term -> bool

(** see {!Term.is_literal_var}. *)
val is_literal_ground: literal -> bool


(** {2 Schema Terms} *)

(** a schema term for a symbol is the most general term for this symbol,
    e.g. for a 3-ary symbol f the schema term is f(x, y, z).
*)

(** is this a schema term? *)
val is_schema_term: term -> bool 

(** create the schema term for a symbol. *)
val create_schema_term: symbol -> term

(** create the schema term for a term. *)
val create_schema_term_from_term: term -> term


(** {2 Clause Properties} *)


(** does the clause contain exactly one positive literal? *)
val is_definit: clause -> bool

(** is the clause Horn? *)
val is_Horn: clause -> bool
  
(** is the clause set? *)
val are_Horn: clause list -> bool

(** is the clause free of function symbols? (Bernays-Schoenfinkle class) *)
val is_BS: clause -> bool

(** are the clauses free of function symbols? *)
val are_BS: clause list -> bool

(** is a clause empty, or contains only false? *)
val contains_empty_clause: clause list -> bool

(** does the clause set contain a (dis)equality literal? *)
val contains_equality: clause list -> bool

(** is the clause a tautology?

    tests for:
    - (t = t) \/ C
    - p \/ -p' \/ C where p and p' are variants
*)
val is_tautology: clause -> bool




(** {2 Comparison} *)

(** are two terms equal? *)
val term_equal : term -> term -> bool

(** are two literals equal? *)
val literal_equal : literal -> literal -> bool

(** are two clauses equal? order of literals does not matter. *)
val clause_equal : clause -> clause -> bool

(** are two clauses equal? order of literals does matter. *)
val clause_approx_equal : clause -> clause -> bool

(** a quick non-total order on terms. *)
val compare_terms: term -> term -> int

(** see {!Term.compare_terms}. *)
val compare_literals: literal -> literal -> int

(** sorts a clause with {!Term.compare_literals}. *)
val sort_clause: clause -> clause



(** {2 Variants} *)


(** are two terms (universal- and parameter-preserving) variants? *)
val are_terms_variants: term -> term -> bool

(** see {!Term.are_terms_variants}. *)
val are_literals_variants: literal -> literal -> bool

(** are two terms variants modulo skolem constants?
    That is, is there a bijective mapping
    of the skolem constants of one literal to the
    skolem constants of the other literal,
    and the literals are universal- and parameter-preverving variants? *)
val are_literals_skolem_variants: literal -> literal -> bool


(** removes variants from a literal set *)
val remove_variants: literal list -> literal list

(** removes duplicates from a literal set *)
val remove_duplicates: literal list -> literal list


(** {2 TPTP Conversions} *)


(** get the i.th finite domain element in tptp format: "1" *)
val tptp_get_domain_element: int -> term

(** replaces all variables in a term by constants,
   so that instead of Darwin's _0 the output is a tptp conform X0. *)
val tptp_replace_var: var -> term

(** performs all transformations necessary to output a term as a tptp term.

    relational function terms of the from r(f, ...)
    are replaced by f(...).
    for r(x, ...), i.e. if there is no fd relation symbol f,
    the function fails with an exception.
*)
val to_tptp_term: term -> term


(** {2 String Representation} *)

(** Example: {b +p(a, _0)}.
    pretty is by default true,
    then for example finite domain relations are printed more compact.
    debug set to true (default is false) will cause various internal debugging
    data about the term to be printed.
*)
val term_to_string: ?pretty:bool -> ?debug:bool -> term -> string

(** Example: {b -p(a, _0)} 
	constraints is by default true, this causes constraint literals to be printed
	as rewrite rules.
*)
val literal_to_string: ?pretty:bool -> ?constraints:bool -> literal -> string

(** Example: {b \{ +p(f(a), _0), -q(b) \}} *)
val clause_to_string: ?pretty:bool -> ?constraints:bool -> clause -> string

(** [tptp_clause_to_tptp_string label clause]
    transforms a clause into a tptp string representation.

    [label] is the label of the clause, e.g. "flattened_0"
*)
val tptp_clause_to_tptp_string: string -> clause -> string

(* a string consisting of a sequence of integers, separated by '.' *)
val position_to_string: position -> string