(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004, 2005, 2006
              The University of Iowa
              Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** term indexing *)

type term = Term.term
type literal = Term.literal
type symbol = Symbol.symbol

type entry_data = Indexing_misc.entry_data

(** raised by {!Term_indexing.iterator.next} if no element is left. *)
exception ITERATOR_EMPTY


(** {6 Constants} *)

(** the offset of an index literal. *)
val index_offset: int

(** the offset of a query literal. *)
val query_offset: int



(** {6 Classes} *)


(** type of data stored in {!Term_indexing.index}. *)
class virtual ['data] data:
object
  method virtual is_equal: 'data -> 'data -> bool
  method virtual to_string: 'data -> string
end
  

(** iterator type over the {!Term_indexing.index}. *)
class virtual ['data] iterator:
object
  (** are more elements left? *)
  method virtual is_empty: bool

  (** advances to and returns the next element.
      @raise Exit if no more elements exist. *)
  method virtual next: 'data
end


(** The interface for a predicate term indexing module.

    To be used for exactly one predicate/function symbol,
    not for several at once.
    Thus adding f(x) and g(x) is not allowed.
    
    Insertion of variants of already inserted terms is allowed
    and must keep the data attached to all variants intact.
    For an explanation for {e p_preserving} and {e i_preserving} see {!Subst}.
*)
class type ['data] predicate_index =
object
  (** {2 Access} *)


  (** add [term] with attached ['data] to the index.

      if [no_duplicates] (false by default) then each term may be stored only once,
      i.e. if a term is added more than once
      only the first entry is stored with its data,
      the others are silently dropped.
  *)
  method add: ?no_duplicates:bool -> ?redundancy_explanation:(State.explanation ref) 
  -> term -> 'data -> unit

  (** removes the entry associated with [term].
      if [data] is given,
      then additionally [data] must be attached the removed variant.

      returns false if no such entry exists.
  *)
  method remove: term -> 'data option -> bool 
    
  (** removes all entries. *)
  method clear: unit
    
  (** the number of elements stored in the index. *)
  method size: int
    

  (** {2 Traversal} *)

  (** like [List.iter].
      function style iterator over all term/data pairs stored in the index. *)
  method iter: (term -> 'data -> unit) -> unit
    
  (** like [List.fold]. *)
  method fold: 'a. ('a -> term -> 'data -> 'a) -> 'a -> 'a

  (** object style iterator over all terms (i.e. the attached data)
      which are generalizations of the [term]. *)
  method get_generalization_iterator: p_preserving:bool -> term -> 'data iterator

  (** object style iterator over all terms (i.e. the attached data)
      which shield the first term from the second term. *)
  method get_shielding_iterator: term -> term -> 'data iterator



  (** {2 Find All} *)
    
  (** returns the data attached to all
      (universal- and parameter-preserving) variant terms of [term]
      (used for testing only). *)
  method find_all_variants: term -> 'data list

  (** returns the data attached to all terms more general than [term]. *)
  method find_all_generalizations: p_preserving:bool -> term -> 'data list

  (** returns the data attached to all terms unifiable with [term]. *)
  method find_all_unifiable: p_preserving:bool -> term -> 'data list
  method find_all_unifiable_subst: p_preserving:bool -> term -> ('data * Subst.subst) list

  method find_all_unifiable_entry_data: p_preserving:bool -> term -> ('data * entry_data) list  
  
  method find_all_unifiable_entry_data_subst: p_preserving:bool -> term -> ('data * entry_data * Subst.subst) list  
  
  (** returns the data attached to all terms which are instances of [term]. *)
  method find_all_instances: p_preserving:bool -> term -> 'data list

  (** returns the data attached to all terms which strongly shield
      the first from the second term
      (see {!Term_indexing.predicate_index.find_strongly_shielding}).
      Note: if the second term is part of the index is counts
      as a shielding term, while the first term does not.
  *)
  method find_all_shielding: term -> term -> 'data list


  (** {2 Find First} *)

  (** returns the data attached to [term]. *)
  method find: term -> 'data option

  (** returns the data attached to a term more general than [term]. *)
  method find_generalization: p_preserving:bool -> term -> 'data option

  (** returns the data attached to a term unifiable with [term]. *)
  method find_unifiable: p_preserving:bool -> term -> 'data option
  
  (** returns the data attached to a term which is an instance of [term]. *)
  method find_instance: p_preserving:bool -> term -> 'data option

  (** [find_strongly_shielding term1 term2]
      
      First part of the productivity check (see the Model Evolution paper for details).

      returns None, if term2 is not strongly shielded from term1
      by the index,
      i.e. there is no true instance of term2 which is more general than term1.
      
      otherwise, returns a shielding entry.

      Note: if the second term is part of the index it counts
      as a shielding term, while the first term does not.
  *)
  method find_strongly_shielding: term -> term -> 'data option

  (** [find_shielding index term1 term2]
      Second part of the productivity check.
      
      returns all index elements that shield term2 from term1,
      i.e. which have a p-preserving instance
      which is also a true instance of term2 and more general than term1.

      Note: if the second term is part of the index it counts
      as a shielding term, while the first term does not.
  *)
  method find_shielding: term -> term -> 'data option
    


  (** {2 Representation} *)

  (** string representation of the index. *)
  method to_string: string
end




(* basically same as above, but uses perfect discrimination tree indexing *)
(* under construction, some methods are stubs *)
class type ['data] perfect_predicate_index =
object
  (** {2 Access} *)


  (** add [term] with attached ['data] to the index.

      if [no_duplicates] (false by default) then each term may be stored only once,
      i.e. if a term is added more than once
      only the first entry is stored with its data,
      the others are silently dropped.
  *)
  method add: ?no_duplicates:bool -> ?redundancy_explanation:(State.explanation ref)
  -> ?clause:Term.clause -> ?literal_position:int 
  -> ?search_flag:(bool ref) -> ?misc_flag:(bool ref) -> term -> 'data -> unit

  (** removes the entry associated with [term].
      if [data] is given,
      then additionally [data] must be attached the removed variant.

      returns false if no such entry exists.
  *)
  method remove: term -> 'data option -> bool 
  
  method remove_clause: term -> Term.clause -> bool
    
  (** analogous to [remove_clause], but returns the removed data *)  
  method extract_clause: term -> Term.clause -> (bool * 'data option)  
  
  (** removes all entries. *)
  method clear: unit
    
  (** the number of elements stored in the index. *)
  method size: int
    

  (** {2 Traversal} *)

  (** like [List.iter].
      function style iterator over all term/data pairs stored in the index. *)
  method iter: (term -> 'data -> unit) -> unit
    
  (** like [List.fold]. *)
  method fold: 'a. ('a -> term -> 'data -> 'a) -> 'a -> 'a

  (** object style iterator over all terms (i.e. the attached data)
      which are generalizations of the [term]. *)
  method get_generalization_iterator: ?subst:Subst.subst -> p_preserving:bool -> term -> 'data iterator

  (** {2 Find All} *)
    
  (** returns the data attached to all
      (universal- and parameter-preserving) variant terms of [term]
      (used for testing only). *)
  method find_all_variants: ?subst:Subst.subst -> term -> ('data * Indexing_misc.entry_data * Subst.subst) list

  (** returns the data attached to all terms more general than [term]. *)
  method find_all_generalizations: ?subst:Subst.subst -> p_preserving:bool -> term -> ('data * Indexing_misc.entry_data * Subst.subst) list

  (** returns the data attached to all terms unifiable with [term]. *)
  method find_all_unifiable: ?subst:Subst.subst -> p_preserving:bool -> term -> ('data * Indexing_misc.entry_data * Subst.subst) list
  
  (** returns the data attached to all terms which are instances of [term]. *)
  method find_all_instances: ?subst:Subst.subst -> p_preserving:bool -> term -> ('data * Indexing_misc.entry_data * Subst.subst) list

  (** {2 Find First} *)

  (** returns the data attached to [term]. *)
  method find: ?subst:Subst.subst -> term -> 'data option

  (** returns the data attached to a term more general than [term]. *)
  method find_generalization: ?subst:Subst.subst -> p_preserving:bool -> term -> 'data option

  (** returns the data attached to a term unifiable with [term]. *)
  method find_unifiable: ?subst:Subst.subst -> p_preserving:bool -> term -> 'data option
  
  (** returns the data attached to a term which is an instance of [term]. *)
  method find_instance: ?subst:Subst.subst -> p_preserving:bool -> term -> 'data option
 
  (** {2 Representation} *)

  (** string representation of the index. *)
  method to_string: string
end







(** a wrapper for {!Term_indexing.predicate_index} over different predicates.

    supposed to be inherited by a specific term index (see e.g. {!Discrimination_tree}).
    for propositional atoms a specialized index is used in any case.

    expects the same arguments as [predicate_index],
    plus the constructor of the actual [predicate_index] type to use.
*)
class ['data] index:
  'data data ->
  ('data data -> 'data predicate_index) ->
  object
    method find: literal -> 'data predicate_index
    method iter: (literal -> 'data predicate_index -> unit) -> unit
    method fold: 'a. ('a -> literal -> 'data predicate_index -> 'a) -> 'a -> 'a
    method is_empty: bool
  end

  
class ['data] perfect_index:
  'data data ->
  ('data data -> 'data perfect_predicate_index) ->
  object
    method find: literal -> 'data perfect_predicate_index
    method iter: (literal -> 'data perfect_predicate_index -> unit) -> unit
    method fold: 'a. ('a -> literal -> 'data perfect_predicate_index -> 'a) -> 'a -> 'a
    method is_empty: bool
  end  
  
  
class ['data] perfect_clause_index:
  'data data ->
  ('data data -> 'data perfect_predicate_index) ->
  object
    method find: literal -> 'data perfect_predicate_index
    method iter: (literal -> 'data perfect_predicate_index -> unit) -> unit
    method fold: 'a. ('a -> literal -> 'data perfect_predicate_index -> 'a) -> 'a -> 'a
    method is_empty: bool
    method add_clause: Term.clause -> ?redundancy_explanation:(State.explanation ref) -> 'data -> unit
	method add_clause_with_subsumption_check: Term.clause -> 
		?redundancy_explanation:(State.explanation ref) -> 
		'data -> bool * ('data list)
	method subsumption_check: Term.clause -> bool * ('data list)	
    method remove_clause: Term.clause -> unit
    method extract_clause: Term.clause -> 'data option    
    method find_clause_generalizations: Term.clause -> 'data list
    method find_clause_instances: Term.clause -> 'data list
    method find_clause_variants: Term.clause -> 'data list
    method find_clause_unifiables: Term.clause -> 'data list
    method find_first_clause_generalization: Term.clause -> 'data
    method find_first_clause_instance: Term.clause -> 'data
    method find_first_clause_variant: Term.clause -> 'data
    method find_first_clause_unifiable: Term.clause -> 'data
 
  end    
  
(** see above, but indexes by terms instead of literals *)  
class ['data] term_index:
  'data data ->
  ('data data -> 'data predicate_index) ->
  object
    method find: term -> 'data predicate_index
    method iter: (term -> 'data predicate_index -> unit) -> unit
    method fold: 'a. ('a -> term -> 'data predicate_index -> 'a) -> 'a -> 'a
    method is_empty: bool
  end

(** see above, but indexes by symbols instead of literals *)
class ['data] symbol_index:
  'data data ->
  ('data data -> 'data predicate_index) ->
  object
    method find: symbol -> 'data predicate_index
    method iter: (symbol -> 'data predicate_index -> unit) -> unit
    method fold: 'a. ('a -> symbol -> 'data predicate_index -> 'a) -> 'a -> 'a
    method is_empty: bool
  end  
  
(** see above, but indexes perfectly by terms instead of literals *)  
class ['data] perfect_term_index:
  'data data ->
  ('data data -> 'data perfect_predicate_index) ->
  object
    method find: term -> 'data perfect_predicate_index
    method iter: (term -> 'data perfect_predicate_index -> unit) -> unit
    method fold: 'a. ('a -> term -> 'data perfect_predicate_index -> 'a) -> 'a -> 'a
    method is_empty: bool
  end
  
(** see above, but indexes perfectly by symbols instead of literals *)  
class ['data] perfect_symbol_index:
  'data data ->
  ('data data -> 'data perfect_predicate_index) ->
  object
    method find: symbol -> 'data perfect_predicate_index
    method iter: (symbol -> 'data perfect_predicate_index -> unit) -> unit
    method fold: 'a. ('a -> symbol -> 'data perfect_predicate_index -> 'a) -> 'a -> 'a
    method is_empty: bool
  end  
