(*
* Copyright (C) 2007 Universitaet Koblenz-Landau
* 
* This file is part of E-KRHyper.
* 
* E-KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* E-KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* E-KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*                 Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*
*)

(* provides functions and representations for clausification
of first-order formula input *)

(* This is in part a workaround. The original KRHyper uses a non-standard
parser, whereas OCaml features and recommends its builtin semi-automated
lex/yacc-based parser construction. The resulting OCaml parsers cannot
accept any arguments, though, and KRHyper relies heavily on its non-global
symbol table, which needs to be given to all relevant functions as an argument.

E-KRHyper therefore does the following for TPTP-FOF input: a standard OCaml 
parser parses the input and turns it into a simplified and non-optimized
formula/'term' representation. In a post-processing step the symbol table is
used to turn the 'terms' into proper Term.terms. Finally everything is clausified,
following an adapted version of the TPTP-clausification algorithm.
*)


let pt t = Term.print_term t
let pe s = print_endline s

let focus_variable = ref (-1)

exception FOUND

(*
type unitary_formula =
	| Quantified_formula of (operator * (variable list) * unitary_formula)
	| Fof_formula of fof_formula
	| Atomic_formula of clausify_term
	
and fof_formula =
	| Assoc_binary_formula of (operator * (unitary_formula list))
	| Nonassoc_binary_formula of (operator * unitary_formula * unitary_formula)
	| Unitary_formula of unitary_formula
	| Negated_unitary_formula of unitary_formula
*)

	
	
(* alternative type system, far more elegant *)
type fof_formula =
	| Assoc_binary_formula of (operator * (fof_formula list))
	| Nonassoc_binary_formula of (operator * fof_formula * fof_formula)
	| Quantified_formula of (operator * (int list) * fof_formula)
	| Atomic_formula of clausify_term
	| Negated_fof_formula of fof_formula
	| Non_formula of fof_formula
	| Conjecture of fof_formula
	
	
			
and proto_term =
	| Const of string
	| Int of int
	| Float of float
	| Var of (int * string)
	| Compound of (string * (proto_term list))
	
and clausify_term =
	{proto_term: proto_term;
	mutable negated: bool;
	mutable term: Term.term}
	
and operator =
	| Bijection
	| Left_implication
	| Right_implication
	| Xor
	| And
	| Or
	| Nand
	| Nor
	| Forall
	| Exists
	| True
	| False
	| Negation	
	
let rec proto_term_to_string pt =
	match pt with
		| Const c -> c
		| Int i -> string_of_int i
		| Float f -> string_of_float f
		| Var (id, _) -> "X_" ^ (string_of_int id)
		| Compound (func, subterms) ->
			let str_subterms =
				(match subterms with
					| head :: tail ->
						(let str_subterms = ref (proto_term_to_string head)
						in
							List.iter
								(fun subterm ->
									str_subterms := !str_subterms ^ ", " ^ (proto_term_to_string subterm))
								tail;
							!str_subterms)
					| _ -> "")
			in
				func ^ "(" ^ str_subterms ^ ")" 	
	
let clausify_term_to_string ct =
	if ct.negated then
		("~" ^ (proto_term_to_string ct.proto_term))
	else
		(proto_term_to_string ct.proto_term)				
				
let rec fof_to_string fof =
	let op_to_string operator =
		match operator with
			| Bijection -> "<=>"
			| Left_implication -> "=>"
			| Right_implication -> "<="
			| Xor -> "<~>"
			| And -> "&&"
			| Or -> "||"
			| Nand -> "~&&"
			| Nor -> "~||"
			| Forall -> "!"
			| Exists -> "?"
			| True -> "True"
			| False -> "False"
			| Negation -> "~"	
	in
	match fof with
		| Assoc_binary_formula (operator, subformulas) ->
			(match subformulas with
				| [] -> "()"
				| [subformula] -> "(" ^ fof_to_string subformula ^ ")"		
				| head :: tail ->
					(let str_formula = ref ("(" ^(fof_to_string head))
					and str_operator = op_to_string operator
					in
						List.iter
							(fun subformula ->
								str_formula := !str_formula ^ " " ^ str_operator ^ " " ^ (fof_to_string subformula))
							tail;
						str_formula := !str_formula ^ ")";
						!str_formula))
	
		| Nonassoc_binary_formula (operator, subformula_1, subformula_2) ->
			"(" ^ (fof_to_string subformula_1) ^ " " ^ (op_to_string operator) ^ " " ^ (fof_to_string subformula_2) ^ ")"
		| Quantified_formula (operator, var_ids, subformula) ->
			(let str_vars =
				(match var_ids with
					| [] -> ""
					| head :: tail ->
						let str_vars = ref ("X_" ^ string_of_int head)
						in
							List.iter
								(fun id ->
									str_vars := !str_vars ^ " " ^ "X_" ^ string_of_int id)
								tail;
							!str_vars)
			in
			"(" ^ (op_to_string operator) ^ " " ^ str_vars ^ ": " ^ (fof_to_string subformula) ^ ")")
		| Negated_fof_formula subformula ->
			"~" ^ (fof_to_string subformula)
		| Non_formula _ -> "<NONFORMULA>"
		| Atomic_formula ct ->
			clausify_term_to_string ct
		| Conjecture fof -> "CONJECTURE: " ^ (fof_to_string fof)
			 
let make_clausify_term proto_term =
	{proto_term = proto_term;
	negated = false;
	term = Term.Unbound}	

let make_negated_clausify_term proto_term =
	{proto_term = proto_term;
	negated = true;
	term = Term.Unbound}	
	
let is_negated clausify_term =
	clausify_term.negated	
	
let complement_clausify_term clausify_term =
	clausify_term.negated <- not clausify_term.negated	
	
let skolem_counter = ref (-1)

let make_skolem_term var_ids =
  skolem_counter := !skolem_counter + 1;
  let skolem_name = "$sk_"^(string_of_int !skolem_counter)
  and variables =
  	List.map
  		(fun id -> Var (id, Term.empty_string))
  		var_ids
  in
  match variables with
  	| [] -> Const skolem_name
  	| _ -> Compound (skolem_name, variables)
  
	
let complement_fof_formula fof_formula =
	match fof_formula with
		| Negated_fof_formula fof -> fof
		| _ -> Negated_fof_formula fof_formula	
	
		
let negate_conjecture fof_formula =
	match fof_formula with
		| Conjecture fof ->
			Negated_fof_formula fof
		| _ -> fof_formula
		
let rec de_arrow formula =
	match formula with
		| Nonassoc_binary_formula (Left_implication, fof1, fof2) ->
		(* 'A => B' becomes 'not A or B' *)
			Assoc_binary_formula
				(Or,
				[(complement_fof_formula
					(de_arrow fof1));
				(de_arrow fof2)])
		| Nonassoc_binary_formula (Bijection, fof1, fof2) ->
		(* 'A <=> B' becomes '(not A or B) and (A or not B)' *)
			let n_fof1 = de_arrow fof1
			and n_fof2 = de_arrow fof2
			in
			let left =
				Assoc_binary_formula
					(Or,
					[(complement_fof_formula n_fof1);
					n_fof2])
			and right =
				Assoc_binary_formula
					(Or,
					[n_fof1;
					(complement_fof_formula n_fof2)])
			in
				Assoc_binary_formula
					(And, [left; right])
		| Nonassoc_binary_formula (Xor, fof1, fof2) ->
			de_arrow 
				(Negated_fof_formula
					(Nonassoc_binary_formula
						(Bijection, fof1, fof2)))
		| Nonassoc_binary_formula _ -> print_endline (fof_to_string formula);
			failwith "Clausify.normalize: unsupported nonassociative binary connective."
		| Assoc_binary_formula (operator, subformulas) ->
			Assoc_binary_formula
				(operator, (List.map de_arrow subformulas))
		| Quantified_formula (operator, var_ids, subformula) ->
			let subformula = de_arrow subformula
			in
				Quantified_formula (operator, var_ids, subformula)
		| Negated_fof_formula fof ->
			let fof = de_arrow fof
			in
				Negated_fof_formula fof
		| Conjecture fof ->
			de_arrow fof
		| _ -> formula
		
let rec pull_negations formula =
(* Pulls negations inwards to the atoms. Implications/bijections must have been removed *)
	let rec pull_negations_1 fof =
		match fof with
			| Negated_fof_formula fof' ->
				pull_negations fof'
			| Assoc_binary_formula (And, subformulas) ->
				Assoc_binary_formula
					(Or,
					(List.map pull_negations_1 subformulas))
			| Assoc_binary_formula (Or, subformulas) ->
				Assoc_binary_formula
					(And,
					(List.map pull_negations_1 subformulas))
			| Atomic_formula ct ->
				(Negated_fof_formula fof)
			| Quantified_formula (Exists, variables, subformula) ->
				Quantified_formula
					(Forall,
					variables,
					(pull_negations_1 subformula))
			| Quantified_formula (Forall, variables, subformula) ->
				Quantified_formula
					(Exists,
					variables,
					(pull_negations_1 subformula))
			| Conjecture fof ->
				pull_negations_1 fof
			| _ ->
				failwith "Clausify.pull_negations: formula not in required form."					
					
	in					
	match formula with
		| Negated_fof_formula fof -> 
			pull_negations_1 fof
		| Assoc_binary_formula (operator, subformulas) ->
			Assoc_binary_formula
				(operator,
				(List.map pull_negations subformulas))
		| Atomic_formula _ -> formula
		| Quantified_formula (operator, variables, subformula) ->
			Quantified_formula 
				(operator,
				variables,
				(pull_negations subformula))
		| Conjecture fof ->
			pull_negations_1 fof
		| _ ->
			failwith "Clausify.pull_negations: formula not in required form."
					

	
	
(* counts variable occurrences, in the most naive fashion: 'Exists X: P(X) and P(X)'
would have three variables, for example *)
let rec count_vars formula = 
	let cnt = ref 0
	and max = ref 0
	in
	let rec count_vars_2 proto_term =
		(match proto_term with
			| Var (id, _) -> 
				(cnt := !cnt + 1;
				if (id > !max) then
					(max := id))
			| Compound (_, subterms) ->
				List.iter
					count_vars_2 
					subterms
			| _ -> ())
	and count_vars_1 formula =
		(match formula with
			| Negated_fof_formula fof ->
				count_vars_1 fof
			| Assoc_binary_formula (_, subformulas) ->
				List.iter count_vars_1 subformulas
			| Nonassoc_binary_formula (_, subformula1, subformula2) ->
				(count_vars_1 subformula1;
				count_vars_1 subformula2)
			| Quantified_formula (_, variables, subformula) ->
				(cnt := !cnt + (List.length variables);
				List.iter
					(fun id ->
						if (id > !max) then
							(max := id))						
					variables;
				count_vars_1 subformula)
			| Atomic_formula ct ->
				(let pt = ct.proto_term
				in
					count_vars_2 pt)
			| Conjecture fof ->
				count_vars_1 fof
			| _ -> ())
	in
		count_vars_1 formula;
		(!cnt, !max)	
	
(* forall-quantifies all free variables in formula
formula requirements: bijections and implications must have been removed,
negations must have been pulled to atoms *)		
let rec quantify_free_vars formula =
	let quantified = []
	and free = Array.make (Term.max_vars) false
	and to_quantify = ref []
	in  
		mark_free_vars quantified free formula;
		Array.iteri
			(fun i is_free ->
				if is_free then
					(to_quantify := i :: !to_quantify))
			free;
		if !to_quantify = [] then
			(formula)
		else
			(Quantified_formula (Forall, !to_quantify, formula))
		
and mark_free_vars quantified free formula =
	match formula with
		| Assoc_binary_formula (operator, subformulas) ->
			List.iter
				(fun subformula ->
					mark_free_vars quantified free subformula)
				subformulas			
		| Quantified_formula (operator, variables, subformula) ->
				mark_free_vars (List.append variables quantified) free subformula
		| Atomic_formula ct ->
			mark_free_vars_term quantified free ct.proto_term
		| Negated_fof_formula subformula ->
			mark_free_vars quantified free subformula
		| Conjecture fof ->
			mark_free_vars quantified free fof
		| _ -> failwith "Clausify.quantify_free_vars: formula not in required form."
			
and mark_free_vars_term quantified free term =
	match term with
		| Var (id, _) ->
			if (not free.(id)) && (not (List.mem id quantified)) then
				(free.(id) <- true)
		| Compound (_, subterms) ->
			List.iter
				(fun subterm ->
					mark_free_vars_term quantified free subterm)
				subterms
		| _ -> ()		
		

			
(* returns a version of [formula] with unique variable names (no variable
name is quantified more than once)
formula requirements: bijections and implications must have been removed,
negations must have been pulled to atoms, no free variables *)		
let rec rename_vars formula =
	let max_vars, max_id = count_vars formula
	in 
	let max_id = ref max_id
	and	master_mapping = Array.make max_vars (-1)
	in 
	let rec rename_vars_term mapping term =
		match term with
			| Var (id, name) ->
				Var (mapping.(id), name)
			| Compound (func, subterms) ->
				Compound
					(func,
					(List.map
						(fun subterm ->
							rename_vars_term mapping subterm)
						subterms))
			| _ -> term		
	in	
	let rec rename_vars_1 mapping formula =
		(match formula with
			| Assoc_binary_formula (operator, subformulas) ->
				(Assoc_binary_formula
					(operator,
					(List.map
						(fun subformula ->
							rename_vars_1 mapping subformula)
						subformulas)))
			| Quantified_formula (operator, variables, subformula) ->
				(let mapping = Array.copy mapping
				in
				let renamed_variables =			
					List.map
						(fun id ->
							if (master_mapping.(id)<0) then
(* not yet in the 'global' list of quantified variables *)
								(master_mapping.(id)<-id;
								mapping.(id) <- id;
								id)
							else
(* this var id has been quantified somewhere already, so:
- determine a new_id
- add the new_id to the global list of quantified ids
- add the new_id to the local mapping to ensure it will be used to replace all occurrences
of the var_id within this subformula *)							
								(max_id := !max_id + 1;
								mapping.(id) <- !max_id;
								master_mapping.(!max_id) <- !max_id;
								!max_id))
						variables
				in
					Quantified_formula
						(operator,
						renamed_variables,
						(rename_vars_1 mapping subformula)))
			| Atomic_formula ct ->
				(let renamed_proto_term =
					(rename_vars_term mapping ct.proto_term)
				in
				let renamed_ct =
					(make_clausify_term renamed_proto_term)
				in
					renamed_ct.negated <- ct.negated;
					Atomic_formula (renamed_ct))
			| Negated_fof_formula subformula ->
				Negated_fof_formula 
					(rename_vars_1 mapping subformula)
			| Conjecture fof ->
				Conjecture
					(rename_vars_1 mapping fof)
			| _ -> failwith "Clausify.rename_vars: formula not in required form.")
	in
		rename_vars_1 (Array.make max_vars (-1)) formula
		

		

(* returns a list of those elements from [to_delete] that do not
occur in [deleter] *)				
let rec delete_commons deleter to_delete =
	delete_commons_1 deleter to_delete []
	
and delete_commons_1 deleter to_delete remains =
	match to_delete with
		| [] -> remains
		| element :: rest ->
			if (List.exists element deleter) then
				(delete_commons_1 deleter rest remains)
			else
				(delete_commons_1 deleter rest (element :: remains))


	

			
			
(* returns a skolemized quantor-free version of [formula]
formula requirements: bijections and implications must have been removed,
negations must have been pulled to atoms, no free variables, no variable
name is quantified more than once *)								
let rec skolemize formula =
	let rec skolemize_term to_skolemize term =
		match term with
			| Var (id, _) ->
				(let skolem_term =
					(try
						(snd(
							List.find
								(fun (index_id, _) ->
									index_id = id)
								to_skolemize))
					with Not_found -> term)
				in
					skolem_term)
			| Compound (func, subterms) ->
				(Compound
					(func,
					(List.map
						(fun subterm ->
							skolemize_term to_skolemize subterm)
						subterms)))
			| _ -> term
	in
	let rec skolemize_1 quantified to_skolemize formula =
		match formula with
			| Assoc_binary_formula (operator, subformulas) ->
				(Assoc_binary_formula
					(operator,
					List.map
						(fun subformula ->
							skolemize_1 quantified to_skolemize subformula)
						subformulas))
			| Quantified_formula (Forall, variables, subformula) ->
					(skolemize_1 
						(List.rev_append variables quantified)
						to_skolemize
						subformula)
			| Quantified_formula (Exists, variables, subformula) ->
				(let new_mappings =
					List.map
						(fun id ->
							(id, 
							(make_skolem_term quantified)))
						variables
				in
					skolemize_1
						quantified
						(List.rev_append new_mappings to_skolemize)
						subformula)
			| Atomic_formula ct ->
				(let skolemized_ct =
					(make_clausify_term
						(skolemize_term to_skolemize ct.proto_term))
				in
					skolemized_ct.negated <- ct.negated;
					Atomic_formula (skolemized_ct))
			| Negated_fof_formula subformula ->
				Negated_fof_formula
					(skolemize_1 quantified to_skolemize subformula)
			| Conjecture fof ->
				skolemize_1 quantified to_skolemize fof
			| _ -> failwith "Clausify.skolemize: formula not in required form."
	in
		(skolemize_1 [] [] formula)

		
		
(* returns a version of [formula] where unnecessary depth has been removed,
i.e. 'A or (B or C)' becomes 'A or B or C'
formula requirements: must be in skolem form *)	
let rec flatten formula =
	let rec flatten_1 operator flattened to_flatten =
		match to_flatten with
			| [] -> flattened
			| formula :: rest ->
				(match formula with
					| Atomic_formula _ ->
						(flatten_1 
							operator 
							(formula :: flattened) 
							rest)
					| Assoc_binary_formula (suboperator, subformulas) ->
						(if (operator = suboperator) then
							(flatten_1
								operator
								(List.rev_append subformulas flattened)
								rest)
						else
							(flatten_1 
								operator 
								(formula :: flattened) 
								rest))
					| Negated_fof_formula _ ->
						(flatten_1 
							operator 
							(formula :: flattened) 
							rest)
					| _ -> failwith "Clausify.flatten: formula not in required form.")
	in				
	match formula with
		| Atomic_formula _ -> formula
		| Assoc_binary_formula (operator, subformulas) ->
			let flat_subformulas =
				flatten_1 operator [] subformulas
			in
				Assoc_binary_formula (operator, flat_subformulas)
		| Negated_fof_formula subformula ->
			Negated_fof_formula
				(flatten subformula)
		| Conjecture fof ->
			flatten fof
		| _ -> failwith "Clausify.flatten: formula not in required form."
		
(* returns a version of [formula] where one subformula with a different operator
than the top operator is distributed
formula requirements: skolem form, flattening recommended *)			
let rec distribute formula =
	let rec distribute_1 operator checked to_check =
		(match to_check with
			| [] -> formula
			| formula :: rest ->
				(match formula with
					| Atomic_formula _ ->
						distribute_1 operator (formula :: checked) rest
					| Assoc_binary_formula (suboperator, subformulas) ->
						if (operator = suboperator) then
							(distribute_1 operator (formula :: checked) rest)
						else
							(let distributees =
								(List.rev_append checked rest)
							in
							let distributed_subformulas =
								List.map
									(fun distributor ->
										Assoc_binary_formula (operator, (distributor :: distributees)))
									subformulas
							in
								Assoc_binary_formula (suboperator, distributed_subformulas))
					| Negated_fof_formula _ ->
						distribute_1 operator (formula :: checked) rest
					| _ -> failwith "Clausify.distribute: formula not in required form."))
	in
	match formula with
		| Atomic_formula _ -> formula
		| Assoc_binary_formula (operator, subformulas) ->
			distribute_1 operator [] subformulas
		| Negated_fof_formula _ -> formula	
		| Conjecture fof ->
			distribute fof
		| _ -> failwith "Clausify.distribute: formula not in required form."

			
(* returns the conjunctive normal form of [formula]
formula requirements: must be in skolem form *)		
let rec normalize formula =
	let all_atomic subformulas =
		(List.for_all
			(fun subformula ->
				match subformula with
					| Atomic_formula _ -> true
					| Negated_fof_formula _ -> true
					| _ -> false)
			subformulas)
	in
	let formula = flatten formula
	in
	match formula with
		| Assoc_binary_formula (And, subformulas) ->
			(let subformulas =
				List.map normalize subformulas
			in
				flatten (Assoc_binary_formula (And, subformulas)))
		| Assoc_binary_formula (Or, subformulas) ->
			(if (all_atomic subformulas) then
				(formula)
			else
				(let subformulas =
					List.map normalize subformulas
				in
				let formula = flatten (Assoc_binary_formula (Or, subformulas))
				in
				let formula = distribute formula
				in
					normalize formula	))
		| Atomic_formula _ ->
			formula
		| Negated_fof_formula _ ->
			formula
		| Conjecture fof ->
			normalize fof
		| _ -> failwith "Clausify.normalize: formula not in required form."			
			
			
let rec get_term symbol_table proto_term =
	(match proto_term with
		| Int int -> Term.Integer int
		| Float float -> Term.Float float
		| Const const -> Term.Constant (Term.find_symbol symbol_table const 0)
		| Compound (func, subterms) ->
			let compound_functor = 
				Term.find_symbol 
					symbol_table 
					func 
					(List.length subterms)
			in
				let subterms = 
					Array.of_list
						(List.map 
							(fun subterm ->
								get_term symbol_table subterm)
							subterms)
				in
					Term.Compound
						(Term.make_compound compound_functor subterms)
		| Var (var, name) -> Term.Variable
						(Term.make_variable ~original_id:name var))

			
(* assigns the proper Term.terms to the clausify_terms in [formula]
formula requirements: skolem form *) 			
let rec assign_terms symbol_table formula =
	match formula with
		| Assoc_binary_formula (_, subformulas) ->
			List.iter
				(fun subformula ->
					assign_terms symbol_table subformula)
				subformulas
		| Atomic_formula ct ->
			ct.term <- get_term symbol_table ct.proto_term
		| Negated_fof_formula subformula ->
			assign_terms symbol_table subformula
		| Conjecture fof ->
			assign_terms symbol_table fof
		| _ ->
			failwith "Clausify.assign_terms: formula not in required form"
			
(* Copied from Convert.ml, to get around modular dependencies. *)			
let rec list_of_term_to_term st terms =
  list_of_term_to_term_1
    (Term.find_symbol st "." 2)
    (Term.find_symbol st "[]" 0)
    terms

and list_of_term_to_term_1 dot nil terms =
  match terms with
    [] -> 
      Term.Constant nil
  | term :: terms1 ->
      Term.Compound 
	(Term.make_compound
	   dot
	   [|term; list_of_term_to_term_1 dot nil terms1|])
			
let rec clausify 
	?(negate_conjecture:bool = true) ?(focus:bool = false) 
	?(partial_answers:bool = false) symbol_table formula = 
	let conjecture = ref false
	and add_false_answer = ref false
	and contains_focus_variable term =
		(	(* print_string (Term.string_of_term term); print_endline "!"; *)
		try
			(Termutils.iter_vars
				(fun variable ->
					if ((Term.variable_number variable) = !focus_variable) then
						(raise FOUND))
				term;
				false)
		with 
			| FOUND ->  true)
	in	
	let rec divide_into_clauses formula =
		(match formula with
			| Atomic_formula _ ->
				let result = [formula]
				in 
					[result]
			| Negated_fof_formula _ ->
				let result = [formula]
				in
					[result]
			| Assoc_binary_formula (Or, subformulas) ->
				let result =
					List.map
						(fun subformula ->
							match subformula with
								| Atomic_formula _ ->
									subformula
								| Negated_fof_formula _ ->
									subformula
								| _ -> failwith "Clausify.extract_clause_terms: formula not in required form.")
						subformulas
				in
					[result]
			| Assoc_binary_formula (And, subformulas) ->
				let clause_terms = ref []
				in
				List.iter
					(fun subformula ->
						match subformula with
							| Atomic_formula _ ->
								let result = [subformula]
								in
									clause_terms := result :: !clause_terms
							| Negated_fof_formula _ ->
								let result = [subformula]
								in
									clause_terms := result :: !clause_terms
							| Assoc_binary_formula (Or, subsubformulas) ->
								let result =
									List.map
										(fun subsubformula ->
											match subsubformula with
												| Atomic_formula _ ->
													subsubformula
												| Negated_fof_formula _ ->
													subsubformula
												| _ -> failwith "Clausify.extract_clause_terms: formula not in required form.")
										subsubformulas
								in
									clause_terms := result :: !clause_terms
							| _ -> failwith "Clausify.extract_clause_terms: formula not in required form.")
					subformulas;
				!clause_terms
			| Conjecture fof ->
				divide_into_clauses fof
			| _ -> failwith "Clausify.extract_clause_terms: formula not in required form.")
	and make_clause_sources literal_list =
		(let rec term_list_to_compound sym terms =
			(match terms with
				| [term] -> term
				| [] -> Term.Unbound
				| term :: rest ->
					Term.Compound
						(Term.make_compound
							sym
							[|term; (term_list_to_compound sym rest)|]))
		in			
		let positive_list = ref []
		and negative_list = ref []
		and check_pos_unit = ref false
		and false_head = ref false
		in
		List.iter
			(fun literal ->
				match literal with
					| Atomic_formula ct ->
						(let term = Term.copy_term_shared ct.term
						in
							positive_list := term :: !positive_list)
					| Negated_fof_formula af ->
						(match af with
							| Atomic_formula ct ->
								(let term = Term.copy_term_shared ct.term
								in
									negative_list := term :: !negative_list)
							| _ -> failwith "Clausify.clausify: formula not in required form.")
					| _ -> failwith "Clausify.clausify: formula not in required form.")
			literal_list;
		let positive = 
			(let compound =
				term_list_to_compound 
					(Term.get_disjunction_symbol symbol_table) 
					!positive_list
			in
				if (compound == Term.Unbound) then
					(false_head := true;
					Term.Constant (Term.get_false_symbol symbol_table))
				else
					(compound))
		and negative = 
			(let compound =
				term_list_to_compound 
					(Term.get_conjunction_symbol symbol_table)
					!negative_list
			in
				if (compound == Term.Unbound) then
					(check_pos_unit := true;
					Term.Constant (Term.get_true_symbol symbol_table))
				else
					(compound))			
		in			(*pt positive;pe " <POS";pt negative;pe "<NEG";*)
			(* positive units should not be treated as 'P :- true' clauses *)
			if (!check_pos_unit && ((List.length !positive_list)=1)) then
				(!positive_list)
			else
				(if (!conjecture && focus && !false_head && (contains_focus_variable negative)) then
					(add_false_answer := true;
					let answer =
						Term.Compound
							(Term.make_compound
								(Term.get_krh_answer_symbol symbol_table)
								[|Term.Variable
									(Term.make_variable ~original_id:(!(Cfg.answer_variable_name)) !focus_variable)|])
					in
					let source =
						Term.Compound
							(Term.make_compound
								(Term.get_rule_symbol symbol_table)
								[|answer; negative|])
					and partial_answer_clause_source_lists =
						if partial_answers then
							(Listutils.get_sublists !negative_list)
						else
							([])
					in
					let partial_answer_clause_sources =
						List.map
							(fun source_list ->
								let body =
									term_list_to_compound 
										(Term.get_conjunction_symbol symbol_table)
										source_list
								and head =
									list_of_term_to_term
										symbol_table
										source_list
								in
									Term.Compound
										(Term.make_compound
											(Term.get_rule_symbol symbol_table)
											[|head; body|]))
							partial_answer_clause_source_lists
					in
						source :: partial_answer_clause_sources) 
				else
					(let source =
						Term.Compound
							(Term.make_compound
								(Term.get_rule_symbol symbol_table)
								[|positive; negative|])
					in	(* Term.print_term source;print_endline "."; *)
						[source])))
	in
	match formula with
(* filter non-proper formulas (like command statements) *)
		| Non_formula non_formula ->
			(	assign_terms symbol_table non_formula;
			let pre_clauses = divide_into_clauses non_formula
			in
				List.concat
					(List.map 
						make_clause_sources 
						pre_clauses))
		| _ ->		(* print_endline (fof_to_string formula);print_endline ""; *)
			let formula =
				(match formula with
					| Conjecture fof ->
						conjecture := true;
						if negate_conjecture then
							(Negated_fof_formula fof)
						else
							(fof)
					| _ -> formula)
			in
			let arrowless_formula =
				de_arrow formula
			in		  (* print_endline (fof_to_string arrowless_formula);print_endline ""; *)
			let negation_pulled_formula =
				pull_negations arrowless_formula
			in		(* print_endline (fof_to_string negation_pulled_formula);print_endline ""; *)
			let quantified_formula =
				quantify_free_vars negation_pulled_formula
			in 		  (* print_endline (fof_to_string quantified_formula);print_endline ""; *)
			let var_renamed_formula =
				rename_vars quantified_formula
			in		 (* print_endline (fof_to_string var_renamed_formula);print_endline ""; *)
			let skolemized_formula =
				skolemize var_renamed_formula
			in		 (* print_endline (fof_to_string skolemized_formula);print_endline ""; *)
			assign_terms symbol_table skolemized_formula;
			let normalized_formula =
				normalize skolemized_formula
			in			(* print_endline (fof_to_string normalized_formula);print_endline ""; *)
			let pre_clauses = divide_into_clauses normalized_formula
			in		(* List.iter (fun pcl -> List.iter (fun pc -> print_endline (fof_to_string pc);print_endline "") pcl) pre_clauses; *)
			let clause_sources = 
				(List.concat
					(List.map 
						make_clause_sources 
						pre_clauses)) 
			in	  (* List.iter (fun cs -> Term.print_term cs;print_endline " <<<") clause_sources; *)
				if !add_false_answer then
					(let false_answer =
						Term.make_false_answer_source symbol_table
					in
						false_answer :: clause_sources)
				else
					clause_sources
