(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(****************************************************************************)
(****************************************************************************)
(** Discrimination Tree Indexing. *)
(****************************************************************************)
(****************************************************************************)

(* See e.g.: "Experiments with Discrimination-Tree Indexing and Path
   Indexing for Term Retrieval" by William McCune, 1990.
*)

exception Failed = Term.Failed

(****************************************************************************)
(* Term Functor (i.e. Index Key) Representation *) 
(****************************************************************************)

(* - Pervasives.compare could be taken as comparison function, provided
     no special ordering is required.
 *)

type hkey =
    Variable |
    Symbol of Term.symbol |
    Integer of int |
    Float of float |
    PVariable of int	(* only used for perfect discrimination tree indexing *)

let print_hkey k =
	match k with
		Variable -> print_string "X"
		| Symbol s -> print_string (Term.symbol_name s);print_string "/";print_int (Term.symbol_arity s)
		| Integer i -> print_int i
		| Float f -> print_float f  
		| PVariable i -> print_string "X_";print_int i
		
let print_fterm fterm =
	List.iter
		(fun k -> print_hkey k;print_string " ")
		fterm   		  
    
let hkey_equal k1 k2 =
  match k1, k2 with
    Variable, Variable ->
      true
  | Symbol s1, Symbol s2 ->
      s1 == s2
  | Integer n1, Integer n2 ->
      n1 = n2
  | Float n1, Float n2 ->
      n1 = n2
  | PVariable v1, PVariable v2 ->
  	  v1 = v2
  | _, _ ->
      false

let term_hkey ?(perfect:bool = false) term =
  match term with
    Term.Compound k ->
      Symbol (Term.compound_functor k)
  | Term.Constant s ->
      Symbol s
  | Term.Variable v ->
  	(if perfect then
  		(PVariable (Term.get_variable_number v))
  	else
      	(Variable))
  | Term.Integer n ->
      Integer n
  | Term.Float n ->
      Float n
  | Term.Unbound ->
      raise (Failure "Bad term used for indexing.")
	
let hkey_arity hkey =
  match hkey with
    Symbol s ->
      Term.symbol_arity s
  | _ -> 0


(****************************************************************************)
(* TableType *)
(****************************************************************************)

module type TableType =
  sig
    type key = hkey
    type 'a t
    val make_table: unit -> 'a t
    val add: 'a t -> key -> 'a -> unit
    val remove: 'a t -> key -> unit
    val find: 'a t -> key -> 'a
    val iter: (key -> 'a -> unit) -> 'a t -> unit
    val iter_and_throw_away: (key -> 'a -> unit) -> 'a t -> unit
  end

module Make (Entry: Index.EntryType) (Table: TableType) =
  struct

(****************************************************************************)
(* Index Data Structure *)
(****************************************************************************)

    type entry = Entry.entry

    type entries = {mutable list: entry list}

    type node =
	DtInnerNode of table
      | DtLeafNode of entries
	    
    and table =
	node Table.t
	  
    type index =
	{mutable root:node;
	 mutable size:int}
	  
    let index_root index = index.root
	
    let make_index () =
      {root=DtInnerNode (Table.make_table ()); size=0}

    let clear index =
      index.root <- DtInnerNode (Table.make_table ());
      index.size <- 0

    let size index =
      index.size

(****************************************************************************)
(* Insertion *)
(****************************************************************************)
	
    let rec insert ?(perfect:bool = false) index term entry = 
      match index_root index with
	DtInnerNode table ->
	  ignore (insert_1 ~perfect:perfect entry term table 1);
	  index.size <- index.size + 1
      | _ ->
	  raise (Failure "Bad index toplevel type.")
	    
    and insert_1 ?(perfect:bool = false) entry term table n = 
      (* n is a counter of the terms to be processed. It is used to
	 find out when we are at the last subterm in the preorder. *)
      let hkey = term_hkey ~perfect:perfect term in
      let hkey_arity = (hkey_arity hkey) in
      let n1 = n + hkey_arity - 1 in
      let node =
	try 
	  Table.find table hkey
	with Not_found ->
	  let node =
	    if n1 = 0 then (* At the last subterm in the preorder. *)
	      DtLeafNode {list=[]}
	    else
	      DtInnerNode (Table.make_table ())
	  in
	  Table.add table hkey node; (* We could also use replace here. *)
	  node
      in
      match n1, node with
	0, DtLeafNode entries ->
	  entries.list <- entry :: entries.list;
	  table, n1 (* Return the table just to stay well typed. *)
      | _, DtInnerNode table1 ->
	  ( match term with
	    Term.Compound k ->
	      insert_1_args
	    ~perfect:perfect
		entry
		(Term.compound_args k)
		hkey_arity
		0
		table1
		n1
	  | _ -> table1, n1 )
      | _ , _ ->
	  raise (Failure "Bad index structure (insert).")
	    
    and insert_1_args ?(perfect:bool = false) entry args arity i table n =
      if i >= arity then
	table, n
      else
	let table1, n1 = insert_1 ~perfect:perfect entry args.(i) table n in
	insert_1_args ~perfect:perfect entry args arity (i+1) table1 n1
	  
	  
(****************************************************************************)
(* Retrieve Whole Content *)
(****************************************************************************)
	  
    let rec iter_all f index =
      if index.size > 0 then
	iter_all_1 f (index_root index)
	
    and iter_all_1 f node =
      match node with
	DtInnerNode table ->
	  Table.iter (fun k v -> iter_all_1 f v) table
      | DtLeafNode entries ->
	  List.iter f entries.list
	    

(****************************************************************************)
(* Retrieve Whole Content And Throw Away While Iterating *)
(****************************************************************************)
	    
    let rec iter_all_and_throw_away f index =
      if index.size > 0 then
	iter_all_and_throw_away_1 f (index_root index)
	
    and iter_all_and_throw_away_1 f node =
      match node with
	DtInnerNode table ->
	  Table.iter_and_throw_away 
	    (fun k v -> iter_all_and_throw_away_1 f v) table
      | DtLeafNode entries ->
	  entries_iter_and_throw_away f entries
	    
    and entries_iter_and_throw_away f entries =
      match entries.list with
	[] -> ()
      | a::l ->
	  entries.list <- l;
	  f a;
	  entries_iter_and_throw_away f entries


(****************************************************************************)
(* Auxiliary Function: Flatten Term *)
(****************************************************************************)
	    
(* - A flattened term is just a list of hkeys in preorder. Skipping is
   done by using hkey_arity.
   
   - This implementation of flatten_term is not tail recursive. It could
   be if pointers were used.
   
   - A flattened term is created for a query term before lookup
   in the access functions. This simplifies the lookup functions.
   
   - The lookup functions work without backtracking and without allocation
   of extra data.
   
   - the optional boolean argument [perfect] switches the flattening of variables
   between an anonymous and a distinguished representation; i.e. given a term
   p(f(X, Y)):
      perfect = false (default)	: p, f, *, *
      perfect = true			: p, f, X, Y
 *)

  
    let rec flatten_term ?(perfect:bool = false) term context =
      flatten_term_1 ~perfect:perfect context term []
	
    and flatten_term_1 ?(perfect:bool = false) context term rest =
      let (term, context) = Term.dereference term context in
      let hkey = term_hkey ~perfect:perfect term in
      match term with
	Term.Compound k ->
	  hkey :: flatten_term_1_args
	  	   ~perfect:perfect
		   context
		   rest
		   (Term.compound_args k)
		   (hkey_arity hkey)
		   0
      | _ -> hkey :: rest
		      
    and flatten_term_1_args ?(perfect:bool = false) context rest args arity i =
      if i >= arity then
	rest
      else
	flatten_term_1
	  ~perfect:perfect
	  context
	  args.(i)
	  (flatten_term_1_args
	  	 ~perfect:perfect
	     context
	     rest
	     args
	     arity
	     (i + 1))
	  
(* Skip some terms in fterm. count is the number of terms to be skipped. *)

    let rec skip_terms_fterm fterm count =
      if count = 0 then
	fterm
      else
	match fterm with 
	  hkey::fterm1 ->
	    skip_terms_fterm fterm1 (count + (hkey_arity hkey) - 1)
	| _ ->
	    raise (Failure "Bad flattened term structure.")
	      
(* Like skip_terms_fterm, but returns the skipped part and the remainder as two separate fterms. *) 
	let split_skip_terms_fterm fterm count =    
		let rec split_skip_terms_fterm_1 skipped fterm count =
	      if count = 0 then
			((List.rev skipped), fterm)
	      else
			(match fterm with 
			  hkey::fterm1 ->
			    split_skip_terms_fterm_1 (hkey :: skipped) fterm1 (count + (hkey_arity hkey) - 1)
			| _ ->
			    raise (Failure "DtIndex.split_skip_fterms: bad flattened term structure."))
		in
			split_skip_terms_fterm_1 [] fterm count
			
	let unflatten_fterm fterm =
		let make_term hkey subterms =
			if (List.length subterms) = (hkey_arity hkey) then
				(match hkey with
					| PVariable v ->
						Term.Variable
							(Term.make_variable v)
					| Symbol s ->
						(if (List.length subterms) = 0 then
							(Term.Constant s)
						else
							(Term.Compound
								(Term.make_compound s (Array.of_list subterms))))
					| Float f ->
						Term.Float f
					| Integer i ->
						Term.Integer i
					| _ -> (print_fterm fterm;print_endline ".";failwith "DtIndex.unflatten_fterm: anonymous variable in fterm."))
			else
				(raise Failed)
		in
(* assumes fterm to be a list/concatenation of flattened terms; a list of the unflattened
terms is returned (i.e. [f; x; g; a; b] -> [f(x); g(a, b)]*)	
		let rec make_term_list fterm =
			match fterm with
				| hkey :: tail ->
					let subterms_fterm, rest_fterm =
						split_skip_terms_fterm tail (hkey_arity hkey)
					in
						let term = make_term hkey (make_term_list subterms_fterm)
						in
							(term :: (make_term_list rest_fterm))
				| _ -> []
		in
(* the initial call; similar to make_term_list, but here fterm represents
a single term, not a concatenation, so only one unflattened term is returned *)	
		match fterm with
			| hkey :: tail ->
				let subterms_fterm, rest_fterm =
					split_skip_terms_fterm tail (hkey_arity hkey)
				in
					if (rest_fterm = []) then
						(make_term hkey (make_term_list subterms_fterm))
					else
						(failwith "DtIndex.unflatten_fterm: bad fterm")
			| _ -> failwith "DtIndex.unflatten_fterm: bad fterm"
	

	    
	    
(****************************************************************************)
(* Retrieve Instance Candidates *)
(****************************************************************************)
	      
    let rec iter_instance_candidates f index term context =
      if index.size > 0 then
	iter_inst f (flatten_term term context) (index_root index)
	
    and iter_inst f fterm node = 
      match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    match hkey with
	      Variable ->
		iter_inst_skip f fterm1 node 1
	    | _ ->
		try
		  let node1 = Table.find table hkey in
		  iter_inst f fterm1 node1
		with Not_found ->
		  ()
	  end
      | [], DtLeafNode entries ->
	  List.iter f entries.list
      | _, _ ->
	  raise (Failure "Bad index structure (instance).")
	    
    and iter_inst_skip f fterm node count =
      if count = 0 then
	iter_inst f fterm node
      else
	match node with
	  DtInnerNode table ->
	    Table.iter
	      (fun hkey node1 ->
		iter_inst_skip f fterm node1 (count + (hkey_arity hkey) - 1))
	      table
	| _ -> raise (Failure "Bad index structure (skip/inst).")
	
   
    let rec perfect_iter_instance_candidates 
    	?(substitution:Term.substitution = Term.empty_substitution) 
    	f index term context =
      if index.size > 0 then
	perfect_iter_inst substitution f (flatten_term ~perfect:true term context) (index_root index)
	
    and perfect_iter_inst substitution f fterm node = 
      match fterm, node with
	hkey::fterm1, DtInnerNode table ->	
	  begin
	    match hkey with
	      Variable ->
	      	()
		| PVariable v ->
			perfect_iter_inst_skip substitution f fterm1 node 1 v []
	    | _ ->
		try
		  let node1 = Table.find table hkey in
		  perfect_iter_inst substitution f fterm1 node1
		with Not_found ->
		  ()
	  end
      | [], DtLeafNode entries ->
	  List.iter (fun entry -> f substitution entry) entries.list
      | _, _ ->
	  raise (Failure "DtIndex.perfect_iter_instance_candidates: bad index structure (instance).")
	    
    and perfect_iter_inst_skip substitution f fterm node count var skipped =
      if count = 0 then
	(let term = unflatten_fterm (List.rev skipped)
	in 
		try
			(let substitution' = Term.add_binding var term substitution
			in
				perfect_iter_inst substitution' f fterm node)
		with
			| Term.Failed -> ())
      else
	match node with
	  DtInnerNode table ->
	    Table.iter
	      (fun hkey node1 ->
		perfect_iter_inst_skip substitution f fterm node1 (count + (hkey_arity hkey) - 1) var (hkey :: skipped))
	      table
	| _ -> raise (Failure "Bad index structure (skip/inst).")	
	
	
	
	      
    let rec iter_variant_candidates f index term context =
      if index.size > 0 then
	iter_variant f (flatten_term term context) (index_root index)
	
    and iter_variant f fterm node = 
      match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
		try
		  let node1 = Table.find table hkey in
		  iter_variant f fterm1 node1
		with Not_found ->
		  ()
	  end
      | [], DtLeafNode entries ->
	  List.iter f entries.list
      | _, _ ->
	  raise (Failure "Bad index structure (variant).")	
	
(****************************************************************************)
(* Retrieve Generalization Candidates *)
(****************************************************************************)
	      
	  
	  
    let rec iter_generalization_candidates f index term context =
      if index.size > 0 then
	iter_gen f (flatten_term term context) (index_root index)
	
    and iter_gen f fterm node =
      match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    begin
	      try
		let node1 = Table.find table Variable in
		iter_gen 
		  f
		  (skip_terms_fterm fterm 1)
		  node1
	      with Not_found ->
		()
	    end;
	    match hkey with
	      Variable ->
		()
	    | _ ->
		try 
		  let node1 = Table.find table hkey in
		  iter_gen f fterm1 node1
		with Not_found ->
		  ()
	  end
      | [], DtLeafNode entries ->
	  List.iter f entries.list
      | _, _ ->
	  raise (Failure "Bad index structure (generalization).")
	    

    let rec perfect_iter_generalization_candidates
    	?(substitution:Term.substitution = Term.empty_substitution) f index term context =	
      if index.size > 0 then
	perfect_iter_gen substitution f (flatten_term ~perfect:true term context) (index_root index)
	
    and perfect_iter_gen substitution f fterm node = 
      match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  (let term = ref Term.Unbound
	  and fterm' = ref []
	  in	
	  Table.iter
	  	(fun hkey' node' -> 
	  		match hkey' with
	  			| PVariable v ->	
	  				(if !term == Term.Unbound then
	  					(let skipped, remainder =
	  						split_skip_terms_fterm fterm 1
	  					in
	  						term := unflatten_fterm skipped;
	  						fterm' := remainder);
	  				try
		  				(let substitution' = 	
		  					Term.add_binding v !term substitution
		  				in	
		  					perfect_iter_gen
		  						substitution'
		  						f
		  						!fterm'
		  						node')
		  			with
		  				| Failed -> ())
		  		| _ -> ())
		  table;
	  match hkey with
	  	| PVariable _ -> ()
	  	| _ ->
	  		(try
	  			(let node' = Table.find table hkey
	  			in
	  				perfect_iter_gen substitution f fterm1 node')
	  		with
	  			| Not_found -> ()))
      | [], DtLeafNode entries ->
	  List.iter (fun entry -> f substitution entry) entries.list
      | _, _ ->
	  (raise (Failure "DtIndex.perfect_iter_gen: bad index structure (generalization)."))
	  
	  
	  
	  	    
(****************************************************************************)
(* Retrieve Unification Candidates *)
(****************************************************************************)
	    
    let rec iter_unification_candidates f index term context =
      if index.size > 0 then
	iter_unif f (flatten_term term context) (index_root index)
	
    and iter_unif f fterm node =
      match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    match hkey with
	      Variable ->
		iter_unif_skip f fterm1 node 1
	    | _ ->
		begin
		  try 
		    let node1 = Table.find table Variable in
		    iter_unif
		      f
		      (skip_terms_fterm fterm 1)
		      node1
		  with Not_found ->
		    ()
		end;
		begin
		  try 
		    let node1 = Table.find table hkey in
		    iter_unif f fterm1 node1
		  with Not_found ->
		    ()
		end
	  end
      | [], DtLeafNode entries ->
	  List.iter f entries.list
      | _, _ ->
	  raise (Failure "Bad index structure (unif).")	
	    
    and iter_unif_skip f fterm node count =
      if count = 0 then
	iter_unif f fterm node
      else
	match node with
	  DtInnerNode table ->
	    Table.iter
	      (fun hkey node1 ->
		iter_unif_skip f fterm node1 (count + (hkey_arity hkey) - 1))
	      table
	| _ -> raise (Failure "Bad index structure (skip/unif).")
	      
	      
(****************************************************************************)
(* Remove Variant *)
(****************************************************************************)
	      
(* - remove_instances seems more useful, but difficult to implement
   with tables, that do not have simultaneous iteration and removing
   operations.
 *)
	      
    exception Found
	
    let is_empty_table table =
      try
	Table.iter 
	  (fun k v -> raise Found) table;
	true
      with Found ->
	false
	  
    let rec remove_variant index ?(equation:bool = false) ?(key:Term.term = Term.Unbound) term =
    	let key =
    		(if (key == Term.Unbound) then
    			(term)
    		else
    			(key))
    	in
      ignore
	(rm_variant ~equation:equation term 
	   (flatten_term key Term.null_context)
	   (index_root index));
      index.size <- index.size - 1;
    if equation then
    	(let flipped_term = Term.binary_flip term
    	in
    	let flipped_key =
    		(if (key == term) then
    			(flipped_term)
    		else
    			(key))
    	in
		ignore
		 (rm_variant ~equation:equation flipped_term 
		   (flatten_term flipped_key Term.null_context)
		   (index_root index));
	      index.size <- index.size - 1)
	
    and rm_variant ?(equation:bool = false) term fterm node =
      match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    try
	      let node1 = Table.find table hkey in
	      if rm_variant ~equation:equation term fterm1 node1 then
		begin
		  Table.remove table hkey;
		  is_empty_table table
		end
	      else
		false
	    with Not_found ->
	      false
	  end
      | [], DtLeafNode entries ->
	  entries.list <- 
	  	(let flipped_term =
	  		if equation then
	  			Term.binary_flip term
	  		else
	  			term
	  	in 
	  		list_remove_variant ~equation:equation term flipped_term entries.list);
	  entries.list = []	  
      | _, _ ->
	  raise (Failure "Bad index structure (generalization).")
	    
    and list_remove_variant ?(equation:bool = false) term flipped_term entries =
      match entries with
	[] ->
	  []
      | entry :: entries1 ->
      let entry_term = Entry.get_term entry
      in
		  if (Term.variant term entry_term)
		  	||
		  	(equation && (Term.variant flipped_term entry_term)) then
		    entries1
		  else
		    entry :: (list_remove_variant ~equation:equation term flipped_term entries1)
		      
		      
(****************************************************************************)
(* Stuff for Debugging *)
(****************************************************************************)
		      
    let dbg_print_hkey hkey =
      match hkey with
	Symbol sym ->
	  print_string (Term.symbol_name sym);
	  print_string "/";
	  print_int (Term.symbol_arity sym)
      | Variable ->
	  print_string "*/0"
      | _ ->
	  print_string "?/0"
	    
    let rec indent d =
      if d <= 0 then
	()
      else
	begin
	  print_string "  ";
	  indent (d - 1)
	end
	  
    let rec dbg_print_index index =
      let node = index_root index in
      dbg_print_index_1 node 0
	
    and dbg_print_index_1 node d =
      match node with
	DtInnerNode table ->
	  Table.iter
	    (fun k v ->
	      indent d;
	      dbg_print_hkey k;
	      print_newline ();
	      dbg_print_index_1 v (d + 1))
	    table
      | DtLeafNode entries ->
	  indent d;
	  print_string "[";
	  List.iter 
	    (fun e ->
	      PrologPrint.print (Entry.get_term e);
	      print_string " ")
	    entries.list;
	  print_string "]\n"

end (* Make *)

(****************************************************************************)
(* TermEntry *)
(****************************************************************************)

(* Implementation of EntryType, that uses terms directly as entries. *)

module TermEntry =
  struct
    type entry = Term.term
    let get_term entry = entry
  end

(****************************************************************************)
(* HKeyHashtbl *)
(****************************************************************************)

(* Implementation of TableType as hash table. *)

module HKeyHashtbl =
  Hashtbl.Make(
  struct
    type t = hkey
    let equal = hkey_equal
    let hash = Hashtbl.hash
  end)

(* Hmm, Hashtbl.hash looks for "meaningful" components of its
   argument. It may be more efficient to use the object identity directly
   for symbols. Is there a function associating integers with object
   identities? *)    

module HashTableImplementation =
  struct
    type key = hkey
    type 'a t = 'a HKeyHashtbl.t
    let make_table () = HKeyHashtbl.create 8
    let add = HKeyHashtbl.add
    let remove = HKeyHashtbl.remove
    let find = HKeyHashtbl.find
    let iter = HKeyHashtbl.iter
    let iter_and_throw_away f t = 
      (* Hashtables from the standard library do not to allow removal
	 of entries while iterating. *)
      HKeyHashtbl.iter f t;
      HKeyHashtbl.clear t
  end

(****************************************************************************)
(* Alist Table *)
(****************************************************************************)

(* Implementation of TableType as association list. *)

module AlistTableImplementation =
  struct

    type key = hkey

    type 'a t =  {mutable alist: (hkey * 'a) list}

    let make_table () = {alist=[]}

    let add table k v = table.alist <- (k,v) :: table.alist

    let rec remove table k =
      table.alist <- remove_assoc k table.alist
    and remove_assoc x = function
	[] -> []
      | (a, b as pair) :: l -> 
	  if hkey_equal a x then l else pair :: remove_assoc x l

    let rec find table k =
      assoc k table.alist
    and assoc x = function
	[] -> raise Not_found
      | (a, b)::l -> if hkey_equal a x then b else assoc x l

    let rec iter f table =
      iter_1 f table.alist
    and iter_1 f = function
	[] -> ()
      | (a,b)::l -> f a b; iter_1 f l

    let rec iter_and_throw_away f table =
      match table.alist with
	[] -> ()
      | (a,b)::l ->
	  table.alist <- l;
	  iter_and_throw_away f table
  end

(****************************************************************************)
(* Exported Combinations *)
(****************************************************************************)

module TermHashIndex =
  Make(TermEntry)(HashTableImplementation)

module TermAlistIndex =
  Make(TermEntry)(AlistTableImplementation)

