(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*				  Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*)


let pt t = PrologPrint.print t
let pe s = print_endline s
let pl l = print_int (List.length l)
let ps s = print_string s
let sp = print_string " "
let pi i = print_int i





type env =
    { symbol_table: Term.symbol_table;
      predicate_table: Predicate.predicate_table;
      negative_units: NegativeUnits.store;
      subterm_table: Ic.subterm_table;
      mutable rules: Term.term list;
      mutable clauses: Ic.clause list;
      mutable facts: Fact.fact list;
      mutable current_state: Tableau.State.state;
      current_state_ref: Tableau.State.state ref;
      mutable states: Tableau.State.state list;
      mutable redundant_store: Ic.redundant_store;
      redundant_store_trail_ref: (Ic.redundant_store list) ref;
      cfg: Cfg.cfg;
      clause_aux: Ic.Aux.aux;
      mutable shared_input_table: SharedInput.table;
      mutable current_result:Tableau.result;
      mutable current_answers:Fact.fact list;
      mutable outside_partition: (bool ref) list;
    }

let make_env cfg = 
  let st = Term.make_symbol_table () 
  and pt = Predicate.make_predicate_table ()
  and nu = NegativeUnits.make_store ()
  and sut = Ic.make_subterm_table ()
  and rs = Ic.make_redundant_store ()
   in
  let ca = Ic.Aux.make_aux cfg pt st nu 
   in	
  Builtin.init_builtins st;
  Evaluable.init_evaluables st;
  ignore(Ic.find_subterm sut (Term.get_equality_symbol st));	(* install the equality symbol as subterm
  right away - it is the only predicate symbol to be used as a subterm as well *)
  ignore(Ic.find_subterm sut (Term.get_krh_binding_symbol st));
  ignore(Ic.find_subterm sut (Term.get_head_literal_wrapper_symbol st));
  ignore(Ic.find_subterm sut (Term.get_body_literal_wrapper_symbol st));
  let krh_dom_pred = Predicate.find_predicate pt (Term.get_krh_dom_symbol st) 
  and krh_partial_answer_pred = Predicate.find_predicate pt (Term.get_krh_partial_answer_symbol st)
  in
  Predicate.ensure_ready_to_assert ~quiet:true krh_dom_pred;
  Predicate.ensure_ready_to_assert ~quiet:true krh_partial_answer_pred;
  let state =
	Tableau.State.obtain_state pt nu sut [] [] [] rs
  in
  let state_ref = ref state
  and rs_ref = ref []
  in
  {symbol_table=st;
   predicate_table=pt;
   negative_units=nu;
   subterm_table=sut;
   rules=[];
   clauses=[];
   facts=[];
   current_state=state;
   current_state_ref=state_ref;
   states=[];
   redundant_store=rs;
   redundant_store_trail_ref=rs_ref;
   cfg=cfg;
   clause_aux=ca;
   shared_input_table=SharedInput.null_table;
   current_result=Tableau.initial_result cfg pt st nu sut ca state_ref rs_ref;
   current_answers=[];
   outside_partition=[];
 }

let get_cfg env =
  env.cfg

let get_clause_aux env =
  env.clause_aux  
  
let get_symbol_table env =
  env.symbol_table

let get_predicate_table env =
  env.predicate_table

let get_rules env =
  env.rules

let get_clauses env =
	env.clauses
	
let get_facts env =
	env.facts	
	
let get_negative_units env =
  env.negative_units
  
let get_subterm_table env =
  env.subterm_table
  
let get_redundant_store env =
	env.redundant_store
	
let get_current_state_ref env =
	env.current_state_ref
	
let get_redundant_store_trail_ref env =
	env.redundant_store_trail_ref

exception FOUND_INT of int
	
let get_lemmas env =
(* 1. finds the index layer containing the lemmas, that is, the layer containing anything derived
before the first branch split
2. returns the units stored in this index layer *)
	(let input_layer_count =
		(match (Tableau.State.get_extent_state !(env.current_state_ref)) with
			| [] ->
				(failwith "Env.get_lemmas: no predicates.")
			| index_state :: _ ->
				Access.Store.layer_count 
					(Access.Store.get_content index_state))
	and current_layer_count =
		(try
			(Predicate.PredicateTable.iter
				(fun _ predicate ->
					raise
						(FOUND_INT 
							(Access.Store.layer_count
								(Access.Reader.get_store
									(Predicate.default_reader predicate)))))
				env.predicate_table;
			failwith "Env.get_lemmas: no predicates.")
		with
			| FOUND_INT count -> count)		
	in	
	let layer_offset = 
		(let states_count =
			(List.length env.states)
		in
			if states_count = 0 then
				(1)
			else 
				(input_layer_count - states_count))
	and lemmas = ref []
	and partial_answer_predicate =
 		Predicate.find_predicate
 			env.predicate_table
 			(Term.get_krh_partial_answer_symbol env.symbol_table)
	in	
	let lemma_layer = pi layer_offset;pe " <O";
		(current_layer_count - (input_layer_count + layer_offset)) 
	in	
		Predicate.PredicateTable.iter
			(fun _ predicate -> 
				if not (predicate == partial_answer_predicate) then
					(Access.Reader.iter_all_nth_layer
						(fun fact ->
							if (Fact.is_redundant fact) then 
								()
							else
								(lemmas := fact :: !lemmas))
						lemma_layer
						(Predicate.default_reader predicate)))
			env.predicate_table;
		!lemmas)	

type statement_type =
    Meta
  | Unit
  | Rule
  | Disjunction
  | False

let classify_statement_functor env sym =
  let st = env.symbol_table in
  if sym == Term.get_meta_functor_symbol st then
    Meta
  else if sym == Term.get_rule_symbol st then
    Rule
  else if sym == Term.get_disjunction_symbol st then
    Disjunction
  else if sym == Term.get_false_symbol st then
    False
  else
    Unit

let ensure_ready_to_assert env symbol =								
  let pred = Predicate.find_predicate env.predicate_table symbol in	
  Predicate.ensure_ready_to_assert pred;
  pred

let ensure_ready_to_run env symbol =
  let pred = Predicate.find_predicate env.predicate_table symbol in
  Predicate.ensure_ready_to_run pred;
  pred

let lookup_predicate env name arity =
  Predicate.PredicateTable.find 
    env.predicate_table
    (Term.find_symbol env.symbol_table name arity)

let abolish_predicate env name arity =
(*Predicate.pript env.predicate_table;*)
  Predicate.abolish
    env.predicate_table
    (Term.find_symbol env.symbol_table name arity);
(* Predicate.pript env.predicate_table; *)
  NegativeUnits.clear env.negative_units
    
  
let find_predicate env name arity =
  Predicate.find_predicate
    env.predicate_table
    (Term.find_symbol env.symbol_table name arity)

let add_fact env fact =
	env.facts <- fact :: env.facts    
    
let remove_fact env fact =
	let rec remove_fact' so_far to_do fact =
		match to_do with
			| [] -> ()
			| fact' :: rest ->
				if (fact' == fact) then
					(env.facts <- (List.append so_far rest))
				else
					(remove_fact' (fact' :: so_far) rest fact)
	and facts = env.facts
	in
		remove_fact' [] facts fact
	
(*    
let add_rule env clause =
  env.rules <- clause :: env.rules
*)

(* If the input contains at least one clause that has a disjunctive
head and at least one clause that is not range restricted, then domain
enumerating clauses will have to be generated. This function checks each
clause for the aforementioned features and sets global flags in the
cfg-object accordingly; these flags will be checked later to make the
decision regarding the creation of domain clauses. *)
let domain_clauses_check env clause =
	let cfg = env.cfg
	in
		if (Cfg.get_disjunctive_input cfg) then ()
		else
			(if (Ic.is_disjunction clause) then
				(Cfg.set_disjunctive_input cfg true));
		if (Cfg.get_non_range_restricted_input cfg) then ()
		else
			(if (Ic.is_range_restricted clause) then ()
			else 
				(Cfg.set_non_range_restricted_input cfg true))
				
				
	
let make_and_add_clause env term =
	let answer_clause =
		if (Convert.term_functor_symbol term) == (Term.get_rule_symbol env.symbol_table) then
			(let head = Convert.term_nth_arg term 0
			in
				(Convert.term_functor_symbol head) == (Term.get_krh_answer_symbol env.symbol_table))
		else
			(false)
	in	
	let clause = Ic.make_clause ~answer_clause:answer_clause env.clause_aux term
	in	
		Ic.register_function_symbols env.symbol_table clause;
		Ic.set_proof clause Fact.InputClauseProof;
		if (Ic.is_indexed env.subterm_table clause) then ()
		else (
			try (
				Ic.np_subsumption_processing_clause 
					env.subterm_table
					env.redundant_store
					clause;
				domain_clauses_check env clause;
				Ic.index_clause_subterms
					env.subterm_table
					clause;
				env.clauses <- clause :: env.clauses)
			with Ic.Redundant -> ())

let add_rule env term =
	make_and_add_clause env term;
	env.rules <- term :: env.rules

   
let add_disjunction env term =
	let disj_term = 
		(Term.Compound
       		(Term.make_compound
	  			(Term.get_rule_symbol env.symbol_table)
	  			[|term; (Term.Constant (Term.get_true_symbol env.symbol_table))|]))
	in		
		add_rule env disj_term

let add_false env =
	let false_term =
		(Term.Compound
       		(Term.make_compound
	  			(Term.get_rule_symbol env.symbol_table)
	  				[|(Term.Constant (Term.get_false_symbol env.symbol_table));
	    			(Term.Constant (Term.get_true_symbol env.symbol_table))|]))
	in
		add_rule env false_term


let clear_rules env =
  env.rules <- [];
  env.clauses <- [];
  let symbols = ref []
  in
  Ic.SubtermTable.iter 
  	(fun symbol _ ->
  		symbols := symbol :: !symbols)
  	env.subterm_table;
  List.iter
  	(fun symbol ->
  		Ic.abolish_subterm env.subterm_table symbol)
  	!symbols;
  NegativeUnits.clear env.negative_units;
  Cfg.set_ehyper env.cfg false;
  Cfg.set_disjunctive_input env.cfg false;
  Cfg.set_non_range_restricted_input env.cfg false;
  List.iter
  	(fun fact ->
  		let term = Fact.get_term fact
  		in
  		let sym = Convert.term_functor_symbol term 
  		in
			Predicate.assert_unit_clause 
				(ensure_ready_to_assert env sym) 
				term
				fact;		
			Ic.index_unit_subterms_no_top
				env.subterm_table
				env.symbol_table
				term
				fact;			
			Term.register_function_symbols
				env.symbol_table
				term)
  	env.facts

let reset_clauses env =
	env.clauses <- []; 
  	Ic.SubtermTable.iter 
  		(fun symbol subterm ->
  			Ic.clear subterm)
  		env.subterm_table;
  	NegativeUnits.clear env.negative_units;
	Cfg.set_ehyper env.cfg false;
	Cfg.set_disjunctive_input env.cfg false;
	Cfg.set_non_range_restricted_input env.cfg false;
  	List.iter 
	(fun rule ->
		make_and_add_clause env rule)
	env.rules
  
  
let begin_shared_input env =
  if env.shared_input_table == SharedInput.null_table then
    env.shared_input_table <- SharedInput.make_table ()
  else
    ()

let end_shared_input env =
  env.shared_input_table <- SharedInput.null_table

let unit_for_env env term =
  if env.shared_input_table == SharedInput.null_table then
    term
  else
    SharedInput.share_subterms env.shared_input_table term

let get_current_result env =
    env.current_result

let set_current_result env model =
  env.current_result <- model

let get_current_answers env =
	env.current_answers
	
let set_current_answers env answers =
	env.current_answers <- answers  
  
(* creates various additional clauses if required by the input *)  
let ensure_required_clauses env =

	if (Cfg.get_pretend_range_restricted_input_flag env.cfg) then
		(Cfg.set_non_range_restricted_input env.cfg false);

(* domain enumerating clauses *)
	if (((Cfg.get_disjunctive_input env.cfg) 
	&& (Cfg.get_non_range_restricted_input env.cfg)))
	then
		(let domain_clause_sources =
			Term.make_domain_clause_sources env.symbol_table
		in 
			List.iter
				(fun source_term -> 
					add_rule env source_term)
				domain_clause_sources);

(* answer predicate computation: 'false :- answer(X)' *)
	if (Cfg.get_tptp_fof_compute_answer_predicate env.cfg) then
		(let false_answer_source =
			Term.make_false_answer_source env.symbol_table
		in	
			add_rule env false_answer_source)	
			
			
let update_current_state env =
	let current_state = (* env.current_state *)
		Tableau.State.obtain_state
			env.predicate_table
			env.negative_units
			env.subterm_table
			env.facts
			env.clauses
			env.rules
			env.redundant_store
	in
		env.current_state <- current_state;
		env.current_state_ref := env.current_state			
			
			
			
let save_state env =
	update_current_state env;
	let current_state = env.current_state
(*		Tableau.State.obtain_state
			env.predicate_table
			env.negative_units
			env.subterm_table
			env.facts
			env.clauses
			env.rules
			env.redundant_store	*)
	in
		env.states <- current_state :: env.states;
		env.redundant_store <- Ic.make_redundant_store ();
		env.current_state <- Tableau.State.add_layer ~dont_add_to_negative_units:true current_state;
		env.current_state_ref := env.current_state;
		Tableau.State.install_state env.current_state
			
let install_state env state =
	env.facts <- Tableau.State.get_facts state;
	env.clauses <- Tableau.State.get_clauses state;
	env.rules <- Tableau.State.get_rules state;
	env.redundant_store <- Tableau.State.get_redundant_store state;
	env.current_state <- state;	
	env.current_state_ref := state;
	Tableau.State.install_state state
	
let install_previous_state env =
(* The layered index makes rolling back to a previous state tricky. Layering adds a new index
layer to each currently existing index. If a new subterm or predicate symbol is introduced
after the saving of the state, the new symbol will start with a fresh single-layer index,
whereas the others have one layer for each saved state. Loading the previous state will
drop the uppermost layer - unless it is the only one, in that case it remains. The result
of this is that when only the mechanisms of the layered index are used, then new symbols
will persist when a state is loaded that was saved before the introduction of these symbols.
Therefore the layer counts of the index stores must be checked before loading the previous state. 
Any single layer index has been introduced in the current state, so it should not persist, and it 
will be deleted with the abolish-commands. *)
	match env.states with
		| state :: states ->
			(Ic.undeclare_redundancies env.redundant_store;
			let predicates_to_abolish = ref []
			and subterms_to_abolish = ref []
			in
				Predicate.PredicateTable.iter
					(fun _ predicate ->
						let store = 
							Access.Reader.get_store 
								(Predicate.default_reader predicate)
						in
							(if ((Access.Store.layer_count store)=1) then
								(predicates_to_abolish := predicate :: !predicates_to_abolish)))
					env.predicate_table;
				Ic.SubtermTable.iter
					(fun _ subterm -> 
						let store =
							Ic.Reader.get_store
								(Ic.get_subterm_reader subterm)
						in
							(if ((Ic.SubtermStore.layer_count store)=1) then
								(subterms_to_abolish := subterm :: !subterms_to_abolish)))
					env.subterm_table;
				List.iter
					(fun predicate ->
						let symbol = Predicate.get_symbol predicate
						in
							Predicate.abolish
								env.predicate_table
								(Term.find_symbol 
									env.symbol_table 
									(Term.symbol_name symbol) 
									(Term.symbol_arity symbol)))
					!predicates_to_abolish;
				List.iter
					(fun subterm ->
						Ic.abolish_subterm
							env.subterm_table
							(Ic.get_subterm_symbol subterm))
					!subterms_to_abolish;
			Cfg.set_reassign_predicates env.cfg true;					
			install_state env state;
			env.states <- states)
		| [] -> ()	
		

		
(*		
let ensure_partitioning env =
	if (get_partitioning_flag env.partitioning) then
		()
*)
		
		
(* DEBUGGING *)		
let print_stats env =
let cnt_pt = ref 0
and cnt_st = ref 0
and cnt_pt_layers = ref 0
and cnt_st_layers = ref 0 
and cnt_pt_extent = ref 0
and cnt_st_extent = ref 0
in
let st_extent_size subterm =
	Ic.Writer.size (Ic.get_subterm_writer subterm)
in
Predicate.PredicateTable.iter 
	(fun _ pred ->
		let store = 
			Access.Reader.get_store 
			(Predicate.default_reader pred)
		in
		cnt_pt_extent := !cnt_pt_extent + (Predicate.extent_size pred);
		cnt_pt_layers:=!cnt_pt_layers + (Access.Store.layer_count store);
		cnt_pt:=!cnt_pt+1) 
	env.predicate_table;
Ic.SubtermTable.iter 
	(fun _ subterm ->
		let store = 
			Ic.Reader.get_store 
			(Ic.get_subterm_reader subterm)
		in
		cnt_st_extent := !cnt_st_extent + (st_extent_size subterm);
		cnt_st_layers:=!cnt_st_layers + (Ic.SubtermStore.layer_count store);
		cnt_st:=!cnt_st+1) 
	env.subterm_table;

let redundant_clauses, redundant_facts =
	(match !(env.redundant_store_trail_ref) with
		| [] -> (-1), (-1)
		| rs :: _ ->
			Ic.get_redundant_store_size rs)
	
in
let output = 
 "\n"
^ "ENV STATS START:\n"
^ "Number of states: "
^ (string_of_int (List.length env.states))
^ "\n"
^ "PredicateTable size: "
^ (string_of_int !cnt_pt)
^ "\n"
^ "SubtermTable size: "
^ (string_of_int !cnt_st)
^ "\n"
^ "Number of rules: "
^ (string_of_int (List.length env.rules))
^ "\n"
^ "Number of clauses: "
^ (string_of_int (List.length env.clauses))
^ "\n"
^ "Number of facts: "
^ (string_of_int (List.length env.facts))
^ "\n"
^ "Number of predicate layers: "
^ (string_of_int !cnt_pt_layers)
^ "\n"
^ "Number of subterm layers: "
^ (string_of_int !cnt_st_layers)
^ "\n"
^ "Number of redundant_stores: "
^ (string_of_int (List.length !(env.redundant_store_trail_ref)))
^ "\n"
^ "Number of redundant clauses: "
^ (string_of_int redundant_clauses)
^ "\n"
^ "Number of redundant facts: "
^ (string_of_int redundant_facts)
^ "\n"
^ "Predicate extent size: "
^ (string_of_int !cnt_pt_extent)
^ "\n"
^ "Subterm extent size: "
^ (string_of_int !cnt_st_extent)
^ "\n"
^ "Number of answers: "
^ (string_of_int (List.length env.current_answers))
^ "\n"
^ "ENV STATS END\n"
in
(*let tempfile, outchannel = Filename.open_temp_file "la_" ".txt"
in
output_string outchannel output;
flush outchannel   *)
pe output