/*
* Copyright (C) 2007 Universitaet Koblenz-Landau
* 
* This file is part of E-KRHyper.
* 
* E-KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* E-KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* E-KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*                 Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*
* This file has been adapted from the file parser_tptp.mly from the Darwin
* theorem prover. The following is the original copyright notice:
*	This file is part of the first order theorem prover Darwin
*	Copyright (C) 2006  The University of Iowa
*	
*	This program is free software; you can redistribute it and/or
*	modify it under the terms of the GNU General Public License
*	as published by the Free Software Foundation; either version 2
*	of the License, or (at your option) any later version.
*	
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*	
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



%{
  

   (* these are only valid for the current clause
     and have to be invalidated with init_clause for every new clause *)

  (* the variable id counter for the currently read term/clause *)
	    
	let var_id = ref 0
	
	let gen_var_id () =
	  var_id := !var_id + 1;
	  (!var_id-1)
	
	let reset_var_id () =
	  var_id := 0      
    
      
  (* mapping of the variable names (e.g. "X") of the currently read term/clause
     to variables. *)
  let (var_map: (string * int) list ref) =
    ref []



  (* reset everything in order to parse a new term/clause *)
  let init_clause () =
    reset_var_id ();
    var_map := []


	
  (* gets the variables associated with a string from the variable mapping
     creates a new mapping for a new variable *)
  let get_var (var_name: string) =
    try 
      (* way faster than List.assoc *)
      snd (
	List.find
	  (fun (var_name', _) ->
             var_name = var_name'
	  )
	  !var_map
      )
    with
      | Not_found ->
	  let new_var_id = 
	    gen_var_id ()
	  in
	    var_map := (var_name, new_var_id) :: !var_map;
	    new_var_id

  (* need to detect if input is fof and contains a conjecture *)
  (* is the input in fof format? *)
  let fof =
    ref false

  (* does the fof input contain a conjecture? *)
  let theorem =
    ref false


%}
  
%token <string> LOWER_WORD
%token <string> UPPER_WORD
%token <string> SINGLE_QUOTED
%token <string> VARIABLE
%token <string> FLOAT
%token <string> INTEGER
%token OR
%token LEFT_PARENTHESIS
%token RIGHT_PARENTHESIS
%token LEFT_BRACE
%token RIGHT_BRACE
%token LEFT_BRACKET
%token RIGHT_BRACKET
%token IMPLICATION
%token CLAUSE_END
%token NEGATION
%token SEMICOLON
%token COMMA
%token EQUALITY
%token EOI
%token META
%token FINDALL
%token NOT_SUBSUMED
%token <string> ARITHMETIC
%token <string> IS
%token <string> MULT
%token <string> PLUS
%token <string> MINUS
%token <string> DIV
%token <string> ABS
%token <string> FLOAT_OP
%token <string> SQRT
%token <string> TRUNCATE
%token <string> SIN
%token <string> COS

%start tme_input
%type <Clausify.proto_term> tme_input

%%

/* parse rules */

tme_input:
  | clause
      { init_clause ();
        $1 }
	  
  | EOI
  	{raise Err.END_OF_FILE}


clause: 
  | clause_head IMPLICATION clause_body CLAUSE_END 
      { Clausify.Compound
			(":-", [$1; $3])}

clause_head: 
  | literal
      { $1 }

clause_body:
  | body_literal
      { $1 }

body_literal:
 | math_literal
	{ $1 }
	
math_literal:
  | math_term ARITHMETIC math_term
	{ Clausify.Compound ($2, [$1; $3]) }	
  | term IS math_term
	{ Clausify.Compound ($2, [$1; $3]) }	

	
math_term:
  | string LEFT_PARENTHESIS math_term RIGHT_PARENTHESIS
	{ Clausify.Compound ($1, [$3]) }
  | number
	{ $1 }
  | variable
	{ $1 }
	

  		



literal:
  | predicate_term
      { $1 }
				
predicate_term:		
  | string LEFT_PARENTHESIS arguments RIGHT_PARENTHESIS
      { Clausify.Compound ($1, $3) }

term: 
  | string LEFT_PARENTHESIS arguments RIGHT_PARENTHESIS
      { Clausify.Compound ($1, $3) }
  | constant
  	{ $1 }
  | LEFT_BRACKET term_list RIGHT_BRACKET
	{ $2 }
  | constant DIV term 
      { Clausify.Compound ($2, [$1; $3]) }  
  | number
	{ $1 }
  | variable
	{ $1 }
  		

  		      
constant:
  | string
      { Clausify.Const $1 }
      
string:
  | LOWER_WORD
  		{ $1 }
  | IS
  		{ $1 }
	

			
term_list:
  | null
  	  { Clausify.Const "[]" }
  | term
	{ let nil = Clausify.Const "[]"
		in
			Clausify.Compound
				(".", [$1; nil]) }
  | term COMMA term_list
	{ Clausify.Compound
		(".", [$1; $3]) }
  | term OR term
	{ Clausify.Compound
		(".", [$1; $3]) }

variable:
  | VARIABLE
      { Clausify.Var (get_var $1) }

arguments:
  | term COMMA arguments
      { $1 :: $3 }
  | term
      { [$1] }

number:
  | FLOAT
	{ Clausify.Float (float_of_string $1) }
  | INTEGER
	{ Clausify.Int (int_of_string $1) }	
  | MINUS FLOAT
	{ Clausify.Float (float_of_string ($1^$2)) }
  | MINUS INTEGER
	{ Clausify.Int (int_of_string ($1^$2)) }
	
null:
      { }	
  

%%
