(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

let pe s = print_endline s

module type LayeredIndex =
  sig
    type entry
    type index
    val make_index : unit -> index
    val clear : ?msg_box:Msg.msg_box -> index -> unit
    val insert : ?perfect:bool -> index -> Term.term -> entry -> unit
    val remove_variant : ?msg_box:Msg.msg_box -> index -> ?equation:bool -> ?key:Term.term -> Term.term -> unit
    val iter_all : (entry -> unit) -> index -> unit
    val iter_all_nth_layer : (entry -> unit) -> int -> index -> unit    
    val iter_all_and_throw_away : ?msg_box:Msg.msg_box -> (entry -> unit) -> index -> unit
    val iter_instance_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
	val perfect_iter_instance_candidates :
		?substitution:Term.substitution -> (Term.substitution -> entry -> unit) 
		-> index -> Term.term -> Term.context -> unit		
    val iter_variant_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_generalization_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
	val perfect_iter_generalization_candidates :
		?substitution:Term.substitution -> (Term.substitution -> entry -> unit) 
		-> index -> Term.term -> Term.context -> unit
    val iter_unification_candidates :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_instance_candidates_ml :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_variant_candidates_ml :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_generalization_candidates_ml :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit
    val iter_unification_candidates_ml :
	(entry -> unit) -> index -> Term.term -> Term.context -> unit	
    val size : index -> int
	
    type state
    val obtain_state : index -> state
    val install_state : state -> unit
    val get_index : state -> index
    val get_content : state -> index    
    val add_layer : state -> state
    val new_layer : index -> state
    val layer_count : index -> int

  end

module Make (LayerType : Index.Index) =
  struct

    type entry = LayerType.entry
    type layer = LayerType.index
    type index1 = (layer list) ref
    type index = (layer list) ref
    type state = index * (layer list)

    let mutable_layer index =
      match !index with
	layer :: _ -> layer
      | _ ->
	  raise (Failure "Attempt to access layered index without layer.")
	  
	let nth_layer index n =
		try
			(List.nth !index n)
		with
			| Not_found ->
				(raise (Failure "Attempt to access non-existent layer in layered index."))	  

    let only_layer index msg_box =	(* print_int (List.length !index);print_endline "!"; *)
    	let error_string = "Layered index should have exactly one layer."
    	in
    	let error_string =
	    	(let prefix = ref "LayeredIndex.only_layer"
    		in
    			List.iter
    				(fun caller ->
    					prefix := !prefix ^ " called by " ^ caller)
    				(Msg.get_callers msg_box);
    			!prefix ^ ": " ^ error_string)
	    in
 (*   	(match caller with
    		| 0 -> print_endline "'only_layer' called by 'clear'"
    		| 1 -> print_endline "'only_layer' called by 'remove_variant'"
    		| 2 -> print_endline "'only_layer' called by 'iter_all_and_throw_away'"
    		| _ -> failwith "'only_layer' called by no-one");    		    		
   *)
      match !index with
	[layer] -> layer	
      | _ ->
	  raise (Failure error_string)
	  
	let layer_count index =
		List.length !index	  
	
    let make_index () = 
      ref [LayerType.make_index ()]
	    
    let clear ?(msg_box:Msg.msg_box = Msg.make_msg_box ()) index =
    	Msg.set_callers msg_box ("clear" :: (Msg.get_callers msg_box));
      LayerType.clear (only_layer index msg_box)

(*    let force_clear index =
    	let base_layer =
    		List.nth !index
    	match !index with
    		| [layer] -> LayerType.clear layer
    		| layer :: _ -> *)
    	  
      
    let remove_variant ?(msg_box:Msg.msg_box = Msg.make_msg_box ()) index = 
    	Msg.set_callers msg_box ("remove_variant" :: (Msg.get_callers msg_box));
      LayerType.remove_variant (only_layer index msg_box)

    let iter_all_and_throw_away ?(msg_box:Msg.msg_box = Msg.make_msg_box ()) f index =
    	Msg.set_callers msg_box ("iter_all_and_throw_away" :: (Msg.get_callers msg_box));
      LayerType.iter_all_and_throw_away f (only_layer index msg_box)

    let insert ?(perfect:bool = false) index t e  =
      (*
      print_string "III ";
      print_int (List.length !index);
      print_endline "";
      *)	 
      LayerType.insert ~perfect:perfect (mutable_layer index) t e

    let iter_all f index = 
      List.iter (fun layer -> LayerType.iter_all f layer) !index
      
    let iter_all_nth_layer f n index =
      LayerType.iter_all f (nth_layer index n)      

    let iter_instance_candidates f index t c =
      List.iter 
	(fun layer -> 
	  LayerType.iter_instance_candidates f layer t c)
	!index

	let perfect_iter_instance_candidates 
		?(substitution:Term.substitution = Term.empty_substitution) f index t c =	
		List.iter
			(fun layer ->
				LayerType.perfect_iter_instance_candidates
					~substitution:substitution
					f layer t c)
			!index
	
	let iter_instance_candidates_ml f index t c =
		LayerType.iter_instance_candidates f (mutable_layer index) t c	
	
    let iter_variant_candidates f index t c =
      List.iter 
	(fun layer -> 
	  LayerType.iter_variant_candidates f layer t c)
	!index	
	
	let iter_variant_candidates_ml f index t c =
		LayerType.iter_variant_candidates f (mutable_layer index) t c
	
		
    let iter_generalization_candidates f index t c =
      List.iter 
	(fun layer -> 
	  LayerType.iter_generalization_candidates f layer t c)
	!index

	let perfect_iter_generalization_candidates 
		?(substitution:Term.substitution = Term.empty_substitution) f index t c =	
		List.iter
			(fun layer ->
				LayerType.perfect_iter_generalization_candidates
					~substitution:substitution
					f layer t c)
			!index	
	
	
	let iter_generalization_candidates_ml f index t c =
		LayerType.iter_generalization_candidates f (mutable_layer index) t c	
	
    let iter_unification_candidates f index t c =
      List.iter 
	(fun layer -> 
	  LayerType.iter_unification_candidates f layer t c)
	!index

	let iter_unification_candidates_ml f index t c =
		LayerType.iter_unification_candidates f (mutable_layer index) t c	
	
    let size index =
      List.fold_left (fun x y -> x + (LayerType.size y)) 0 !index

    let obtain_state index =
      (index, !index)
      
    let get_index (index, content) =
      index	  
      
    let get_content (index, content) =
      ref content
	
    let install_state (index, content) = 
      index := content

    let add_layer (index, content) = 
      (index, (LayerType.make_index ()) :: content)
	
    let new_layer index =
      (index, [LayerType.make_index ()])

  end
