(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

let pt t = Term.print_term t
let pe s = print_endline s
let pl l = print_int (List.length l)
let sp = print_string " "
let ps s = print_string s
let pi i = print_int i



(****************************************************************************)
(* Handing of Declarations - Converting from and to Terms *)
(****************************************************************************)

let indextype_of_spec term =
  match Convert.term_to_string term with
    "alist_dt" ->
      Opt.AlistDt
  | "hash_dt" ->
      Opt.HashDt
  | "none" ->
      Opt.NoIndex
  | _ ->
      raise (Err.Error "Bad index type specifier.")

let indexargs_of_spec term =
  match term with
    Term.Constant sym ->
      begin
	match Term.symbol_name sym with
	  "all" ->
	    Opt.WholeTerm
	| _ ->
	    raise (Err.Error "Bad index args specifier.")
      end
  | Term.Integer int ->
      Opt.Argument int
  | _ ->
      let args = Convert.term_to_array_of_int term in
      let len = Array.length args in
      if len > 1 then
	Opt.Arguments args
      else if len = 1 then
	Opt.Argument args.(0)
      else
	Opt.WholeTerm

let process_index_decl env term =
  let (name, arity) = 
    Convert.predspec_pair (Convert.term_nth_arg term 0) in
  let pred = (Env.find_predicate env name arity) in
  let args = Convert.term_to_alist (Convert.term_nth_arg term 1) in
  let indextype_spec =
    try List.assoc "type" args
    with Not_found ->
      raise (Err.Error "Missing 'type' specifier in index declaration.")
  and indexargs_spec =
    try List.assoc "args" args
    with Not_found ->
      raise (Err.Error "Missing 'args' specifier in index declaration.")
  in
  Predicate.add_declaration
    pred
    (Opt.IndexDecl
       ((indextype_of_spec indextype_spec),
	(indexargs_of_spec indexargs_spec)))

let process_mode_decl env term =
  let template = (Convert.term_nth_arg term 0) in
  match template with
    Term.Compound compound ->
      let (name, arity) =
	Convert.term_functor_pair template in
      let pred = (Env.find_predicate env name arity)
      and args = Term.compound_args compound
      and mode = Array.make arity Opt.OutputGround in
      for i = 0 to arity - 1 do
	match Convert.term_to_string args.(i) with
	  "b" ->
	    mode.(i) <- Opt.OutputGround
	| "f" ->
	    mode.(i) <- Opt.OutputFree	    
	| _ ->
	    raise (Err.Error "Bad argument in 'mode' declaration template.")
      done;
      Predicate.add_declaration
	pred
	(Opt.ModeDecl mode)
  | _ ->
      ()

let process_disc_hints_decl env term =
  let (name, arity) = 
    Convert.predspec_pair (Convert.term_nth_arg term 0) in
  let pred = (Env.find_predicate env name arity) in
  Predicate.add_declaration
    pred
    (Opt.DiscriminationHintsDecl
       (Convert.term_to_array_of_float (Convert.term_nth_arg term 1)))

let decl_to_term env predicate decl =
  let st = Env.get_symbol_table env in
  match decl with
    Opt.IndexDecl (indextype, indexargs) ->
      begin
	let itype =
	  Term.Constant
	    (Term.find_symbol
	       st
	       ( match indextype with
		 Opt.HashDt -> "hash_dt"
	       | Opt.AlistDt -> "alist_dt"
	       | Opt.NoIndex -> "none")
	       0) in
	let iargs =
	  match indexargs with
	    Opt.WholeTerm -> Term.Constant (Term.find_symbol st "all" 0)
	  | Opt.Argument n -> Term.Integer n
	  | Opt.Arguments ns -> Convert.array_of_int_to_term st ns in
	Term.Compound
	  (Term.make_compound
	     (Term.find_symbol st "index" 2)
	     [|Convert.symbol_to_pred_spec_term
		 st
		 (Predicate.get_symbol predicate);
	       (Convert.alist_to_term
		  st
		  [("type", itype); ("args", iargs)])|])
      end
  | Opt.DiscriminationHintsDecl vals ->
      Term.Compound
	(Term.make_compound
	   (Term.find_symbol st "discrimination_hints" 2)
	   [|Convert.symbol_to_pred_spec_term
	       st
	       (Predicate.get_symbol predicate);
	     Convert.array_of_float_to_term st vals|])
  | Opt.ModeDecl mode ->
      let arity =  (Array.length mode) in
      let args = Array.make arity Term.Unbound in
      for i = 0  to arity - 1 do
	args.(i) <-
	  Term.Constant 
	    (Term.find_symbol
	       st 
	       ( match mode.(i) with
		 Opt.OutputGround -> "b"
	       | Opt.OutputFree -> "f"
	       | Opt.InputGround -> "i" )
	       0)
      done;
      Term.Compound
	(Term.make_compound
	   (Term.find_symbol st "mode" 1)
	   [|(Term.Compound
		(Term.make_compound
		   (Predicate.get_symbol predicate)
		   args))|])

let print_declarations env pred =
  List.iter
    (fun decl -> 
      PrologPrint.print
	(decl_to_term env pred decl);
      print_endline ".")
    (List.rev (Predicate.get_declarations pred))


(****************************************************************************)
(* Commands *)
(****************************************************************************)

let model_count = ref 0

let scan_format string len i =
  if !i >= len then
    raise (Err.Error "No format action specified.");
  begin
    match String.get string !i with
      'm' ->
	print_int !model_count
    | 'n' ->
	print_newline ()	
    | '~' ->
	print_char '~'
    | 'z' ->
	flush stdout
    | _ ->
	raise (Err.Error "Unknown format action.")
  end;
  i := !i + 1

let write_formatted string =
  let len = String.length string
  and i = ref 0 in
  while !i < len
  do
    match String.get string !i with
      '~' ->
	i := !i+1;
	scan_format string len i
    | c -> 
	i := !i+1;
	print_char c;
  done



let ensure_branch_reset env =
  Cfg.set_running_flag (Env.get_cfg env) false;
  List.iter
  	(fun redundant_store ->
  		Ic.undeclare_redundancies redundant_store)
  	!(Env.get_redundant_store_trail_ref env);
  let result = Env.get_current_result env in
  if Tableau.is_open result then
    (Msg.msg 1 "Abandoning current model.");
  Env.set_current_result env (Tableau.abandon_result result);
  Env.set_current_answers env [];
  model_count := 0

let abandon_cmd =
  (fun env term ->
    ensure_branch_reset env)
	     
    

    
    
let make_clauses env = 
		(IcSet.make_eval_plain_clauses
		     (Env.get_cfg env)
		     (Env.get_predicate_table env)
		     (Env.get_symbol_table env)
		     (Env.get_negative_units env)
		     (Env.get_rules env))

		     		

let szs_output env sat =
(*	let szs_problem_name = Cfg.get_szs_problem_name (Env.get_cfg env)
	in	*)
	let cfg = Env.get_cfg env
	in
	if (Cfg.get_szs_output_flag cfg) then
		(print_endline "";
		if (Cfg.get_szs_fof_flag cfg) then
(* fof input *)		
			(if (Cfg.get_szs_conjecture_flag cfg) then
(* theorem/contains conjecture *)
				(if sat then
					(print_endline "SZS status CounterSatisfiable")
				else
					(print_endline "SZS status Theorem"))
			else
				(if sat then
					(print_endline "SZS status Satisfiable")
				else
					(print_endline "SZS status Unsatisfiable")))
		else
(* clause input *)		
			(if sat then
				(print_endline "SZS status Satisfiable")
			else
				(print_endline "SZS status Unsatisfiable")))


		
let next_cmd =
  (fun env term ->
    let t0 = Sys.time () 
    and sat = ref true
    in
    Cfg.set_running_flag (Env.get_cfg env) true;
    let result =
      Tableau.next_result (Env.get_current_result env) in
    if Tableau.is_open result then
      (
	model_count := !model_count + 1;
	Msg.msg1 1 "*** RESULT: Fixedpoint, %8.3fs." (Sys.time () -. t0);
	   )
    else
      begin
    if !model_count=0 then
    	(sat := false);   
	model_count := 0;
	Msg.msg1 1 "*** RESULT: Closed, %8.3fs." (Sys.time () -. t0)
      end;
	Cfg.set_running_flag (Env.get_cfg env) false;      
    Env.set_current_result env result;
    Tableau.info_msg result;
    szs_output env !sat)



    
(****************************************************************************)
(* Timers *)
(****************************************************************************)

exception Timeout of bool

let set_timeout_handling env =
	let timeout_handler () =
		(match (Cfg.get_timeout_termination_method (Env.get_cfg env)) with
			| 0 -> 
				(raise (Timeout true))
			| 1 ->
				(ensure_branch_reset env;
				Tableau.State.install_state !(Env.get_current_state_ref env);
				raise (Timeout false))
			| 2 ->
				(raise (Timeout false))
			| _ ->
				(failwith "Process.set_timeout_handling: unsupported timeout termination method."))
	in
		Sys.set_signal 
			Sys.sigalrm 
			(Sys.Signal_handle
				(fun s ->
					Msg.msg 1 "Wallclock timeout reached.";
					timeout_handler ()));

		Sys.set_signal 
			Sys.sigvtalrm 
			(Sys.Signal_handle
				(fun s ->
					Msg.msg 1 "Virtual timeout reached.";
					timeout_handler ()))    
    
	 
				
let start_wallclock_timer seconds =
	if (seconds > 0.0) then
		ignore
			(Unix.setitimer Unix.ITIMER_REAL
				{Unix.it_interval = 0.0;
				Unix.it_value  = seconds})
		
let start_virtual_timer seconds =
	if (seconds > 0.0) then
		ignore
			(Unix.setitimer Unix.ITIMER_VIRTUAL
				{Unix.it_interval = 0.0;
				Unix.it_value  = seconds})	
		      
let stop_wallclock_timer () =
	ignore
		(Unix.setitimer Unix.ITIMER_REAL
			{Unix.it_interval = 0.0;
			Unix.it_value  = 0.0})    
	      
let stop_virtual_timer () =
	ignore
		(Unix.setitimer Unix.ITIMER_VIRTUAL
			{Unix.it_interval = 0.0;
			Unix.it_value  = 0.0})  	      

(****************************************************************************)
(* Process Statement *)
(****************************************************************************)

exception EXIT
exception Found
exception Interrupted
exception Switched_input_type of int

(* Hmm, this isn't used at all. *)
let set_signals () =
  Sys.set_signal
    Sys.sigint 
    (Sys.Signal_handle
       (fun s ->
	 Msg.msg 1 "Interrupted";
	 raise Interrupted))
	 
	      
	      	
let unit_is_indexed predicate symbol_table new_term = (*	PrologPrint.print new_term;print_string "   ";	*)
	let equals = Term.get_equality_symbol symbol_table
	and sym = Convert.term_functor_symbol new_term
	in
		if equals == sym then (
			let flip_term = Term.binary_flip new_term
			in
				try
		    		Access.Reader.iter_unification_candidates 
		      			(fun fact ->
		      				let term = Fact.get_term fact
		      				in 
								if (Term.variant_chk term new_term) then
				  					raise Found
								else
				  					())
		      			(Predicate.default_reader predicate)
		      			new_term
		      			Term.null_context;
		      		Access.Reader.iter_unification_candidates 
		      			(fun fact ->
		      				let term = Fact.get_term fact
		      				in 
								if (Term.variant_chk term flip_term) then
				  					raise Found
								else
				  					())
		      			(Predicate.default_reader predicate)
		      			flip_term
		      			Term.null_context;
		    		false
		  		with Found ->
		    		true
		  )
		  else(			(*	PrologPrint.print new_term;print_string "   ";		*)
		  		try
		    		Access.Reader.iter_unification_candidates 
		      			(fun fact -> 
							if (Term.variant_chk (Fact.get_term fact) new_term) then
			  					raise Found
							else
			  					())
			      		(Predicate.default_reader predicate)
			      		new_term
			      		Term.null_context;
		    		false
		  		with Found ->
		    		true)	 
 
		    		
let get_current_clauses env =		    		
   let clauses = ref []
   in
   Ic.SubtermTable.iter
	  (fun symbol subterm -> 
	    try 
	      Ic.Reader.iter_all
			(fun entry -> 
				match entry with
					Ic.SubtermEntry.Clause(clause, _, _, _) ->
			  			if (Ic.is_redundant clause)||(List.memq clause !clauses) then 
		  					()
			  			else 
			  				(clauses := clause :: !clauses)
			  		|_ -> ())
			(Ic.get_subterm_reader subterm)
	    with Not_found ->
              (* May be thrown by Predicate.default_reader predicate *)
	      ())
  (Env.get_subterm_table env);
  !clauses
		    		
		    		
let process_giving_up env =
	let cfg = Env.get_cfg env
	in
		if (Cfg.get_szs_output_flag cfg) && (Cfg.get_running_flag cfg) then
			(print_endline "";
			print_endline "SZS status GaveUp")
		else
			();
(*		Tableau.State.install_state !(Env.get_current_state_ref env);*)
		Cfg.set_running_flag cfg false
			 
	 
let rec process_statement ?(dont_range_restrict:bool = false) env term =	
  Term.renumber_variables term; 
  let cfg = Env.get_cfg env
  in
  if Cfg.get_echo_flag cfg then
    begin
      PrologPrint.print term;
      print_string ".";
      print_newline ();
    end
  else
    ();
  let sym = Convert.term_functor_symbol term in
  match Env.classify_statement_functor env sym with
    Env.Meta ->
      (try
	process_meta_statement env (Convert.term_nth_arg term 0)
      with 
      	| Timeout terminate ->
      		(process_giving_up env;
      		if terminate then 
      			(exit 0)
      		else
      			(true))
      	| e ->
			ensure_branch_reset env;
			raise e)
  | _ ->
	(let term = 
		(if ((Cfg.get_finite_model_transformation_flag cfg)
			&& (not dont_range_restrict)) then
			(Termutils.range_restrict (Env.get_symbol_table env) term)
		else
			(term))
	in
	let sym = Convert.term_functor_symbol term in
	match Env.classify_statement_functor env sym with
  | Env.Meta ->
	failwith "Process.process_statement: bad range restriction result."
  | Env.Unit -> 
      ensure_branch_reset env;
(* Equation units are treated as as rules. *)
      let equals = (Term.get_equality_symbol (Env.get_symbol_table env)) 
      in
	      	if sym == equals then (
				Env.add_rule env term
			)
			else (
				if (Cfg.get_hyperextension_flag cfg) then
					(try 
						(let new_term = (Env.unit_for_env env term)
						in	
						let fact = (Fact.make_input_fact new_term)
						and pred = (Env.ensure_ready_to_assert env sym)
						in						
							Ic.np_subsumption_processing_unit
								(Cfg.get_full_branch_redundancy_flag cfg)
								(Env.get_redundant_store env)
								pred
								fact;			
							Predicate.assert_unit_clause 
								pred 
								new_term
								fact;			
							Ic.index_unit_subterms_no_top
								(Env.get_subterm_table env)
								(Env.get_symbol_table env)
								new_term
								fact;			
							Term.register_function_symbols
								(Env.get_symbol_table env)
								(Fact.get_term fact);
							Env.add_fact env fact;
							if not (Termutils.is_ground new_term) then
								(Cfg.set_non_range_restricted_input cfg true)								
						) 
					with Ic.Redundant -> ())
				else
					(let eq_term =
						Term.Compound
							(Term.make_compound
								equals
								[|term; Term.Unbound|])
					in
						Env.add_rule env eq_term));
	(*  Env.update_current_state env;*)
      true
  | Env.Rule ->
      ensure_branch_reset env;
      Env.add_rule env term;
	 (* Env.update_current_state env;      *)
      true
  | Env.Disjunction ->
      ensure_branch_reset env;
      Env.add_disjunction env term;
	  (* Env.update_current_state env;      *)
      true
  | Env.False ->
      ensure_branch_reset env;
      Env.add_false env;
	  (* Env.update_current_state env;      *)
      true)

      
and process_meta_statement env term =
  match Convert.term_functor_pair term with
    ("halt", 0) ->
      false
  | _ ->
      try
	Sys.set_signal
	  Sys.sigint 
	  (Sys.Signal_handle
	     (fun s -> raise Interrupted));
	process_meta_statement_1 env term;
	Sys.set_signal Sys.sigint Sys.Signal_default;
	true
    with 
    	| Interrupted ->
			(Msg.msg 1 "Interrupted.";
			true)

			 
and ensure_finite_domain env =
(* part two of the finite model transformation, performs the "Restriction to Finite Domain"-step if required *)
	if (Cfg.get_finite_model_transformation_flag (Env.get_cfg env)) then
		(let domain_restricting_clause_sources =
			Term.make_domain_restricting_clause_sources 
				(Cfg.get_finite_model_transformation_domain_size (Env.get_cfg env)) 
				(Env.get_symbol_table env)
		in
			List.iter
				(fun term -> (*pt term;pe "!";*)
					ignore(process_statement ~dont_range_restrict:false env term))
				domain_restricting_clause_sources)
	else
		()

and run_cmd =
  (fun env term ->	
    let cfg = Env.get_cfg env
    in
    (match (Cfg.get_timeout_termination_method cfg) with
    	| 2 ->
    		(Tableau.State.install_state !(Env.get_current_state_ref env))
    	| _ -> ());
(*    if (Cfg.get_running_flag cfg) then*)	(* if this is left in, then '#(run). #(run).' will
crash on a satisfiable problem, since the negative units will accumulate one layer per run,
yet must be zero-layered at the start of a run.
Could removing it cause problems elsewhere? *)
    	(ensure_branch_reset env);
    Env.update_current_state env;	
    ensure_branch_reset env; 	(* see previous comment: maybe drop this one entirely instead? *)
    ensure_finite_domain env;	
    Env.ensure_required_clauses env; 	
    Env.update_current_state env; (* Again - otherwise krh-constants won't get added layers. *)
    (* Gc.compact ();	*)
    (*Env.print_stats env;*)
    let t0 = Sys.time () 
    and sat = ref true
    and clauses = (Env.get_clauses env)
    and predicate_table = (Env.get_predicate_table env)
    in
    if (Cfg.get_reassign_predicates cfg) then
	    (List.iter
	    	(fun clause -> (*pt (Ic.get_source clause);pe "!!!";*)
	    		Ic.assign_predicates predicate_table clause)
	    	clauses;
	    Cfg.set_reassign_predicates cfg false);
	Cfg.set_running_flag cfg true;	
    let result =
      Tableau.first_result
        cfg
        predicate_table
        (Env.get_symbol_table env)
		(Env.get_negative_units env)
		(Env.get_subterm_table env)
		(Env.get_clause_aux env)
        clauses
        (Env.get_current_state_ref env)
        (Env.get_redundant_store_trail_ref env)
    in	
    if Tableau.is_open result then
      begin
	model_count := !model_count + 1;
	Msg.msg1 1 "*** RESULT: Fixedpoint, %8.3fs." (Sys.time () -. t0)
      end
    else
      begin
	model_count := 0;
	sat := false;
	Msg.msg1 1 "*** RESULT: Closed, %8.3fs." (Sys.time () -. t0)
      end;
    Cfg.set_running_flag cfg false;  
    Env.set_current_result env result;
    Env.set_current_answers 
    	env 
    	(Tableau.get_answers result);
    Tableau.info_msg result;
    szs_output env !sat)
	
and run_relevant_cmd =
  (fun env term ->
    let (name, arity) =
      Convert.predspec_pair (Convert.term_nth_arg term 0) 
    in
    Env.update_current_state env;    
    ensure_branch_reset env;
    ensure_finite_domain env;
    Env.ensure_required_clauses env; 
    Env.update_current_state env; 
    let t0 = Sys.time ()
    and sat = ref true
    in
    let pred = Env.find_predicate env name arity
    and clauses = (Env.get_clauses env)
    and predicate_table = (Env.get_predicate_table env)
    and cfg = (Env.get_cfg env)
    in
    if (Cfg.get_reassign_predicates cfg) then
	    (List.iter
	    	(fun clause ->
	    		Ic.assign_predicates predicate_table clause)
	    	clauses;
	    Cfg.set_reassign_predicates cfg false);
	Cfg.set_running_flag cfg true;	    
    let result =
      Tableau.first_result
        cfg
        predicate_table
        (Env.get_symbol_table env)
		(Env.get_negative_units env)
		(Env.get_subterm_table env)
		(Env.get_clause_aux env)
		(IcSet.select_relevant_clauses clauses pred)
		(Env.get_current_state_ref env) 
		(Env.get_redundant_store_trail_ref env)
	in
    if Tableau.is_open result then
      begin
	model_count := !model_count + 1;
	Msg.msg1 1 "*** RESULT: Fixedpoint, %8.3fs." (Sys.time () -. t0)
      end
    else
      begin
	model_count := 0;
	sat := false;
	Msg.msg1 1 "*** RESULT: Closed, %8.3fs." (Sys.time () -. t0)
      end;
	Cfg.set_running_flag cfg false;      
    Env.set_current_result env result;
    Env.set_current_answers 
    	env 
    	(Tableau.get_answers result);
    Tableau.info_msg result;
    szs_output env !sat)
			

and cmd_table =
  [
   (("help", 0),
    ("'help' prints documentation.",
     (fun env term -> help env)));
   (("run", 0),
    ("'run' computes the first model or a refutation. If a model is found,\n\
      it becomes the current model.",
     run_cmd));
   (("run", 1),
    ("'run(p/n)' is like run/0, but only the subset of input rules\n\
     that is relevant for the given predicate is used.", 
     run_relevant_cmd));
   (("next", 0),
    ("'next' computes the next model. If a model is found, it becomes the\n\
      current model.",
     next_cmd));
   (("abandon", 0),
    ("'abandon' abandons the current model.",
     abandon_cmd));
     
   (("compute_answer_for_tptp_fof_variable", 1),
    ("'compute_answer_for_tptp_fof_variable('<id>')' makes E-KRHyper\n\
    	store the bindings for the variable <id> during the derivation\n\
    	process.\n\
    	For this to have any effect, the variable must occur in a tptp\n\
    	first order conjecture.\n\
    	Once a tableau refutation has been found, the binding(s) can be\n\
    	listed using the command 'print_answer_units'. A number of\n\
    	'$answer'/1-units will be listed, each having as its argument one\n\
    	term that was bound to <id>.\n\
    	Only one variable can be treated this way, as this command causes\n\
    	the transformation of input clauses.\n\
    	The command must be given before the TPTP-conjecture is read.\n\
    	The negation of TPTP-conjectures must be enabled (this is the case by default).\n\
    	The command is a combination of the commands 'set_tptp_fof_answer_variable'\n\
    	and 'set_flag(tptp_fof_compute_answer_predicate, true)'.",
    (fun env term ->
    	Cfg.set_tptp_fof_answer_variable
    		(Term.string_of_term 
    			(Convert.term_nth_arg term 0));
    	Cfg.set_tptp_fof_compute_answer_predicate
    		(Env.get_cfg env)
    		true)));
   
   (("save_input_state", 0),
    ("'save_input_state' saves the current state of the set input clauses.\n\
		If further input clauses are added after this, the set of input clauses\n\
		can be rolled back to the saved state using 'load_previous_input_state'.\n\
		Multiple states can be saved consecutively. Multiple 'load_previous_input_state'-\n\
		commands can then be used to iterate through the saved states in reverse order.",
     (fun env term ->
     	Env.save_state env)));     		
    		  
   
   (("load_previous_input_state", 0),
    ("'load_previous_input_state' returns the set of input clauses to the state\n\
		saved by the most recent 'save_input_state'-command.\n\
		Multiple 'previous_input_state'-commands can then be used to iterate\n\
		through the saved states in reverse order.\n\
		The command has no effect if no state has been saved yet.",
     (fun env term ->
     	ensure_branch_reset env;
     	Env.install_previous_state env)));     	
     	
   (("szs_problem_name", 1),
    ("'szs_problem_name/1' enters the problem name required for SZS-conforming output.",
     (fun env term ->
     	Cfg.set_szs_problem_name
     		(Env.get_cfg env)
     		(Convert.term_to_string
     			(Convert.term_nth_arg term 0)))));     
     
	(("print_current_subterm_table_keys", 0),
		("Debugging function, prints the keys of the current subterm_table.",
		(fun env term ->
		     Ic.SubtermTable.iter
		  		(fun symbol subterm -> ps (Convert.symbol_to_string symbol);pe ".")
		  		(Env.get_subterm_table env))));
		  		
		  		
		  		
		  		
		  		
		  		
		  		
   (("print_current_compiled_rules", 0),
    ("'print_current_compiled_rules' prints the set of compiled rules in the current\n\
       model.",
     (fun env term ->
       (let clauses = ref []
       and reds = ref 0
       and redscl = ref []
       in
       Ic.SubtermTable.iter
		  (fun symbol subterm -> 
		    try 
		      Ic.Reader.iter_all
				(fun entry -> 
					match entry with
						Ic.SubtermEntry.Clause(clause, _, _, _) ->
				  			if (Ic.is_redundant clause)||(List.memq clause !clauses) then 
				  				(if Ic.is_redundant clause then 
				  					(if not (List.memq clause !redscl) then 
				  						(redscl := clause :: !redscl; 
				  						reds := !reds + 1)))
				  			else (clauses := clause :: !clauses)
				  		|_ -> ())
				(Ic.get_subterm_reader subterm)
		    with Not_found ->
	              (* May be thrown by Predicate.default_reader predicate *)
		      ())
	  (Env.get_subterm_table env);
	  Info.pretty_print_clauses_as_terms !clauses;
	 ))));
	 
   (("print_current_clauses_contain_impure_vars", 0),
    ("'print_current_clauses_contain_impure_vars' prints 'Impure variables: yes.' if at least one current\n\
       clause contains at least one variable shared between head literals and not found\n\
       in the body, and 'Impure variables: no.' otherwise.",
     (fun env term ->
       (let clauses = get_current_clauses env
       in
	  let result =
	  	List.exists
	  		Ic.has_non_range_restricted_shared_head_vars
	  		clauses
	  in
	  	if result then
	  		(Msg.msg 1 "Impure variables: yes.")
	  	else
	  		(Msg.msg 1 "Impure variables: no.");
	 ))));	 
	 
   
(("print_current_equation_units", 0),
    ("'print_current_equation_units' prints the set of equation units in the current\n\
       model.",
     (fun env term ->
       (let clauses = ref []
       in
       Ic.SubtermTable.iter
		  (fun symbol subterm ->
		    try 
		      Ic.Reader.iter_all
				(fun entry -> 
					match entry with
						Ic.SubtermEntry.Clause(clause, _, _, _) ->
				  			if (not (Ic.is_eq_unit clause))||(Ic.is_redundant clause)||(List.memq clause !clauses) then ()
				  			else (clauses := clause :: !clauses)
				  		|_ -> ())
				(Ic.get_subterm_reader subterm)
		    with Not_found ->
	              (* May be thrown by Predicate.default_reader predicate *)
		      ())
	  (Env.get_subterm_table env);
	  Info.print_clauses_as_terms !clauses))));
 	  
(("print_answer_units", 0),
    ("'print_answer_units' prints the set of answer predicate units that\n\
       have been computed so far.",
     (fun env term ->
     	let answers = Env.get_current_answers env
     	in
     		List.iter
     			(fun fact ->
				if (fact==Fact.NullFact) then ()
     				else
					(PrologPrint.print
						(Fact.get_term fact);
					print_endline "."))
     			answers)));	  

(("print_partial_answer_units", 0),
	("'print_partial_answer_units' prints the set of partial answer predicate\n\
	units that have been computed so far.\n\
	A partial answer unit takes the form\n\
		'$partial_answer'(<successful_literals>, <bindings>, <unsuccessful_literals>).\n\
	where <literals> is a Prolog-style list of those literals from a conjecture\n\
	based clause that unified with the current extent (displayed using the original\n\
	variable names), and where <bindings> is a Prolog-style list of special terms\n\
	that express the substitution used for the unification.\n\
	Each such binding term takes the form '$bound_to'(<variable>, <term>),\n\
	where <variable> is a variable occuring in <literals> and <term> is the term\n\
	substituting the <variable>.\n\
	<unsuccessful_literals> is a list of those literals that were not unified,\n\
	in the order used by E-KRHyper. The first literal in this list is thus the\n\
	literal that failed to unify, while the remaining elements were never tested.",
     (fun env term ->
     	let partial_answer_symbol =
     		Term.get_krh_partial_answer_symbol
     			(Env.get_symbol_table env)
     	in
     	let partial_answer_predicate =
     		Predicate.find_predicate
     			(Env.get_predicate_table env)
     			partial_answer_symbol
		in
			try
				(Access.Reader.iter_all
					(fun entry ->
						let term_string =
							Term.string_of_term
								~pretty_lists:true
								~prolog_compatible:true
								~use_original_var_ids:true								
								(Fact.get_term entry)
						in
							print_endline (term_string^"."))
				(Predicate.default_reader partial_answer_predicate))
			with Not_found ->
				())));
     			  
   (("print_extent", 0),
    ("'print_extent' prints the extents of all predicates in the current\n\
       model.",
     (fun env term ->
       (Predicate.PredicateTable.iter
	  (fun symbol predicate ->
	    try 
	      Access.Reader.iter_all
		(fun entry -> 
		  	if (Fact.is_redundant entry) then ((*ps "RED: ";PrologPrint.print (Fact.get_term entry);print_endline "."*))
		  	else (
				PrologPrint.print (Fact.get_term entry);
		  		print_endline "."))
		(Predicate.default_reader predicate)
	    with Not_found ->
              (* May be thrown by Predicate.default_reader predicate *)
	      ())
	  (Env.get_predicate_table env))
     )));
     
(("print_lemmas", 0),
	("EXPERIMENTAL!\n\
	'print_lemmas' prints the subset the extent that has been\n\
	computed before the first split and which thus holds in\n\
	all tableau branches (although some units may be subsumed\n\
	in some branches).",
     (fun env term ->
     	let lemmas =
     		Env.get_lemmas env
     	in
     		List.iter
     			(fun fact ->
     				let term_string =
							Term.string_of_term
								~pretty_lists:true
								~prolog_compatible:true
								~use_original_var_ids:true								
								(Fact.get_term fact)
					in
						print_endline (term_string^"."))
     			lemmas)));	     
     
(("print_tptp_fof_lemma", 0),
	("EXPERIMENTAL!\n\
	'print_tptp_fof_lemma' is essentially the same as 'print_lemmas',\n
	but the output is printed as a single first-order axiom in the\n
	TPTP format.",
     (fun env term ->
     	let lemmas =
     		Env.get_lemmas env
     	in	
     	let lemma_strings =
     		List.map
     			(fun fact ->
					Term.string_of_term
						~pretty_lists:true
						~prolog_compatible:true
						~use_original_var_ids:true								
						(Fact.get_term fact))
     			lemmas
     	in
     	let conjunction_string =
     		String.concat
     			" & "
     			lemma_strings
     	in
     		print_endline ("fof('ekrh_incomplete_derivation_lemma', lemma, ("
     			^ conjunction_string
     			^ "))."))));	     
     			
     			
(("print_model", 0),
    ("'print_model' prints the current model, which means\n\
     all positive unit clauses in the current branch.",
     (fun env term ->
     (*	let krh_dom_symbol = Term.get_krh_dom_symbol (Env.get_symbol_table env)
	in *)
     	let clauses = ref []
     	in
       (Predicate.PredicateTable.iter
	  (fun symbol predicate ->
	  	if false (* symbol == krh_dom_symbol *) then ()
	  	else
		    (try 
		      Access.Reader.iter_all
			(fun entry -> 
			  	if (Fact.is_redundant entry) then ((*ps "RED: ";PrologPrint.print (Fact.get_term entry);
			  		print_endline "."*))
			  	else (
					PrologPrint.print (Fact.get_term entry);
			  		print_endline "."))
			(Predicate.default_reader predicate)
		    with Not_found ->
	              (* May be thrown by Predicate.default_reader predicate *)
		      ()))
	  (Env.get_predicate_table env));
	  (Ic.SubtermTable.iter
		  (fun symbol subterm ->
		    try 
		      Ic.Reader.iter_all
				(fun entry -> 
					match entry with
						Ic.SubtermEntry.Clause(clause, _, _, _) ->
				  			if (not (Ic.is_eq_unit clause))||(Ic.is_redundant clause)||(List.memq clause !clauses) then ()
				  			else (clauses := clause :: !clauses)
				  		|_ -> ())
				(Ic.get_subterm_reader subterm)
		    with Not_found ->
		      ())
	  (Env.get_subterm_table env);
	  Info.print_clauses_as_terms !clauses)
     )));     
     
     
   (("print_extent", 1),
    ("'print_extent(p/n)' prints the extent of the predicate p/n in the\n\
      current model.",
     (fun env term ->
      begin
	let (name, arity) = 
	  Convert.predspec_pair
	    (Convert.term_nth_arg term 0) in
	try 
	  Access.Reader.iter_all
	    (fun entry -> 
	      PrologPrint.print (Fact.get_term entry);
	      print_endline ".")
	    (Predicate.default_reader (Env.lookup_predicate env name arity))
	with Not_found ->
	  print_endline "% Predicate not found."
      end)));
   (("print_extent_with_proofs", 1),
    ("'print_extent_with_proofs(p/n)' prints a Fact-Proof pair for each\n\
       fact in the extent of the predicate p/n in the current model.\n\
      [Works not yet with disjunctions].",
	   (fun env term ->
	     begin
	       let (name, arity) = 
		 Convert.predspec_pair
		   (Convert.term_nth_arg term 0) in
	       try 
		 Access.Reader.iter_all
		   (fun entry -> 
		     PrologPrint.print (Fact.get_term entry);
		     print_endline " - ";
		     PrologPrint.print
		       (Fact.proof_term_for_fact
			  (Env.get_symbol_table env)
			  entry);
		     print_endline ".")
		   (Predicate.default_reader
		      (Env.lookup_predicate env name arity))
	       with Not_found ->
		 print_endline "% Predicate not found."
	     end)));
   (("print_proof", 0),
    ("'print_proof' prints the proof if the current result is closed.\n\
       Output format is the same as of 'print_extent_with_proofs'.\n\
      [Output for non-Horn problems and open results is still experimental].",
     (fun env term ->
       let r = Tableau.get_refutation (Env.get_current_result env) in
       PrologPrint.print 
	 (Tableau.refutation_to_term
	    (Env.get_symbol_table env)
	    r);
       print_endline("."))));
   (("print_szs_proof", 0),
    ("'print_szs_proof' prints the proof if the current result is closed.\n\
       Output format is similar to that of 'print_proof', except that\n\
       extra control clauses are added to make the output suitable for the\n\
       'post_szs.pl' postprocessor.\n\
      [Output for non-Horn problems and open results is still experimental].",
     (fun env term ->
     if !model_count=0 then (
       let r = Tableau.get_refutation (Env.get_current_result env) in
       
       	match r with
       		Tableau.NullRefutation ->
       			(print_endline "szs_unfinished.")
       		|_ ->
       (print_endline "szs_refutation_start.";
       PrologPrint.print 
	 (Tableau.refutation_to_term
	    (Env.get_symbol_table env)
	    r);
       print_endline(".");
       print_endline "szs_refutation_stop."))
     else
     	(print_endline "szs_satisfiable.");)));       
   (("print_proof_1", 0),
    ("'print_proof_1' prints the proof if the current result is closed.\n\
       Output format is the same as of 'print_extent_with_proofs'.\n\
      [Output for non-Horn problems and open results are still experimental].\
      \n [Deprecated - old version of print_proof]",
      (fun env term ->
       let proof = Tableau.get_l_proof (Env.get_current_result env) in
       PrologPrint.print 
	 (Fact.proof_term_for_l_proof
	    (Env.get_symbol_table env)
	    proof);
       print_endline("."))));
   (("print_negative_units", 0),
    ("'print_negative_units' prints the negative clauses of the current\n\
       model which are maintained for heuristic purposes, such as complement\n\
      splitting.",
      (fun env term ->
       let rule_sym =
	 (Term.get_rule_symbol (Env.get_symbol_table env))
       and false_term =
	 Term.Constant 
	   (Term.get_false_symbol (Env.get_symbol_table env))
       in
       NegativeUnits.iter_all
	 (fun fact ->
	   PrologPrint.print 
	     (Term.Compound 
		(Term.make_compound rule_sym [|false_term; 
					       (NegativeUnits.FactEntry.get_term fact)|]));
	   print_endline ".")
	 (Env.get_negative_units env))));
   (("print_rules", 0),
    ("'print_rules' prints all rules.",
     (fun env term ->
       (List.iter
	  (fun rule ->
	    PrologPrint.print rule;
	    print_endline ".")
	  (Env.get_rules env)))));
   (("print_declarations", 1),
    ("'print_declarations(p/n)' prints the currently effective declarations\n\
      for the predicate p/n.",
     (fun env term ->
      begin
	let (name, arity) = 
	  Convert.predspec_pair
	    (Convert.term_nth_arg term 0) in
	try 
	  let pred = Env.lookup_predicate env name arity in
	  print_declarations env pred
	with Not_found ->
	  print_endline "% Predicate not found."
      end)));
   (("print_declarations", 0),
    ("'print_declarations' prints the currently effective declarations.",
     (fun env term ->
       (Predicate.PredicateTable.iter
	  (fun symbol predicate -> print_declarations env predicate)
	  (Env.get_predicate_table env)))));
   (("print_settings", 0),
    ("'print_settings' prints the currently effective flag and parameter\n\
       settings.",
     (fun env term ->
      Cfg.print_settings (Env.get_cfg env))));
   (("print_compiled_rules", 0),
    ("'print_compiled_rules' prints all rules after preprocessing,\n\
       especially body literals are sorted heuristically and obeying\n\
       builtins. This command may be useful to generate input for other\n\
       preprocessing operations such as a magic sets transformation.\n\
       [Currently does not properly print literals with findall/3 and\n\
        negation operators other than \\+/1.]",
     (fun env term ->
       ensure_branch_reset env;   
(*       let state =
       	Tableau.State.obtain_state
       		(Env.get_predicate_table env)
	 	 	(Env.get_negative_units env)
	 	 	(Env.get_subterm_table env)
	 	 	(Env.get_facts env)
	 	 	(Env.get_clauses env)
	 	 	(Env.get_rules env)
	 	 	(Env.get_redundant_store env)
	 	in *)
       Tableau.print_compiled_rules
         (Env.get_cfg env)
         (Env.get_predicate_table env)
         (Env.get_symbol_table env)
	 	 (Env.get_negative_units env)
	 	 (Env.get_subterm_table env)
	 	 (Env.get_clause_aux env)
         (Env.get_clauses env)
         (Env.get_current_state_ref env)
         (Env.get_redundant_store_trail_ref env))));
   (("set_tptp_fof_answer_variable", 1),
    ("'set_tptp_fof_answer_variable(s)' declares the variable with the\n\
    name 's' to be the answer variable used for the computation of the\n\
    '$answer'-predicate units. See 'tptp_fof_compute_answer_predicate'.\n\
    The string argument for the variable name must be given in single quotes.\n\
    The variable name must be set before the tptp formula input is read.\n\
    The default name of the answer variable is 'FOCUS'.",
    (fun env term ->
    	Cfg.set_tptp_fof_answer_variable
    		(Term.string_of_term 
    			(Convert.term_nth_arg term 0)))));
   (("show_symbols", 0),
    ("'show_symbols' prints information about symbols in the symbol table.",
     (fun env term ->
      Info.print_symbols_info (Env.get_symbol_table env))));
   (("show_predicates", 0),
    ("'show_predicates' prints information about predicates with respect\n\
      to the current model.",
     (fun env term ->
      Info.print_predicates_info (Env.get_predicate_table env))));
   (("show_extent_size", 1),
    ("'show_extent_size(p/n)' prints information about the extent size of\n\
       predicate p/n in the current model.",
     (fun env term ->
      let (name, arity) = 
	Convert.predspec_pair
	  (Convert.term_nth_arg term 0) in
      print_string "% ";
      print_int
	(Predicate.extent_size (Env.lookup_predicate env name arity));
      print_endline "")));
   (("show_stratification", 0),
    ("show_stratification prints information about the stratification\n\
      of input rules performed by the system.",
    (fun env term ->
      let strata = IcSet.stratify (Env.get_clauses env)
      and n = ref 0 in
      print_string "% ========== Number of strata: ";
      print_int (List.length strata);
      print_endline " ==========";
      List.iter
	(fun stratum ->
	  print_string "% ========== Stratum: ";
	  print_int !n;
	  let stratum_size = (List.length stratum) in
	  print_string " (";
	  print_int stratum_size;
	  if (stratum_size = 1) then print_string " clause)" else print_string " clauses)";
	  print_endline " ==========";
	  List.iter 
	    (fun clause ->
	      PrologPrint.print (Ic.get_source clause);
	      print_endline ".")
	    stratum;
	  n := !n + 1)
      	strata)));
   (("set_flag", 2),
    ("'set_flag(Flag, Boolean)' assigns the given truth value to the flag.\n\
       The boolean can be specified as 'true'/'false' or 'on'/'off'.",
     (fun env term ->
       Cfg.set_flag 
	 (Env.get_cfg env)
	 (Convert.term_to_string (Convert.term_nth_arg term 0))
	 (match (Convert.term_to_string (Convert.term_nth_arg term 1)) with 
	   "true" | "t" | "on" -> true
	 | "false" | "f" | "off" -> false
	 | _ -> raise
	       (Err.Error "Value of set_flag does not specify a boolean.")))));
   (("set_parameter", 2),
    ("'set_parameter(Parameter, Integer)' assigns the given integer value to\n\
       the parameter.",
       (fun env term ->
       let parameter =
       		(Convert.term_to_string (Convert.term_nth_arg term 0))
       	in
	 Cfg.set_parameter
	   (Env.get_cfg env)
	   parameter
	   (Convert.term_to_int (Convert.term_nth_arg term 1));
	 	match parameter with
	 		| "timeout_termination_method" ->
	 			(set_timeout_handling env)
	 		| _ -> ())));
   (("index", 2),
    ("'index(p/n, [type=Type, args=Args])' declares an index for the\n\
       predicate p/n. Type is one of 'none', 'alist_dt' or 'hash_dt'.\n\
       Args is either 'all' (index on the whole term), an integer n (index\n\
       on the nth argument) or a list of integers (index on a subterm\n\
       composed of the respective arguments in the given ordering).",
	   (fun env term ->
	     ensure_branch_reset env;
	     process_index_decl env term)));
   (("discrimination_hints", 2),
    ("'discrimination_hints(p/n, Hints)' declares Hints as discrimination\n\
      hints for the predicate p/n. Hints is a list of floats between\n\
      0.0 and 1.0 and has length n.",
     (fun env term ->
       ensure_branch_reset env;
       process_disc_hints_decl env term)));
   (("mode", 1),
    ("'mode(Template)' declares the ``mode'' of a predicate. It is used\n\
       for body literal ordering. 'Template' is a term with\n\
       functor and arity of the predicate. The arguments are 'f', which\n\
       means that the respective argument can have nonground values in\n\
       facts with that predicate, or 'b', which means that it has always\n\
       ground values. By default a mode that assigns 'b' to all argument\n\
       positions is assumed. Note that, unless builtins are used, mode\n\
       declarations play only a heuristic role. The effect of different\n\
       mode declarations for a predicate is undefined.",
    (fun env term ->
      ensure_branch_reset env;
      process_mode_decl env term)));
   (("begin_shared_input", 0),
    ("'begin_shared_input' enables sharing of subsequent input terms.\n\
       Input terms are only shared within dynamic begin_shared_input/\n\
       end_shared_input contexts. The behavior for nested\n\
       begin_shared_input/end_shared_input contexts is undefined.",
	   (fun env term ->
	     Env.begin_shared_input env)));
   (("end_shared_input", 0),
    ("'end_shared_input' disables sharing of subsequent input terms\n\
       and frees resources used for sharing.",
	   (fun env term ->
	     Env.end_shared_input env)));
   (("clear_builtin", 1),
    ("'clear_builtin(s/n)' removes any special evaluation of the builtin/evaluable\n\
    function or predicate symbol 's' with arity n.",
    (fun env term ->
			ensure_branch_reset env;
			let (name, arity) = 
				Convert.predspec_pair
					(Convert.term_nth_arg term 0)
			in
				Term.clear_builtin
					(Env.get_symbol_table env)
					name
					arity)));
					
   (("abolish", 1),
    ("'abolish(p/n)' abolishes predicate p/n. This means its extent and\n\
      its declarations are removed.",
     (fun env term ->
			ensure_branch_reset env;
			
			let (name, arity) = 
				Convert.predspec_pair
					(Convert.term_nth_arg term 0)
			in	
			let symbol =
				Term.find_symbol
					(Env.get_symbol_table env)
					name
					arity
			in		
			let predicate =
				Predicate.find_predicate
					(Env.get_predicate_table env)
					symbol
			and entries_to_delete = ref []
			in					
				(try		
					(Access.Reader.iter_all
						(fun entry ->
							entries_to_delete := entry :: !entries_to_delete)
						(Predicate.default_reader predicate))
				with Not_found -> ());	

				Env.abolish_predicate env name arity;	
(* Abolishing a predicate removes it from the index and from the predicate table. The
compiled clauses that contain the abolished predicate end up with a dangling predicate
that is inaccessible from the index. These clauses must have their predicates
reassigned to fresh predicates in the predicate table. A flag is set to mark this need. When
the tableau is about to be computed with one of the 'run' commands, the flag will be checked
and the predicates will be reassigned if necessary. *)
				Cfg.set_reassign_predicates (Env.get_cfg env) true;
				List.iter
					(fun entry ->
						(Ic.remove_unit_variants
							(Env.get_subterm_table env)
							(Env.get_symbol_table env)
							entry;	
						Env.remove_fact env entry))
					!entries_to_delete)));
					
	(("assign_rule_to_predicate", 2),
		("Used for knowledge-base partitioning. 'assign_rule_to_predicate(Rule, P/A)'\n\
		declares that Rule may be required for the computation of predicate A with arity\n\
		A. Rule must be an input clause that has already been given to E-KRHyper; this\n\
		command cannot be used to enter input clauses, only to assign a special status to\n\
		already entered clauses. Rule should be typed exactly the same way as it was\n\
		entered into E-KRHyper, minus the terminating full-stop.\n\
		This command is an experimental preprocessing-step. It may be computationally\n\
		intensive when many clauses have been entered into the system.",
		(fun env term ->
			let rule, predspec =
				Convert.term_pair term
			in
			let pred_name, pred_arity =
				Convert.predspec_pair predspec
			in
			let pred_sym =
				Term.find_symbol
					(Env.get_symbol_table env)
					pred_name
					pred_arity
			in
			let pred =
				Predicate.find_predicate
					(Env.get_predicate_table env)
					pred_sym
			in
			let search_clause =
				Ic.make_clause
					(Env.get_clause_aux env)
					rule
			in
			let search_source =
				Ic.get_source search_clause
			in
				List.iter
					(fun clause ->
						if (Term.term_ident search_source (Ic.get_source clause)) then
							(Predicate.set_assigned_clause_flags
								pred
								((Ic.get_assignment_flag clause)::(Predicate.get_assigned_clause_flags pred))))
					(Env.get_clauses env))));					
					
   (("clear_rules", 0),
    ("'clear_rules' removes all rules.",
     (fun env term ->
       ensure_branch_reset env;
       Env.clear_rules env)));
   (("load", 1),
    ("'load(Filename)' processes the statements in the specified file.\n\
      Filename can be a relative pathname. If the load command is called\n\
      from an input file, the pathname is resolved using the directory of\n\
      that file. If load is called from standard input, the pathname is\n\
      resolved relative to the current working directory.",	   
     (fun env term ->
       let filename = Convert.term_to_string (Convert.term_nth_arg term 0) in
       process_file env filename)));
   (("load_shell_output", 1),
    ("'load_shell_output(ShellCommand)' runs the given shell command\n\
     in parallel with the krh program. The standard output of the shell\n\
     command is processed as input by krh. The shell command is interpreted\n\
     by /bin/sh.\n\
     Example: #(load_shell_command('canonicalize.pl myfile.tme'))",
     (fun env term ->
       let shell_cmd = Convert.term_to_string (Convert.term_nth_arg term 0) in
       let input = Unix.open_process_in shell_cmd 
       and input_type =
       	(match (Cfg.get_input_type (Env.get_cfg env)) with
       		| 0 -> 1
       		| 1 -> 1
       		| 2 -> 2
       		| 3 -> 3
       		| _ -> failwith "Process.cmd_table: unsupported input type")
       	in
       process_in_channel env input_type input)));
   (("onto_file", 2),
    ("'onto_file(StatementList, Filename)' subsequently processes each\n\
       Statement in StatementList with redirecting standard output to the\n\
       specified file.",
     (fun env term ->
       let input_stmts = (Convert.term_nth_arg term 0) in
       let stmts =
	 if Convert.term_is_list input_stmts then
	   Convert.term_list_to_list input_stmts
	 else
	   begin
	     Msg.msg 1 ("Warning: Deprecated use of onto_file/2.");
	     [input_stmts]
	   end
       and filename = Convert.term_to_string (Convert.term_nth_arg term 1) in
       Msg.msg 1 ("Writing file: "^filename^".");
       Fromonto.onto_file
	 (fun () -> 
	   List.iter (fun stmt -> ignore (process_statement env stmt)) stmts)
	 filename
     )));
   (("progn", 1),
    ("'progn(StatementList)' subsequently processes each statement in\n\
       the list.",
     (fun env term ->
       List.iter
	 (fun stmt ->
	   ignore (process_statement env stmt))
	 (Convert.term_list_to_list
	    (Convert.term_nth_arg term 0)))));
   (("do_models", 1),
    ("'do_models(StatementList)' processes the statements in StatementList\n\
       like progn but for each model that is found. During processing of\n\
       the statements, the current model is set to the respective model.",
     do_models_cmd));
   (("write", 1),
    ("'write(String)' writes the content of the given string to the\n\
       standard output (similar to write/1 in Prolog). String is a\n\
       Prolog-syntax atom. Special sequences, starting with the tilde (~),\n\
       are interpreted as formatting actions:\n\
       \n    ~~    Output the tilde itself.\
       \n    ~m    Output the number of the current model.\
       \n    ~z    Flush the output stream.",
     (fun env term ->
       (write_formatted
	  (Convert.term_to_string (Convert.term_nth_arg term 0))))));
   (("assert", 1),
    ("'assert(Statement)' processes Statement.",
     (fun env term ->
      ignore (process_statement env (Convert.term_nth_arg term 0)))));
   (("time", 1),
    ("'time(StatementList)' subsequently processes each Statement in\n\
       StatementList and prints information about the execution time.",
    (fun env term ->
      let input_stmts = (Convert.term_nth_arg term 0) in
      let stmts =
	if Convert.term_is_list input_stmts then
	  Convert.term_list_to_list input_stmts
	else
	  begin
	    Msg.msg 1 ("Warning: Deprecated use of time/1.");
	    [input_stmts]
	  end
      in
      let t0 = Sys.time () in
      List.iter (fun stmt -> ignore (process_statement env stmt)) stmts;
      let t1 = Sys.time () in
      print_string "% CPU Time: ";
      print_float (t1 -. t0);
      print_endline " seconds.")));
   (("set_verbosity", 1),
    ("'set_verbosity(Integer)' assigns a verbosity value, that determines\n\
      which messages are printed. It should range from -1 (no messages)\n\
      to 3 (maximum messages).",
      (fun env term ->
	Cfg.set_verbosity
	  (Convert.term_to_int (Convert.term_nth_arg term 0)))));
   (("set_message_output", 1),
    ("'set_message_output(OutputStream)' determines the output stream on\n\
      which messages are printed. By default messages are printed to the\n\
      standard error stream. Since standard error and standard output use\n\
      different buffers internally, they are not synchronized. OutputStream\n\
      can be 'stdout', 'stderr' or a filename.",
    (fun env term ->
      Cfg.set_msg_output
	(Convert.term_to_string (Convert.term_nth_arg term 0)))));
   (("start_virtual_timer", 1),
    ("'start_virtual_timer(Seconds)'  starts a virtual timer, which will terminate\n\
    E-KRHyper after <Seconds> seconds.",
    (fun env term ->
	(let seconds = Convert.term_to_float (Convert.term_nth_arg term 0)
	in
		start_virtual_timer seconds))));
   (("start_wallclock_timer", 1),
    ("'start_wallclock_timer(Seconds)' starts a wallclock timer, which will terminate\n\
    E-KRHyper after <Seconds> seconds.",
    (fun env term ->
	(let seconds = Convert.term_to_float (Convert.term_nth_arg term 0)
	in
		start_wallclock_timer seconds))));		
   (("stop_virtual_timer", 0),
    ("'stop_virtual_timer' stops the virtual timer.",
    (fun env term ->
		(stop_virtual_timer ()))));
   (("stop_wallclock_timer", 0),
    ("'stop_wallclock_timer' stops the wallclock timer.",
    (fun env term ->
		(stop_wallclock_timer ()))));
   (("clear_builtins", 0),
    ("'clear_builtins' disables builtin predicates, arithmetic evaluation\n\
       and logic operators with exception of: :-/2, ','/2, ';'/2, 'false'/0, and 'true'/0.\n\
       It must be called before any facts or rules are read in.",
      (fun env _ ->
	 (let st = (Env.get_symbol_table env) in
	 Evaluable.clear_evaluables st;
	 Builtin.clear_builtins st;
(*	 Term.set_true_symbol st (Term.gensym ());	*)
	 Term.set_stratified_neg_as_failure_symbol st (Term.gensym ());
	 Term.set_stratified_neg_as_failure_alternate_symbol
	   st (Term.gensym ());
	 Term.set_stratified_neg_as_failure_not_subsumed_symbol
	   st (Term.gensym ());
	 Term.set_findall_symbol st (Term.gensym ());
	 Term.set_neg2_symbol st (Term.gensym ())))));
   (("sys_gc", 0),
    ("'sys_gc' forces an OCaml garbage collection.",
     (fun env term ->
      Gc.full_major ())));
   (("sys_print_stat", 0),
    ("'sys_print_stat' prints some OCaml system statistics. See\n\
      documentation of Gc.print_stat in the OCaml standard library for\n\
      more information.",
     (fun env term -> Gc.print_stat stdout)));
   (("man", 1),
    ("'man(Section)' prints the specified section of the documentation.\n\
      Currently the sections 'general', 'commands', 'flags' and 'parameters'\n\
      are supported.",
    (fun env term ->
      let page = Convert.term_to_string (Convert.term_nth_arg term 0) in
      match page with
	"general" -> man_general env
      | "commands" -> man_commands env
      | "flags" -> man_flags env
      | "parameters" -> man_parameters env
      | _ -> print_endline "<No such documentation page available>")));
   (("print_version", 0),
    ("'print_version' prints a fact version(Version) with Version an atom\n\
       indicating the version of ekrh.",
	   (fun env term -> 
	     print_string ("version('" ^ (Release.release ()) ^ "').");
	     print_newline ())));
   (("exit", 0),
    ("'exit' terminates ekrh with status code 0.",
     (fun env term -> 
       Msg.msg 1 "Bye.";
       exit 0 )))]

and process_meta_statement_1 env term =
  let (_, cmd) =
    try
      List.assoc
	(Convert.term_functor_pair term)
	cmd_table
    with Not_found ->
      raise (Err.Error "Unknown command.") in
  cmd env term

and bigheader x =
  print_string "========================================";
  print_string "======================================\n";
  print_string x;
  print_string "\n========================================";
  print_string "======================================\n\n"

and smallheader x =
  print_string "----------------------------------------";
  print_string "--------------------------------------\n";
  print_string x;
  print_string "\n----------------------------------------";
  print_string "--------------------------------------\n"

and man_general env =
  bigheader "Invocation";
  Printf.printf "    krh [-tty] [FILE]...\n\n";
  Printf.printf 
    "Process the statements in the given files.\n\
     \n\
     With no FILE, or when FILE is -, process standard input.\n\
     \n\
     If standard input is processed, and it is a terminal, the system\n\
     prints a prompt whenever it is ready to read a statement.\n\
     \n\
     Signal SIGINT (CTRL-C) can be used to abort processing of a proving/\n\
     model generation task. In a terminal CTRL-D (end of file) can be\n\
     used to quit the program.\n\
     \n\
     Options:\n\
     \n\
     -tty\
     \n        Forces to handle the standard input in terminal mode,\
     \n        also if it is not actually connected to a terminal.\n\n";
  bigheader "Statements";
  Printf.printf "A statement is either of the form\n\n  #(Command).\n\n\
    which effects that the given command will be executed. Or it is a\n\
    clause that will be asserted.\n\n\
    Statements must be written in canonical ISO-Prolog syntax. Lists\n\
    may also be written in Prolog list notation, using square brackets and\n\
    the vertical bar. Strings in double-quotes are not supported.\n\n";
  flush stdout

and man_commands env =
  bigheader "Commands";
  Printf.printf "Note: Commands starting with 'print_' produce Prolog \
    readable output.\n\n";
  List.iter
    (fun ((name, arity), (doc, _)) ->
      smallheader (Printf.sprintf "Command: %s/%d" name arity);
      Printf.printf "%s\n\n" doc)
    cmd_table;
  flush stdout

and man_flags env =
  bigheader "Flags";
  Cfg.print_flags_doc (Env.get_cfg env);
  flush stdout

and man_parameters env =
  bigheader "Parameters";
  Cfg.print_parameters_doc (Env.get_cfg env);
  flush stdout
    
and help env =
  man_general env;
  man_commands env;
  man_flags env;
  man_parameters env

and os_verified filename = 
(*	filename	*)
(* This was supposed to become a smart file opener that can handle difficult filenames. Some
characters are not allowed in filenames under certain operating systems, so an OS may rename
such files, for example with the offending character replaced by '_' in Windows. 
However, references to the file (for example in a '#(load/1)'-commands in a config-file)
will still use the original name and thus cause E-KRHyper not to find the file. 
This file opener was supposed to fix the problem in Windows and Cygwin, but it turns out that
in practice the required pattern matchings become so convoluted that the function is not safe for use.
For example, while Windows does not allow the colon ':' in filenames, the character is allowed in
path names. Given a path name string, it is not trivial to determine whether a colon is acceptable
or not. It may be feasible to do so, but a function that catches such anomalies will grow fast,
and it will be difficult to tell whether it is capable of handling all anomalies in all operating
systems, adding a lot of bulk and uncertainty just to be smart about a tiny detail. 
It seems preferrable to require the user to use a proper file naming scheme. *)
  	match Sys.os_type with
  		| "Unix" ->
  			(filename)
  		| _ ->
			(let bad_chars = ['\\'; '/'; ':'; '*'; '?'; '"'; '<'; '>'; '|'; '$']
			in	
				for index = 0 to ((String.length filename) - 1) do
					(if (List.exists (fun bc -> (filename.[index] = bc)) bad_chars) then
						(filename.[index] <- '_'))
					done;
				filename)	
  
and process_file env filename = 
  let cfg = Env.get_cfg env in 
  let old_cd = Cfg.get_current_directory cfg in
  let filename = filename (*os_verified filename*)
  in
  let absfilename =
    if Filename.is_relative filename then
      (Filename.concat old_cd filename)
    else
      (filename) in
  Msg.msg 1 ("Processing file: "^absfilename^".");
  let input_type =
  	(match (Cfg.get_input_type cfg) with
  		| 0 ->
  			(if (Filename.check_suffix absfilename ".p") 
  			|| (Filename.check_suffix absfilename ".ax")
  			|| (Filename.check_suffix absfilename ".tptp") then 
  				(2) 
  			else (1))
  		| 1 -> 1
  		| 2 -> 2
		| 3 -> 3
  		| _ -> failwith "Process.process_file: unsupported input type.")
  	and input = open_in_bin absfilename 
  	in		
	  try 
	    Cfg.set_current_directory cfg (Filename.dirname absfilename);
	    process_in_channel env input_type input;
	    Cfg.set_current_directory cfg old_cd
	  with e ->
	    Cfg.set_current_directory cfg old_cd;
	    close_in input;
	    raise e

	    	    
and process_next_tme env stream =
	let symbol_table =
		Env.get_symbol_table env
	and proto_term =
		Stream.next stream
	in	
	let term =
		Clausify.get_term
			symbol_table
			proto_term	   
	in
		process_statement env term

			
and process_next_tptp env stream =
	let symbol_table =
		Env.get_symbol_table env
	and cfg =
		Env.get_cfg env
	and fof =
		Stream.next stream
	in
	let terms =
		Clausify.clausify
			~negate_conjecture:(Cfg.get_tptp_fof_negate_conjectures cfg)
			~focus:(Cfg.get_tptp_fof_compute_answer_predicate cfg)
(*			~partial_answers:(Cfg.get_tptp_fof_compute_partial_answers cfg)	*) (* too slow for large input *)
			symbol_table
			fof
	in	(* List.iter (fun term -> Term.print_term term;pe " <BEFORE") terms; *)
		List.for_all
			(fun term -> 
				process_statement env term)
			terms
			
and process_next_canonical_tme env stream =
 	let term =
		Stream.next
			stream	   
	in
		process_statement env term

	    
and process_in_channel env input_type input =
  	(match input_type with
  		| 1 ->	(* tme-input *)
  			(let proto_term_stream =
			    (TmeRead.tme_parser
			    	(Lexing.from_channel input))
		     in 
			     try
			     	(while true do
		     			(try
			     			(if (process_next_tme env proto_term_stream) then
			     				()
			     			else
			     				(exit 0))
			     		with
			     			| Timeout terminate ->
			     				(process_giving_up env;
			     				if terminate then
			     					(exit 0)));
			     	done)
				with Stream.Failure -> ();
				close_in input)		
		| 2 ->	(* tptp-input *)
			(let fof_stream =
				(TptpRead.tptp_parser
					(Env.get_cfg env)
					(Lexing.from_channel input))
			in
				try
			     	(while true do
			     		(try
				     		(if (process_next_tptp env fof_stream) then
				     			()
				     		else
				     			(exit 0))
				     	with
				     		| Timeout terminate ->
				     			(process_giving_up env;
				     			if terminate then
				     				(exit 0)));
			     	done)
				with Stream.Failure -> ();
				close_in input)
  		| 3 ->	(* classic tme input, canonical Prolog-style *)
  			(let statement_stream =
			    (PrologRead.prolog_parser
			    	(Env.get_symbol_table env)
			    	(Stream.of_channel input))
		     in 
			    try
			     	(while true do
			     		(try
				     		(if (process_next_canonical_tme env statement_stream) then
				     			()
				     		else
				     			(exit 0))
				     	with
				     		| Timeout terminate ->
				     			(process_giving_up env;
				     			if terminate then
				     				(exit 0)));
			     	done)
				with Stream.Failure -> ();
				close_in input)
		| _ -> failwith "Process.process_in_channel: unsupported input type.")


		 
      
and do_models_cmd =
  (fun env term ->
    let stmts = 
      Convert.term_list_to_list
	(Convert.term_nth_arg term 0) in
    ensure_branch_reset env;
    let t0 = Sys.time () 
    and sat = ref false
    and clauses = (Env.get_clauses env)
    and predicate_table = (Env.get_predicate_table env)
    and cfg = (Env.get_cfg env)
    in
    if (Cfg.get_reassign_predicates cfg) then
	    (List.iter
	    	(fun clause ->
	    		Ic.assign_predicates predicate_table clause)
	    	clauses;
	    Cfg.set_reassign_predicates cfg false);
    let result =
      ref (Tableau.first_result
        cfg
        predicate_table
        (Env.get_symbol_table env)
		(Env.get_negative_units env)
		(Env.get_subterm_table env)
		(Env.get_clause_aux env)
        clauses
        (Env.get_current_state_ref env)
        (Env.get_redundant_store_trail_ref env)) 
    in
    while (Tableau.is_open !result) do
      sat := true;
      model_count := !model_count + 1;
      Msg.msg1 1 "*** RESULT: Fixedpoint, %8.3fs." (Sys.time () -. t0);
      szs_output env !sat;
      Env.set_current_result env !result;
      Tableau.info_msg !result;
      List.iter
	(fun stmt ->
	  ignore (process_statement env stmt))
	stmts;
      result := Tableau.next_result !result;
    done;
    model_count := 0;
    Msg.msg1 1 "*** RESULT: Closed, %8.3fs." (Sys.time () -. t0);
    Env.set_current_result env !result;
    Env.set_current_answers
      	env
      	(Tableau.get_answers !result);
    Tableau.info_msg !result;
    szs_output env !sat)


    
 
    
    
(****************************************************************************)
(* Interactive Processing. *)
(****************************************************************************)

let rec connected_to_terminal () =
  isatty (Unix.descr_of_in_channel stdin)
   
and isatty fd =
  match Sys.os_type with
    "Unix" ->
      (try 
	ignore (Unix.tcgetattr fd);
	true
      with Unix.Unix_error (Unix.ENOTTY,_,_) 
      |  Unix.Unix_error (Unix.EINVAL,_,_) -> 
	  false
      | Unix.Unix_error (e, _, _) ->
	  raise (Failure ("Isatty hack failure: "^(Unix.error_message e))))
  | _ -> false

let prompt () =
  print_string "KRH> ";
  flush stdout
    
  
(* The original version by Christoph Wernhard, with his solution to OCaml's
stdin stream problem. Kept for reference. *)  
let old_process_stdin env force_tty =
  let tty = force_tty || connected_to_terminal () in
  if tty then
    begin
      Msg.msg 1 ("Welcome to KRH - " ^ (Release.release ()));
      Msg.msg 1 "Enter #(help). for help."
    end;
  let p =
    ref (PrologRead.prolog_parser
	   (Env.get_symbol_table env)
	   (Stream.of_channel stdin)) in
  let loop = ref true in
  while !loop
  do
    if tty then
      prompt ();
    try
      let term =
	try
	  Stream.next !p
	with Stream.Failure ->
	  (* ### how to find out if stdin is at EOF ? Stream.is_empty
	     is not suited, since it seems to parse the next term. *)
	  loop := false;
	  raise Stream.Failure
	| e ->
	  p := (PrologRead.prolog_parser
		  (Env.get_symbol_table env)
		  (Stream.of_channel stdin));
	  raise (Failure ("Syntax Error: "^(Printexc.to_string e)))
      in
      loop := process_statement env term
    with e -> (* ### Error handling *)
      if !loop then
	Msg.msg 0 ("Error: "^(Printexc.to_string e))
      else
	()
  done;
  if tty then
    Msg.msg 1 "Bye."

 
let check_input_type cfg =
	let previous_input_type =
		Cfg.get_previous_input_type cfg
	and input_type =
		Cfg.get_input_type cfg
	in
		if (previous_input_type = input_type) then 
			()
		else
			(Cfg.set_previous_input_type cfg input_type;
			raise (Switched_input_type input_type))

let get_current_input_string () =
	let current_input_string = ref ""
	in
	(try
		(while true do
			(current_input_string :=
				(!current_input_string ^ (String.make 1 (input_char stdin))))
		done)
	with 
		| End_of_file -> ());
	!current_input_string
	
(* Current version, featuring the different input types. *)
let process_stdin env force_tty =
	let tty = (force_tty || connected_to_terminal ())
	and inner_loop = ref true
(*	and outer_loop = ref true	*)
	and cfg = Env.get_cfg env
	and handle_switched_input_type input_type =
		(match input_type with
			| 0 | 1 -> Msg.msg 1 "Switched input type to native TME."
			| 2 -> Msg.msg 1 "Switched input type to TPTP."
			| 3 -> Msg.msg 1 "Switched input type to canonical TME."
			| _ -> failwith "Process.process_stdin: unsupported input type.")
	in	
	let handle_stdin_exception exc =
		(match exc with
			| Stream.Failure -> 
				(inner_loop:=false)		
			| Switched_input_type input_type ->
				(handle_switched_input_type input_type)	
			| Timeout terminate ->
				(process_giving_up env;
				if terminate then
					(exit 0))	
			| Interrupted ->
				(exit 0)
			| e -> 
				(print_endline (Printexc.to_string e)))
	in	
		if tty then
			(Msg.msg 1 ("Welcome to KRH - " ^ (Release.release ()));
			 Msg.msg 1 "Enter #(help). for help.");
		while !inner_loop do
			(try
				(match (Cfg.get_input_type cfg) with
					| 0 | 1 ->
						(let proto_term_stream =
							(TmeRead.tme_parser
								(Lexing.from_channel stdin))
						in
							(try
								(while !inner_loop do	
									(check_input_type cfg;
									if tty then prompt ();
									inner_loop := process_next_tme env proto_term_stream);
								done)	
							with 
								| e -> 
									(handle_stdin_exception e)))							
					| 2 ->
						(let fof_stream =
							(TptpRead.tptp_parser
								cfg
								(Lexing.from_channel stdin))
						in
							(try
								(while !inner_loop do
									(check_input_type cfg;
									if tty then prompt ();
									inner_loop := process_next_tptp env fof_stream);
								done)
							with 
								| e -> 
									(handle_stdin_exception e)))	
					| 3 -> 
						(let statement_stream =
							(PrologRead.prolog_parser
								(Env.get_symbol_table env)
								(Stream.of_channel stdin))
						in
							(try
								(while !inner_loop do
									(check_input_type cfg;
									if tty then prompt ();
									inner_loop := process_next_canonical_tme env statement_stream);
								done)
							with 
								| e -> 
									(handle_stdin_exception e)))
					| _ -> failwith "Process.process_stdin: unsupported input type.")
			with 
				| Timeout terminate ->
					(process_giving_up env;
					if terminate then 
						(exit 0)))
					
		done;
		if tty then
    		(Msg.msg 1 "Bye.")
	 
		
(****************************************************************************)
(* Invocation. *)
(****************************************************************************)

let is_option string =
  String.length string > 1 && String.get string 0 = '-'

let main () =
  print_endline "";		(* flush stdout *)
  let force_tty = ref false 
  and env = Env.make_env (Cfg.make_cfg ())
  and n = (Array.length Sys.argv) - 1
  and i = ref 1 in
  set_timeout_handling env;	
  while !i <= n && is_option Sys.argv.(!i)  do
    begin
      let option = Sys.argv.(!i) in
      match option with
	"-tty" ->
	  force_tty := true;
      | _ ->
	  prerr_string "ekrh: invalid option '";
	  prerr_string option;
	  prerr_endline "'.";
	  exit 1
    end;
    i := !i + 1
  done;
  if !i > n then
    process_stdin env !force_tty
  else
    (try
	    (for j = !i to n do
	      let arg = Sys.argv.(j) in
	      if arg = "-tty" then
		force_tty := true
	      else
		if arg = "-" then
		  (process_stdin env !force_tty)
		else
		  (process_file env arg)
	    done)
	 with
	 	| Timeout terminate ->
	 		(process_giving_up env;
	 		if terminate then
	 			(exit 0)
	 		else
	 			(process_stdin env !force_tty)))

  
let interact () =
  let env = Env.make_env (Cfg.make_cfg ()) in
  process_stdin env true

