(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(* States - originally only required for internal tableau processing, but E-KRHyper needs
them for input layering. *)

module State :
	sig
		type state
		val obtain_state : Predicate.predicate_table -> NegativeUnits.store 
			-> Ic.subterm_table -> (Fact.fact list) -> (Ic.clause list) 
			-> (Term.term list) -> Ic.redundant_store -> state
(* adds a layer to the indexes; when used for layering of input (#(save_input_state) command,
then no layer should be added to negative_units (set parameter to true), since the negative
units store is only ever used/filled during the derivation - it has nothing to do with the input,
and layering it beforehand will cause errors *)			
		val add_layer : ?dont_add_to_negative_units:bool -> state -> state
		val install_state : state -> unit
		val get_facts : state -> Fact.fact list
		val get_clauses : state -> Ic.clause list
		val get_rules : state -> Term.term list
		val get_extent_state : state -> Access.Store.state list		
		val get_redundant_store : state -> Ic.redundant_store
	end


(** Evaluation of clause sets. *)

(** Operations are performed on hypertableau like structures. The fact base is
    modified by side effects: at a point in time, there is a single 
    "current result" (i.e. "current model") (with respect to a 
    [Predicate.predicate_table] and [NegativeUnits.store]).
*)

(** Represents an open hypertableau fragment or a closed hypertableau. *)
type result

(** Create a [result] object. The [result] object is not open (see
  [is_open]). It is used to initialize an environment structure 
  (see module [Env]). It may be seen as representing input facts,
  without inference rules applied.
*) 
val initial_result : Cfg.cfg ->
  Predicate.predicate_table -> Term.symbol_table ->
  NegativeUnits.store -> Ic.subterm_table -> Ic.Aux.aux -> State.state ref -> (Ic.redundant_store list) ref -> result

(** Derive the first model (or a closed tableau). *)
val first_result : Cfg.cfg -> Predicate.predicate_table -> Term.symbol_table ->
  NegativeUnits.store -> Ic.subterm_table -> Ic.Aux.aux -> IcSet.clauses -> 
  State.state ref -> (Ic.redundant_store list) ref -> result

(** Derive the next model (or a closed tableau) with respect to the
   given model. If the given [result] is not open, it is returned directly.
   *)
val next_result : result -> result

(** Reset the fact base to its original state. The effect of calling
   [next_result] on the input [result] after calling this function is
   undefined. The [result] object returned is not open (see [is_open]).
   Notice that [first_result] and [next_result] automatically reset the
   fact base if they return a closed result.
 *)
val abandon_result : result -> result

(** True iff [result] is open (i.e. is a model). If [result] is not
 open, and has been returned by [first_result] or [next_result], it is closed
 (i.e. represents a refutation). *)
val is_open : result -> bool

(** Print information messages, such as statistics, about [result] to
   the message output. Reported values may only be accurate if this
   function is called immediately after the result has been returned
   by [first_result] or [next_result].
 *)
val info_msg : result -> unit

type refutation =
	NullRefutation
  | PrimitiveRefutation of Fact.proof
  | ComposedRefutation of (refutation list)

val get_refutation : result -> refutation

val get_answers : result -> Fact.fact list

val refutation_to_term : Term.symbol_table -> refutation -> Term.term



(** Deprecated *)
val get_l_proof : result -> Fact.l_proof

(* Debugging *)
val print_compiled_rules : 
   Cfg.cfg -> Predicate.predicate_table -> Term.symbol_table ->
   NegativeUnits.store -> Ic.subterm_table -> Ic.Aux.aux -> IcSet.clauses -> State.state ref -> (Ic.redundant_store list) ref -> unit

   
