(*
* Copyright (C) 2007 Universitaet Koblenz-Landau
* 
* This file is part of E-KRHyper.
* 
* E-KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*                 Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*
* This file has been adapted from the file lexer_tme.mll from the Darwin
* theorem prover. The following is the original copyright notice:
*	This file is part of the first order theorem prover Darwin
*	Copyright (C) 2006  The University of Iowa
*	
*	This program is free software; you can redistribute it and/or
*	modify it under the terms of the GNU General Public License
*	as published by the Free Software Foundation; either version 2
*	of the License, or (at your option) any later version.
*	
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*	
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(* A lexer for TME input *)

{ 

open TmeParser

let prev_column_index =
  ref 1

let current_column_index =
  ref 1

let prev_line_index =
  ref 1

let current_line_index =
  ref 1

let current_token =
  ref ""

let update_column_index (value: int) =
  prev_column_index := !current_column_index;
  current_column_index := value


let update_line_index () =
  prev_line_index := !current_line_index;
  current_line_index := !current_line_index + 1;
  update_column_index 1


let count_new_lines (string: string) : unit =
  String.iter
    (fun char ->
       if char = '\n' then
	 update_line_index ()
       else
	 update_column_index (!current_column_index + 1)
    )
    string

let update_token (token: string) =   
  current_token := token;
  update_column_index (!current_column_index + String.length token)

exception PARSE_ERROR

let lexing_error (error: string) (token: string) =
  print_endline (error
	    ^ " at line " ^ string_of_int !prev_line_index
	    ^ " column " ^ string_of_int !prev_column_index
	    ^ ":\n" ^ token);
  raise PARSE_ERROR

let parse_error () =
  print_endline ("Parse error"
	    ^ " at line " ^ string_of_int !prev_line_index
	    ^ " column " ^ string_of_int !prev_column_index
	    ^ ":\n" ^ !current_token);
  raise PARSE_ERROR

}




(* regular expressions *)

let one_line_comment =
  '%' [^ '\n' '\r']* ('\n' | "\r\n" | eof)

let multi_line_comment =
  "/*" ( [^ '*'] | ('*' [^ '/']) )* "*/"

let multi_line_comment_unclosed =
  "/*" ( [^ '*'] | ('*' [^ '/']) )* eof

(* let variable =
  [ 'A'-'Z' '_' ] [ 'a'-'z' 'A'-'Z' '0'-'9' '_']*
  
let symbol =
  (*[ 'a'-'z' '0'-'9' ':' '*' '+' '-' '/'] [ 'a'-'z' 'A'-'Z' '0'-'9' '_']* *)
  [ 'a'-'z' '0'-'9'] [ 'a'-'z' 'A'-'Z' '0'-'9' '_']*
*)
let implication =
  ":-" | "<-"

let negation =
  "~" | "\\+" | "not"

let arithmetic =
  "<" | "=:=" | "=<" | "=\\=" | ">" | ">="

let is =
	"is"  
  
let numeric = ['0' - '9']
let non_zero_numeric = ['1' - '9']
let integer = ('0' | non_zero_numeric numeric*)
let fraction = '.' numeric (numeric* | numeric* "e" numeric numeric* | numeric* "e-" numeric numeric*)
let float = integer fraction

let lower_alpha = ['a' - 'z' '#' '$'] (* '#' '$' '%' '&' '*' '+' '-' '.' '/' ':' '<' '=' '>' '@' '\\' '^' '`' '~'] *)
let upper_alpha = ['A' - 'Z']
let alpha_numeric = ( lower_alpha | upper_alpha | numeric | '_' )
let upper_word = upper_alpha alpha_numeric*
let lower_word = lower_alpha alpha_numeric*
let single_quoted = "'" ([^ '\\' '\''] | '\\' "'" | "\\")* "'"
let variable = (upper_alpha | '_') alpha_numeric*


rule token =
    parse
      | [' ' '\t' '\r']              { update_token (Lexing.lexeme lexbuf);
				       token lexbuf } (* skip blanks *)
      | ['\n']                       { update_line_index ();
				       current_token := Lexing.lexeme lexbuf;
				       token lexbuf } (* skip new lines *)
      | one_line_comment             { update_line_index ();
				       current_token := Lexing.lexeme lexbuf;
				       token lexbuf } (* skip comment *)
      | multi_line_comment           { count_new_lines (Lexing.lexeme lexbuf);
				       current_token := Lexing.lexeme lexbuf;
				       token lexbuf } (* skip comment *)
      | multi_line_comment_unclosed  { prev_column_index := !current_column_index;
				       prev_line_index := !current_line_index;
				       lexing_error "Unclosed Comment" (Lexing.lexeme lexbuf) }
	  (* end of input - for channels, strings, ... *)
      | arithmetic               { update_token (Lexing.lexeme lexbuf); ARITHMETIC (Lexing.lexeme lexbuf) }
      | is			               { update_token (Lexing.lexeme lexbuf); IS (Lexing.lexeme lexbuf) }      
      | eof                          { update_token (Lexing.lexeme lexbuf); EOI }
      | variable                     { update_token (Lexing.lexeme lexbuf); VARIABLE (Lexing.lexeme lexbuf) }
      | single_quoted                { update_token (Lexing.lexeme lexbuf); SINGLE_QUOTED(Lexing.lexeme lexbuf) }  
      | float 	               { update_token (Lexing.lexeme lexbuf); FLOAT(Lexing.lexeme lexbuf) }
      | integer 	               { update_token (Lexing.lexeme lexbuf); INTEGER(Lexing.lexeme lexbuf) }  
      | '|'                          { update_token (Lexing.lexeme lexbuf); OR }
      | '('                          { update_token (Lexing.lexeme lexbuf); LEFT_PARENTHESIS }
      | ')'                          { update_token (Lexing.lexeme lexbuf); RIGHT_PARENTHESIS }
      | '{'                          { update_token (Lexing.lexeme lexbuf); LEFT_BRACE }
      | '}'                          { update_token (Lexing.lexeme lexbuf); RIGHT_BRACE }
	  | '['                          { update_token (Lexing.lexeme lexbuf); LEFT_BRACKET }
      | ']'                          { update_token (Lexing.lexeme lexbuf); RIGHT_BRACKET }     
      | implication                  { update_token (Lexing.lexeme lexbuf); IMPLICATION }
      | negation                     { update_token (Lexing.lexeme lexbuf); NEGATION }
      | '.'                          { update_token (Lexing.lexeme lexbuf); CLAUSE_END }
      | ';'                          { update_token (Lexing.lexeme lexbuf); SEMICOLON }
      | ","                          { update_token (Lexing.lexeme lexbuf); COMMA }
      | '='                          { update_token (Lexing.lexeme lexbuf); EQUALITY }
      | '#'                          { update_token (Lexing.lexeme lexbuf); META }
      | "not_subsumed"               { update_token (Lexing.lexeme lexbuf); NOT_SUBSUMED }
      | "findall"               { update_token (Lexing.lexeme lexbuf); FINDALL }
      | '*'               		{ update_token (Lexing.lexeme lexbuf); MULT (Lexing.lexeme lexbuf) }
      | '+'               		{ update_token (Lexing.lexeme lexbuf); PLUS (Lexing.lexeme lexbuf) }
      | '-'               		{ update_token (Lexing.lexeme lexbuf); MINUS (Lexing.lexeme lexbuf) }
      | '/'               		{ update_token (Lexing.lexeme lexbuf); DIV (Lexing.lexeme lexbuf) }
(*      | "abs"               		{ update_token (Lexing.lexeme lexbuf); ABS (Lexing.lexeme lexbuf) }
      | "float"             		{ update_token (Lexing.lexeme lexbuf); FLOAT_OP (Lexing.lexeme lexbuf) }
      | "sqrt"       		{ update_token (Lexing.lexeme lexbuf); SQRT (Lexing.lexeme lexbuf) }      
      | "truncate"       		{ update_token (Lexing.lexeme lexbuf); TRUNCATE (Lexing.lexeme lexbuf) }
      | "sin"       		{ update_token (Lexing.lexeme lexbuf); SIN (Lexing.lexeme lexbuf) }
      | "cos"       		{ update_token (Lexing.lexeme lexbuf); COS (Lexing.lexeme lexbuf) }      *)
      | lower_word                   { update_token (Lexing.lexeme lexbuf); LOWER_WORD(Lexing.lexeme lexbuf) }
      | upper_word                   { update_token (Lexing.lexeme lexbuf); UPPER_WORD(Lexing.lexeme lexbuf) }      
      | _                            { prev_column_index := !current_column_index;
				       prev_line_index := !current_line_index;
				       lexing_error "Invalid Input" (Lexing.lexeme lexbuf) }


{ 

(* footer *)

}
