/*
* Copyright (C) 2007 Universitaet Koblenz-Landau
* 
* This file is part of E-KRHyper.
* 
* E-KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* E-KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* E-KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>,
*                 Bjoern Pelzer <bpelzer@uni-koblenz.de>.
*
* This file has been adapted from the file parser_tptp.mly from the Darwin
* theorem prover. The following is the original copyright notice:
*	This file is part of the first order theorem prover Darwin
*	Copyright (C) 2006  The University of Iowa
*	
*	This program is free software; you can redistribute it and/or
*	modify it under the terms of the GNU General Public License
*	as published by the Free Software Foundation; either version 2
*	of the License, or (at your option) any later version.
*	
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*	
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


%{
  
  (* these are only valid for the current clause
     and have to be invalidated with init_clause for every new clause *)

  (* the variable id counter for the currently read term/clause *)
	    
	let var_id = ref 0
	
	let gen_var_id () =
	  var_id := !var_id + 1;
	  (!var_id-1)
	
	let reset_var_id () =
	  var_id := 0      
    
      
  (* mapping of the variable names (e.g. "X") of the currently read term/clause
     to variables. *)
  let (var_map: (string * int) list ref) =
    ref []



  (* reset everything in order to parse a new term/clause *)
  let init_clause () =
    reset_var_id ();
    var_map := []


	
  (* gets the variables associated with a string from the variable mapping
     creates a new mapping for a new variable *)
  let get_var (var_name: string) =
    try 
      (* way faster than List.assoc *)
      snd (
	List.find
	  (fun (var_name', _) ->
             var_name = var_name'
	  )
	  !var_map
      )
    with
      | Not_found ->
	  let new_var_id = 
	    gen_var_id ()
	  in
	    var_map := (var_name, new_var_id) :: !var_map;
	    new_var_id

  (* need to detect if input is fof and contains a conjecture *)
  (* is the input in fof format? *)
  let fof =
    ref false
    
  (* does the fof input contain a conjecture? *)
  let conjecture =
    ref false

%}
  
%token LEFT_PARENTHESIS
%token RIGHT_PARENTHESIS
%token LEFT_BRACKET
%token RIGHT_BRACKET
%token DOT
%token NEGATION
%token COLON
%token COMMA
%token EQUALITY
%token DISEQUALITY
%token EOI
%token FOF
%token CNF
%token INCLUDE
%token <string> SINGLE_QUOTED
%token <string> DOLLAR_WORD
%token <string> DOLLAR_DOLLAR_WORD
%token <string> DISTINCT_OBJECT
%token <string> LOWER_WORD
%token <string> UPPER_WORD
%token <string> UNSIGNED_INTEGER
%token <string> SIGNED_INTEGER
%token <string> REAL
%token DOLLAR_TRUE
%token DOLLAR_FALSE
%token DOLLAR
%token AND
%token OR
%token FORALL
%token EXISTS
%token BIJECTION
%token LEFT_IMPLICATION
%token RIGHT_IMPLICATION
%token XOR
%token META
%token SLASH

%start tptp_input
%type <Clausify.fof_formula> tptp_input


%%


/* parse rules */

tptp_input:
  | annotated_formula DOT
      { $1 }

  | include_ DOT
      { $1 }
      
  | command DOT
  	  { $1 }

  | EOI
  	  {if !fof || !conjecture then
  	  	(raise (Err.END_OF_FILE_WITH_DATA (!fof, !conjecture)))
  	   else
  	  	(raise Err.END_OF_FILE)}

command:
	| META LEFT_PARENTHESIS command_term_top RIGHT_PARENTHESIS 
		{init_clause ();
		Clausify.Non_formula
			(Clausify.Atomic_formula
				(Clausify.make_clausify_term
					(Clausify.Compound
						("#", [$3]))))}  		
												
												
annotated_formula:
  | fof_annotated
      { fof := true;
      	$1 }

  | cnf_annotated
      { $1 }

fof_annotated:
  | FOF LEFT_PARENTHESIS name COMMA formula_role COMMA fof_formula annotations RIGHT_PARENTHESIS 
    { init_clause ();
      if $5 = "conjecture" then
    	(conjecture := true;
    	Clausify.Conjecture $7)
      else
        ($7) }

fof_formula:
  | binary_formula
    { $1 }

  | unitary_formula
    { $1 }


binary_formula:
  | nonassoc_binary
    { $1 }

  | assoc_binary
    { $1 }

nonassoc_binary:
  | unitary_formula binary_connective unitary_formula
    { match $2 with
    	| Clausify.Right_implication ->
    		Clausify.Nonassoc_binary_formula
    			(Clausify.Left_implication, $3, $1)
    	| _ ->
    		Clausify.Nonassoc_binary_formula
    			($2, $1, $3) }

binary_connective:
  | BIJECTION
    { Clausify.Bijection }
  | LEFT_IMPLICATION
    { Clausify.Left_implication }
  | RIGHT_IMPLICATION
    { Clausify.Right_implication }
  | XOR
    { Clausify.Xor }
  | NEGATION OR
    { Clausify.Nor }
  | NEGATION AND
    { Clausify.Nand }

assoc_binary:
  | or_formula
    { Clausify.Assoc_binary_formula
    	(Clausify.Or, $1) }
  | and_formula
    { Clausify.Assoc_binary_formula
    	(Clausify.And, $1) }

or_formula:
  | unitary_formula OR more_or_formula
    { $1 :: $3 }

more_or_formula:
  | unitary_formula
    { [$1] }
  | unitary_formula OR more_or_formula
    { $1 :: $3 }

and_formula:
  | unitary_formula AND more_and_formula
    { $1 :: $3 }

more_and_formula:
  | unitary_formula
    { [$1] }
  | unitary_formula AND more_and_formula
    { $1 :: $3 }

unitary_formula:
  | quantified_formula
    { $1 }
  | unary_formula
    { $1 }
  | LEFT_PARENTHESIS fof_formula RIGHT_PARENTHESIS
    { $2 }
  | atomic_formula
    { match $1 with
    	| Clausify.Atomic_formula ct ->
    		if (Clausify.is_negated ct) then
	    		(Clausify.complement_clausify_term ct;
	    		Clausify.Negated_fof_formula $1)
	    	else
	    		($1)
	    | _ -> $1}

quantified_formula:
  | quantifier LEFT_BRACKET variable_list RIGHT_BRACKET COLON unitary_formula
    { 	let var_ids =
    		List.map
    			(fun var -> 
    				match var with
    					| Clausify.Var (id, _) ->
    						id
    					| _ -> failwith "TptpParser: bad variable.")
    			$3
    	in
    		Clausify.Quantified_formula ($1, var_ids, $6) }

quantifier:
  | FORALL
    { Clausify.Forall }
  | EXISTS
    { Clausify.Exists }

variable_list:
  | variable
    { [$1] }
  | variable COMMA variable_list
    { $1 :: $3 }

unary_formula:
  | unary_connective unitary_formula
    { match $1 with
    	| Clausify.Negation ->
    		(match $2 with
    			| Clausify.Negated_fof_formula fof ->
    				fof
    			| _ -> 
    				Clausify.Negated_fof_formula $2)
    	| _ -> failwith "tptpParser: unsupported unary connective" }

unary_connective:
  | NEGATION
    { Clausify.Negation }


cnf_annotated:
  | CNF LEFT_PARENTHESIS name COMMA formula_role COMMA LEFT_PARENTHESIS cnf_formula RIGHT_PARENTHESIS annotations RIGHT_PARENTHESIS 
      // ignore everything except for the formula
      {
	let clause = 
	  $8
	in
	  init_clause ();
	  clause
      }

formula_role:
  | LOWER_WORD
    { $1 }

annotations:
  | null
      { "" }

  | COMMA source optional_info
      { "" }



cnf_formula:
  | disjunction
      { $1 }

disjunction:
  | literal OR disjunction
      { Clausify.Assoc_binary_formula
      		(Clausify.Or, [$1; $3]) }

  | literal
      { $1 }



literal:
  | LEFT_PARENTHESIS atomic_formula RIGHT_PARENTHESIS
      { $2 }
      
  | atomic_formula
      { $1 }

  | NEGATION atomic_formula
      { match $2 with
      		| Clausify.Negated_fof_formula fof ->
      			fof
      		| _ -> Clausify.Negated_fof_formula $2 }


atomic_formula:
  | plain_atom
      { let ct = $1
      	in
      		if (Clausify.is_negated ct) then
      			(Clausify.complement_clausify_term ct;
      			Clausify.Negated_fof_formula
      				(Clausify.Atomic_formula ct))
      		else
      			(Clausify.Atomic_formula $1) }	
  | defined_atom
      { let ct = $1
      	in
      		if (Clausify.is_negated ct) then
      			(Clausify.complement_clausify_term ct;
      			Clausify.Negated_fof_formula
      				(Clausify.Atomic_formula ct))
      		else
      			(Clausify.Atomic_formula $1) }	
  | system_atom
      { let ct = $1
      	in
      		if (Clausify.is_negated ct) then
      			(Clausify.complement_clausify_term ct;
      			Clausify.Negated_fof_formula
      				(Clausify.Atomic_formula ct))
      		else
      			(Clausify.Atomic_formula $1) }
     			
plain_atom:
  | plain_term_top
      { Clausify.make_clausify_term $1 }

arguments:
  | term
      { [ $1 ] }

  | term COMMA arguments
      { $1 :: $3 }
      
command_arguments:
  | command_term
      { [ $1 ] }

  | command_term COMMA command_arguments
      { $1 :: $3 }      
      

defined_atom:
  | DOLLAR_TRUE
      { Clausify.make_clausify_term 
      		(Clausify.Const "true") }

  | DOLLAR_FALSE
      { Clausify.make_clausify_term 
      		(Clausify.Const "false") }

  | term EQUALITY term
      { Clausify.make_clausify_term 
      		(Clausify.Compound ("=", [$1; $3])) }

  | term DISEQUALITY term
      { Clausify.make_negated_clausify_term 
   			(Clausify.Compound ("=", [$1; $3])) }

system_atom:
  | system_term_top
      { $1 }


term:
  | function_term
      { $1 }

command_term:
  | variable
  	  { $1 }
  	  
  | constant SLASH number
      { Clausify.Compound ("/", [(Clausify.Const $1); $3]) }  	  
  	  
  | constant
      { Clausify.Const $1 }

  | functor_ LEFT_PARENTHESIS command_arguments RIGHT_PARENTHESIS
      { 
	  	Clausify.Compound ($1, $3)
      }      
  
  
function_term:
  | LEFT_BRACKET function_term_list_arguments RIGHT_BRACKET
  	  { $2 }
  	  
  | plain_term
      { $1 }

  | defined_term
      { $1 }

  | system_term
      { $1 }

plain_term_top:
  | constant
      { Clausify.Const $1 }

  | functor_ LEFT_PARENTHESIS arguments RIGHT_PARENTHESIS
      { 
	  	Clausify.Compound ($1, $3)
      }

command_term_top:
	| constant SLASH number
		{ Clausify.Compound ("/", [(Clausify.Const $1); $3]) }

	| constant
		{ Clausify.Const $1 }
		
	| functor_ LEFT_PARENTHESIS command_arguments RIGHT_PARENTHESIS
		{ Clausify.Compound ($1, $3) }      
      
plain_term:
  | constant
      { Clausify.Const $1 }
      
  | variable
  	  { $1 }

  | functor_ LEFT_PARENTHESIS arguments RIGHT_PARENTHESIS
      { 
	  	Clausify.Compound ($1, $3)
      }

function_term_list_arguments:
  | null
  	  { Clausify.Const "[]" }
  	  
  | function_term
  	  { let empty = Clausify.Const "[]"
  	  	in
  	    	Clausify.Compound (".", [$1; empty]) }
  	  
  | function_term COMMA function_term_list_arguments       
      { Clausify.Compound (".", [$1; $3]) }
      
  | function_term OR function_term       
      { Clausify.Compound (".", [$1; $3]) }    
        
constant:
  | atomic_word
      { $1 }

functor_:
  | atomic_word
      { $1 }

defined_term:
  | number
      { $1 }

  | DISTINCT_OBJECT
      { Clausify.Const "<UNSUPPORTED>" }

system_term_top:
  | atomic_system_word
      { Clausify.make_clausify_term 
      		(Clausify.Const $1) }

  | atomic_system_word LEFT_PARENTHESIS arguments RIGHT_PARENTHESIS
      { Clausify.make_clausify_term 
      		(Clausify.Compound ($1, $3)) }

      		
      		
      		
system_term:
  | atomic_system_word
      { Clausify.Const $1 }

  | atomic_system_word LEFT_PARENTHESIS arguments RIGHT_PARENTHESIS
      { Clausify.Compound ($1, $3) }



variable:
  | UPPER_WORD
      { let var = get_var $1
      	in
      		if $1 = !(Cfg.answer_variable_name) then
      			(Clausify.focus_variable := var)
      		else
      			();
      		Clausify.Var (var, $1) }



source:
  | general_term
      { "" }

optional_info:
  | COMMA useful_info
      { "" }

  | null
      { "" }

useful_info:
  | general_list
      { "" }
      

include_:
  | INCLUDE LEFT_PARENTHESIS file_name formula_selection RIGHT_PARENTHESIS 
      { let command =
      		Clausify.Compound
      			("load", [Clausify.Const $3])
      	in
	      	Clausify.Non_formula
				(Clausify.Atomic_formula
					(Clausify.make_clausify_term
						(Clausify.Compound
							("#", [command])))) }

formula_selection:
  | COMMA LEFT_BRACKET name_list RIGHT_BRACKET
      { $3 }

  | null
      { [] }

name_list:
  | name
      { [$1] }

  | name COMMA name_list
      { $1 :: $3 }



general_term:
  | general_data
      { $1 }

  | general_data COLON general_term
      { Clausify.Const "<UNSUPPORTED>" }

  | general_list
      { Clausify.Const "<UNSUPPORTED>" }

general_data:
  | atomic_word
      { Clausify.Const $1 }

  | atomic_word LEFT_PARENTHESIS general_arguments RIGHT_PARENTHESIS
      { Clausify.Compound ($1, $3) }

  | number { $1 }

  | DISTINCT_OBJECT
      { Clausify.Const $1 }

general_arguments:
  | general_term
      { [$1] }

  | general_term COMMA general_arguments
      { $1 :: $3 }

general_list:
  | LEFT_BRACKET RIGHT_BRACKET
      { [] }

  | LEFT_BRACKET general_term_list RIGHT_BRACKET
      { $2 }

general_term_list:
  | general_term
      { [$1] }

  | general_term COMMA general_term_list
      { $1 :: $3 }


name:
  | atomic_word
      { Clausify.Const $1 }

  | UNSIGNED_INTEGER
      { Clausify.Int (int_of_string $1) }

atomic_word:
  | LOWER_WORD
      { $1 }

  | SINGLE_QUOTED
      { let str = String.sub $1 1 ((String.length $1) - 2)
        in	
        	str }

atomic_system_word:
  | DOLLAR_DOLLAR_WORD
      { "$$"  }

number:
  | REAL
      { Clausify.Float (float_of_string $1) }

  | SIGNED_INTEGER
      { Clausify.Int (int_of_string $1) }

  | UNSIGNED_INTEGER
      { Clausify.Int (int_of_string $1) }

file_name:
  | atomic_word
      { $1 }
  | DISTINCT_OBJECT
  	  { let str = String.sub $1 1 ((String.length $1) - 2)
        in	
        	str }

null:
      { }
      
      
/*
debug_formula:
	| META
	{Clausify.Atomic_formula
				(Clausify.make_clausify_term
					(Clausify.Const
						"#"))}						

debug_cterm:
	| META
	{Clausify.make_clausify_term
					(Clausify.Const
						"#")}						

debug_const:
	| META
	{Clausify.Const	"#"}						
*/      

%%




