/**
 * Copyright (c) 2007 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-21.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "Inference.hpp"
#include "InferenceInfo.hpp"
#include "TPTP.hpp"
#include "Unit.hpp"
#include "Int.hpp"
#include "Problem.hpp"
#include "Refutation.hpp"
#include "VampireException.hpp"
using namespace Shell;
string TPTP::toString (const Atom& a)
{
   if (a.isEquality()) {
      return a.args().head().toString() + "="  + a.args().second().toString();
   }
   else {
      return (a.functor().isEqualityProxy() ? "$e" : a.functor().name()) 
	 + a.args().toString(",");
   }
}
string TPTP::toString (const Clause& c)
{
  if (c.isEmpty()) {
    return "$false";
  }
  string result = string("(") + toString(c.literals().head());
  Iterator<Literal> lits (c.literals().tail());
  while (lits.hasNext()) {
    result += string(" | ") + toString(lits.next());
  }
  return result + ")";;
} 
string TPTP::toString (const Formula& f)
{
  static string names [] =
    { "", " & ", " | ", " => ", " <=> ", " <~> ", 
      "~", "!", "?", "$false", "$true"};
  Connective c = f.connective();
  string con = names[(int)c];
  switch (c) {
  case ATOM:
    return toString(f.atom());
  case AND:
  case OR: 
    {
      string result = "(" + toString(f.args().head(),c);
      Iterator<Formula> arg (f.args().tail());
      while (arg.hasNext()) {
	result += con + toString(arg.next(),c);
      }
      return result + ")";
    }
  case IMP:
  case IFF:
  case XOR:
    return "(" + toString(f.left(),c) + con + toString(f.right(),c) + ")";
  case NOT:
    return "(" + con + toString(f.uarg(),c) + ")";
  case FORALL:
  case EXISTS:
    {
      string result = "(" + con + "[" + f.vars().head().toString();
      Iterator<Var> vs(f.vars().tail());
      while (vs.hasNext()) {
	result += string(",") + vs.next().toString();
      }
      return result + "] : " + toString(f.qarg(),c) + ")";
    }
  case FALSE:
  case TRUE:
    return con;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
string TPTP::toString (const Literal& l)
{
  return string(l.isPositive() ? "" : "~") + toString(l.atom());
} 
string TPTP::toString (const Unit& u)
{
  static string inputTypes[] = {"plain","axiom","negated_conjecture","derived"};
  UnitList parents (u.inference()->premises());
  string nm =  Int::toString(u.number());
// u.name() != "" ? u.name() : 
              
  string parString = "[";
  Iterator<Unit> ps1 (parents);
  while (ps1.hasNext()) {
    parString += Int::toString(ps1.next().number());
    if (ps1.hasNext()) {
      parString += ",";
    }
  }
  parString += "]";

  Inference* inf = u.inference();

  string infString = inf->rule() == IR_KERNEL ?
    static_cast<const KernelInferenceInfo*>(inf->info())->rulesAsString() :
    Inference::toString(inf->rule());
  
  string roleString = inputTypes[u.inputType()];

  switch (inf->rule()) {
  case IR_INPUT: 
	infString = ", file(input, " + u.name() + ")";
	if(u.inputType() == 0) {
	   roleString = "axiom";
	}
	break;
  case IR_PREDICATE_DEFINITION:
  case IR_EQUALITY_PROXY_AXIOM1:
	infString = ", introduced(definition, [vampire_rule('" + infString + "')])";
	roleString = "definition";
	break;
  case IR_EQUALITY_PROXY_AXIOM2: // have to replace \/ with | 
	infString = ", introduced(definition, [vampire_rule('equality proxy axiom ~E(x,y) | x=y')])";
	roleString = "definition";
	break;
  default:
	infString = ", inference('" + infString + "',[],"
	   + parString + ")";
  }

  switch (u.unitType()) {
  case CLAUSE:
	return string("cnf(") + nm + ", " + 
           roleString + ", " + toString(u.clause()) + infString + ").\n";
  case FORMULA:
    return string("fof(") + nm + ", " + 
           roleString + ", " + toString(u.formula()) + infString + ").\n";
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
string TPTP::toStringAsInference (const Unit& u)
{
  return toString(u);
} 
string TPTP::toString (const Formula& f, Connective outer)
{
  return f.parenthesesRequired(outer) ?
         string("(") + toString(f) + ")" :
         toString(f);
} 
string TPTP::toString (const Refutation& ref)
{
  string result = "%=========== Refutation ==========\n";
  Iterator<Unit> units(ref.refutation());
  while (units.hasNext()) {
    result += toStringAsInference(units.next());
  }
  return result + "%======= End of refutation =======\n";
} 
string TPTP::toString (const Problem& problem)
{
  string result("");
  UnitChain::Iterator us (problem.units());
  while (us.hasNext()) {
    result += toString(us.next()) + "\n\n";
  }
  return result;
} 
