# Copyright (c) 2007 Christian Heimes. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#
#  3. Neither the name of Infrae nor the names of its contributors may
#     be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INFRAE OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Get process informations

The module contains for implementations:

 - an Unsupported implementation that raises UnsupportedError
 - a Linux implementation that read the data from /proc
 - a Unix/POSIX implementation that parses the output of ps
 - a Windows implementation that uses ctypes to get the infos from psapi.dll

API
===
  getPids() - list of ints/longs
  getPidNames() - mapping pid -> name
  getPidDetails(pid) - detailed informations about a process
  getPidDetails('self') - detailed informations about current process

>>> cur = os.getpid()
>>> exe = sys.executable
>>> pids = getPids()
>>> pids > 1
True
>>> isinstance(pids[0], (int, long))
True
>>> cur in pids
True

>>> mapping = getPidNames()
>>> cur in mapping
True
>>> mapping[cur].lower() in sys.executable
True

>>> getPidDetails('self')['name'] == getPidDetails(cur)['name']
True
>>> getPidDetails(cur)['name'] == mapping[cur]
True

>>> for impl in (WinEnumProcesses(), LinuxProcReader(), PsParser()):
...     if impl.supported():
...         isinstance(impl.getPids(), list) and None
...         isinstance(impl.getPidNames(), dict) and None
"""
__author__ = "Christian Heimes"
__version__ = "$Id$"
__revision__ = "$Revision$"
__all__ = ['getPids', 'getPidNames', 'getPidDetails', 'supportedOS',
           'UnsupportedError']

import logging
import os
import os.path
import sys


LOG = logging.getLogger("processinfo")
NULL = "\x00"

_plat = sys.platform.startswith
if _plat('win') or _plat('cygwin'):
    from ctypes import windll, c_ulong, sizeof, c_buffer, byref
    PSAPI = windll.psapi
    KERNEL = windll.kernel32
    PROCESS_QUERY_INFORMATION = 0x0400
    PROCESS_VM_READ = 0x0010
    PROCESS_FLAGS = PROCESS_QUERY_INFORMATION | PROCESS_VM_READ
else:
    from subprocess import Popen
    from subprocess import PIPE

class UnsupportedError(OSError):
    """OS or command not supported error
    """
    pass

class Unsupported(object):
    """Unsupported OS implementation

    Raises L{UnsupportedError}
    """
    __slots__ = ()

    @classmethod
    def supported(cls):
        """Supported flag property"""
        return False

    @classmethod
    def log(cls):
        """Log information about the implementation"""
        LOG.warning("Unsupported OS. Neither proc filesystem nor 'ps' works.")

    @classmethod
    def getPids(cls):
        """Get a list of pids

        @return: a list of pid numbers
        @rtype: list(int)
        """
        raise UnsupportedError

    @classmethod
    def getPidNames(cls):
        """Get a mapping of pid -> name

        @return: mapping of pid -> name
        @rtype: dict(int:str)
        """
        raise UnsupportedError

    @classmethod
    def getPidDetails(pid):
        """Get detailed informations about a process

        @param pid: pid number or 'self'
        @type pid: int or basestring
        @return: detailed information about a process
        @rtype: dict(str:data)
        """
        raise UnsupportedError

class PsParser(object):
    """Parse the output of the ps command
    """
    __slots__ = ()

    CMD = "ps -e --no-header --cols=1024"
    PIDNAMES = "%s --format=pid,ucmd" % CMD
    PIDS = "%s --format=pid" % CMD

    @classmethod
    def supported(cls):
        """Supported flag property"""
        try:
            cls._exec(cls.PIDS)
        except Exception: # catch all!
            return False
        else:
            return True

    @classmethod
    def log(cls):
        """Log information about the implementation"""
        LOG.debug("Using the 'ps' command on POSIX os")

    @classmethod
    def getPids(cls):
        """Get a list of pids
        """
        stdout = cls._exec(cls.PIDS)
        if stdout is None:
            return None
        pids = []
        for line in stdout:
            try:
                pid = int(line.strip())
            except ValueError:
                pass
            else:
                pids.append(pid)
        return pids

    @classmethod
    def getPidNames(cls):
        """Get a list of pid -> name
        """
        stdout = cls._exec(cls.PIDNAMES)
        if stdout is None:
            return None
        mapping = {}
        for line in stdout:
            line = line.strip()
            idx = line.find(' ')
            pid, name = line[:idx], line[idx+1:]
            try:
                pid = int(pid)
            except ValueError:
                pass
            else:
                mapping[pid] = name
        return mapping

    @classmethod
    def getPidDetails(cls, pid):
        """Get detailed informations about a process

        TODO
        """
        if pid == 'self':
            pid = os.getpid()
        raise UnsupportedError("not implemented yet")

    @staticmethod
    def _exec(cmd):
        """Execute command cmd

        The method waits until the command has finished. It returns None of 
        something went wrong.

        @param cmd: Command to execute
        @type cmd: str
        @return: None or stdin as file like object
        """
        try:
            popen = Popen(cmd, shell=True, bufsize=-1, stdout=PIPE,
                          env = {'LC_ALL' : 'C'})
            rc = popen.wait()
        except (OSError, ValueError):
            LOG.exception("Failed to execute '%s'" % cmd)
            return None
        else:
            if rc != 0:
                LOG.error("'%s' returned with error code %i" % (cmd, rc))
                return None
            else:
                return popen.stdout

class LinuxProcReader(object):
    """Get process informations under Linux by reading /proc

    Tested under Linux, may work on other POSIX os with /proc, too.
    """
    __slots__ = ()

    PROC = "/proc"

    @classmethod
    def supported(cls):
        return os.path.isfile("%s/self/status" % cls.PROC)

    @classmethod
    def log(cls):
        """Log information about the implementation"""
        LOG.debug("Using the proc filesystem on Linux")

    @classmethod
    def getPids(cls):
        """Get a list of pids
        """
        pids = []
        for name in os.listdir(cls.PROC):
            if os.path.isdir(cls.PROC + '/' + name):
                try:
                    pids.append(int(name))
                except ValueError:
                    pass
        return pids

    @classmethod
    def getPidNames(cls):
        """Get a list of pid -> name
        """
        mapping = {}
        for pid in cls.getPids():
            name = cls._readProcStatus(pid, searchkey='name')
            if name is not None:
                mapping[pid] = name
        return mapping

    def getPidDetails(cls, pid):
        """Get detailed informations about a process
        """
        details = cls._readProcStatus(pid)
        if details is None:
            return None
        return details

    @classmethod
    def _readProcStatus(cls, pid, searchkey=None):
        """Read and parse status informations for PID pid

        pid - pid as long or int or 'self'

        If searchkey is None the method returns a mapping of lower keys
        to values (stripped).
        If searchkey is given than the method immediatly returns the value
        or returns None if they key can't be found.
        """
        mapping = {}
        status = '%s/%s/status' % (cls.PROC, pid)
        try:
            # read entiry file to avoid race conditions
            lines = open(status, 'r').readlines()
        except IOError:
            LOG.exception("%s not found" % status)
            return None
        for line in lines:
            try:
                key, value = line.split(':\t')
            except ValueError:
                continue
            key = key.lower()
            value = value.strip()
            #if value.endswith(' kB'):
            #    value = value[:-3]
            if key == searchkey:
                return value
            mapping[key.lower()] = value
        if searchkey is not None:
            return None
        return mapping


class WinEnumProcesses(object):
    """""Get process informations under Win32 with psapi.dll

    Based on enumprocesses from Eric Koome 
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/305279
    """
    __slots__ = ()

    @classmethod
    def supported(cls):
        try:
            cls.getPids()
        except Exception: # catch all!
            return False
        else:
            return True

    @classmethod
    def log(cls):
        """Log information about the implementation"""
        LOG.debug("Using ctypes on Windows")

    @classmethod
    def getPids(cls):
        """Get a list of pids
        """
        arr = c_ulong * 256
        lpidProcess = arr()
        cb = sizeof(lpidProcess)
        cbNeeded = c_ulong()

        # Call Enumprocesses to get hold of process id's
        PSAPI.EnumProcesses(byref(lpidProcess), cb, byref(cbNeeded))
        nReturned = cbNeeded.value/sizeof(c_ulong()) # number of processes returned
        return [pid for pid in lpidProcess][:nReturned]

    @classmethod
    def getPidNames(cls):
        """Get a list of pid -> name
        """
        hModule = c_ulong()
        count = c_ulong()
        modname = c_buffer(51)
        mapping = {}
        for pid in cls.getPids():
            # Get handle to the process based on PID
            hProcess = KERNEL.OpenProcess(PROCESS_FLAGS, False, pid)
            if hProcess:
                PSAPI.EnumProcessModules(hProcess, byref(hModule),
                                         sizeof(hModule), byref(count))
                PSAPI.GetModuleBaseNameA(hProcess, hModule.value, modname,
                                         sizeof(modname))
                try:
                    name = u"".join([c for c in modname if c != NULL])
                except UnicodeError, msg:
                    LOG.exception("Can't decode name of pid %s" % pid)
                else:
                    mapping[pid] = name
                modname[:] = sizeof(modname) * NULL
                KERNEL.CloseHandle(hProcess)

        return mapping

    @classmethod
    def getPidDetails(cls, pid):
        """Get detailed informations about a process
        """
        if pid == 'self':
            pid = os.getpid()

        hModule = c_ulong()
        count = c_ulong()
        modname = c_buffer(51)
        hProcess = KERNEL.OpenProcess(PROCESS_FLAGS, False, pid)
        if not hProcess:
            return None

        PSAPI.EnumProcessModules(hProcess, byref(hModule),
                                 sizeof(hModule), byref(count))
        PSAPI.GetModuleBaseNameA(hProcess, hModule.value, modname,
                                 sizeof(modname))
        try:
            name = u"".join([c for c in modname if c != NULL])
        except UnicodeError, msg:
            LOG.exception("Can't decode name of pid %s" % pid)
        else:
            name = None

        KERNEL.CloseHandle(hProcess)
        return {'name' : name}


# Initialize global methods
_enumProcesses = None
for impl in (WinEnumProcesses(), LinuxProcReader(), PsParser()):
    if impl.supported():
        impl.log()
        _enumProcesses = impl
        break

if _enumProcesses is None:
    LOG.error("System %s is not supported" % sys.platform)
    _enumProcesses = Unsupported()

getPids = _enumProcesses.getPids
getPidNames = _enumProcesses.getPidNames
getPidDetails = _enumProcesses.getPidDetails
supportedOS = _enumProcesses.supported()

def test_suite():
    import unittest
    from doctest import DocTestSuite
    return unittest.TestSuite((
        DocTestSuite()
        ))

if __name__ == '__main__':
    import unittest
    logging.basicConfig()
    unittest.main(defaultTest="test_suite")
    print getPids()
    print getPidNames()
