import os
import sys

from utils import props

import tptp
import suspender
import selection
from selection import SelectionParams

class BillingError(Exception):
	pass

def get_cpu_time():
	times=os.times()
	return reduce(lambda x,y:x+y, times[0:4])
def get_wall_time():
	return os.times()[4]

start_wall_time=get_wall_time()

class Timer(object):
	phases=[
		(0.06, SelectionParams()),
		(0.06, SelectionParams(select_proof_axioms=True)),
		(0.40, SelectionParams(select_proof_axioms=True)),
		(0.40, SelectionParams(select_proof_axioms=True, benevolence=1.2)),
		(0.80, SelectionParams(select_proof_axioms=True, benevolence=1.5)),
		(1, SelectionParams(select_proof_axioms=True, benevolence=2)),
		(1, SelectionParams(select_proof_axioms=True, benevolence=3)),
		]

	def __init__(self, cpu_limit, problem_groups, engine):
		self.__class__.inst=self
		
		self.cpu_limit=cpu_limit
		self.wall_limit=cpu_limit*2
		self.problem_groups=problem_groups
		self.engine=engine
		
		self.unsolved=set()
		for prb_set in problem_groups.itervalues():
			self.unsolved.update(prb_set)
		
		self.cpu_adjustation=0.0
	
		self.billed_object=None
		self.billing_zero_point=None
		self.billing_start_wall=None
	
	def start_billing(self, obj):
		if self.billed_object:
			raise BillingError('Another object being billed.')
		self.billed_object=obj
		self.billing_zero_point=self.time_remaining()
		self.billing_start_wall=get_wall_time()
	def stop_billing(self, obj, assume_full_cpu_usage=False):
		if self.billed_object==None:
			raise BillingError('There is no billing active.')
		if self.billed_object!=obj:
			raise BillingError('Given object is not being billed.')
		
		own_elapsed=self.billing_zero_point-self.time_remaining()
		if assume_full_cpu_usage:
			wall_elapsed=get_wall_time()-self.billing_start_wall
			self.cpu_adjustation+=wall_elapsed-own_elapsed
			secs=wall_elapsed
		else:
			secs=own_elapsed
		
		if isinstance(obj, tptp.Problem):
			obj.time_spent+=secs
		elif isinstance(obj, tptp.AxiomSetSet):
			problems=self.problem_groups[obj].intersection(self.unsolved)
			cnt=len(problems)
			if cnt==0:
				return
			inc=secs/cnt
			for prb in problems:
				prb.time_spent+=inc
		else:
			raise BillingError('Given object cannot be billed.')
		
		self.billed_object=None
		
	
	
	def wall_time_remaining(self):
		elapsed=get_wall_time()-start_wall_time
		return self.wall_limit-elapsed
	def cpu_time_remaining(self):
		elapsed=get_cpu_time()+self.cpu_adjustation
		return self.cpu_limit-elapsed
	def time_remaining(self):
		return min(self.cpu_time_remaining(), self.wall_time_remaining())

	def problem_time_remaining(self, problem):
		rem=self.time_remaining()
		cnt=len(self.unsolved)
		
		#avg_spent=reduce(float.__add__, [prb.time_spent for prb in self.unsolved])/cnt
		return rem/cnt#+avg_spent-problem.time_spent
	
	def perform_phase(self, problem):
		'''Performs single phase of proof attempts. If last phase was finished, returns False.'''
		if problem.current_phase>=len(self.phases):
			return False
		time_perc,params = self.phases[problem.current_phase]

		problem.current_phase=problem.current_phase+1
		
		forms=selection.select_formulas(problem, params)
		forms_fname=tptp.write_forms_to_file(problem, forms)
		
		#we don't get avaiable time earlier as the previous lines could take a long time
		time=time_perc*self.problem_time_remaining(problem)
		
		if time<=0:
			return False
		
		print '---remains %d cpu, %d wall, %d problems to solve' % \
		  	   	(self.cpu_time_remaining(), self.wall_time_remaining(), len(self.unsolved))
		print '-running phase %d (%s) prover on %s, run time limit is %d s.' % \
			 	(problem.current_phase, params, problem, time)


		self.start_billing(problem)
		self.engine.run_into_file(problem.out_fname, problem_file=forms_fname, time=time)
		self.stop_billing(problem, True)
		
		self.handle_finished_prover_run(problem)
			
		return True

	def perform_phase_on_group(self, axiom_set_set):
		problems=self.problem_groups[axiom_set_set].intersection(self.unsolved)
		res=False
		for prb in problems:
			try:
				res|=self.perform_phase(prb)
			except:
				print "Unexpected error:", str(sys.exc_info()[0])
				
		return res
			
	def work(self):
		ax_grps=self.problem_groups.keys()
		ax_grps.sort(lambda x,y:cmp(x.size, y.size))
		while True:
			any_run=False
			for ax_grp in ax_grps:
				any_run|=self.perform_phase_on_group(ax_grp)
			print "Phase finished."
			if not any_run or not self.unsolved:
				break


	def handle_finished_prover_run(self, problem):
		if self.engine.proof_complete(problem.out_fname):
			self.engine.update_proof_axioms(problem.out_fname, problem.axiom_set_set)
			self.unsolved.remove(problem)
			print 'Problem %s solved in phase %s in total time of %d s.' % \
			 	   (problem, problem.current_phase, problem.time_spent)
		
	@classmethod
	def get_instance(cls):
		return cls.inst

def work(*args, **kwargs):
	timer=Timer(*args, **kwargs)
	timer.work()	
	
	
	