/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "EqualityProxy.hpp"
#include "Signature.hpp"
#include "Problem.hpp"
#include "Inference.hpp"

using namespace Shell;

EqualityProxy::EqualityProxy (Signature& signature,Options::EqualityProxy opt)
  : _equality(signature.equality()),
    _option(opt)
{
  switch (opt) {
  case Options::EP_R:
  case Options::EP_RS:
  case Options::EP_RST:
    _RST = true;
    break;
  default:
    _RST = false;
    break;
  }

  string proxy("$$e");
  for (;;) {
    PredicateSymbol* symb = signature.existsPred(proxy,2);
    if (symb) { 
      if (symb->isEqualityProxy()) { 
	_proxy = symb;
	return;
      }
      proxy += "e";
    }
    else { 
      _proxy = &signature.createPred(proxy,2);
      _proxy->makeEqualityProxy();
      return;
    }
  }
} 
/**
 * Apply the equality proxy transformation to the problem.
 * It works as follows:
 * <ol>
 *   <li>all literals x = y are replaced by E(x,y);</li>
 *   <li>if _option is in {R,RS,RST} all literals s=t are replaced by E(s,t);</li>
 *   <li>all literals s != t are replaced by ~E(s,t);</li>
 *   <li>the clause E(x,x) is added;</li>
 *   <li>if _option is in {RS,RST} the symmetry clause ~E(x,y) \/ E(y,x) is added;</li>
 *   <li>if _option=RST the transitivity clause
 *       ~E(x,y) \/ ~E(y,z) \/ E(x,z) is added;</li>
 *   <li>if _option is not in {R,RS,RST} the clause ~E(x,y) \/ x = y is added;</li>
 * </ol>
 */
void EqualityProxy::apply (Problem& problem)
{
  UnitChain::DelIterator us(problem.giveUnits());
  // all literals x = y are replaced by E(x,y)
  // if _option is in {R,RS,RST} all literals s=t are replaced by E(s,t)
  // all literals s != t are replaced by ~E(s,t)
  while (us.hasNext()) {
    Unit u(us.next());
    LiteralList ls(u.clause().literals());
    apply(ls);
    if (ls != u.clause().literals()) { 
      Unit result(IR_EQUALITY_PROXY_REPLACEMENT,Clause(ls),u);
      us.replace(result);
    }
  }

  // the clause E(x,x) is added
  Term x(Var(0));
  Literal lit(true,Atom(*_proxy,TermList(x,TermList(x))));
  LiteralList lits(lit);
  Clause reflexivity(lits);
  problem.addUnit(Unit(reflexivity,
		       new Inference(IR_EQUALITY_PROXY_AXIOM1)));

  // if _option is in {RS,RST} the symmetry clause ~E(x,y) \/ E(y,x) is added
  switch (_option) {
  case Options::EP_RS:
  case Options::EP_RST:
    {
      Term y(Var(1));
      Literal lft(false,Atom(*_proxy,TermList(x,TermList(y))));
      Literal rht(true,Atom(*_proxy,TermList(y,TermList(x))));
      LiteralList lits(lft,LiteralList(rht));
      Clause symmetry(lits);
      problem.addUnit(Unit(symmetry,
			   new Inference(IR_EQUALITY_PROXY_AXIOM1)));
    }
    break;

  default:
    break;
  }
  // if _option=RST the transitivity clause ~E(x,y) \/ ~E(y,z) \/ E(x,z) is added
  switch (_option) {
  case Options::EP_RST:
    {
      Term y(Var(1));
      Term z(Var(2));
      Literal neg1(false,Atom(*_proxy,TermList(x,TermList(y))));
      Literal neg2(false,Atom(*_proxy,TermList(y,TermList(z))));
      Literal pos(true,Atom(*_proxy,TermList(x,TermList(z))));
      LiteralList lits(neg1,LiteralList(neg2,LiteralList(pos)));
      Clause transitivity(lits);
      problem.addUnit(Unit(transitivity,
			   new Inference(IR_EQUALITY_PROXY_AXIOM1)));
    }
    break;

  default:
    break;
  }

  // if _option is not in {R,RS,RST} the clause ~E(x,y) \/ x = y is added
  if (! _RST) {
    Term y(Var(1));
    TermList xy(x,TermList(y));
    Literal lft(false,Atom(*_proxy,xy));
    Literal rht(true,Atom(_equality,xy));
    LiteralList lits2(lft,LiteralList(rht));
    Clause proxy(lits2);
    problem.addUnit(Unit(proxy,
			 new Inference(IR_EQUALITY_PROXY_AXIOM2)));
  }
} 
void EqualityProxy::apply (LiteralList& ls)
{
  if (ls.isEmpty()) {
    return;
  }
  Literal hd(ls.head());
  apply(hd);
  LiteralList tl(ls.tail());
  apply(tl);
  if (hd != ls.head() || tl != ls.tail()) {
    ls = LiteralList(hd,tl);
  }
} 

void EqualityProxy::apply (Literal& l)
{
  Atom a(l.atom());
  if (! a.isEquality()) {
    return;
  }
  TermList ts(a.args());
  if (l.isPositive()) {
    if (! _RST && (ts.head().tag() != VAR ||
		   ts.second().tag() != VAR)) {
      return;
    }
  }
  else {
    if (! _RST && (ts.head().tag() == VAR ||
		   ts.second().tag() == VAR)) {
      return;
    }
  }
  l = Literal(l.sign(),Atom(*_proxy,ts));
} 
