/**
 * Copyright (c) 2006 Andrei Voronkov and the University of Manchester.
 *
 * This version of Vampire is intended to be used only for CASC-J3.
 * It is not a full version of Vampire.
 *
 * The only reason you can view the source code of this software is
 * that it is required by the conditions of CASC. You are not allowed
 * to use, modify or distribute this software or any part of it in any form.
 * To obtain a licence to use this software please check Vampire's
 * Web page. If you are interested in Vampire's implementation you
 * can check publications on Vampire.
 */

#include "TPTP.hpp"
#include "Unit.hpp"
#include "Inference.hpp"
#include "InferenceInfo.hpp"
#include "Int.hpp"
#include "Problem.hpp"
#include "Refutation.hpp"
#include "VampireException.hpp"

using namespace Shell;
string TPTP::toString (const Atom& a)
{
  return a.functor().name() + a.args().toString(",");
} 
string TPTP::toString (const Clause& c)
{
  if (c.isEmpty()) {
    return "$false";
  }
  string result = toString(c.literals().head());
  Iterator<Literal> lits (c.literals().tail());
  while (lits.hasNext()) {
    result += string("\n  | ") + toString(lits.next());
  }
  return result;
} 
string TPTP::toString (const Formula& f)
{
  static string names [] =
    { "", " & ", " | ", " => ", " <=> ", " <~> ", 
      "~", "!", "?", "$false", "$true"};
  Connective c = f.connective();
  string con = names[(int)c];
  switch (c) {
  case ATOM:
    return toString(f.atom());
  case AND:
  case OR: 
    {
      string result = string("(") + toString(f.args().head());
      Iterator<Formula> arg (f.args().tail());
      while (arg.hasNext()) {
	result += con + toString(arg.next());
      }
      return result + ")";
    }
  case IMP:
  case IFF:
  case XOR:
    return string("(") + toString(f.left()) + con + toString(f.right()) + ")";
  case NOT:
    return string("(") + con + toString(f.uarg()) + ")";
  case FORALL:
  case EXISTS:
    {
      string result = string("(") + con + "[" + f.vars().head().toString();
      Iterator<Var> vs(f.vars().tail());
      while (vs.hasNext()) {
	result += string(",") + vs.next().toString();
      }
      return result + "] : " + toString(f.qarg()) + ")";
    }
  case FALSE:
  case TRUE:
    return con;
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
string TPTP::toString (const Literal& l)
{
  Atom atom = l.atom();
  if (atom.isEquality()) {
    string op = l.isPositive() ? "=" : "!=";
    return atom.args().toString(op);
  }
  return string(l.isPositive() ? "" : "~") + toString(atom);
} 
string TPTP::toString (const Unit& u)
{
  static string inputTypes[] = {"axiom","conjecture","hypothesis","derived"};
  string nm = u.name() != "" ? 
              u.name() : 
              string("u_") + Int::toString(u.number());
  switch (u.unitType()) {
  case CLAUSE:
    return string("cnf(") + nm + "," + 
           inputTypes[u.inputType()] + ",\n  " +
           toString(u.clause()) + ").";
  case FORMULA:
    return string("fof(") + nm + "," + 
           inputTypes[u.inputType()] + ",\n  " +
           toString(u.formula()) + ").";
#if DEBUG_SHELL
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 
string TPTP::toStringAsInference (const Unit& unit)
{
  Inference* inf = unit.inference();
  InferenceRule rule = inf->rule();

  string prefix;
  string main;

  switch (unit.unitType()) {
  case CLAUSE:
    prefix = "cnf";
    main = toString(unit.clause());
    break;

  case FORMULA:
    prefix = "fof";
    main = toString(unit.formula());
    break;
  }

  string kind = "plain";
  if (rule == IR_INPUT) {
    switch (unit.inputType()) {
    case AXIOM:
      kind = "axiom";
      break;
    case ASSUMPTION:
      kind = "hypothesis";
      break;

    case CONJECTURE:
      kind = "negated_conjecture";
      break;
    }
  }

  string parents("");
  Iterator<Unit> ps(inf->premises());
  while (ps.hasNext()) {
    parents += Int::toString(ps.next().number());
    if (ps.hasNext()) {
      parents += ",";
    }
  }

  string ruleName = inf->rule() == IR_KERNEL ?
         static_cast<const KernelInferenceInfo*>(inf->info())->rulesAsString() :
         Inference::toString(inf->rule());

  return prefix + "(" + Int::toString(unit.number()) + "," + kind + ",\n"
    + "    " + main + ",\n"
    + "    inference('" + ruleName + "',[" + parents + "])).\n";

//   result += Int::toString(number()) + ", " + 
//             infString + "] " + 
//             stars + "\n";
//   switch (unitType()) {
//   case CLAUSE:
//     result += clause().toString();
//     break;
//   case FORMULA:
//     result += formula().toString();
//     break;
//   }
//   return result + "\n";
} 
string TPTP::toString (const Formula& f, Connective outer)
{
  return f.parenthesesRequired(outer) ?
         string("(") + toString(f) + ")" :
         toString(f);
} 
// string TPTP::toString (const Refutation& ref)
// {
//   string result = "%=========== Refutation ==========\n";
//   Iterator<Unit> units(ref.refutation());
//   while (units.hasNext()) {
//     result += toStringAsInference(units.next());
//   }
//   return result + "%======= End of refutation =======\n";
// } 
string TPTP::toString (const Problem& problem)
{
  string result("");
  UnitChain::Iterator us (problem.units());
  while (us.hasNext()) {
    result += toString(us.next()) + "\n\n";
  }
  return result;
} 
