/**
 * @file VampireMode.cpp
 * Defines several functions used in the Vampire Mode
 *
 * @since 03/03/2004 Manchester
 */

// uncomment the following if the problem properties must be saved as SQL
// queries to the Spider database
// #define SAVE_SPIDER_PROPERTIES 1

#if SAVE_SPIDER_PROPERTIES
#include "TheoryFinder.hpp"
#endif

#include <climits>

#include "Environment.hpp"
#include "Kernel.hpp"
#include "Options.hpp"
#include "Preprocess.hpp"
#include "Problem.hpp"
#include "Property.hpp"
#include "Refutation.hpp"
#include "Signature.hpp"
#include "Statistics.hpp"
#include "Tracer.hpp"
#include "VampireMode.hpp"
#include "XML.hpp"

using namespace Shell;

/**
 * Initialise the Vampire mode.
 * @since 12/09/2004 Manchester airport.
 */
VampireMode::VampireMode (Options& options,const VarList& answerVars)
  : Run(options,answerVars)
{
} // VampireMode::VampireMode

/**
 * Running in the standard Vampire mode.
 *
 * @since 26/09/2002 Manchester
 * @since 28/06/2003 Manchester, statistics result registration added
 * @since 27/08/2003 Vienna, changed to remove explicit handling of test numbers
 */
void VampireMode::run ()
{
  TRACER ("VampireMode::run");
  RunSaver saver(this);

  XML::DeepElement* proofAttempt = 0;

  if (_log) {
    proofAttempt = new XML::DeepElement("proof_attempt");
    _log->addChild(proofAttempt);
    proofAttempt->addChild(_options.toXML());
  }

  UnitList units;
  Environment::collectUnits(units);

  Problem problem(_options.problemName(),units);
  // preprocess the problem
  Property property(problem);
  property.scanProblem();
#if SAVE_SPIDER_PROPERTIES
  TheoryFinder tf(problem,&property);
  tf.search();
  cout << property.toSpider() << "\n";
  exit(0);
#endif

  Preprocess prepro(problem,_options,property);
  prepro.preprocess();

  // create a kernel
  Kernel kernel(_options,Signature::sig);
  int timeLimit = _options.timeLimitInDeciseconds() ?
                  _options.timeLimitInDeciseconds() :
                  INT_MAX;

  RunResult result = kernel.run(timeLimit,problem,_timer,proofAttempt);
  switch (result) {
  case REFUTATION:
    cout << "Refutation found! Thanks to Tanya!\n";
    cout << "%%%%% SZS status "
	 << (property.hasTheorem() ? "Theorem" : "Unsatisfiable")
	 << " %%%%%\n";
//     if (! _silent) {
//       cout << "Refutation found. Thanks to Tanya!\n";
//     }
    if (_options.proof() != Options::PROOF_OFF) {
      Refutation ref (kernel,problem,_options);
      if (_log) {
	proofAttempt->addChild(ref.toXML());
      }
      if (_options.outputMessages()) {
	ref.prettyPrint(cout,1);
      }
    }
    break;

  case SATISFIABLE:
//     if (! _silent) {
//       cout << "Satisfiability detected\n";
//     }
    break;

  case UNKNOWN:
//     if (! _silent) {
//       cout << "Refutation not found\n";
//     }
    break;
  }

   // output statistics
   Statistics statistics(kernel);
//    if (_log) {
//      XML::Element* xmlStat = statistics.toXML();
//      if (xmlStat) {
//       proofAttempt->addChild(xmlStat);
//     }
//   }

   string resultString = Run::toString(result);
//   if (_log) {
//     proofAttempt->addChild(new XML::TextElement("result",resultString));
//     _log->addChild(new XML::TextElement("result",resultString));
//   }

   statistics.setResult(resultString);
   if (! _silent) {
//      _options.output(cout);
     statistics.output(cout,_options);
   }

  BK::DestructionMode::makeFast();
  kernel.closeSession(); // not really necessary here
}  // VampireMode::run


/**
 * Report allocation failure.
 * @since 21/03/2004 Manchester
 */
void VampireMode::reportAllocationFailure ()
{
  if (! _silent) {
    cout << "Refutation not found\n";
  }
} // VampireMode::reportAllocationFailure
