(*----------------------------------------------------------------------(C)-*)
(* Copyright (C) 2006-2008 K. Korovin and The University of Manchester. 
   This file is part of iProver - a theorem prover for first-order logic.

   iProver is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   iProver is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
   See the GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with iProver.  If not, see <http://www.gnu.org/licenses/>.         *)
(*----------------------------------------------------------------------[C]-*)

open Lib

let start_time = Unix.gettimeofday ()    

(*---------Input Options---------------------*)

type input_options = 
    {
     mutable problem_file        : string;
     mutable iprover_path        : string;
     mutable iprover_options     : string; 
     mutable time_limit          : float;
   }

let current_options = 
  {
   problem_file        = "";
   iprover_path        = "";
   iprover_options     = "";   
   time_limit          = 0.
 }

let bool_str   = "<bool>"
let string_str = "<string>"
let int_str    = "<int>"
let float_str    = "<float>"
let inf_pref  = "\n    "
let example_str = inf_pref^"example: " 

(*---------*)
let args_error_msg opt_name str = 
  ("Input Options: "^opt_name^" unsupported value \'"^str^"\'")  
					
let default_arg_fun s = 
  current_options.problem_file <- s

(*---------*)

let opt_iprover_path_str  = "--iprover_path"

let opt_iprover_path_fun s =
  current_options.iprover_path <- s

let opt_iprover_path_inf = 
  string_str^
  inf_pref^"example: --iprover_path /home/iprover/\n"
(*"Ex. --iprover_command_str \"/home/iprover/iproveropt --eporver_path /home/iprover/ \"\n"*)


(*---------*)

let opt_iprover_options_str  = "--iprover_options"

let opt_iprover_options_fun s =
  current_options.iprover_options <- s

let opt_iprover_options_inf = 
  string_str^
  inf_pref^"example: --iprover_options \"--eporver_path /home/eprover/ \"\n"




(*---------*)
let opt_time_limit_str = "--time_limit"

let opt_time_limit_fun f = 
  current_options.time_limit <- f

let opt_time_limit_inf =
  float_str^
  inf_pref^"Time limit for all problems, should be specified and greater than 0.\n"



let spec_list =
  [
   (opt_iprover_path_str, Arg.String(opt_iprover_path_fun),opt_iprover_path_inf);
   (opt_iprover_options_str, Arg.String(opt_iprover_options_fun),opt_iprover_options_inf);
   (opt_time_limit_str, Arg.Float(opt_time_limit_fun),opt_time_limit_inf)
 ]

(*----------------------*)
let usage_msg = ("ltb_iprover_script [options] [problem_file_name]"^"\n"
		 ^"problem_file_name contains pairs of file names: input file name output file name\n")
 

let read_args() = 
  Arg.parse spec_list default_arg_fun usage_msg
    
let () = 
  read_args(); 
  if current_options.problem_file = "" 
  then
    (Arg.usage spec_list usage_msg;
     failwith "No problem file to solve")
  else 
    if not (current_options.time_limit > 0.) 
    then 
      (Arg.usage spec_list usage_msg;
       failwith "time limit should be specified and greater than 0.")
    else ()

(*-------Filling the problem list----------*)


type problem_status = 
    Not_tried| Started | Finished

let problem_status_to_string  = function
  |Not_tried -> "Not_tried"
  |Started   -> "Started"
  |Finished  -> "Finished"
  

type out_prover_status = 
    Satisfiable | CounterSatisfiable 
  | Theorem | Unsatisfiable 
  | Unknown

let out_prover_status_to_string = function 
  |Undef -> "Undef"
  |Def (s) -> 
      match s with
      | Satisfiable         -> "Satisfiable"
      | CounterSatisfiable  -> "CounterSatisfiable"
      | Theorem             -> "Theorem"
      | Unsatisfiable       -> "Unsatisfiable"
      | Unknown             -> "Unknown"


type problem = 
    {
     mutable in_file           : string;
     mutable out_file          : string;
     mutable problem_status    : problem_status;
     mutable out_prover_status : out_prover_status param;
   }

let problem_to_string p = 
  "in_file: "^p.in_file^"\n"
  ^"out_file: "^p.out_file^"\n"
  ^"problem_status: "^(problem_status_to_string p.problem_status)^"\n"
  ^"out_prover_status: "^(out_prover_status_to_string p.out_prover_status)^"\n"

let problem_list_to_string l =
  list_to_string (fun l -> ("\n-------------------\n"^
			    (problem_to_string l))) l "\n"  


let create_problem in_file out_file = 
  {in_file           = in_file; 
   out_file          = out_file;
   problem_status    = Not_tried;
   out_prover_status = Undef
 }


let input_problems_ref     = ref ([]: problem list)   

let finished_problems_ref  = ref ([]: problem list) 
let not_tried_problems_ref = ref ([]: problem list)
let num_of_not_tried_ref   = ref 0

let current_global_time_limit_ref = ref current_options.time_limit

let read_problems () = 
  let in_channel = 
    try
      open_in current_options.problem_file
    with 
      Sys_error x -> 
	(out_str ("\nCannot open file: "^(current_options.problem_file)^"\n");
	 raise (Sys_error x))
  in
  let problem_list_ref = ref [] in
  try
    while true  
    do
      let s = input_line in_channel in
      match (Str.split (Str.regexp "[ \t]+") s)
      with 
      | [s_in;s_out] ->
	  problem_list_ref:=((create_problem s_in s_out)::(!problem_list_ref))
      | [] -> ()
      | _ -> failwith "Input problem file should contain pairs of file names"
    done;
  with 
    End_of_file ->
      (close_in in_channel;
       input_problems_ref     := List.rev !problem_list_ref;
       not_tried_problems_ref := !input_problems_ref;
       num_of_not_tried_ref := List.length !not_tried_problems_ref)

let _ = read_problems ()

(*let _ = out_str (problem_list_to_string !input_problems_ref)*)
(*let _ = out_str (current_options.iprover_command_str^"\n")*)



(*------------From unix.ml--------------*)

(* Close file descriptor ignoring errors *)
let safe_close fd =
  try Unix.close fd with Unix.Unix_error(_,_,_) -> ()

(* Redirect stdin, stdout and stderr *)
let perform_redirections new_stdin new_stdout new_stderr =
  let newnewstdin = Unix.dup new_stdin in
  let newnewstdout = Unix.dup new_stdout in
  let newnewstderr = Unix.dup new_stderr in
  safe_close new_stdin;
  safe_close new_stdout;
  safe_close new_stderr;
  Unix.dup2 newnewstdin Unix.stdin; Unix.close newnewstdin;
  Unix.dup2 newnewstdout Unix.stdout; Unix.close newnewstdout;
  Unix.dup2 newnewstderr Unix.stderr; Unix.close newnewstderr

(* Create process in a new process group *)
let create_process_newsid cmd args new_stdin new_stdout new_stderr =
  match Unix.fork() with
    0 ->
      begin try
        perform_redirections new_stdin new_stdout new_stderr;
	ignore (Unix.setsid ());
        Unix.execvp cmd args
      with _ ->
        exit 127
      end
  | id -> id


(*--------------------Signals Timers---------------*)

exception Time_out_real
exception Time_out_virtual

let set_sys_signals () = 
  let signal_handler signal =
    if (signal = Sys.sigint || signal = Sys.sigterm || signal = Sys.sigquit) 
    then raise Termination_Signal
    else 
      if (signal = Sys.sigalrm) 
      then raise Time_out_real
      else 
	if (signal = Sys.sigvtalrm)
	then raise Time_out_virtual
	else failwith "Unknown Signal"
  in  	  
  Sys.set_signal Sys.sigint     (Sys.Signal_handle signal_handler);
  Sys.set_signal Sys.sigterm    (Sys.Signal_handle signal_handler);
  Sys.set_signal Sys.sigquit    (Sys.Signal_handle signal_handler);
  Sys.set_signal Sys.sigalrm    (Sys.Signal_handle signal_handler);
  Sys.set_signal Sys.sigvtalrm  (Sys.Signal_handle signal_handler)


  
let _ = set_sys_signals ()

let set_timer time_limit = 
  (if time_limit > 0. 
  then 
    let _= 
	Unix.setitimer Unix.ITIMER_REAL
	{
   	 Unix.it_interval = 0.;
   	 Unix.it_value  = time_limit;
       }  in ()
  ) 
 

(*-----------Time limits--------------*)

let current_time () = Unix.gettimeofday ()

let used_time () = (current_time ()) -. start_time
    
let time_left () = current_options.time_limit -. (used_time ())

exception Run_out_of_time

let get_problem_time_limit () =
  if  !num_of_not_tried_ref = 0 
  then failwith "!num_of_not_tried_ref should not be 0 here"
  else
   let limit =  (time_left ())/.(float_of_int !num_of_not_tried_ref) in
   if limit <= 0.5 
   then raise Run_out_of_time
   else limit

  

(*----------Global Open Processes--------------*)

let open_processes_list_ref = ref []

let add_process pid = 
  open_processes_list_ref:= pid::!open_processes_list_ref

let remove_top_process () = 
  match !open_processes_list_ref with 
  |[] -> ()
  |h::tl ->
      open_processes_list_ref:= tl

let kill_process_group pid = 
  try 			      
    (* Kill processes in process group *)
    Unix.kill (-pid) Sys.sigkill;			      
  with 
    Unix.Unix_error(Unix.ESRCH, _, _) -> ()

let kill_all_processes () = 
  List.iter kill_process_group !open_processes_list_ref


(*------Problem run with time limit------------*)


let iprover_cmd_str () = 
  current_options.iprover_path^"iproveropt"

   
let run_problem_time_limit problem time_limit =
  out_str ("\n--------------------------------------\n");
  out_str ("Run problem: "^(problem.in_file)^"\n");
  out_str ("Time limit: "^(string_of_float time_limit)^"\n");
  let problem_start_time = current_time () in
  let out_channel = 
    try
      open_out (problem.out_file)
    with 
      Sys_error x -> 
	(out_str ("\nCannot open file: "^(current_options.problem_file)^"\n");
	 raise (Sys_error x))
  in
(*!! first arg should be program name !!*)
  let args_list = 
    (iprover_cmd_str ())::
    ((Str.split (Str.regexp " ") current_options.iprover_options)@
     [problem.in_file]) in
  set_timer time_limit; 
  let process_id = 
    create_process_newsid 
      (iprover_cmd_str ()) 
      (Array.of_list args_list)
      Unix.stdin 
      (Unix.descr_of_out_channel out_channel)
      Unix.stdout
  in
  add_process process_id;
  let finished problem = 
    (problem.problem_status <- Finished;
     close_out out_channel;
      remove_top_process ();
     out_str ("Time used: "
	      ^(string_of_float (current_time () -.problem_start_time ))^"s\n")
    )
  in
  try 
    (let _process_status =  Unix.waitpid [] process_id in
    finished problem
    )
  with 
    Time_out_real | Time_out_virtual
    ->
      (kill_process_group process_id;
       (try 
	 let _p_status = Unix.waitpid [] process_id in ()
       with  
	 Unix.Unix_error(Unix.ESRCH, _, _) -> ()
       );       
       finished problem
      )




let main () =
  try 
    let rec f () = 
      match !not_tried_problems_ref with 
      |[] -> 
	  out_str ("All problems finished, time used: "
		   ^(string_of_int (int_of_float (current_time() -. start_time)))^"s\n")
	    
      |problem::tl -> 	
	  (run_problem_time_limit problem (get_problem_time_limit ()); 
	 not_tried_problems_ref := tl;
	   finished_problems_ref:= problem::(!finished_problems_ref);
	   num_of_not_tried_ref:=!num_of_not_tried_ref-1;
	   f ())
	    
    in
    try
      f()
    with Run_out_of_time -> 
      out_str ("Run out of time, num of unfinished problems:  "
	       ^(string_of_int !num_of_not_tried_ref)^"\n")
  with 
    Termination_Signal -> 
      (out_str ("Termination Signal\n"
		^"Killing all child process...");
       kill_all_processes ();
       out_str ("Done\n");	 )
let _ = main ()
    
