(*----------------------------------------------------------------------(C)-*)
(* Copyright (C) 2006-2008 K. Korovin and The University of Manchetser. 
   This file is part of iProver - a theorem prover for first-order logic.

   iProver is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   iProver is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
   See the GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with iProver.  If not, see <http://www.gnu.org/licenses/>.         *)
(*----------------------------------------------------------------------[C]-*)

open Lib

type symb   = Symbol.symbol
type term   = Term.term 
type clause = Clause.clause

let symbol_db_ref  = Parsed_input_to_db.symbol_db_ref
let term_db_ref    = Parsed_input_to_db.term_db_ref

(* one should run flat_signature (once) before flattening *)
let flat_signature () = 
  let f symb = 
    if ((Symbol.get_type symb) = Symbol.Fun ) &&
      (Symbol.get_num_input_occur symb) > 0
    then 
      let new_symb_name = ("iProver_Flat_"^(Symbol.get_name symb)) in
      let flat_symb = 
	Symbol.create_from_str_arity 
	  new_symb_name Symbol.Pred ((Symbol.get_arity symb) +1) in
      Symbol.assign_property Symbol.Flat flat_symb;
      let add_flat_symb = SymbolDB.add_ref flat_symb symbol_db_ref in
      Symbol.assign_flattening symb add_flat_symb
    else ()
  in 
  SymbolDB.iter f !symbol_db_ref



module TermHashKey = 
  struct
    type t    = term
    let equal = (==)
    let hash  = Term.get_fast_key
  end 

module TermHash = Hashtbl.Make(TermHashKey)


(* var_env is a hash table  mapping non-var term into vars *)
(* in order to identify vars for the equal terms*)
(* var_env is also used to create flat terms later     *)
(* max_var is max var used                             *)
(* the input term itself also added to the the var_env *)

let rec term_flattening_env var_env max_var_ref term = 
  match term with 
  | Term.Var _ -> ()
  | Term.Fun (symb, args, _) -> 
      if (TermHash.mem var_env term) 
      then ()
      else 
	(
	 Term.arg_iter (term_flattening_env var_env max_var_ref) args;
	 if (Symbol.get_type symb) = Symbol.Fun 
	 then
	   (
	    max_var_ref:= Var.get_next_var !max_var_ref;
	   let max_var_term = 
	     TermDB.add_ref (Term.create_var_term !max_var_ref) term_db_ref in
	   TermHash.add var_env term max_var_term
	   )
	 else ()
	)


let flat_term_to_var var_env max_var_ref t = 
  if (Term.is_var t) 
  then t
  else 
    (
     try 
       term_flattening_env var_env max_var_ref t;
       TermHash.find var_env t
     with Not_found -> failwith "flat_term_to_var: Not_found"
    )
	

let order_term_var tv1 tv2 = 
  if (Var.compare (Term.get_var tv1) (Term.get_var tv2)) > 0 
  then (tv1,tv2) 
  else (tv2,tv1)


(* we obtain flat def. of terms in var_env and   *)
(* a normilised subst. corresponding to x \not = y *)
(* subst is confluent *)
let flat_lit_env var_env max_var_ref neg_var_subst_ref lit = 
  if (Term.is_neg_lit lit)
  then 
    begin
      let atom = Term.get_atom lit in 
      match atom with 
      | Term.Fun (symb, args, _) -> 
	  if (symb == Symbol.symb_equality) 
	  then 
	  (*flat neg eq*)  
	    let [t1;t2] = Term.arg_to_list args in
(*	    let rec fl t1 t2 = *)
	    if t1==t2 then ()
	    else	      
	      let var_t1 = flat_term_to_var var_env  max_var_ref t1 in
	      let var_t2 = flat_term_to_var var_env  max_var_ref t2 in		   
	      let norm_t1 = 
		Subst.find_normalised !neg_var_subst_ref var_t1 in 
	      let norm_t2 = 
		Subst.find_normalised !neg_var_subst_ref var_t2 in
	      if  norm_t1==norm_t2 
	      then ()
	      else 
		begin
		  let (big_t, small_t) = order_term_var norm_t1 norm_t2 in
		  neg_var_subst_ref:= 
		    Subst.add (Term.get_var big_t) small_t !neg_var_subst_ref
		end		    
		  (* atom is not equality *)
	  else
	    term_flattening_env var_env max_var_ref atom
      | Term.Var _ -> failwith "flat_lit_env: atom cannot be a var"
    end
(* positive lit*)
  else 
    term_flattening_env var_env max_var_ref lit
	  

let rec get_max_var_term current_max_var_ref term =  
  match term with 
  |Term.Fun (_, args,_) ->  
      Term.arg_iter (get_max_var_term current_max_var_ref) args
  |Term.Var (v,_) -> 
      if (Var.compare v !current_max_var_ref) > 0
      then 
	(current_max_var_ref := v)
      else ()
	  
let get_max_var clause = 
  let var_ref  = ref (Var.get_first_var ()) in
  Clause.iter (get_max_var_term var_ref) clause;
  !var_ref

let add_var_term var = TermDB.add_ref (Term.create_var_term var) term_db_ref
    
let add_fun_term symb args = 
  TermDB.add_ref (Term.create_fun_term symb args) term_db_ref

let equality_term t s = 
  let args = [t;s] in
  add_fun_term Symbol.symb_equality args 

let dis_equality t s = 
  add_fun_term Symbol.symb_neg [(equality_term t s)]

let flat_clause clause = 
  let var_env = TermHash.create 19 in
  let max_var_ref = ref (get_max_var clause) in
  let neg_var_subst_ref = ref (Subst.create ()) in
  Clause.iter (flat_lit_env var_env max_var_ref neg_var_subst_ref) clause;
(* create var term corresponding to the term *)
  let term_to_var_term term =
    try  
      (Subst.find_normalised  !neg_var_subst_ref 
	 (TermHash.find var_env term))
    with Not_found ->
      if (Term.is_var term)
      then Subst.find_normalised  !neg_var_subst_ref term 
      else
	failwith ("term_to_var_term: Not_found term: "
				^(Term.to_string term))
  in
  let flat_lit rest lit = 
    let atom = Term.get_atom lit in 
    match atom with 
    | Term.Fun (symb, args, _) -> 
	if (symb == Symbol.symb_equality) 
	then
	  if (Term.is_neg_lit lit) 
	  then
(* all neg eq are falttend to x\not y which are added to neg_var_subst_ref *)
(* and will be added to the rest later *)
	    rest 
	  else
	    (*positive eq, terms replaced by definitions *)
	    let [t1;t2] = Term.arg_to_list args in
	    (* replace *)
	    (equality_term  (term_to_var_term t1) (term_to_var_term t2))::rest
	else 
(* non equlaity literal *)
	  let new_atom = 
	    let new_args = Term.arg_to_list (Term.arg_map term_to_var_term args) in
	    add_fun_term symb new_args 
	  in   
	  if (Term.is_neg_lit lit) 
	  then 
	    let new_lit = add_fun_term Symbol.symb_neg [new_atom] in
	    new_lit::rest
	  else
	    new_atom::rest
    | Term.Var _ -> failwith "flat_lit: atom cannot be var"
  in
  let flat_part =  Clause.fold flat_lit [] clause in
  let get_env_part term var_term rest = 
    match term with 
    | Term.Fun (symb, args, _) -> 
	let new_atom = 
	  let new_symb = Symbol.get_flattening symb in
	  let new_var_term = 	    
	    (Subst.find_normalised  !neg_var_subst_ref var_term)
	  in	
(* the value of the function is the first argument of the relation*)	
	  let new_args = new_var_term::
	    (Term.arg_to_list (Term.arg_map term_to_var_term args)) 
	  in
	  add_fun_term new_symb new_args 
	in 
	let new_lit = add_fun_term Symbol.symb_neg [new_atom] in
	new_lit::rest
    | Term.Var _ -> failwith "get_env_part should not be var term"
  in
  let env_part = TermHash.fold get_env_part var_env [] in
   Clause.normalise term_db_ref (Clause.create (env_part@flat_part))


let flat_clause_list clause_list = 
  List.map flat_clause clause_list 


(*---------------Axioms-------------------------*)

let add_domain_constant i = 
  let dom_symb_name = ("iProver_Domain_"^(string_of_int i)) in
  let dom_symb = 
    Symbol.create_from_str_arity 
      dom_symb_name Symbol.Fun 0 in
  Symbol.assign_property Symbol.DomainConstant dom_symb;
  let add_dom_symb = SymbolDB.add_ref dom_symb symbol_db_ref in
  add_fun_term add_dom_symb []
 

let add_domain_pred i = 
  let dom_symb_name = ("iProver_Domain_Pred_"^(string_of_int i)) in
  let dom_symb = 
    Symbol.create_from_str_arity 
      dom_symb_name Symbol.Pred 0 in
  Symbol.assign_property Symbol.DomainPred dom_symb;
  let add_dom_symb = SymbolDB.add_ref dom_symb symbol_db_ref in
  add_fun_term add_dom_symb []
    

let add_domain_constants first last = 
  let dom_const_list_ref = ref [] in
  for i=first to last 
  do
    let add_dom_const = add_domain_constant i in
    dom_const_list_ref:=add_dom_const::!dom_const_list_ref
  done;
  !dom_const_list_ref
  

let dis_eq_axioms t term_list = 
  let f rest_f s = 
    if not (s==t) 
    then 	
      (Clause.create [(dis_equality t s)])::rest_f
    else 
      rest_f
  in 
  List.fold_left f [] term_list 

(* for al pairs t s in term_list adding t!=s *)
(* we need to add both t!=s and s!=t since we don not add symmetry axioms*)
let dis_eq_axioms_list term_list = 
  let f rest t =     
    (dis_eq_axioms t term_list)@rest
  in
  List.fold_left f [] term_list 




let domain_axioms dom_pred dom_terms = 
  let dom_clause symb rest_clauses =     
    if (Symbol.is_flat symb)
    then
      let rec get_var_args rest current_var i = 
	if i = 0 then List.rev rest
	else 
	  get_var_args 
	    ((add_var_term current_var)::rest) (Var.get_next_var current_var) (i-1)
      in     
      let var_args = 
	get_var_args [] (Var.get_first_var ()) ((Symbol.get_arity symb)-1) 
      in 
      let f rest dom_term = 
	(add_fun_term symb (dom_term::var_args))::rest
      in
      ( Clause.normalise term_db_ref  (Clause.create (dom_pred::(List.fold_left f [] dom_terms))))::rest_clauses
    else 
      rest_clauses
  in
  SymbolDB.fold dom_clause !symbol_db_ref [] 

(* optimized version with restricted domain for first constants *)
let domain_axioms_opt dom_pred dom_terms = 
  let split_constants_other s (const_rest,flat_pred_rest)  = 
    if (Symbol.is_constant s) && (Symbol.get_num_input_occur s) >0 
    then (
(*      out_str ((Symbol.to_string s)
	       ^" num_occ: "^(string_of_int (Symbol.get_num_input_occur s))^"\n"); *)
      ((s::const_rest),flat_pred_rest))
    else 
      if (Symbol.is_flat s) && (Symbol.get_arity s) >1 
      then 	
	(const_rest,(s::flat_pred_rest))	
      else
	(const_rest,flat_pred_rest)
  in
  let (const_list,other_flat_pred) = 
    SymbolDB.fold split_constants_other !symbol_db_ref ([],[]) in
  let cmp c1 c2 = 
    -(compare (Symbol.get_num_input_occur c1) (Symbol.get_num_input_occur c2)) in 
  let ordered_const_list = List.sort cmp const_list in
  let oredered_flat_const_list = List.map Symbol.get_flattening ordered_const_list
  in
  let rec i_const_lits rest_dom_terms i symb = 
    match rest_dom_terms with 
    | [] -> []
    | h::tl ->
	if (i =0) then []
	else 
	  ((add_fun_term symb [h])::(i_const_lits tl (i-1) symb))
  in  
  let get_axioms_const (rest,i) symb = 
    let new_clause = 
      Clause.normalise term_db_ref (Clause.create (dom_pred::(i_const_lits dom_terms i symb))) in
    ((new_clause::rest),i+1)
  in
  let (axioms_const,_) = 
    List.fold_left get_axioms_const ([],1) oredered_flat_const_list in
(* for the rest *)
  let dom_clause  rest_clauses symb =     
      let rec get_var_args rest current_var i = 
	if i = 0 then List.rev rest
	else 
	  get_var_args 
	    ((add_var_term current_var)::rest) (Var.get_next_var current_var) (i-1)
      in     
      let var_args = 
	get_var_args [] (Var.get_first_var ()) ((Symbol.get_arity symb)-1) 
      in 
      let f rest dom_term = 
	(add_fun_term symb (dom_term::var_args))::rest
      in
       (Clause.normalise term_db_ref (Clause.create (dom_pred::(List.fold_left f [] dom_terms))))::rest_clauses
  in
  let axioms_flat_rest =  List.fold_left dom_clause [] other_flat_pred in
  axioms_const@axioms_flat_rest
  
