%% File: leancop20_rando.pl  -  Version: 1.1  -  Date: 20 June 2008
%%
%%         "Make everything as simple as possible, but not simpler."
%%                                                 [Albert Einstein]
%%
%% Purpose: leanCoP: A Lean Connection Prover for Classical Logic
%%
%% Authors: Jens Otten, Thomas Raths (extensions for randoCoP)
%% Web:     www.leancop.de
%%
%% Usage: prove(M,P).    % where M is a set of clauses and P is
%%                       %  the returned connection proof
%%                       %  e.g. M=[[q(a)],[-p],[p, -q(X)]]
%%                       %  and  P=[[q(a)],[[-(q(a)),p],[[-(p)]]]]
%%        prove(F,P).    % where F is a first-order formula and
%%                       %  P is the returned connection proof
%%                       %  e.g. F=((p,all X:(p=>q(X)))=>all Y:q(Y))
%%                       %  and  P=[[q(a)],[[-(q(a)),p],[[-(p)]]]]
%%        prove2(F,S,P). % where F is a formula, S is a subset of
%%                       %  [nodef,def,conj,reo(I),scut,cut,comp(J)]
%%                       %  (with numbers I,J) defining attributes
%%                       %  and P is the returned connection proof
%%
%% Copyright: (c) 1999-2008 by Jens Otten, Thomas Raths
%% License:   GNU General Public License


:- [def_mm].    % load program for clausal form translation
:- lib(iso).    % load library for ISO compatibility
:- lib(random). % load random library (for randoCoP)
:- set_flag(occur_check,on). % global occur check on
:- dynamic(pathlim/0), dynamic(lit/3).


%%% prove matrix M / formula F

prove(F,Proof) :- prove2(F,[cut,comp(7)],Proof).

prove2(F,Set,Proof) :-
    (F=[_|_] -> M=F ; make_reord_matrix(F,M,Set)),
    retract_all(lit(_,_,_)), (member([-(#)],M) -> S=conj ; S=pos),
    assert_clauses(M,S), prove(1,Set,Proof).

prove(PathLim,Set,Proof) :-
    \+member(scut,Set) -> prove([-(#)],[],PathLim,[],Set,[Proof]) ;
    lit(#,C,_) -> prove(C,[-(#)],PathLim,[],Set,Proof1),
    Proof=[C|Proof1].
prove(PathLim,Set,Proof) :-
    member(comp(Limit),Set), PathLim=Limit -> prove(1,[],Proof) ;
    (member(comp(_),Set);retract(pathlim)) ->
    PathLim1 is PathLim+1, prove(PathLim1,Set,Proof).

%%% leanCoP core prover

prove([],_,_,_,_,[]).

prove([Lit|Cla],Path,PathLim,Lem,Set,Proof) :-
    Proof=[[[NegLit|Cla1]|Proof1]|Proof2],
    \+ (member(LitC,[Lit|Cla]), member(LitP,Path), LitC==LitP),
    (-NegLit=Lit;-Lit=NegLit) ->
       ( member(LitL,Lem), Lit==LitL, Cla1=[], Proof1=[]
         ;
         member(NegL,Path), unify_with_occurs_check(NegL,NegLit),
         Cla1=[], Proof1=[]
         ;
         lit(NegLit,Cla1,Grnd1),
         ( Grnd1=g -> true ; length(Path,K), K<PathLim -> true ;
           \+ pathlim -> assert(pathlim), fail ),
         prove(Cla1,[Lit|Path],PathLim,Lem,Set,Proof1)
       ),
       ( member(cut,Set) -> ! ; true ),
       prove(Cla,Path,PathLim,[Lit|Lem],Set,Proof2).

%%% write clauses into Prolog's database

assert_clauses([],_).
assert_clauses([C|M],Set) :-
    (Set\=conj, \+member(-_,C) -> C1=[#|C] ; C1=C),
    (ground(C) -> G=g ; G=n), assert_clauses2(C1,[],G),
    assert_clauses(M,Set).

assert_clauses2([],_,_).
assert_clauses2([L|C],C1,G) :-
    append(C1,C,C2), assert(lit(L,C2,G)), append(C1,[L],C3),
    assert_clauses2(C,C3,G).


%%% -----------------------------------------
%%% extensions for randoCoP (by Thomas Raths)


%%% make_reord_matrix(+Formula,?Matrix,Settings)
%%% reorders axioms and literals in formula

make_reord_matrix(F,M,Set) :-
    ( member(reord(I),Set), F=(A=>C), I>0 ->
      R is I*1000, randomise(R), term2list(A,Ax),
      rand_perm(Ax,Ax1), list2term(Ax1,A1), F1=(A1=>C) ; F1=F ),
    make_matrix(F1,M1,Set),
    ( member(reord(I),Set), I>0 ->
      R is I*1000, randomise(R), reordering_lits(M1,M) ; M=M1 ).

%% reordering_lits(+Lists,?ReorderedLists)
%% reorders literals in lists

reordering_lits([],[]) :- !.
reordering_lits([A|R],[B|S]) :-
    rand_perm(A,B), reordering_lits(R,S).

%% term2list(+Term,?List)
%% transforms axioms into list

term2list(T,[H|R]) :-
    T=..[','|Axioms], !, Axioms=[H|[S]], term2list(S,R).
term2list(T,[T]).

%% list2term(+List,?Term)
%% transforms list into axioms

list2term([H],H).
list2term([H,H1|R],T) :- list2term([H1|R],T1), T=(H,T1).
