%% File: randocop_sched.pl  -  Version: 1.1  -  Date: 15 July 2008
%%
%% Purpose: Fixed strategy scheduling for randoCoP
%%
%% Authors: Jens Otten, Thomas Raths
%% Web:     www.leancop.de
%%
%% Usage:   randocop_sched(F,T,R). % proves formula F with the time
%%                                 % limit T and returns result R
%%
%% Copyright: (c) 2007-2008 by Jens Otten, Thomas Raths
%% License:   GNU General Public License


:- nodbgcomp.         % switch off debugging mode
:- [leancop20_rando]. % load randoCoP core prover
:- [leancop_proof].   % load program for proof presentation
:- [leancop_tptp2].   % load program for TPTP input syntax
:- local(include/1), dynamic(include/1).
:- dynamic(cputime_limit/2).


%%% randoCoP strategy scheduling

randocop_sched(ProblemFile,TimeLimit,Result) :-
    TotalTime is 1.10*TimeLimit, cputime_init,
    cputime(0,TotalTime,fail, (
     [ProblemFile], include_files,
     ( fof(_,_,_), \+fof(_,conjecture,_) -> Conj=n ; Conj=y ),
     ( leancop_tptp2(F) -> Problem=F ; f(Problem) ),
     ( Conj=y -> Problem1=Problem ; Problem1=(~Problem) ),
     leancop_equal(Problem1,Problem2),
     strategy(Settings,Time,Complete,TimeLimit),
     cputime(1,TotalTime,fail,
      make_reord_matrix(Problem2,Matrix,Settings)
     ),
     ( cputime(2,Time,TimeOut=true,
        prove2(Matrix,Settings,Proof)
       ) ->
       var(TimeOut),
       ( Conj=y -> Result='Theorem' ; Result='Unsatisfiable' ) ;
       Complete=comp,
       ( Conj=y -> Result='Non-Theorem' ; Result='Satisfiable' )
     ),
     output_result(ProblemFile,Matrix,Proof,Result,Conj)
    ) ).

% define fixed scheduling strategy

strategy([cut,comp(7)],5,comp,_).
strategy(Settings,Time,Complete,TimeLimit) :-
    Imax is (TimeLimit-10)//5,
    strategy_loop(Settings,Time,Complete,0,Imax).
strategy([def],TimeLimit,comp,TimeLimit).

strategy_loop(Set,Time,Comp,I,Imax) :-
    ( Set=[cut,scut,reord(I)],     Time=3, Comp=incomp ) ;
    ( Set=[def,conj,cut,reord(I)], Time=2, Comp=incomp ) ;
    ( I1 is I+1, I1<Imax, strategy_loop(Set,Time,Comp,I1,Imax) ).

% print status and connection proof

output_result(ProblemFile,Matrix,Proof,Result,Conj) :-
    ( Conj=y -> Art=' is a ' ; Art=' is ' ),
    write(ProblemFile), write(Art), write(Result), nl, flush(stdout),
    ( Result='Theorem' -> Out=' for ' ;
      Result='Unsatisfiable' -> Out=' for negated ' ; true ),
    ( var(Out) -> true ; write('Start of proof'), write(Out),
      write(ProblemFile), nl, leancop_proof(Matrix,Proof),
      write('End of proof'), write(Out), write(ProblemFile), nl ).

% load include files

include_files :-
    getenv('TPTP',PathString), atom_string(Path,PathString),
    retract(include(File)), concat_atom([Path,'/',File],Filename),
    [Filename], fail.
include_files.

% call a goal with a CPU time limit

cputime(IntTag,TimeLimit,TOutGoal,Goal) :-
    cputime(Time), Time1 is Time+TimeLimit,
    assert(cputime_limit(IntTag,Time1)),
    block(Goal,Tag,true) ->  retract(cputime_limit(IntTag,_)),
    ( var(Tag) -> true ; (atom(Tag);Tag=IntTag) -> call(TOutGoal) ;
      exit_block(Tag) ) ; retract(cputime_limit(IntTag,_)), fail.

cputime_init :-
    set_event_handler(cputime_up,cputime_up/0),
    event_after_every(cputime_up,1).

cputime_up :-
    cputime_limit(Tag,TimeOut), cputime(Time),
    Time>TimeOut -> exit_block(Tag), fail.
cputime_up.
