/*
 Copyright 2006, 2007, 2008 Hao Xu
 ertranne@hotmail.com
 
 This file is part of OSHL-S.

 OSHL-S is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 OSHL-S is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Foobar; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package chercher;

import java.util.Arrays;
import java.util.TreeSet;

/**
 *
 * @author ertri
 */
public class IntegerTree {
    // input
    int[][] coefficients;
    int[] bound;
    int[] varBound;
    boolean[] allzero;
    boolean[] tightable;
    private boolean noSolution;

    // output
    public static class IntTreeNode {

        /** for regular node the value,
         * for head node the current value of the tail node
         */
        public int value;
        /** depth
         * 
         */
        int i;
        /** for regular node, tokens = backlink.tokens - values * coefficients[i];
         * for head node, tokens = backlink.tokens
         * 
         */
        int[] tokens;
        /** the head node which serves as state for computation in this level
         * 
         */
        IntTreeNode head;
        /** link to the previous level
         * 
         */
        public IntTreeNode backlink;

        public IntTreeNode getBacklink() {
            return backlink;
        }
        /** cache for the next node
         * 
         */
        IntTreeNode next = null;
        /** previous node
         * 
         */
        IntTreeNode prev = null;
        /** cache for the first child node
         * 
         */
        IntTreeNode subtrees = null;

        /** regular node
         * 
         * @param value
         * @param tokens
         * @param prev
         */
        public IntTreeNode(int value, int[] tokens, IntTreeNode prev) {
            this.value = value;
            this.i = prev.head.i;
            this.tokens = tokens;
            this.head = prev.head;
            this.prev = prev;
            this.backlink = prev.head.backlink;
        }

        /** root node
         * 
         * @param bound
         */
        public IntTreeNode(int[] bound) {
            this.i = -1;
            this.tokens = bound;
            this.value = -1;
            this.backlink = null;
            this.head = this;
        }

        /** head node
         * 
         * @param backlink
         */
        public IntTreeNode(IntTreeNode backlink) {
            this.backlink = backlink;
            this.head = this;
            this.tokens = backlink.tokens;
            this.i = backlink.i + 1;
            this.value = -1;
        }
    }
    IntTreeNode root;

    public IntegerTree(int[][] coefficients, int[] bound, boolean[] tightable, int[] varBound) {
        this.coefficients = coefficients;
        this.bound = bound;
        this.tightable = tightable;
        this.varBound = varBound;
        this.allzero = new boolean[coefficients[0].length];
        this.root = new IntTreeNode(bound);
        for (int i = 0, m = coefficients[0].length; i < m; i++) {
            allzero[i] = true;
            for (int k = 0, n = bound.length; k < n; k++) {
                if (coefficients[k][i] != 0) {
                    allzero[i] = false;
                    break;
                }
            }
        }
        for (int i=0, n = bound.length;i<n;i++) {
            if(bound[i]<0) {
                noSolution = true;
            }
        }
        testOverflow();
    }

    public IntTreeNode getRoot() {
        return root;
    }

    /**
     * 
     * @param curr
     * @return null this is a leaf node
     *         otherwise the head of list of child nodes
     */
    public IntTreeNode getChildNode(
            IntTreeNode curr) {
        if (curr.subtrees == null) {
            if (curr.i == coefficients[0].length - 1) {
                return null;
            } else {
                curr.subtrees = new IntTreeNode(curr);
                return curr.subtrees;
            }
        } else {
            return curr.subtrees;
        }
    }

    /**
     * 
     * @param curr
     * @return null no node in current level
     *         otherwise the next node in the current level
     */
    public IntTreeNode getNextNode(IntTreeNode curr) {
        int[] coeff;
        int[] ntokens;
        int value;
        if (curr == root) {
            return null;
        } else if (curr.next != null) {
            return curr.next;
        } else if (curr.i == coefficients[0].length - 1) {
            // use head.value as the count for the equation that should be tight
            IntTreeNode head = curr.head;
            if (head.value == Integer.MAX_VALUE) {
                return null;
            } else {
                ntokens = new int[curr.head.tokens.length];
                value = -1;
                TreeSet<Integer> values = new TreeSet<Integer>();
                final int i = head.i; // variable index
                outer:
                while (true) {
                    head.value++;
                    if (head.value >= coefficients.length) {
                        break;
                    } else if (!tightable[head.value]) {
                        continue;
                    }
                    coeff = coefficients[head.value];
                    if (coeff[i] != 0) {
                        if (head.tokens[head.value] % coeff[i] == 0) {
                            value = head.tokens[head.value] / coeff[i];
                            if (value <= varBound[i]) {
                                for (int k = 0, n = head.tokens.length; k < n; k++) {
                                    ntokens[k] = head.tokens[k] - value * coefficients[k][i];
                                    if (ntokens[k] < 0) {
                                        continue outer;
                                    }
                                }
                                values.add(value);
                            }
                        }
                    } else {
                        if (head.tokens[head.value] == 0) {
                            // If the coefficient for the trailing variable is zero and 
                            // the equation is tight already, then values starting from
                            // 0 to the first that falsifies some inequality or exceeds
                            // the variable bound should be tried and
                            // only tried once. After this, no need to try other inequalities
                            // because this loop covers all possible values of the trailing
                            // variable that may satisfy all inequalities.
                            if (allzero[i]) {
                                values.add(-1);
                            } else {
                                value = 0;
                                while (true) {
                                    if (value > varBound[i]) {
                                        break outer;
                                    }
                                    for (int k = 0, n = head.tokens.length; k < n; k++) {
                                        ntokens[k] = head.tokens[k] - value * coefficients[k][i];
                                        if (ntokens[k] < 0) {
                                            break outer;
                                        }
                                    }
                                    values.add(value++);
                                }
                            }
                        }
                    }
                }
                head.value = Integer.MAX_VALUE;
                IntTreeNode node = head;
                for (Integer v : values) {
                    if (v == -1) {
                        head.next = new IntTreeNode(v, head.tokens, head);
                        break;
                    }
                    ntokens = new int[head.tokens.length];
                    for (int k = 0, n = head.tokens.length; k < n; k++) {
                        ntokens[k] = head.tokens[k] - v * coefficients[k][i];
                    }
                    node.next = new IntTreeNode(v, ntokens, node);
                    node = node.next;
                }
            }
            return head.next;
        } else {
            IntTreeNode head = curr.head;
            if (head.value == Integer.MAX_VALUE) {
                return null;
            }
            ntokens = new int[curr.head.tokens.length];
            if (allzero[head.i]) {
                value = -1;
                head.value = Integer.MAX_VALUE;

            } else {
                head.value++;
                value = head.value;
                if (value > varBound[head.i]) {
                    head.value = Integer.MAX_VALUE;
                    return null;
                } else {
                    for (int j = 0, n = coefficients.length; j < n; j++) {
                        coeff = coefficients[j];
                        if (coeff[head.i] != 0) {
                            if (value <= head.tokens[j] / coeff[head.i]) {
                                ntokens[j] = head.tokens[j] - value * coeff[head.i];
                            } else {
                                head.value = Integer.MAX_VALUE;
                                return null;
                            }
                        } else {
                            ntokens[j] = head.tokens[j];
                        }
                    }
                }
            }
            IntTreeNode nnode = new IntTreeNode(value, ntokens, curr);
            curr.next = nnode;
            return nnode;
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final IntegerTree other = (IntegerTree) obj;
        if (!Arrays.deepEquals(coefficients, other.coefficients)) {
            return false;
        }
        if (!Arrays.equals(this.bound,other.bound)) {
            return false;
        }
        if (!Arrays.equals(this.varBound,other.varBound)) {
            return false;
        }
        if (!Arrays.equals(this.tightable, other.tightable)) {
            return false;
        }
        return true;
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 41 * hash + (this.coefficients != null ? Arrays.deepHashCode(coefficients) : 0);
        hash = 41 * hash + (this.bound != null ? Arrays.hashCode(bound) : 0);
        hash = 41 * hash + (this.varBound != null ? Arrays.hashCode(varBound) : 0);
        hash = 41 * hash + (this.tightable != null ? Arrays.hashCode(tightable) : 0);
        return hash;
    }

    /** prune the current node and return the previous node
     * 
     * @param curr the node to be pruned
     * @return null whole tree pruned
     *         otherwise the previous node
     */
    public IntTreeNode prune(IntTreeNode curr) {
        if(curr == root) {
            return root;
        }
        IntTreeNode prev = curr.prev;
        prev.next = curr.next; // prev is never null because the existence of head node
        if (curr.next != null) {
            curr.next.prev = prev;
        }
        if (curr == null) {
            return null;
        } else if (prev == prev.head &&
                prev.value == Integer.MAX_VALUE) {
            return prune(prev.backlink);
        } else {
            return prev;
        }
    }

    public IntTreeNode nextSolution(IntTreeNode curr) {
        if(noSolution) {
            return null;
        }
        IntTreeNode node;
        boolean down;
        boolean up;
        if (curr == null) {
            node = root;
            down = true;
            up = false;
        } else {
            node = curr;
            down = false;
            up = false;
        }
        while (true) {
            if (down) {
                //System.out.println("down");
                IntTreeNode child = getChildNode(node);
                if (child != null) {
                    down = up = false;
                    node = child;
                } else {
                    return node;
                }
            } else if (up) {
                //System.out.println("up");
                if (node == root) {
                    return null;
                } else {
                    node = node.backlink;
                    down = up = false;
                }
            } else {
                //System.out.println("next");
                IntTreeNode next = getNextNode(node);
                if (next == null) {
                    if (node.i == coefficients.length - 1 && node == node.head) { // prune tree if no solution found
                        node = prune(node.backlink);
                        down = false;
                        up = false;
                    } else {
                        down = false;
                        up = true;
                    }
                } else {
                    down = true;
                    up = false;
                    node = next;
                }
            }
        }

    }

    private void testOverflow() {
        for (int i = 0; i < this.coefficients.length; i++) {
            if(tightable[i]) {
                int sum = 0;
                for (int j = 0; j < this.coefficients[i].length; j++) {
                    if (coefficients[i][j] == 0) {
                    } else if(varBound[j] == Integer.MAX_VALUE) {
                        overflow = true;
                        return;
                    } else {
                        sum += coefficients[i][j] * varBound[j];
                    } 
                }
                if(sum>bound[i]) {
                    overflow = true;
                    return;
                }
            }
        }
        overflow = false;
    }
    boolean overflow;

    public boolean overflow() {
        return overflow;
    }
}


